#!/usr/bin/perl

strict;

# no strict;	# uncomment this line if running under mod_perl, also see below for mod_perl

# don't want any redefinition to sneak into the code during development/debugging
# use warnings qw(redefine);

# if you redefine any subroutines you will want to debug with no warnings
no warnings qw(redefine);

# too many ENV & CGI variables are undefined for this to be useful
no warnings qw(uninitialized);

#
#	 WODA - Web Oriented Database
#
# A program and library of perl functions to manage semi-relational datausebase
# using WWW and cgi interface. Includes functions to add, edit, browse and
# search records using WWW. Database structure is described in a separate
# file.
#
# Usage:
#	see http://www.ddatabase.com/
#	see http://woda.sourceforge.net/ (updated)
#
# Author:
#	Ziga Turk (www.zturk.com) adn others.
#
# Copyright:
#	Copyright (c) 1995-2025 Ziga Turk,
#	University of Ljubljana, Faculty of Civil Engineering and Geodesy, IKPIR.
#	All rights reserved.
#
#	Copyright (c) 2009 Malcolm Fitzgerald, Xavier Brochard
#       and all named contributors in the software.
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU Affero General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
# 
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU Affero General Public License for more details.
# 
#       You should have received a copy of the GNU Affero General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#       See license.txt for explanations
#
# Hacking:
#       see http://launchpad.net/woda for development version
#       Please, format your code using http://perltidy.sf.net like this:
#       perltidy --warning -logfile-gap --maximum-line-length=150
#
#########################################################################

# compatibility with definitions that used
# unpackaged WODA and do main() ... new way is marked below

# main program
sub main {

    WODAUK'mainError() if $ENV{'MOD_PERL'};

    if ( defined &WODAUK'run ) {    #UK
        if ( defined %WBB ) {
            WODAUK'run( *WBB, *WBF );    #UK
        }
        else {
            WODAUK'run( *WBBase, *WBField );    #UK
        }
    }
    else {
        if ( defined %WBB ) {
            run( *WBB, *WBF );
        }
        else {
            run( *WBBase, *WBField );
        }
    }
}

# to use WODA as a package under MOD_PERL
# uncomment the package line below
# beware that this makes life quite complicated for
# any user defined functions !
#
# package WODAUK;	#UK

$VERSION = '4.610';    # set version
wbSetINC();            # add dir of this file to @INC

# all server-wide configuration parameters are in this sub
# to change configuration, edit the file local/woda.conf

# define config file
sub mainConfig {

    if ( -e "$WodaFileDir/local/woda.conf" ) {
        do "$WodaFileDir/local/woda.conf";
    }
    elsif ( -e "$WodaFileDir/../en/local/woda.conf" ) {
        do "$WodaFileDir/../en/local/woda.conf";
    }
    else {
        wbFail( 'woda.conf not available', "File $WodaFileDir/local/woda.conf not available\n" );    #UK
    }

}

############# No need to change anything below ##################

# primary routine
sub run {

    $STDOUT = select;

    # reset all variables for the sake of modPerl

    if ( $ENV{'MOD_PERL'} ) {
        mainClean();

    }
    elsif ( $ENV{'SERVER_SOFTWARE'} =~ m/woda/i ) {

        # do mainClean();

    }

    $StartTime = time;
    logTime("Perl compile overhead");

    # compatibility with old versions of data definition file

    local ( *WBB, *WBF ) = @_;

    # --- where is this script, what are its functions
    # --- moved above Config so that config can use what whoami found

    wbWhoami();
    ( $SCRIPTPACKAGE, @x ) = caller;

    # --- configure WODA

    mainConfig();

    # this is a MODperlis hack
    #
    # 	  findMYSubs();
    #
    # 	 	if ($#MYSubs > 0 && defined &WODA'run) {	# one is main!
    # 			wbFail ("Subroutines are not allowed in package $SCRIPTPACKAGE",
    # 	 		"Author of this database should re-write the
    # 	 		 routines into a separate package. Any WODA variables should be
    # 	 		 prefixed with WODA' eg. $WODA'rec{'name'}.");
    # 	 	}

    # --- command line first

    mainCmdline();

    # --- get command parameters - CGI or command line

    mainPathInfo();

    # --- OK, now we have the WBBs !!!
    # --- pull in the required file
    # --- we are not yet in the DATADIR !!!

    $rf = $WBB{'requireFile'};

    if ( $rf ne '' ) {

        if ( $rf =~ m|^/| ) {

            # nothing

        }
        elsif ( $rf =~ m/^\w/ ) {

            $rf = "$TEMPLATEDIR/lib/$x";

        }
        else {

            $rf = $WBB{'dataDir'} . '/' . $rf;

        }

        if ( -e $rf ) {
            require $rf;
        }
        else {

            # this should not be fatal because it kills the datadir defined stuff
            $Log .= "ERROR: required file $rf is missing\n";
        }
    }

    # --- determine the location of datadir ... needed below

    mainDataDir();

    # --- is there a _common.pl file at ../$DATADIR ... if so, require it
    # --- make sure that file is not accessible by browsers
    # --- load order is: dd.txt, dd.pl, requirefile, _common.pl

    $x = "$DATADIR/../_common.pl";
    if ( $DATADIR && -e $x ) {
        require $x;
        unless ( -e "$DATADIR/../$HtaccessFile" ) {
            open( h, ">$DATADIR/../$HtaccessFile" );
            print h $HtaccessDeny;
            close(h);
        }
    }

    # --- set the defaults

    mainDefaults();

    # --- random seed

    time =~ m/....$/;
    srand( $& * $$ );

    # --- if dataDir is defined, but not set, create it!

    if ($DATADIR) {
        if ( !-d $DATADIR ) {
            wbCheckDirectories();
        }
        chdir($DATADIR) || wbFail("Cannot chdir to $DATADIR");
    }

    # --- JavaScript compatibility

    $JS = 0;

    if ( $ENV{'HTTP_USER_AGENT'} =~ m/Mozilla\/./i ) {
        $JS = 1.2;
    }

    if ( $ENV{'HTTP_USER_AGENT'} =~ m/Mozilla\/3|Mozilla\/2/i ) {
        $JS = 1.0;
    }

    if ( $WBB{'JavaScript'} ne '' ) {
        $JS = $WBB{'JavaScript'} if $WBB{'JavaScript'} < $JS;
    }

    # fill %CGI with parameters

    # Obsolete code: with Perl 5.10 multiline (singleline) matching with $* is replaced
    #                by a /m (or not) in regexp. Changed regexp are in htParseMulti
    # $* = 1;	     # multi-line matching
    htParseMulti();

    # $* = 0;	     # single line matching

    htReadCookies();

    if ( $CGI{'_isindex'} && !$Page ) {
        $Page = "Show";
        $CGI{'_id'} = $CGI{'_isindex'};
    }

    if ( !$Page ) {
        $Page = "Home";
    }

    if ( $CGI{'_isindex'} && !$Page ) {
        $Page = "Show";
        $CGI{'_id'} = $CGI{'_isindex'};
    }

    $Action   = "cgi" . $Page;
    $MyAction = "my" . $Page;

    ############ LOGGING IN ############

    wbLogin();

    # can group do the action (admin can do anything!)

    if ( $n = &wbDenied( $Group, $Page ) ) {
        printDeny($n);
    }

    ########### ARE WE CLOSED ###########

    if ( -e "_cache/closed.ip" ) {
        open( H, "_cache/closed.ip" );
        $tip = <H>;
        close(H);
        if ( $tip ne $RA && $tip ne $ENV{'HTTP_WODA_END_USER'} ) {
            open( H, "_cache/closed.txt" );
            @msg = <H>;
            $msg = join( '', @msg );
            close(H);
            wbError( "Database is closed", "$msg" );    #UK
        }
        else {
            $WBB{'dbTitle'} .= "[CLOSED FOR GENERAL PUBLIC]";    #UK
        }
    }

    ######### IS THE USER A ROBOT ? #########

    if (
        ( $Page !~ m/Show|Home|About|Browse|RobotToc/ )
        && (   -e "$DATADIR/_tmp/robot.$RA"
            || $RA =~ m/$ROBOTIP/
            || $ENV{'HTTP_USER_AGENT'} =~ m/$ROBOTAGENT/i )
      )
    {
        redirectRobot();
        wodaExit();
    }

    ########### IS THE LOAD TOO HIGH ########

    if ( $UPTIME && $WBB{'quotaLoad'} ne '' ) {
        $x = `$UPTIME`;
        $x =~ m/averages:\s+([0-9\.]+)/;
        $load = $1;
        @x = split( /\n/, $WBB{'quotaLoad'} . "\n" );
        foreach $x (@x) {
            ( $l, $p ) = split( /\s+/, $x, 2 );
            next if $l eq '';
            next if $p eq '';
            next if $l > $load;
            next if $Page !~ m/$p/;
            wbError( "Load is too high", "Please try to reload $Page later" );    # UK
        }
    }

    ########### CACHEING ##########

    $NOCACHE = '.|/Adm|Form';                                                     # do not cache anything !!!

    $request = "_$ENV{'PATH_INFO'}_$ENV{'QUERY_STRING'}_";
    $request =~ s|/|\.|g;

    unless ( $request =~ m|$NOCACHE|
        || 'admin' =~ m|$groups|
        || $ENV{'REQUEST_METHOD'} eq "POST" )
    {                                                                             # something is not cached
        $file = "$DATADIR/_cache/$request";
        if ( -s $file ) {
            open( h, $file );
            @x = <h>;
            close h;
        }
        else {
            open( STDOUT, ">$file" );
            binmode(STDOUT);
            mainAction();
            close(STDOUT);

            open( STDOUT, ">-" );
            binmode(STDOUT);
            open( h, $file );
            @x = <h>;
            close h;
        }
        print @x;
    }
    else {
        mainAction();
    }

    wodaExit();

    # add calls to some end user functions so that they will
    # be pulled into the main module if the split version is used
    # xxx how about pulling them in the evalSafe module ???
    # also they could be used from any user defined function
    # which is called through safeEva anyway ??

    if (0) {
        &NOOP();
        &ID();
        &TODAY();
        &QRY();
        &GET();
        &PUT();
        &RELATED();
        &DECODE();
        &LINK();
        &ROWS();
        &FLD();
        &KEY();
        &FETCH();
        &PIC();
        &IDS();
        &NEXT();
        &URL();
        &PUTN();
    }
}

# redirect the robot to robot index
sub redirectRobot {

    cgiRobotToc();
    $Err = 'bot';
    printFoot();
    wodaExit();

    # --- thats it!

    $Err    = "RRD";
    $ErrMsg = "Redirect $ENV{'HTTP_USER_AGENT'} to /robots";
    wbLog();

    # --- shall we update static pages (once every 6 hrs)

    if ( &wbExpired( "$HTMLDIR/robots/index.htm", "", 6 ) ) {
        admStatic();
    }

    # --- redirect the robot there

    print "Location:$SCRIPTBASE$WBB{'htmlURL'}/robots/index.htm\n\n";

    wodaExit();
}

# set paths to data sources
sub mainDataDir {

    # DATADIRDIR include drive letter

    $DATADIRDIR =~ s/^[a-z]\://i;

    # compatibility with old layout of files

    if ( $WBB{'homeURL'} && !$WBB{'htmlDir'} ) {
        $WBB{'htmlURL'} = $WBB{'homeURL'};
        $WBB{'htmlDir'} = $WBB{'dataDir'};
    }

    # set htmlDir to dataDir if undefined

    $WBB{'htmlDir'} = $WBB{'htmlDir'} || $WBB{'dataDir'};

    # if htmlURL is undefined, try to set it up automatically

    $na = "/htmlURL/is/not/defined";

    if ( !$WBB{'htmlURL'} || $WBB{'htmlURL'} eq $na ) {

        $x = $ENV{'DOCUMENT_ROOT'};
        if ( $x ne '' && $WBB{'htmlDir'} =~ m|$x|i ) {    # ignore case for mac IE5
            $x = $';                                      # unmatched part
            $x =~ s/\/$//;                                # remove trailing slash
            $x = "/$x";                                   # prepend start slash
            $x =~ s|//|/|;                                # remove double slashes
            $x =~ s|^[a-z]\:||i;                          # remove c: at start

            $WBB{'htmlURL'} = $x;
        }
        else {

            # this error may not be critical
            $WBB{'htmlURL'} = $na;
        }
    }

    $DATADIR = $WBB{'dataDir'};
    $HTMLDIR = $WBB{'htmlDir'};

    # expand datadirdir

    $DATADIR =~ s|\$DATADIRDIR|$DATADIRDIR|;
    $HTMLDIR =~ s|\$DATADIRDIR|$DATADIRDIR|;

    if ( $DATADIR eq '' ) {
        $SCRIPTFILE =~ m|[^/]*$|;
        $DATADIR = "$DATADIRDIR/$&";
        $HTMLDIR = $DATADIR;
    }

    # make sure they are absolute
    # if they are not they are relative to
    # scriptdir

    if ( $DATADIR && $DATADIR !~ m|^/| ) {
        $DATADIR = "$SCRIPTDIR/$DATADIR";
    }

    if ( $HTMLDIR && $HTMLDIR !~ m|^/| ) {
        $HTMLDIR = "$SCRIPTDIR/$HTMLDIR";
    }

    $TMPDIR = "$DATADIR/_tmp";
}

# set default values for essential variables
sub mainDefaults {

    $ROBOTIP = $ROBOTIP || '193.2.92.xx';    # for testing purposes
    $ROBOTAGENT = $ROBOTAGENT
      || 'scooter|mercator|altavista|google|spider|slurp|msnbot|aipbot|crawler|bot';

    # --- some pointers to home of Woda

    $WodaHomeURL = 'http://woda.sourceforge.net';
    $WodaHelpURL = "$WodaHomeURL/help.htm";

    # --- defaults and compatibility issues

    $WBB{'sort;DEFAULT'}   = $WBB{'listRow'}  unless $WBB{'sort;DEFAULT'};
    $WBB{'format;DEFAULT'} = $WBB{'listHtml'} unless $WBB{'format;DEFAULT'};
    $WBB{'formatNameDEFAULT'} = 'Default'
      unless $WBB{'formatNameDEFAULT'} ne '';
    if ( $WBB{'formatRowDEFAULT'} eq '' && $WBB{'format;DEFAULT'} ne '' ) {

        # $WBB{'formatRowDEFAULT'} = $WBB{"format;DEFAULT"};
        $WBB{'formatHeadDEFAULT'} = "<ol>";
        $WBB{'formatFootDEFAULT'} = "</ol>";
    }
    elsif ( $WBB{'formatRowDEFAULT'} eq '' && $WBB{'format;DEFAULT'} eq '' ) {    #MF failsafe
        $WBB{'formatHeadDEFAULT'} = "<ol>";
        $WBB{'formatFootDEFAULT'} = "</ol>";
    }

    $WBB{'maxHits'}  = 10000 unless $WBB{'maxHits'};
    $WBB{'dfltHits'} = 10    unless $WBB{'dfltHits'};
    $WBB{'generatedHomePage'} = "index.htm" if $WBB{'homePage'} eq '1';
    $WBB{'homePage'}          = "index.htm" if $WBB{'homePage'} eq '1';
    $WBB{'groups'} = $WBB{'groups'} || $WBB{'users'};
    $WBB{'grepBigger'} = '300000';

    # --- where are woda files visible by http

    $WODAURL = '/woda';

    # --- icons

    if ( $WBB{'iconURL'} ne '' ) {
        $ICONURL = $WBB{'iconURL'};
    }
    elsif ( $WBB{'iconSet'} ne '' ) {
        $ICONURL = '/woda/icons/' . $WBB{'iconSet'};
    }
    else {
        $ICONURL = '/woda/icons/default';
    }

    $ICONOPEN = $WBB{'iconOpen'} || 'open';

    $ICONPAR      = 'border="0" height="24" width="24" class="ICON"';
    $BIGICONPAR   = 'border="0" height="48" width="48" class="ICON"';
    $SMALLICONPAR = 'border="0" height="12" width="12" class="ICON"';

    # --- 3d lines

    $p    = qq(<img alt="" src="$ICONURL/pixel.gif" width="1" height="1">);
    $Pix4 = qq(<img alt="" src="$ICONURL/pixel.gif" width="4" height="1">);

    $nCriteria = 5;

    # --- powered by stuff
    # --- license does not allow you to change settings of the $POWER... variables!

    $POWERURL  = $POWERURL  || $WodaHomeURL;
    $POWERTEXT = $POWERTEXT || "Powered by WODA $VERSION";
    $POWERIMG  = $POWERIMG  || "$ICONURL/poweredby.gif";
    $POWERGEN  = $POWERGEN  || "WODA $VERSION - $WodaHomeURL";
    $POWERSIZE = $POWERSIZE || 'width="76" height="24"';

    # internationalisation

    $WBB{'intlCharset'} = $WBB{'intlCharset'} || 'UTF-8';                #UK
    $WBB{'intlCollate'} = ''                  || $WBB{'intlCollate'};    #UK
    $WBB{'intlLower'}   = ''                  || $WBB{'intlLower'};      #UK
    $WBB{'intlAscii'}   = ''                  || $WBB{'intlAscii'};      #UK
    $WodaLanguage       = 'en'                || $WBB{'language'};       #UK
    $WodaLanguageAuthor = 'Ziga Turk &lt;ziga.turk@fagg.uni-lj.si&gt;';  #UK
}

# adds name of current woda directory to @INC

sub wbSetINC {

    local ( $package, $filename, $line ) = caller;

    $filename =~ m|/[^/]*$|;
    unshift( @INC, $` );
}

# what file is executed, which WODA library, which language ...

sub wbWhoami {

    local ( $WodaPackage, $filename, $line ) = caller;
    local ( $i, $c, $p, $v, $x );

    $filename =~ m|[^/]*$|;
    $WodaFile    = $&;    # what is the name of this script
    $WodaFileDir = $`;    # in which dir is this script
    chop($WodaFileDir);   # remove the trailing /
    $x = $WodaFile;       # same as the file
    $x =~ s/woda/include/;    # replace woda with include
    $x =~ s/\.pl//;           # replace .pl with nothing
    $WodaLibDir = "$WodaFileDir/$x";

    $SCRIPTFILE = $0;         # command being executed
    $SCRIPTFILE =~ tr/\\/\//;        # use forward slashes
    $SCRIPTFILE =~ s/^[a-z]\://i;    # remove c: d: or similar
    $SCRIPTFILE =~ m|/[^/]*$|;       # match directory
    $SCRIPTDIR = $`;

    if ( $WodaFileDir !~ m|^/| ) {    # not absolute
        $WodaFileDir = $SCRIPTDIR . '/' . $WodaFileDir;
    }

    $WodaFileDir =~ s|/[^/]+/\.\.||g;    # remove /something/..

    $PERL = $PERL || $^X;                # first argument from command line
    if ( $PERL !~ /perl/i || $PERL eq 'perl' || $SCRIPTFILE =~ m/\Q$PERL/ ) {    # looks like $^X reported wrong
        $PERL = '/usr/local/bin/perl';
        $PERL = '/usr/bin/perl' unless -e $PERL;
        $PERL = '' unless -e $PERL;
    }
    $PERLV      = $] + 0.0;                                                      # what Perl are we running for perl >= 5.6
    $PERLV      = $^V if $PERLV >= 5;                                            # old format, before Perl 5.6
    $TESTCGIDIR = $SCRIPTDIR . "/test";
}

# command line interface
#
# synopsis:
# /www/server/cgi/script -x PATH_INFO QUERY_STRING USERNAME:PASSWORD
# perl db/use -x table/Search
# eg: /www/server/cgi/script -x Search 'search=name' 'admin:leo12'

sub mainCmdline {

    if ( $ARGV[0] eq '-x' ) {

        if ( $WBB{'dataDir'} eq '' ) {
            if ( $DATADIRDIR && $ARGV[1] =~ m|^(.+)(/[A-Z].*)| ) {
                $DATADIR = "$DATADIRDIR/$1";
            }
            elsif ( $DATADIRDIR && $ARGV[1] =~ m|^([a-z][^/]*)(.*)| ) {
                $DATADIR = "$DATADIRDIR/$1";
            }
            else {
                print '<p class="ERROR">To run from command line, WBB{dataDir} must be defined</p>';
                wodaExit();
            }
        }
        else {

            # this is needed for admGetEnv to work
            $DATADIR = $WBB{'dataDir'};
        }

        $HTMLFILE = 1;
        $CMDLINE  = 1;
        ( $ENV{'SERVER_PROTOCOL'}, $ENV{'SERVER_NAME'}, $ENV{'SERVER_PORT'}, $ENV{'SCRIPT_NAME'} ) = &admGetEnv();

        if ( !$ENV{'SERVER_PROTOCOL'} ) {
            print '<p class="ERROR">Cannot access environment of ', $SCRIPTFILE,
              ' at ', $DATADIR, '. Was it ever accessed through the Web?</p>', "\n";    #UK
            wodaExit();
        }

        $ARGV[1] = "/$ARGV[1]" unless $ARGV[1] =~ m|^/|;                                # one forgets /Search
        $ARGV[1] =~ s|//|/|;                                                            # remove a possible double slash
        $ENV{'PATH_INFO'}    = $ARGV[1];
        $ENV{'QUERY_STRING'} = $ARGV[2];
        $ENV{'QUERY_STRING'} =~ s/^'//;                                                 # DOS does not remove this from command line parameters
        $ENV{'QUERY_STRING'} =~ s/'$//;                                                 # same here

        unless ( $ENV{'REQUEST_METHOD'} ) {                                             # true command line
                                                                                        # simulte cookie for wbLogin; default is admin
            $COOKIE{_login} = $ARGV[3] || 'admin:';                                     # Group defined in wbLogin
        }
        $ENV{'REMOTE_ADDR'} = $ENV{'REMOTE_ADDR'} || '127.0.0.1';
        $ENV{'REQUEST_METHOD'} = 'GET';
    }
}

# parse PATH_INFO information
# PATH_INFO = [/group:[password]][/Action]
# set
#	$SCRIPT::the constant part of the URL
#	$URLGroup::username supplied on URL
#	$URLPass::password supplied on URL
#	$Referer::the full URL of this page
#

sub mainPathInfo {

    # set RH and RA

    $RH = $ENV{'REMOTE_HOST'};
    $RA = $ENV{'REMOTE_ADDR'};

    # make sure SN has the host name as typed in the browser !

    $SN = $ENV{'HTTP_HOST'} || $ENV{'SERVER_NAME'};

    # compose the base of the script http://www.bla.com

    $ENV{'SERVER_PROTOCOL'} =~ m|/|;
    $p = $` || $ENV{'SERVER_PROTOCOL'};
    $p =~ tr/[A-Z]/[a-z]/;
    $p = 'https'
      if ( $ENV{'HTTPS'} || $ENV{'HTTP_NR_SSL'} && $p eq 'http' );    ## HTTPS ## mf
    $SCRIPTBASE = $p . '://' . $SN . ':' . $ENV{'SERVER_PORT'};
    $SCRIPTBASE =~ s/\n|\r//g;
    $SCRIPTBASE =~ s/:80$//;                                          # remove :80 !

    #	   like /cgi/script?/Search?since=999

    if ( $ENV{'QUERY_STRING'} =~ m/\?/ ) {

        # /cgi/script?/Search?since=999 for microsoft

        $pi = $`;
        $qs = $';

        $SCRIPTNAME = $ENV{'SCRIPT_NAME'};
        $SCRIPTNAME = "/$SCRIPTNAME" unless $SCRIPTNAME =~ m|^/|;

    }
    elsif ( $ENV{'SERVER_SOFTWARE'} =~ /Microsoft/ ) {

        # /cgi/script?/Search

        $pi = $ENV{'QUERY_STRING'};
        $qs = '';

        $SCRIPTNAME = $ENV{'SCRIPT_NAME'};
        $SCRIPTNAME = "/$SCRIPTNAME" unless $SCRIPTNAME =~ m|^/|;

    }
    elsif ( $ENV{'SCRIPT_NAME'} =~ m|^/[A-Z]| ) {

        # alexander has identified this as a problem
        # see http://www.notyourhomework.net/cgi-bin/woda/faqs.pl/Show?_id=8443
    }
    else {

        # /cgi/script/Search?since=999

        $pi = $ENV{'PATH_INFO'};
        $qs = $ENV{'QUERY_STRING'};

        # /cgi/script?/Login ... should work on non IIS as well !

        if ( $qs =~ m|^/| ) {
            $pi = $qs;
            $qs = '';
        }

        $SCRIPTNAME = $ENV{'SCRIPT_NAME'};
        $SCRIPTNAME = "/$SCRIPTNAME" unless $SCRIPTNAME =~ m|^/|;
    }

    # base for all relative URLs

    $SCRIPT = $SCRIPTNAME;

    # add a ? after SCRIPT name for Microsoft

    $SCRIPT .= '?' if $ENV{'SERVER_SOFTWARE'} =~ /Microsoft/i;

    # parse the PATH_INFO part for password and ToFrame

    @x = split( /\//, $pi );

    shift(@x);    # first is blank, started with / !

    # called via a helper cgi script

    $COOKIESCOPE = $SCRIPTNAME;
    $DATADICT    = $SCRIPTFILE;
    $DDTYPE      = $DDTYPE || 'CGI';

    # L=language&U=username&P=pass

    if ( $x[0] =~ /\=/ ) {
        $p = $1 if $x[0] =~ m/P=([^\&]*)/;
        $u = $1 if $x[0] =~ m/U=([^\&]*)/;
        $SCRIPT = "$SCRIPT/U=$u&amp;P=$p";    # encode the ampersand #mf
        $URL_login = "$u:$p" if $u;
        shift(@x);
    }

    # what remains is the action to be taken

    $Page = $x[0];

    # ToFrame ?

    if ( $Page =~ m/ToFrame(.)$/ ) {
        ;
        $TargetDefault = 'target="_parent"';
        $TargetSelf    = 'target="_self"';
        $InFrame       = $1;
        $Page          = $`;
    }

    # how to show this page

    $SCRIPTURL = "$SCRIPTBASE$SCRIPT/$Page";
    $SCRIPTURL .= '?' . $ENV{'QUERY_STRING'} if $ENV{'QUERY_STRING'};

    if ( $ENV{'REQUEST_METHOD'} eq 'GET' ) {
        $Referer = "referer=" . &htEscape($SCRIPTURL);
    }
    else {
        $Referer = '';
    }
}

# fatal error for main

sub mainError {

    wbFail(
        "Improper use of WODA.", "The database definition file
should end in a call like <em>do main(*WBB,*WBF)</em> or <em>do WODAUK'run(*WBB,*WBF)</em>
if used under MOD_PERL.
"
    );
}

# cleans all variables in WODA's package
# With Woda Server ...
# Perl 4 causes Page fault
# Perl 5 OK
# this is for MOD_PERL ONLY !!!

sub mainClean {

    ( $package, @x ) = caller;

    #print STDOUT $package . "\n";
    local (*_stab) = eval("*_$package");                 # if $] <5;
    local (*_stab) = eval("*${package}::") if $] >= 5;

    while ( ( $_key, $_val ) = each(%_stab) ) {
        local (*_entry) = $_val;
        next unless $_key =~ m/^[A-Z]/;                  # starts with a letter
        next if $_key =~ m/^ENV|^INC|^ARGV|^SIG|^WBB|^WBF|^VERSION|^STDOUT/;
        next if $_key =~ m/::/;

        #print STDERR "$_key\n";
        undef %_entry if defined %_entry;
        undef @_entry if defined @_entry;
        undef $_entry if defined $_entry;
    }

}

# sets @MYSubs

sub findMYSubs {

    undef @MYSubs;
    local (*stab) = eval("*_$SCRIPTPACKAGE")     if $PERLV < 5;
    local (*stab) = eval("*${$SCRIPTPACKAGE}::") if $PERLV >= 5;

    while ( ( $key, $val ) = each(%stab) ) {
        local (*entry) = $val;
        if ( defined &entry ) {
            push( @MYSubs, $key );
        }
    }
}

# last thing that WODA sees!

sub wodaExit {

    recCloseAll();

    # print $Log; ... actions after footer !!!

    if ( $ENV{'MOD_PERL'} ) {
        Apache'exit(0);    # close ' for syntax highlighting
    }
    elsif ( $ENV{'SERVER_SOFTWARE'} =~ m/woda/i ) {
        exit(0);
    }
    else {
        exit(0);
    }
}

# determines
#	$User
#	$Group
#	$LogPass
# based on:
#	 URLGroup ... group set on URL ??? se kje ???
# URLPass ... password set on URL
# COOKIE{_user,_email,_login,_group,_voucher}
# Login user, determines $User, $Group and $LogPass

sub wbLogin {

    local ( $group, $pass ) = @_;
    $Group = $User = '';

    # --- we use command line, URL or COOKIES, in that order

    if ( $group eq ' ' ) {

        $dGroup  = '';
        $LogPass = '';
        $Log .= "Logging with parameters - empty\n";

    }
    elsif ( $group ne '' ) {

        $dGroup  = $group;
        $LogPass = $pass;
        $Log .= "Logging with parameters as $group\n";

    }
    elsif ( $URL_login ne '' ) {

        ( $dGroup, $LogPass ) = split( /:/, $URL_login, 2 );
        $Log .= "Logging with URL as '$dGroup'\n";

    }
    else {

        ( $dGroup, $LogPass ) = split( /:/, $COOKIE{'_login'}, 2 );
        $Log .= "Logging with COOKIES as '$dGroup'\n";

    }

    # --- scope of cookies

    $cscope = $WBB{'userScope'} || $COOKIESCOPE;

    # --- determine passwd(groups)file

    if ( $WBB{'groups'} eq '' ) {    # use default of WODA
        @passwd = ($ADMUSER);
    }
    else {
        @passwd = split( /\n/, $WBB{'groups'} );
    }

    # --- identify as client if localhost and

    if ( !grep( /^client\s/, @passwd ) && $ENV{'HTTP_USER_AGENT'} =~ m/woda/i ) {
        local ($x);
        $x = $ENV{'SERVER_ADDR'};
        push( @passwd, "client\t$x|127.0.0.1\t" );
    }

    # --- determine its relevant part

    if ( $dGroup ne '' ) {
        @passwd = grep( /^$dGroup\s/, @passwd );
    }

    # --- lets move down the list

    foreach $_ (@passwd) {
        ( $group, $patt, $passwd ) = split( /\s+/, $_, 3 );
        $passwd =~ s/\s*//g;    # remove whitespace from password
        $patt   =~ s/\s+//;
        $patt   =~ tr/#/./;     # compatibility reasons

        if ( $passwd eq $LogPass
            && ( ( $RH && $RH =~ m/$patt/ ) || $RA =~ m/$patt/ ) )
        {

            # --- we have a match ... admin should not happen via IP

            if ( $dGroup || $group ne 'admin' ) {
                $Group = $group;
                $Log .= "Logged as $Group using groups setting\n";
                last;
            }
        }
    }

    # --- if group still undetermined, try userTable

    if ( $Group eq '' && $dGroup ne '' && $WBB{'userTable'} ) {
        $t = int( time / 1000 );    # close / for syntax highlighting

        # did he supply a good voucher

        if (   $COOKIE{_group}
            && $COOKIE{_voucher}
            && &wbTestPass( "$ra$t:$dGroup", $COOKIE{_voucher} ) )
        {
            $Group = $COOKIE{_group};
            $User  = $dGroup;
            $Log .= "Logged using valid voucher\n";
            return;
        }

        # no, test him te slow way
        #$dataDir = &wbTables('DIR',$WBB{'userTable'});
        #unless ($dataDir && -d $dataDir) {
        #do wbFail ("Directory for table $WBB{'userTable'} not found\n");
        #}

        $Log .= "Reading from $WBB{'userTable'} for $dGroup\n";
        %rec = &GET( $WBB{'userTable'}, $dGroup );

        if (%rec) {

            # novo 4.155

            $Log .= "recpass='$rec{'_password'}'\n";

            if ( $rec{'_password'} && &wbTestPass( $LogPass, $rec{'_password'} )
                || ( $rec{'_password'} eq '' && $LogPass eq '' ) )
            {

                # password OK
                if ( $WBB{'userAllow'} eq '' || &evalSafe( $WBB{'userAllow'} ) ) {    # setting
                    $User    = $dGroup;
                    $Group   = $rec{'group'} || $WBB{'userTableGroup'} || 'user';
                    $voucher = &wbCryptPasswd("$ra$t:$dGroup");
                    $cg      = &htFormatCookie( '_group', $Group, $cscope, 'SESSION' );
                    $cv      = &htFormatCookie( '_voucher', $voucher, $cscope, 'SESSION' );
                    &htDemandCookie( $cu, $cg, $cv );
                    $Log .= "Verified username and password in a database, setting voucher\n";
                    return;
                }
                else {
                    $Log .= "userAllow failed\n";
                    $userAllowFailed = 1;
                }
            }
            else {
                $Log .= "Login in userTable failed - bad password for $dGroup\n";
            }
        }
        else {
            $Log .= "Login in userTable failed - bad username $dGroup\n";
        }
        undef %rec;
    }
    else {
        $Log .= "Login without userTable\n";
    }

    # --- we did not log in via user table

    $Group = $Group || 'guest';    # make him at least guest

    if ( $COOKIE{_user} =~ m/^_/ ) {
        $User = $COOKIE{_user};
    }
    else {
        $User = '_anon_' . int( rand(1000000) );
    }

    htDemandCookie( &htFormatCookie( '_user', "$User", $cscope ) )
      unless $COOKIE{_user} eq $User;

    # --- promote guest to friend if

    if (   $Group eq 'guest'
        && $COOKIE{_user}  =~ m/^_\d+$/
        && $COOKIE{_email} =~ m/.\@./ )
    {
        $Log .= "Promoting to friend\n";
        $Group = 'friend';
    }
}

# resets cookies by re-calling the current page

sub cgiResetCookie {

    $name   = $CGI{'cookie'};
    $cscope = $WBB{'userScope'} || $COOKIESCOPE;
    $cookie = &htFormatCookie( $name, "b", $cscope, "RESET" );

    printLocation( $CGI{'referer'}, $cookie );
}

# is the /Whatever?something allowed

sub wbDeniedURL {

    local ( $dgroup, $durl ) = @_;
    local ( $x1, $x2 );

    # extract action part!

    ( $x1, $x2 ) = split( /\?/, $durl, 2 );
    $x1 =~ m|/([^/]*)$|;
    $durl = $1;

    return &wbDenied( $dgroup, $durl );
}

# syntactic shortcut for NOT denied

sub wbAllowed {

    return !&wbDenied(@_);

}

# is /Whatever allowed
# parameters in CGI!

sub wbDenied {

    local ( $dgroup, $dpage, $dparam ) = @_;
    local (@x);

    if ( $HIDDENCGI && $dpage =~ m/$HIDDENCGI/ ) {

        # dont wbFail because it exits, this called to check URLs too
        $Log .= ("Hidden $dpage\n");
        return 1;
    }

    #	if ($userAllowFailed) {
    #		return 0 unless $dpage =~ /^Login$|^LoginForm$|^$|^About$/;
    #	}
    #
    return 0
      if $dpage =~ /^Login$|^LoginForm$|^LoginOrRegister$|^$|^AboutUser$/;

    #
    #	to je bilo prej za if stavkom ... ampak Logi se dovoli VEDNO!
    #

    # deny access of others to the record

    if (   $RGroup eq 'other'
        && $dpage =~ m/Edit|Modify|Delete/
        && $dgroup ne 'admin' )
    {
        return 1;
    }

    # make sure rights end with \n

    $x = $WBB{'rights'};
    $c = chop($x);
    if ( $c ne "\n" ) {
        $x .= "$c\n";
    }
    else {
        $x .= "\n";
    }

    # Agent and Adm URLs are denied by default

    $x .= <<EOM;
allow	client	^X
deny	guest	^Basket|^SearchHistory|^Form
deny	.*	^Agent
deny	.*	^Adm|^X
deny	.*	^Search	format=XML|format=CSV|format=TAB
EOM

    @x = split( /\n/, $x );

    $querystring = $dparam;    # will not use CGI!
    foreach $_ (@x) {

        # deny guest Search sex
        ( $what, $group, $patt, $qpatt ) = split( /\s+/, $_, 4 );
        $patt =~ s/\s+//;
        if ( $dpage =~ m/$patt/ && $dgroup =~ m/$group/ ) {
            if ( $qpatt eq '' ) {
                return 1 if $what eq 'deny' && $dgroup ne 'admin';
                return 0 if $what eq 'allow';
            }
            else {

                # 4th column ... value of parameter

                # make string ... once!
                if ( $querystring eq '' ) {
                    foreach $key ( keys(%CGI) ) {
                        next if length( $CGI{$key} ) > 500;
                        $querystring .= "&$key=";
                        $querystring .= $CGI{$key};
                    }
                }

                next if $querystring eq '';    # applies if qs is there

                # $Log .= "Deny $querystring $qpatt\n";
                if ( $querystring =~ m/$qpatt/ ) {
                    $y = $&;
                    $x = $`;
                    $x =~ m/[^\&]*$/;
                    $DeniedPar = "$&$y";
                    return 10 if $what eq 'deny' && $dgroup ne 'admin';
                    return 0 if $what eq 'allow';
                }
            }
        }
    }

    # allow by default !
    return 0;
}

# do whatever the user wanted
# /cgi-bin/script/Wanted is handled by
#
# function myWanted ... if defined
# function cgiSomething in local/cgiSomething.pl ... if exists
# function cgiWanted ... if defined
# function cgiWanted in include/cgiWanted.pl (built ins have precedence over contrib)
# function cgiWanted in contrib/cgiWanted.pl
# function cgiWanted in ../en/contrib/cgiWanted.pl

sub mainAction {

    # another compatibility trick

    if ( !$DATADIR || $Action eq 'cgiUpgrade24' || $CMDLINE || -e '_data' ) {

        # data laid out right
    }
    else {
        forwardTo("Upgrade24");
    }

    # sub mySomething defined
    if ( defined &$MyAction ) {
        &$MyAction();

        # file cgiSomething in local
    }
    elsif ( -e "$WodaFileDir/local/$Action.pl" ) {    # overloaded separate file ?
        require "$WodaFileDir/local/$Action.pl";
        &$Action()

          # sub cgiSomething defined
    }
    elsif ( defined &$Action ) {
        &$Action();

        # file cgiSomething in include dir
    }
    elsif ( -e "$WodaLibDir/$Action.pl" ) {           # separate file ?
        require "$WodaLibDir/$Action.pl";
        &$Action();

        # file cgiSomething in contrib dir
    }
    elsif ( -e "$WodaFileDir/contrib/$Action.pl" ) {    # overloaded separate file ?
        require "$WodaFileDir/contrib/$Action.pl";
        &$Action()

          # file cgiSomething in ../en/contrib dir
    }
    elsif ( -e "$WodaFileDir/../en/contrib/$Action.pl" ) {    # overloaded separate file ?
        require "$WodaFileDir/../en/contrib/$Action.pl";
        &$Action()

    }
    else {
        $m = <<EOM;                                           #UK
Maybe file '$WodaLibDir/$Action.pl' is missing.
<br>Maybe full path to the above file is missing.
<br>Maybe the upper/lower case of the filename (under UNIX) is wrong.
EOM
        wbFail( "Action '$Action' not implemented", $m );     #UK
    }

    printFoot();
    wodaExit();
}

# forward use to another action (passed as parameter)

sub forwardTo {
    printLocation("$SCRIPTBASE$SCRIPT/$_[0]");
}

# do a cgi action ... load appropriate file first

sub wbDo {

    $_ = shift(@_);
    "$WodaLibDir/$_.pl" unless defined &$_;    # load
    &$_(@_);                                   # execute
}

# print sidebar on the /Home page

sub formatHomeMenu {

    local ($cols) = @_;

    local ($r);

    @ub = split( /\n|\,/, $WBB{'userMenuBrowse'} );
    @uf = split( /\n|\,/, $WBB{'userMenuFind'} );
    @ud = split( /\n|\,/, $WBB{'userMenuAdd'} );
    @ua = split( /\n|\,/, $WBB{'userMenuPersonal'} );
    @uo = split( /\n|\,/, $WBB{'userMenuOther'} );

    $r = <<'EOM';
<div class="MENU">
<!-- 
Use these to add your own items to menus:
$WBB{'userMenuBrowse'}
$WBB{'userMenuFind'}
$WBB{'userMenuAdd'}
$WBB{'userMenuOther'}
Use $WBB{'rights'} to remove from menus!
-->
EOM

    $r .= <<EOM if $cols;
<table width="100%" border="0"><tr><td width="25%" valign="top">
EOM

    # --- A-Z indexes

    undef @az;
    if ( $WBB{'AZindex'} ) {
        @x = split( /\n/, $WBB{'AZindex'} );
        foreach $_ (@x) {
            local ( $id, $x, $desc ) =
              split( /\s+/, $_, 3 );    # make variables my #mf
            push( @az, $desc, "$SCRIPT/BrowseAZ?name=$id" );
        }
    }

    # --- browsing stuff

    undef @fields;
    wbBrowseDefaults();
    $f = join( ', ', @fields );

    if ( $#fields >= 0 ) {
        foreach $field (@fields) {
            if ( $WBF{ $field, 'typePar' } =~ m/TREE/ ) {
                $t = "Hierarchy of '$field'";    #UK
                push( @az, $t, "$SCRIPT/BrowseTree?field=$field" );
            }
        }

        # $t = "By one keyword from fields $f"; #UK
        $t = "By one keyword";                   #UK
        push( @az, $t, "$SCRIPT/BrowseContents" );

        $t = "By multiple keywords";             #UK
        push( @az, $t, "$SCRIPT/BrowseKeywords" );

    }

    if ( $#fields >= 1 ) {
        $t = "Combinations";                     #UK
        push( @az, $t, "$SCRIPT/BrowseMatrixList" );
    }

    # --- ring

    if ( $x = &wbRingField() ) {
        $t = "Ring of pages pointed by $x";      #UK
        push( @az, $t, "$SCRIPT/BrowseRing" );
    }

    # --- conditional personal items

    if ( -s "$DATADIR/_friend/$User.srch" ) {
        $t = "Search history";                   #UK
        push( @ua, $t, "$SCRIPT/SearchHistory" );    #UK
    }

    if ( -s "$DATADIR/_friend/$User.forms" ) {
        $t = "Saved forms";                          #UK
        push( @ua, $t, "$SCRIPT/FormsList" );        #UK
    }

    if ( -s "$DATADIR/_friend/$User.bskt" ) {
        $x = $WBB{'basketName'};
        $t = "$x";                                   # e.g. Show shopping list
        push( @ua, $t, "$SCRIPT/BasketShow?editable=1" );
    }

    # --- FIND MENU

    $x = <<EOM ;                                     #UK
Find
Simple search
What's new
Advanced search
A record by key
EOM
    @x = split( /\n/, $x );

    $r .= &formatMenuTitle(qq(<img src="$ICONURL/search.gif" $ICONPAR alt="search" title="search"> $x[0]));
    $r .= &formatMenuItems( "$x[1]", "$SCRIPT/SearchForm", "$x[2]", "$SCRIPT/NewForm", "$x[3]", "$SCRIPT/AdvancedSearchForm", "$x[4]",
        "$SCRIPT/SearchIDForm", @uf );

    $r .= '</td><td width="25%" valign="top">' if $cols;

    # --- BROWSE MENU

    $x = <<EOM ;    #UK
Browse
All records
EOM

    @x = split( /\n/, $x );

    $r .= &formatMenuTitle(qq(<img alt="browse" src="$ICONURL/browse.gif" $ICONPAR title="browse"> $x[0]));
    $r .= &formatMenuItems( @az, @ub, $x[1], "$SCRIPT/Search?", );

    $r .= '</td><td width="25%" valign="top">' if $cols;

    # --- ADD

    $x = <<EOM ;    #UK
Add
Add new $WBB{'recordTitle'}
EOM

    @x = split( /\n/, $x );
    $r .= &formatMenuTitle(qq(<img alt="add" src="$ICONURL/add.gif" $ICONPAR title="add"> $x[0]));
    $r .= &formatMenuItems( "$x[1]", "$SCRIPT/Add", @ud );

    # --- PERSONAL

    $x = <<EOM ;    #UK
Personal
View login record
Change password
Login or register
About $User
Database administration
EOM

    @x = split( /\n/, $x );

    if ( $Group eq 'admin' || $Group eq 'guest' || $Group eq 'friend' ) {
        $uicon = "u$Group";
    }
    else {
        $uicon = "uother";
    }

    $u = &loginRecord();
    if ( $u eq '' ) {
        $x[1] = '';
        $x[2] = '';
        $x[3] = '';
    }
    else {
        $cp = $u;
        $cp =~ s|/Show|/PassChange|;
        $x[3] = 'Logout';
    }

    $r .= &formatMenuTitle(qq(<img src="$ICONURL/$uicon.gif" $ICONPAR alt="face" title="face"> $x[0]));
    $r .= &formatMenuItems( "$x[1]", $u, "$x[2]", $cp, "$x[3]", "$SCRIPT/LoginOrRegister?$Referer",
        @ua, "$x[4]", "$SCRIPT/AboutUser", "$x[5]", "$SCRIPT/AdmMenu" );

    $r .= '</td><td width="25%" valign="top">' if $cols;

    $x = <<EOM ;    #UK
Other
General information
Recent users
Database structure
Office 2003
$POWERTEXT
EOM
    @x = split( /\n/, $x );

    $r .= &formatMenuTitle(qq(<img src="$ICONURL/info.gif" $ICONPAR alt="info" title=info""> $x[0]));
    $r .= &formatMenuItems(
        "$x[1]", "$SCRIPT/About",  "$x[2]", "$SCRIPT/FriendsOnline", "$x[3]", "$SCRIPT/DisplayStructure",
        "$x[4]", "$SCRIPT/Office", @uo,     "$x[5]",                 "$POWERURL"
    );

    $r .= '</td></tr></table>' if $cols;
    $r .= "</div>\n";
}

# display home page, provide user access to cgi tools
sub cgiHome {

    local ( $MENU, $SEARCH, $ABOUT, $DATA, $LOGIN, $help );
    my $t = $WBB{'title;Home'} || 'Main Menu';    #UK
    printHead($t);

    $SEARCH = &wbSearchForm();
    $ABOUT  = '<p>' . $WBB{'about'};
    if ( $WBB{'homePageData'} ne '' || length( $WBB{'homePageFormat'} ) > 500 ) {
        $DATA  = &formatHomeData();
        $MENU  = &formatHomeMenu();                                      # one column
        $LOGIN = &formatLoginBox('width="100%"') if $WBB{'userTable'};
    }
    else {
        $DATA  = '';
        $MENU  = &formatHomeMenu(4);                                     # columns
        $LOGIN = &formatLoginBox() if $WBB{'userTable'};
    }

    $help = <<EOM;
<!-- 
	if you don't like how this page looks
	set a static page with WBB{'homePage'} or
	format it differently with WBB{'homePageFormat'}
	change the data part with WBB{'homePageData'}
	and add menu items with WBB{'userMenuXxxxx'}
-->
EOM

    if ( $WBB{'homePageFormat'} ne '' ) {

        print &evalMaybe( $WBB{'homePageFormat'} );

    }
    elsif ($DATA) {

        print <<EOM;
$help
<table cellpadding="0" cellspacing="0" border="0" width="100%">
<tr valign="top">
<td valign="top">$LOGIN $MENU</td>
<td STYLE="padding-left:5px">
	$SEARCH

	$ABOUT

	$DATA

</td>
</tr>
</table>
EOM

    }
    else {

        print <<EOM;
$help
<table border="0"><tr><td width="100%">$SEARCH</td><td>$LOGIN</td></tr></table>
$ABOUT
$MENU
EOM

    }

    # obsolete, robots files are dynamic since 4.596
    #	if ($WBB{'robots'}) {
    #		print "<a href=$WBB{'htmlURL'}/robots/index.htm>&nbsp;</a>";
    #	}
}

# formats the data part of the home page

sub formatHomeData {

    local ($r);
    local ($cfile);

    $cfile = "$DATADIR/_cache/$Group-homedata.txt";

    # we don't cache home page data any more!
    if ( &wbExpired($cfile) || &wbExpiredDD($cfile) ) {

        open( cfile, ">$cfile" );
        select(cfile);

        $d = $WBB{'homePageData'} || 5;

        if ( $d =~ m/^\d+$/ ) {

            print "<h3>Last $d Entries:</h3>";    #UK

            $CGI{'search'} = '';
            $CGI{'sort'}   = 'TIME';
            $CGI{'max'}    = $d;
            cgiSearch('H');

        }
        else {

            $CGI{'field'} = $d;
            $CGI{'order'} = '';
            cgiBrowseTree('H');

        }
        select($STDOUT);
        close(cfile);
    }

    open( cfile, $cfile );
    @x = <cfile>;
    $r = join( '', @x );
    return $r;
}

# displays the about pages

sub cgiAbout {

    printHead("About $WBB{'dbTitle'}");    #UK

    undef @p;
    undef @v;
    @p = (
        "About this database",             #UK
        "Database manager",                #UK
        "Database engine",                 #UK
        "Translations by",                 #UK
        "Access count"                     #UK
    );

    ( $ac, $since ) = &wbAccessData();
    $since = &formatYMDHMS($since);
    $times = "Accessed $ac times since $since";       #UK
    $m     = &formatMailto( $WBB{'managerEmail'} );

    @v = ( $WBB{'about'}, "$WBB{'manager'} $m", "<a href=\"$WodaHomeURL\">WODA</a> $VERSION $WodaLanguage</a>", $WodaLanguageAuthor, $times );

    print &formatNameVal( @p, @v );
}

# display information about the user

sub cgiAboutUser {

    printHead("About user $User of $WBB{'dbTitle'}");    #UK

    undef @p;
    undef @v;
    foreach $x ( sort( keys(%COOKIE) ) ) {
        push( @v, "$x = $COOKIE{$x}" );
        push( @p,
            qq(<a href="$SCRIPT/ResetCookie?cookie=$x&$Referer"><img $ICONPAR align="top" src="$ICONURL/del.gif" alt=delete title=delete></a>) );  #UK
    }

    $cooks = &formatNameVal( @p, @v );

    undef @p;
    undef @v;
    @p = (
        "",
        "User ID",                                                                                                                                 #UK
        "Group ID",                                                                                                                                #UK
        "",
        "Cookies",                                                                                                                                 #UK
        ""
    );

    @v = (
        '<b>Login data:</b>',                                                                                                                      #UK
        $User,
        $Group,
        '<b>Cookies</b> (delete with utmost care)',                                                                                                #UK
        $cooks,
        '<b>Histories</b>'                                                                                                                         #UK
    );

    if ( -s "$DATADIR/_friend/$User.srch" ) {
        push( @p, 'Search history' );                                                                                                              #UK
        push( @v, qq(<a href="$SCRIPT/SearchHistory"><img src="$ICONURL/search.gif" $ICONPAR alt="search" title="search"></a>) );
    }

    if ( -s "$DATADIR/_friend/$User.forms" ) {
        push( @p, 'Saved searches' );                                                                                                              #UK
        push( @v, qq(<a href="$SCRIPT/FormsList"><img src="$ICONURL/save.gif" $ICONPAR alt="saved forms" title="saved forms"></a>) );
    }

    if ( -s "$DATADIR/_friend/$User.bskt" ) {
        $x = $WBB{'basketName'};
        push( @p, $x );
        push( @v, qq(<a href="$SCRIPT/BasketShow?editable=1"><img src="$ICONURL/basket.gif" $ICONPAR alt="basket" title="basket"></a>) );
    }

    $x = &formatLoginRecordLink();

    if ( $x ne '' ) {
        push( @p, '' );
        push( @v, '<b>For logged-in users only:</b>' );    #UK

        push( @p, 'Show personal information' );           #UK
        push( @v, $x );

        $x =~ s|/Show|/Edit|g;
        push( @p, 'Edit personal information' );           #UK
        push( @v, $x );

        $x =~ s|/Edit|/PassChange|g;
        push( @p, 'Change password' );                     #UK
        push( @v, $x );

    }

    print '<div class="RECORDS">';
    print &formatNameVal( @p, @v );
    print '</div>';
}

#include javascript for htmlarea, rich text editor for text area

sub prepareHtmlArea {

    if ( $ENV{'HTTP_USER_AGENT'} =~ m/msie [6789]|msie 5.5|gecko/i ) {
        $HtmlAreaWorks = 1;
        $MetaFields    = <<EOM;    #UK
<!-- are .js files installed there ? -->
<script type="text/javascript" src="$WODAURL/htmlarea/htmlarea.js"></script>
<script type="text/javascript" src="$WODAURL/htmlarea/lang/en.js"></script>
<script type="text/javascript" src="$WODAURL/htmlarea/dialog.js"></script>
<script type="text/javascript" src="$WODAURL/htmlarea/popupwin.js"></script>
<style type="text/css">
\@import url($ICONURL/../htmlarea/htmlarea.css);
</style>
EOM

    }
    else {
        $HtmlAreaWorks = 0;
    }
}

# Add new record

sub cgiAdd {

    prepareHtmlArea();

    undef %rec;

    $x = $WBB{'recordTitle'} || '<!-- WBB{recordTitle} not defined -->';
    printHead("New $x");

    $reset = "Undo edits";    #UK

    if ( $WBB{'verifyEdits'} ) {
        $submit = "Continue to Preview ...";    #UK
    }
    else {
        $submit = "Save into the database";     #UK
    }

    $x = &wbForm();                             # sets has userfile

    $enctype = "";
    $enctype = 'ENCTYPE="multipart/form-data"' if $HasAttachmentField;

    print <<EOM;
<form class="FORM" name="main" $enctype action="$SCRIPT/Changed" method="post">
$x
<p><a name="SUBMIT">
<input class="BUTTON" type="submit" value="$submit">
<input class="BUTTON" type="reset"  value="$reset"></a>
</form>
EOM
    print evalMaybe( $WBB{'formFoot'} );
}

# show a form to correct bad fields
# show a form to correct fields that fail verification during data entry

sub correct {

    $x     = &wbForm();
    $reset = "Undo edits";    #UK

    if ( $WBB{'verifyEdits'} ) {
        $submit = "Continue to Preview ...";    #UK
    }
    else {
        $submit = "Save into the database";     #UK
    }

    $enctype = "";
    $enctype = 'ENCTYPE="multipart/form-data"' if $HasAttachmentField;

    print <<EOM;
<form class="FORM" name="main" $enctype action="$SCRIPT/Changed" method="post">
$x
<p>
<input type="hidden" name="_id" value="$CGI{'_id'}">
<input type="hidden" name="_accessPassword" value="$CGI{'_accessPassword'}">
<input class="BUTTON" type="submit" value="$submit">
<input class="BUTTON" type="reset" value="$reset">
</form>
EOM
}

# eval the WBB{'key'}, fixes it and returns value of key
# returns blank on keys which are automatically computed (&NEXT)

sub evalWBBkey {

    local ( $x, $y );

    $x = $WBB{'key'};
    return '' if $x =~ m/\&NEXT/;

    $y = &evalSafe($x);    # setting
    return &wbFixId($y);

}

# next in step in sequence after cgiAdd.

sub cgiChanged {

    prepareHtmlArea();

    # set cookies for sticky stuff

    $x = &wbCookieCGI();

    # headers

    if ( $WBB{'verifyEdits'} ) {
        printHead( "Confirm Edits", $x );    #UK
    }
    else {
        printHead( "Data Submitted", $x );    #UK
    }

    if ( $CGI{'_password'} ne $CGI{'_password1'} ) {
        print '<p class="ERROR">';
        print "The two passwords do not match!";    #UK
        return;
    }

    $oldID = $CGI{'_id'};

    # --- authenticate the record

    if ($oldID) {
        recRead($oldID);
        if ( %rec
            && !&wbTestPass( $CGI{'_accessPassword'}, $rec{'_password'} ) )
        {
            if ( &isOwner() ) {

                # OK ... we own the old record
            }
            else {
                wbBadPass($oldID);
                return;
            }
        }

        if ( $CGI{'_at'} && $rec{'_at'} ne $CGI{'_at'} ) {
            wbError(
                "Data in form expired",
"The data changed while you were editing it. Your changes are lost! Perhaps another user was editing the same record at the same time. Or you got to an old Edit screen by using browser's back and forward buttons.",
                '',
                -2
            );    # UK
            return;
        }

        # xxx admin group should not corrupt password !!!

        if ($ADMINPASS) {
            print '<p class="WARN">';
            print "Bad password entered by administrator.<br>Existing password will be preserved.</p>\n";    #UK
            $oldRec{'_password'} = $rec{'_password'};
        }

        # preserve values we were not allowed to edit into oldrec

        foreach $x ( keys(%rec) ) {
            $oldRec{$x} = $rec{$x} unless defined $CGI{$x};                                                  # was not on form

            next if $WBF{ $x, 'modifies' } eq '';                                                            # everyone modifies
            next if $Group =~ m/admin|$WBF{$x,'modifies'}/;                                                  # I modify
            $oldRec{$x} = $rec{$x};                                                                          # I don't modify
        }
    }

    # assign a temporaly ID
    # this ID will be a random code always

    wbCheckDir($TMPDIR);
    $tmpID = &tmpID();

    # covert data from CGI variables into %rec
    # also saves attachment files

    wbCgiToRec($tmpID);

    # change record

    changeRec();

    # add/change password

    $rec{'_password'} = &wbFixPass( $CGI{'_lock'}, $CGI{'_password'} );

    # merge in old values

    foreach $x ( keys(%oldRec) ) {
        $rec{$x} = $oldRec{$x};
    }

    # store the .rec file

    recWriteFile("$TMPDIR/$tmpID.rec");

    # compute possible new ID if adding new record OR key not NEXT
    # here is too early for &NEXT key !

    $newID = &evalWBBkey();

    #print "TEST - newID = $newID";

    if ( $newID eq '' ) {

        $newID = $oldID;    # existing? random key reused

    }
    elsif ( $newID eq $oldID ) {

        #OK

    }
    elsif ( &recExists($newID) >= 0 ) {

        wbError( "Record exists", "Record with the same key data ($newID/$oldID) exists <a href=\"$SCRIPT/Show?_id=$newID\">here</a>." );    #UK
        return;

    }
    elsif ( $oldID eq '' ) {

        #OK

    }
    else {

        # we need confirmation!
        $WBB{'verifyEdits'} = 1;

        $b = "Go back";    #UK
        $b = <<EOM;
<form>
<input class="BUTTON" type=BUTTON value="$b" onClick="history.go(-1);">
</form>
EOM

        $t = <<EOM;        #UK
You changed key data in this record.
<p>
If you ignore this warning,
a new record will be added to the database.
<p>
The new record will include any attachments of this record.
<p>
Relations to/from the record will all be broken.
<p>
The old record would have to be deleted!
EOM
        wbWarn( "Warning, key changed", "$t$b" );    #UK

    }

    $err        = &recValidate();
    $rec{'_at'} = time;
    $html       = &wbRecHtml();
    undef $rec{'_at'};

    if ( $WBB{'verifyEdits'} ) {
        print <<EOM ;                                #UK
<p>Nothing has been saved (yet). Please review what you entered
and confirm at the bottom of this page.
EOM
    }
    else {
        print <<EOM unless $err;                     #UK
<p>This is what you entered:
EOM
    }

    print <<EOM unless $err;                         #UK
<p>
$html
<p>Please note that the attachments 
and perhaps the related data may not be properly
displayed at this point!
EOM

    if ($err) {
        $x = "<img src=\"$ICONURL/del.gif\" title=Error alt=Error $ICONPAR>";
        $err =~ s/\n/<br>/g;

        if ( $Group eq 'admin' ) {
            print <<EOM;    #UK
<p>$html</p>
<p class="WARN">$err<br>... but as admin you are allowed to make errors!</p>
EOM
        }
        else {
            print <<EOM;    #UK
<h3 class="ERROR">Errors were discovered !</h3>

<p>Please correct the fields with errors below.
They are marked with $x and <font class="ERROR">painted like this</font>.</p>
EOM
            correct();
            return;
        }

    }
    else {
        print "<p>There are no errors in your input, but please <b>manually</b> check <b>all links</b>!\n";    #UK
    }

    $x = "Save changes";                                                                                       #UK
    $b = "Go back";                                                                                            #UK
    $t = "Nothing is saved yet. Please select:";                                                               #UK

    if ( $WBB{'verifyEdits'} ) {

        # two step store

        print <<EOM;
<p>
<form name="main" action="$SCRIPT/Store">
<center>
$t<br><br>
<input name="tmpID" value="$tmpID" type="hidden">
<input name="newID" value="$newID" type="hidden">
<input class="BUTTON" type="submit" method="post" value="$x">
<input class="BUTTON" type=BUTTON value="$b" onClick="history.go(-1);">
</center>
</form>
<p>
EOM

    }
    else {

        undef %CGI;
        $CGI{'tmpID'} = $tmpID;
        $CGI{'newID'} = $newID;
        cgiStore();

    }
}

# responsible for modifying data for fields that have the 'change' option defined

sub changeRec {

    wbSortFields();

    foreach $field (@IFields) {
        if ( $WBF{ $field, 'change' } ne '' ) {
            $_ = $rec{$field};
            $rec{$field} = &evalSafe( $WBF{ $field, 'change' } );
        }
    }

}

# User registration, Step 1

sub cgiRegisterForm {

    printHead("Registration step 1/3");    #UK

    if ( $Group eq 'guest' ) {
        print <<EOM;                       #UK
<p>By telling the application your email address,
your status will be promoted to '<b>friend</b>' and more
options and functionality will become available to you.
Simply fill in the form below!

<p>Typically, even more options are available to users who have a username and password
and can <a href="$SCRIPT/LoginForm">login</a> into this service.

<p><b>Instructions to complete your registration will be emailed to
the address given, so enter a correct one!</b>
<p>
EOM

        $t    = 'Registration form';                                                     #UK
        $r    = 'Register!';                                                             #UK
        @n    = ( 'Your nickname (optional):', 'Your email address (required):' );       #UK
        @v    = ( '<input name="name" size="10">', '<input name="email" size="20">' );
        $x    = &formatNameVal( @n, @v );
        $form = <<EOM;
<form action="$SCRIPT/Reg2">
$x
<input class="BUTTON" type="submit" value="$r"></form>
EOM

        print &formatWindow( $t, $form );

    }
    else {

        printYouare();
        print "<p>There is no need for you to register.\n";    #UK

    }
}

# print user information

sub printYouare {

    local ($x);

    print <<EOM;                                               #UK
<h3>Information</h3>
<p>Currently you are user <b>$User</b> from group <b>$Group</b>. You may <a href="$SCRIPT/AboutUser">view or edit</a> the information about you.
EOM

}

# returns a link to record in users table
# code shared with loginRecord TODO code factoring
sub formatLoginRecordLink {

    if ( $WBB{'userTable'} && $Group !~ m/friend|guest|admin/ ) {
        $url = &wbTables( 'URL', $WBB{'userTable'} );
        $u = &wbFixId($User);
        return <<EOM;
<a href="$url/Show?_id=$u">$url/Show?_id=$u</a></p>
EOM
    }
    else {
        return '';
    }

}

# display login form for users

sub cgiLoginForm {

    $NOLOGICON = 1;

    local ( $cgiProg, $a );

    $t = "Login / Logout";    #UK
    printHead($t);

    #	if ($SCRIPT =~ m|(/[^/:]+\:[^/]*)|) {
    if ( $URL_login =~ m/(.+)/ ) {
        $t = <<EOM;           #UK
You cannot login as another user if you are specifying the username on the URL. Remove the '$1'
portion of the URL and try again.
EOM

        wbError($t);
        return;
    }

    if ( $Group ne 'guest' ) {    # don't show guests information # mf
        printYouare();
    }
    else {

        #  do not show links to register if RegisterForm is  denied, # mf
        if ( !&wbDenied( 'guest', 'RegisterForm' ) ) {

            if ( $a = $WBB{'userTable'} ) {

                $url = &wbTables( 'URL', $a );

                $r = rand();

                my $app = $APPNAME || 'web site';    #UK

                print <<EOM;                         #UK
<H3>New Users are Welcome</H3><p>
If you would like to use the $app please <a href="$url/Add" target="$r"><b>fill out this registration form</b></a> then log-in.
EOM
            }

        }

        #print "<H3>Old users:</H3><p>\n";	#UK # mf

        print <<EOM;                                 #UK
<H3>Welcome</H3>
<p>
To log in enter your username and password in the spaces below.
</p>
EOM

    }

    $t   = 'Login';                                                             #UK
    $t1  = 'Logout';                                                            #UK
    $u   = $CGI{'username'};
    $out = qq(<input class="BUTTON" type="submit" value="$t1" name="logout">)
      if $Group ne 'guest' && $Group ne 'friend';

    @p = ( "Username", "Password", "Log me in forever", "", "" );               #UK
    @v = (
        qq(<input size="8" name="groupname" value="$u">),
        '<input size="8" name="password" type="password">',
        '<input name="save" type="checkbox">',
        qq(<input name="referer" value="$CGI{'referer'}" type="hidden">),
        qq(<input name="button" class="BUTTON" type="submit" value="$t"> $out),
    );

    $x = &formatNameVal( @p, @v );

    $w = $CGI{'win'} || 'x';

    $body = <<EOM;
<form action="$SCRIPT/Login" method="post">
$x
<input name="win" value="$w" type="hidden"></form>
EOM

    print &formatWindow( $t, $body, 'BOX' );

}

# display form asking user to login or to register

sub cgiLoginOrRegister {

    if ( ( $Group ne 'guest' ) || &wbDenied( 'guest', 'RegisterForm' ) ) {
        cgiLoginForm();
        return;
    }

    printHead("Login or register");    #UK

    printYouare();

    $ref = &htEscape( $CGI{'referer'} );

    print <<EOM;                       #UK
<p>You may wish to:
<ul>
<li><a href="$SCRIPT/LoginForm?referer=$ref">Login</a> (you need to know a username and password!).
<li><a href="$SCRIPT/RegisterForm?referer=$ref">Register</a> (we'll ask for your email address).
</ul>
EOM

}

# determines whether the browser uses cookies

sub wbCanCookie {

    if ( $ENV{'HTTP_COOKIE'} eq '' ) {
        $x = <<EOM;    #UK
It seems that you are not letting this application
to set cookies or your browser is not supporting
them at all.
<p>
Please enable cookies and reload this page.
<p>
Cookie is a small piece of information
stored <b>in your PC</b>. Your browser sends it
to us whenever you use this application, so that
we can maintain who you are.This is quite
essential for some operations.
<p>
We let you review and/or delete any cookies this application set
in the 'AboutUser' page.
EOM

        wbError( "Cookies disabled", $x );

    }
}

# User Registration, step 2

sub cgiReg2 {

    wbCanCookie();

    $to   = $CGI{'email'};
    $name = $CGI{'name'};

    unless ( &isEmail($to) ) {
        $to = &htQuote($to);
        wbError( "Bad email address", "'$to' is not a valid email address." );    #UK
    }

    # --- known email already ?

    $key = &getRegisteredKey($to);
    if ($key) {
        ( $to, $name ) = &getRegisteredData($key);
        loginRegistered( $key, $to, $name );
        printHead("Registration renewed");                                        #UK
        print <<EOM;                                                              #UK
<p>$to registered before. The registration will be
remembered (using a cookie) in this browser as well. 
EOM
        return;
    }

    printHead("Registration: Step 2/3");                                          #UK

    $from = $WBB{'managerEmail'};
    $key  = int( rand(100000000) );
    $url  = "$SCRIPTBASE$SCRIPT/Reg3?key=$key";

    $subject = "REGISTRATION: " . $WBB{'dbTitle'};                                #UK

    $headers = <<EOM;
From: $from
Reply-to: $from
To: $to
Subject: $subject
EOM

    $message = <<EOM;                                                             #UK
Hi,

To conclude the registration,
please open this URL:

$url

or enter this value:

$key

into the form shown at

$SCRIPTBASE$SCRIPT/Reg3

Automatically,

$WBB{'manager'},
manager.
EOM

    # write file to tmp dir

    open( h, ">$TMPDIR/$key.reg" );
    binmode(h);
    print h "$to\t$name\n" || wbFail("Could not write .reg file");
    close(h);

    # prepare message

    $message = "$headers\n$message";

    wbFail( "Error sending mail", "Email could not be sent. <br />Check the -f flag, eg, -f postmaster@xyz.net" ) if &sendMail($message);

    print <<EOM;    #UK
<p>Soon you will receive an email with these headers:
<pre>$headers
</pre>
<p>In the email, you will learn a key,
so that you will be able to complete your
registration at <a href="$SCRIPT/Reg3">$SCRIPTBASE$SCRIPT/Reg3</a>.

<p>Until then you can continue to use
this application just as before.
EOM
}

# User Registration, Step 3

sub cgiReg3 {

    $key = $CGI{'key'};

    if ( $key eq '' ) {
        printHead("Registration: Step 3/3");    #UK
        print '<p>';
        $form = <<EOM;                          #UK
<form action="$SCRIPT/Reg3">
Key that you received by email: <input name="key" size="10">
<input class="BUTTON" type="submit" value="OK">
</form>
EOM
        print &formatWindow( 'Enter key', $form );    #UK
        return;
    }

    $file = "$TMPDIR/$key.reg";

    open( h, $file )
      || wbError( "Invalid key", "Key $key is not valid or has been already used." );    #UK
    ( $email, $name ) = split( /\t/, <h> );
    close(h);
    unlink($file);

    chop($name);

    ( $email2, $x ) = &getRegisteredData($key);
    if ( $email2 eq $email ) {
        wbError( "You registered before.", "$email is already registered" );
    }
    elsif ( $email2 ne '' ) {

        # someone else is here with the same key!
        while ( $key = int( rand(100000000) ) ) {
            ( $email2, $x ) = &getRegisteredData($key);
            last if $email2 eq '';
        }
    }

    setRegistered( $key, $email, $name );

    # --- fake that server received these cookies now

    loginRegistered( $key, $email, $name );

    # --- output the page

    printHead("Registration: Step 3/3");    #UK

    print <<EOM;
<p>Congratulations!

<p>You are now registered as '_$key'.
This is <em>not</em> your username.
<p>This information is stored in a cookie.
If you delete a cookie or would like to
access this database from another workstation,
register again. We will remember your email address!
EOM
}

# log in user using cookie data

sub loginRegistered {

    local ( $key, $email, $name ) = @_;

    $COOKIE{_user}  = "_$key";
    $COOKIE{_email} = $email;
    $COOKIE{_name}  = $name;

    wbLogin();

    $cscope = $WBB{'userScope'} || $COOKIESCOPE;
    $uc = &htFormatCookie( '_user',  "_$key",  $cscope );
    $ec = &htFormatCookie( '_email', "$email", $cscope );
    $nc = &htFormatCookie( '_name',  "$name",  $cscope );
    &htDemandCookie( $uc, $ec, $nc );
}

# sets information about a registered user

sub setRegistered {

    local ( $key, $email, $name ) = @_;
    local ($t);

    $t = time;
    open( h, ">>$DATADIR/_cache/register.txt" );
    flock( h, 2 );
    binmode(h);
    print h "$key\t$email\t$name\t$t\n";
    close(h);
}

# returns an email address of a registered user with id=$k

sub getRegisteredData {

    local ($k) = @_;

    open( h, "$DATADIR/_cache/register.txt" );
    while (<h>) {
        if (m/^$k\t/) {
            ( $key, $email, $name ) = split( /\t/, $_, 3 );
            chop($name);
            return $email;
        }
    }
    return '';
}

# returns key a registered user

sub getRegisteredKey {

    local ($e) = @_;    # email address

    open( h, "$DATADIR/_cache/register.txt" );
    while (<h>) {
        if (m/\t$e\t/) {
            return $`;
        }
    }
    return '';
}

# Print Form which will allow admin to send email to registered friends

sub cgiAdmRegisteredMailForm {

    return
      unless &saveFormCan('AdmRegisteredMailForm');    # sets $SaveFormButton, sets %CGI

    printHead("Send email to registered friends");

    print <<EOM;
$SaveFormForm

<form class="FORM" action="$SCRIPT/AdmRegisteredMailSend" method="post">
EOM

    $me = $CGI{'from'} || $WBB{'managerEmail'};

    @prompts = ( 'Search:', 'From:', 'Cc:', 'Subject:', 'Message:', 'Where-to:', 'Action:' );
    @flds = (
        qq(<input name="search" size="40" value="$CGI{'search'}" >),
        qq(<input name="from" size="40" value="$me">),
        qq(<input name="cc" size="40" value="$me">),
        qq(<input name="subject" size="40" value="$CGI{'subject'}">),
        qq(<textarea wrap="PHYSICAL" name="message" rows="15" cols="40">$CGI{'message'}</textarea>),
        qq(<select name="where"><option>SCREEN<option>MAIL</select> Try 'SCREEN' first to check that the right content is sent!),
        '<input class="BUTTON" type="submit" value="Send"> <input class="BUTTON" type="reset" value="reset" Form> ' . $SaveFormButton
    );

    print &formatNameVal( @prompts, @flds );
    print "</form>\n";
}

# Send email to registered friends

sub cgiAdmRegisteredMailSend {

    return unless &saveFormToolbar();

    printHead("Sending Serial Mail to Registered Friends");
    @all = &regSelect( $CGI{'search'} );
    $x   = $#all + 1;

    print "<p>Sending $x messages.<ol>\n";

    foreach $_ (@all) {
        chop $_;
        ( $id, $to, $n ) = split( /\t/, $_, 3 );

        $mess = <<EOM;
From: $CGI{'from'}
Cc: $CGI{'from'}
To: $to
Reply-to: $from
Subject: $CGI{'subject'}

$CGI{'message'}
EOM

        if ( &isEmail($to) ) {

            # OK
        }
        else {
            $to = &htQuote($to);
            print '<li><em class="ERROR">Bad destination address ', $to, ' </em>', "\n";
            next;
        }

        if ( $CGI{'where'} eq 'MAIL' ) {
            if ( !&sendMail($mess) ) {
                print "<li>To: $to\n";
            }
            else {
                print '<li><em class="ERROR">Error sending mesage</em>', "\n";
            }
        }
        else {
            print "<li><pre>$mess</pre><hr>\n";
        }
    }

    print "</ol>\n";
}

# select a user from list of registered users

sub regSelect {

    local ($filter) = @_;

    open( h, "$DATADIR/_cache/register.txt" );
    @all = <h>;
    close(h);
    @all = grep ( /$filter/i, @all );
    return @all;
}

# List registered users, with a search form

sub cgiAdmRegisteredPrint {

    printHead("Registered users");
    $x = <<EOM;
<form action="$SCRIPT/AdmRegisteredPrint">
<br>
<input name="search" value="$CGI{'search'}">
<input class="BUTTON" type="submit" value="Search">
<br><span class="HINT">This is a simple Perl regular expression, NOT a standard Woda search expression!</span>
</form>
EOM

    print &formatWindow( "Enter search parameters", $x );

    print <<EOM;
<p>
<table cellspacing="1">
<tr align="right" class="DATA1" >
  <th>no</th>
  <th>id</th>
  <th>email address</th>
  <th>nickname</th>
  <th>files</th>
  <th>actions</th>
</tr>
EOM

    @all = &regSelect( $CGI{'search'} );
    $no  = 1;
    foreach $_ (@all) {
        chop $_;
        ( $id, $e, $n ) = split( /\t/, $_, 3 );
        $files = '';
        foreach $type ( 'srch', 'agnt', 'bskt' ) {
            if ( -s "$DATADIR/_friend/_$id.$type" ) {
                $files .= qq(<a href="$SCRIPT/AdmRegisteredData?id=_$id&amp;type=$type">$type</a> );
            }
        }

        $logfile = &htEscape( &logFile() );
        $llink   = qq(<a href="$SCRIPT/AdmLogfilePrint?logfile=$logfile&amp;search=_$id">log</a>);

        $c = ( $no + 1 ) % 2;

        print <<EOM;
<tr align="right" class="DATA$c">
  <td>$no</td>
  <td>$id</td>
  <td><a href="mailto:$e">$e</a></td>
  <td>$n</td>
  <td>$files</td>
  <td>$llink</td>
</tr>
EOM

        $no++;

    }

    print "</table>\n\n";

    print <<EOM;
<form action="$SCRIPT/AdmRegisteredMailForm">
<input type="hidden" name="search" value="$CGI{'search'}">
<input class="BUTTON" type="submit" value="Send them email ...">
</form>
EOM
}

# display information about 'Friends'

sub cgiAdmRegisteredData {

    $type = $CGI{'type'};
    $id   = $CGI{'id'};

    printHead("$type file of user $id");
    print "<pre>\n";

    open( h, "$DATADIR/_friend/$id.$type" );
    while (<h>) {
        print $_;
    }
    close(h);

    print "\n\n</pre>\n";
}

# log in !
# login user

sub cgiLogin {

    if ( $CGI{'new'} ne '' ) {

        if ( $a = $WBB{'userTable'} ) {
            $url = &wbTables( 'URL', $a );
            if ($url) {
                print "Location: $url/Add\n\n";
            }
            else {
                wbFail( "Error", "Table $a for login information not found" );
            }
        }
        else {
            wbError( "Error", "Adding new users is not possible" );
        }

    }

    if ( $CGI{'logout'} ne '' ) {
        $CGI{'groupname'} = '';
        $CGI{'password'}  = '';
    }

    $cscope = $WBB{'userScope'} || $COOKIESCOPE;

    if ( $CGI{'groupname'} ne '' ) {
        wbLogin( $CGI{'groupname'}, $CGI{'password'} );    # return $Group and $User
        $ok = 0;
        $ok = 1
          if ( $Group eq $CGI{'groupname'} || $User eq $CGI{'groupname'} );
        if ( $CGI{'save'} ) {
            $t = '';                                       # forever 2020
        }
        else {
            $t = 'SESSION';
        }

        $cookie = &htFormatCookie( '_login', "$CGI{'groupname'}\:$CGI{'password'}", $cscope, $t );

        # otherwise _user is from anonymous part of session

        $cookie .= &htFormatCookie( '_user', "$CGI{'groupname'}", $cscope, $t );

    }
    else {    # logout
        wbLogin( 0, 0 );
        $cookie = &htFormatCookie( '_login', "$CGI{'groupname'}\:$CGI{'password'}", $cscope, "RESET" );
        $cookie .= &htFormatCookie( '_group',   "a", $cscope, "RESET" );
        $cookie .= &htFormatCookie( '_voucher', "b", $cscope, "RESET" );

        $cookie .= &htFormatCookie( '_user', "c", $cscope, "RESET" );

        $ok = 1;

        # ensure these values are set so that toolbar
        # icon displays the proper status
        # mf
        $Group = 'guest';
        $User  = '_anon_' . int( rand(1000000) );

    }

    if ( $CGI{'logout'} ne '' ) {    # say hello or goodbye #mf # original user name should be keep for goodbye #xb
        $tsay = "Goodbye $xUser";    #UK
    }
    else {
        $tsay = "Hello";             #UK
        $tsay .= " $User" unless $User =~ m/_anon_/;    #UK
    }

    if ($ok) {

        # clean $SCRIPT

        if ( $SCRIPT =~ m|/u\=| ) {
            $SCRIPT =~ s|/u\=[^/]+||;
            $CGI{'referer'} =~ s|/u\=[^/]+||;
            $clean = 1;
        }
        else {
            $clean = 0;
        }

        printHead( "$tsay\n", $cookie );    # return another $t, so I replaced $t with specific variable $tsay #xb
    }
    else {
        printHead("Login failed\n");        #UK
        if ($userAllowFailed) {
            wbError( "Account disabled", "Your account is disabled or has not been activated yet." );    #UK
        }
        else {

            $m = <<EOM;                                                                                  #UK
<p>The password and the username do not match.
Note that the usernames are typically lower-case only.
Passwords are case sensitive!
EOM

            if ( $a = $WBB{'userTable'} ) {
                my $fixid;
                $fixid = &wbFixId( $CGI{'groupname'} );
                $url = &wbTables( 'URL', $a );
                $m .= <<EOM;                                                                             #UK
<p>If you forgot the password, please
<a href="$url/PassR1?_id=$fixid">click here</a>.
EOM
            }

            wbError( "Bad password", $m );                                                               #UK
        }
    }

    #	print &formatTmpWinClose() unless $CGI{'win'} eq 'self';
    $hr = "$SCRIPTNAME/U=$CGI{'groupname'}&amp;P=$CGI{'password'}/Home";

    if ( $CGI{'referer'} ne '' ) {
        $go = <<EOM;                                                                                     #UK
Try the page <a href="$CGI{'referer'}">$CGI{'referer'}</a> again.
EOM
    }
    elsif ( $clean == 0 ) {                                                                              #mf
        $go = <<EOM;                                                                                     #UK
<form>
<input class="BUTTON" type="BUTTON" value="go back" onClick="history.go(-2);">
</form>
<p>and reload that page.
EOM
    }

    if ( $CGI{'logout'} ne '' ) {                                                                        # show different text #mf
        my $app = $APPNAME || 'web site';
        print <<EOM;                                                                                     #UK
<p>You have logged out of the $app. $tsay.
EOM
    }
    else {
        $h = &homeURL();
        $x = "<li>Go to the home page of <strong>$WBB{'dbTitle'}</strong>.";                             #UK
        $go .= qq(<a href="$h">$x</a>);
        $show = '';

        if ( $User ne $Group ) {

            # when group is user it is confusing, so we only show when group is different     #mf
            $show = "<li>You are part of the '$Group' group.";                                           #UK
        }

        #         temporaly removed because of too much informations and links in the login page and redundancy with Home #xb
        #         $url = &loginRecord();
        #         if ($url) {
        #             $show .= qq(<li>View the information that we have stored in <a href="$url">your personal record</a>.);    #UK
        #         }
        #         else {
        #             $show = '';
        #         }

        print <<EOM;    #UK
<p>$tsay! Login Successful!
<ul>
<li>$go</li>
$show
</ul>
EOM
    }

    if ( $ENV{'HTTP_COOKIE'} eq '' ) {
        $x = <<EOM;     #UK
This application may not work well,
because you disabled the use of cookies.
Please enable them!
<p>
To log in without cookies, please use 
<a href="$hr">URL
based login</a>.
EOM
        wbWarn( "Warning: Cookies disabled in your browser", $x );    #UK
    }

    return;

    print <<EOM;                                                      #UK
<p> You can now go to <a href="$h">home page</a> of this database and use it as $user from $Group.

<p class="WARN">COOKIE based login/logout fails with some versions of Internet Explorer. Try something like:
<br><a href="$hr" target="woda_opener">$hr</a>.
This one opens in a new window!
EOM
}

# reset password for user, step 1

sub cgiPassR1 {

    $did = $CGI{'_id'};

    printHead("Reseting password for $did");    #UK
    recRead($did);

    $show = qq(<a href="$SCRIPT/Show?_id=$did">$did</a>);

    if ( $rec{'_id'} eq '' ) {
        wbError( "Password for $did cannot be reset", "Are you sure record $show exists at all? Try to <a href=\"$SCRIPT/SearchForm\">find it</a>!" )
          ;                                     #UK
        return;
    }
    elsif ( $rec{'_password'} =~ m/^[ugh]:/ ) {
        wbError( "Record $did is not locked with password", "$show is associated to a user, group, hostname or IP number. It cannot be changed" ); #UK
        return;
    }

    $to = '';
    wbSortFields();
    foreach $field (@Fields) {
        if ( $WBF{ $field, 'type' } eq 'EMAIL' && $rec{$field} =~ m/\@/ ) {
            $to = $rec{$field};
            last if $WBF{ $field, 'typePar' } =~ m/OWNER/;
        }
    }

    if ( $to eq '' ) {
        wbError( "Valid email required", "Record $show does not contain a valid email address. Reseting password is not possible" );    #UK
        return;

    }

    $from = $WBB{'managerEmail'};
    $key  = int( rand(100000000) );
    $url  = "$SCRIPTBASE$SCRIPT/PassR2?key=$key";

    $subject = "Password Recovery: " . $WBB{'dbTitle'};                                                                                 #UK

    $headers = <<EOM;
From: $from
Reply-to: $from
To: $to
Subject: $subject
EOM

    $message = <<EOM;                                                                                                                   #UK
Hi,

To really change the password,
please open this URL:

$url

or enter this value:

$key

into the form shown at

$SCRIPTBASE$SCRIPT/PassR2

Automatically,

$WBB{'manager'},
manager.
EOM

    # write file to tmp dir

    open( h, ">$TMPDIR/$key.pwd" );
    binmode(h);
    print h "$did\n" || wbFail("Could not write .pwd file");
    close(h);

    # prepare message

    $message = "$headers\n$message";

    if ( $e = &sendMail("$headers\n$message") ) {
        wbFail( "Error sending mail", "Email could not be sent. Error was $e" );
    }

    print <<EOM;    #UK
<p>Soon you will receive an email to all the email addresses
in record $show.

<p>For security reasons we cannot disclose these addresses,
but perhaps this application lets you see record
$show anyway.

<p>In the email, you will learn a key,
so that you will be able to complete
the password recovery at 
<a href="$SCRIPT/PassR2">$SCRIPTBASE$SCRIPT/PassR2</a>.

<p>If the emails in the record are not yours, we can not help you!

<p>If something goes wrong, try to contact $WBB{manager} at $WBB{managerEmail}.
EOM
}

# reset password for user, step 2

sub cgiPassR2 {

    printHead("Step 2/3 of the password change process");    #UK

    $key = $CGI{'key'};

    print <<EOM;                                             #UK
<form action="$SCRIPT/PassR3" method="post">
Key you got over the email: <input name="key" value="$key"><br>
New password: <input name="pass1" type="password"><br>
New password again: <input name="pass2" type="password"><br>
<input class="BUTTON" type="submit" value="Change password">
</form>
EOM
}

# displays a form to change password

sub cgiPassChange {

    $id = $CGI{'_id'};

    printHead("Changing password for user $id");    #UK

    if ( $id eq '' ) {
        $i = qq(<input name="_id">);
    }
    else {
        $i = qq(<input name="_id" value="$id" type="hidden">$id);
    }

    @p = (
        "Username:",                                #UK
        "Old password:",                            #UK
        "New password:",                            #UK
        "New password again:",                      #UK
        ""
    );

    @i = (
        $i,
        '<input name="oldPass" type="password">',
        '<input name="pass1" type="password">',
        '<input name="pass2" type="password">',
        qq(<input class="BUTTON" type="submit" value="Change password">)
    );

    $x = &formatNameVal( @p, @i );

    print <<EOM;
<form action="$SCRIPT/PassR3" method="post">
$x
</form>
EOM

}

# changes password step 3 ... allows the change either via key or valid current password

sub cgiPassR3 {

    printHead("Step 3/3 of the password change process");    #UK

    $key = $CGI{'key'};
    $p1  = $CGI{'pass1'};
    $p2  = $CGI{'pass2'};

    if ( $p1 ne $p2 ) {
        wbError( "Passwords do not match", "The two new passwords should be the same!" );    #UK
    }

    if ($key) {
        open( h, "$TMPDIR/$key.pwd" ) || wbError("Bad key");
        $did = <h>;
        close(h);
        $did =~ s/\n//g;
    }
    else {
        $did     = $CGI{'_id'};
        $oldpass = $CGI{'oldpass'};
    }

    recRead($did);
    if ( $rec{'_id'} eq '' ) {
        wbFail( "Bad id", "Record with id='$did' does not exist" );
    }

    if ( !$CGI{'key'} ) {
        if ( &wbTestPass( $CGI{'oldPass'}, $rec{'_password'} ) ) {

            #ok
        }
        else {
            wbError( "Bad password", "Old password is not a valid current password for user $did." );    #UK
        }
    }

    if ( $rec{'_id'} eq '' ) {
        wbError( "Cannot read record $did", "Looks like it was deleted in the meantime" );
    }
    else {
        $rec{'_at'}       = time;
        $rec{'_password'} = &wbCryptPasswd($p1);
        recWrite();
        print <<EOM;
<p>Congratulations!
<p>The password for record <a href="$SCRIPT/Show?_id=$did">$did</a>
was changed. You can now use this password to access the record.
EOM
        unlink "$TMPDIR/$key.pwd";
    }
}

# returns a nicely formatted %hash

sub formatHash {

    local ( $name, $quote, @hash ) = @_;
    local ( $out, $n, $v, %x );

    $out .= <<EOM;
<table>
<tr class="DATAH" >
<th COLSPAN=2>$name</th>
</tr>
EOM

    %x = @hash;
    foreach $n ( sort( keys(%x) ) ) {
        $v = $x{$n};
        $v = &htQuote($v) if $quote;
        $out .= <<EOM;
<tr class="DATA1" >
<td align="right">$n</td>
<td align="left">&nbsp;$v</td>
</tr>
EOM
    }

    $out .= <<EOM;
</table>
EOM

    undef %x;
    return $out;
}

# Useful Data for Debugging and Troubleshooting

sub cgiAdmDebug {

    printHead("Useful Data for Debuging and Troubleshooting");

    #	print "<h3>CGI</h3>";

    print &formatHash( "CGI", 1, %CGI );
    print "<p>";
    print &formatHash( "ENV", 1, %ENV );
    print "<p>";
    print &formatHash( "COOKIE", 1, %COOKIE );
    print "<p>";
    print &formatHash( "REQUIRED FILES", 1, %INC );
    print "<p>";

    for $var ( 'WodaFile', 'WodaLanguage', 'WodaLibDir', 'WodaFileDir', 'SCRIPTFILE', 'SCRIPTDIR', 'PERLV', '^X', 'DATADICT' ) {
        $y{ '$' . $var } = eval "\$$var";
    }

    print &formatHash( "WHO AM I", 1, %y );

    print "<br>---END OF DEBUG DATA--<hr>\n";

}

# ask user to verify before deleting record

sub cgiDeleteWarn {

    $id = $CGI{'_id'};
    printHead("About to delete record $id");    #UK

    $d = "Delete";                                                #UK
    $c = "Cancel";                                                #UK
    $t = "Please confirm that you want to delete record $id.";    #UK

    wbWarn( "Are you sure?", <<EOM);                              #UK
$t
<form action="$SCRIPT/DeleteIt">
<input name="_id" value="$id" type="hidden">
<input class="BUTTON" type="submit" value="$d">
<input class="BUTTON" type=BUTTON value="$c" onClick="history.go(-1);">
</form>
EOM
}

# display form to delete record

sub cgiDelete {

    $id = $CGI{'_id'};
    recRead($id);

    if ( &isOwner() ) {

        cgiDeleteWarn();

    }
    elsif ( $x = &wbTestPass( '//11//', $rec{'_password'} ) ) {

        # no need to type in a password ...

        cgiDeleteWarn();

    }
    else {
        printHead("Delete record $CGI{'_id'}");    #UK

        ( $p, $a ) = ( "Password you assigned to this record", "Delete" );    #UK

        print &formatWindow( "$p", <<EOM);
<form action="$SCRIPT/DeleteIt" method="post">
<p>
<input type="password" size="10" name="_accessPassword">
<br>
<input type="hidden" name="_id" value="$CGI{'_id'}">
<p>
<input class="BUTTON" type="submit" value="$a!">
</form>
EOM
    }
}

# Delete a record

sub cgiDeleteIt {

    printHead("Delete results");    #UK

    $id = $CGI{'_id'};
    if ( $id eq '' ) {
        wbError( "Error: Blank ID", "Records with blank ID cannot be deleted.<br>ID may be blank, because the {key} setting is corrupt." );
    }

    recRead($id);
    if ( $rec{'_id'} eq '' ) {
        wbError( "Error: Record does not exist", "Record $id does not exist.<br>Perhaps it has already been deleted." );
    }

    if ( !&wbTestPass( $CGI{'_accessPassword'}, $rec{'_password'} ) ) {
        wbBadPass($id);
        return;
    }

    # related files (.bak into original)

    $err = &recDelete($id);
    if ($err) {
        print '<h3 class="ERROR">Delete failed</h3>';    #UK
        print "$err\n";
        return;
    }
    else {
        print "<h3>Record deleted</h3>";                 #UK
        wbDatabaseChanged();
    }
}

#			print <<EOM ; # U K !!!
#<h3>Thank you for cleaning $WBB{'dbTitle'}.</h3>
#<h3>Undo ?</h3>
#
#To undo the deletion just made, press below:
#<form action="$SCRIPT/Undo">
#<input name=file value="$CGI{'to'}" type="hidden">
#<center><input class="BUTTON" type="submit" value="Undelete!"</center>
#</form>
#EOM
#
#	}
#}

# print a deny message

sub printDeny {

    local ($n) = @_;

    $title = "Access denied";    #UK
    $t     = '';

    if ($userAllowFailed) {

        $message = <<EOM;        #UK
As <b>$User</b> you are not allowed to
use this database. Try to access it as 
'guest'.
EOM

    }
    else {

        if ( $n == 10 ) {
            $p = "<br>with parameter <b>$DeniedPar</b>";    #UK
        }

        if (   &wbAllowed( 'friend', $Page )
            && &wbAllowed( $Group, 'RegisterForm' ) )
        {
            $t = qq(<li><a href="$SCRIPT/RegisterForm">register</a> as a friend.</li>);    #UK
        }

        # always show login box

        if (
               1
            || ( !&wbDenied( 'user', $Page ) )
            || ( $WBB{'userTableGroup'}
                && !&wbDenied( $WBB{'userTableGroup'}, $Page ) )
          )
        {
            $t .= qq(<li><a href="$SCRIPT/LoginForm">login</a> with your username here\n);    #UK
            $t .= &formatLoginBox('');
            $t .= '<br clear="all"></li>';
        }

        if ( !&wbDenied( 'guest', $Page ) ) {
            $t .= qq(<li><a href="$SCRIPT/Login?logout=1">logout</a>.</li>);                  #UK
        }

        if ($t) {
            $t =
              'To be granted access to that page you can select one of the following options:<ol style="list-style-type: none">' . $t . '</ol>';   #UK
        }
        else {

            # never happens
        }

        $message .= <<EOM;                                                                                                                         #UK
<div>Users from group <b>$Group</b> are not allowed to request page <b>$Page</b> $p.
<br>$t
... or ...</div>
EOM

    }

    $Err = 'NA';
    wbError( $title, $message );
}

# values on URL are appended to the record's values

sub cgiAppend {

    # see cgiEditForm for changes

    cgiEdit();

}

# Edit a record

sub cgiEdit {

    #ziga	$id = &wbFixId($CGI{'_id'});

    $id = $CGI{'_id'};

    recRead($id);

    if ( &isOwner() ) {

        cgiEditForm();

    }
    elsif ( $x = &wbTestPass( '//11//', $rec{'_password'} ) ) {

        # no need to type in a password ...

        cgiEditForm();

    }
    else {
        printHead("Edit record $CGI{'_id'}");    #UK

        ( $p, $a ) = ( "Password you assigned to this record", "Edit ..." );    #UK

        print &formatWindow( "$p", <<EOM);
<form name="main" action="$SCRIPT/EditForm" method="post">
<p>
<input type="password" size="10" name="_accessPassword">
<br>
<input type="hidden" name="_id" value="$CGI{'_id'}">
<p>
<input class="BUTTON" type="submit" value="$a">
</form>
EOM
    }
}

# can the record be edited, deleted ...

sub wbCanModifyRecord {

    local ($pass) = @_;

    return 1 if $Group eq 'admin';
    return 1 if &isOwner();
    return 0 if &isOwned();
    return 1 if &wbTestPass( $pass, $rec{'_password'} );
    return 0;
}

# form for editing record

sub cgiEditForm {

    prepareHtmlArea();

    #ziga	$id = &wbFixId($CGI{'_id'});

    $id = $CGI{'_id'};

    printHead("Edit $id");    #UK
    recRead($id);

    if ( $Page eq 'Append' ) {

        foreach $cgi ( keys(%CGI) ) {
            next if $cgi =~ m/^_/;
            $rec{$cgi} .= $CGI{$cgi};
        }

    }

    if (%rec) {
        $x       = time;
        $oldPass = $rec{'_password'};
        if ( &wbCanModifyRecord( $CGI{'_accessPassword'} ) ) {

            # OK

        }
        else {

            wbBadPass($id);
            return;

        }

    }
    else {
        print '<p class="WARN">Record did not exist</p>';
        undef %rec;

        # empty form will be displayed
        # assumes that rec{'id'} is defined ???
        # $rec{'id'} = $CGI{'_id'};
        # $rec{'id'} = &wbFixId($CGI{'_id'});
        # just in case the _id is rec{'id'}
    }

    $x     = &wbForm();
    $reset = "Undo edits";    #UK

    if ( $WBB{'verifyEdits'} ) {
        $submit = "Continue to Preview ...";    #UK
    }
    else {
        $submit = "Save into the database";     #UK
    }

    $enctype = "";
    $enctype = 'ENCTYPE="multipart/form-data"' if $HasAttachmentField;

    print <<EOM;
<form class="FORM" name="main" $enctype action="$SCRIPT/Changed" method="post" target="_self">
$x
<p>
<input type="hidden" name="_id" value="$CGI{'_id'}">
<input type="hidden" name="_at" value="$rec{'_at'}">
<input type="hidden" name="_accessPassword" value="$CGI{'_accessPassword'}">
<input class="BUTTON" type="submit" value="$submit">
<input class="BUTTON" type="reset" value="$reset">
</form>
EOM
    print &evalMaybe( $WBB{'formFoot'} );
}

# when displaying records in a loop, go to prev or next record

sub cgiGo {

    $_id    = $CGI{'_id'};
    $to     = $CGI{'to'};
    $sort   = $CGI{'sort'};
    $search = $CGI{'search'};

    if ( $sort eq '' ) {
        wbFail("Undefined sort");
    }

    $cfile = "$search-$sort";
    $cfile = "$DATADIR/_qry/" . &htEscape($cfile) . ".go";
    $cfile =~ s/%//g;
    if ( &wbExpired($cfile) ) {
        ( $allHits, *ids, *fields ) = &recSelect( '', $search, $sort, RAW, '_id' );
        open( h, ">$cfile" ) || wbFail($cfile);
        binmode(h);
        print h "_id\n";
        print h join( "\n", @ids );
        print h "\n";
        close(h);
    }
    else {
        open( h, $cfile );
        $x   = <h>;
        @ids = <h>;
        close(h);
        foreach $_ (@ids) { chop }
    }

    if ( $to eq 'first' ) {
        $i = 0;
    }
    elsif ( $to eq 'last' ) {
        $i = $#ids;
    }
    else {
        $i = 0;
        foreach $id (@ids) {
            last if $id eq $_id;
            $i++;
        }

        $i++ if $to eq 'next';
        $i-- if $to eq 'previous';

    }

    if ( $i < 0 ) {
        $i = $#ids;
    }
    elsif ( $i > $#ids ) {
        $i = 0;
    }

    $_id = $ids[$i];
    $CGI{'_id'} = $_id;

    wbDo("cgiShow");

}

# tool to assist locating new records

sub cgiNewForm {

    printHead("News Form");    #UK

    $submit   = "   L i s t   ";                                                                                                      #UK
    $newsin   = "news in the last";                                                                                                   #UK
    $unit     = "days.";                                                                                                              #UK
    $newsince = "news since my last listing of the news";                                                                             #UK
    $nocookie = "On their next visit, users of browsers that support cookies also see option `What's new since my last visit?`\n";    #UK

    if ( $COOKIE{NEWS} ) {

        $x = &formatYMDHMS( $COOKIE{NEWS} );
        $newsince .= " <b>$x</b>.";

        print <<EOM;
<form class="FORM" method="get" action="$SCRIPT/Search">
<input class="BUTTON" type="submit" value="$submit">
$newsince
<input name="since" value="$COOKIE{NEWS}" type="hidden">
<input type="hidden" name="sort" value="TIME">
<input type="hidden" name="cookie" value="NEWS">
</form>
EOM

    }
    else {
        print "<p><b>$nocookie</b><p>";
    }

    print <<EOM;
<form class="FORM" method="get" action="$SCRIPT/Search">
<input class="BUTTON" type="submit" value="$submit">
$newsin
<input name="days" size="3" value="7">
<input type="hidden" name="sort" value="TIME">
<input type="hidden" name="cookie" value="NEWS">
$unit
</form>
EOM

}

# print existing parameters for CGI & ENV

sub cgiParam {

    printHead("Cgi parameters");

    print "<pre>\n";
    foreach $x ( keys(%CGI) ) {
        print "<b>$x</b>=$CGI{$x}\n";
    }

    foreach $x ( keys(%ENV) ) {
        print "<b>$x</b>=$ENV{$x}\n";
    }

}

# like search but returns results as plain text

sub cgiQry {

    print "Content-type: text/plain\n\n";
    wbDo( "cgiSearch", 'HF' );
    wodaExit();
}

# used by command line to get related data
# uses mySearch if it is defined

sub cgiXsearch {

    print "Content-type: text/plain\n\n";
    wbDo( "cgiSearch", 'HF' );

    # cgiSearch(HF) exits woda!
}

# print topmost search frame

sub printSearchFrames {

    $how = 'COLS' if $CGI{'frames'} =~ m/^V(.*)$/;
    $how = 'ROWS' if $CGI{'frames'} =~ m/^H(.*)$/;
    $proc = $1 || 50;

    $qs = $ENV{'QUERY_STRING'};
    $qs =~ s/frames\=.//;

    print $HTTP200OK . "Content-type: text/html\n\n";
    print <<EOM;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">
<head>
<title>Search Results in Frames</title>
</head>
<frameset $how="$proc%,*">
<frame src="$SCRIPT/SearchToFrameL?$qs" name="woda_list">
<frame src="$ICONURL/poweredby.gif" name="woda_detail">
</frameset>
<noframes>
EOM

    print <<EOM;    #UK
Your browser should support frames to use this type of search.
EOM

    print "</noframes>\n";
    wodaExit();
}

# store search

sub wbRememberSearch {

    if ( &wbAllowed( $Group, 'SearchHistory' ) && $CGI{'search'} ne '' ) {
        wbSearchHistoryAdd( $CGI{'search'} );
    }

}

# generates appropriate header for the particulare type of search

sub printSearchHttpHeader {

    local ( $format, $yh, $yf ) = @_;
    local ( $html, $x );

    if ( $format =~ m/^(\d+)$/ && $WBB{"formatHttpHead$1"} ne '' ) {
        $x = $WBB{"formatHttpHead$1"};
        $x =~ s/\n*$//g;    # remove trailing newlines
        print "$x\n\n" if $yh;
        $html = $yh = $yf = 0;    # no more head or foot stuff for sure
    }
    elsif ( $format =~ m/^CSV/ ) {
        print "Content-type: text/plain\n\n" if $yh;
        $html = $yh = $yf = 0;    # no more head of foot stuff for sure
    }
    elsif ( $format =~ m/^TAB$/ ) {
        print "Content-type: text/plain\n\n" if $yh;
        $html = $yh = $yf = 0;    # no more head of foot stuff for sure
    }
    elsif ( $format =~ m/^RAW$/ ) {
        print "Content-type: text/plain\n\n" if $yh;
        $html = $yh = $yf = 0;    # no more head of foot stuff for sure
    }
    elsif ( $format =~ m/^XML/ ) {
        printXmlHead() if $yh;
        $html = $yh = $yf = 0;    # no more head of foot stuff for sure
    }
    elsif ( $format =~ m/^RSS/ ) {
        print "Content-type: text/xml\n\n" if $yh;
        $html = $yh = $yf = 0;    # no more head of foot stuff for sure
    }
    else {
        $html = 1;
    }

    return ( $html, $yh, $yf );

}

# USES:
#	CGI{'noform'}	... if not blank, no form will be printed
#	 CGI{'search'}	... what
#	 CGI{'sort'}	... DEFAULT, TIME, sort1, sort2 ...
#	 CGI{'format'}	... DEFAULT, LONG, format1, format2 ...
#	CGI{'first'}	... 0
#	CGI{'max'}	... 20
#	CGI{'days'}	... younger than x days
#	CGI{'fields'}	... space or comma separated list
#	CGI{'since'}	... silently changed to {_at} > since
#	CGI{'efields'}
# Parameters:
#	 how = HF or H or F or empty
#	 H means not HEAD, F not FOOT
# searches for records

sub cgiSearch {

    return unless &saveFormToolbar();

    local ($how) = @_;

    $yh = $yf = 0;
    $yh = 1 unless $how =~ m/H/;
    $yf = 1 unless $how =~ m/F/;

    $CGI{'search'} =~ tr/\n/ /;    # many search fields on line

    # frames

    if ( $CGI{'frames'} =~ /^H|^V/ ) {
        printSearchFrames();
    }
    if ($InFrame) {
        $MINIHEAD      = 1;
        $TargetDetail  = 'target="woda_detail"';
        $SuffixDetail  = "ToFrameD";
        $TargetDefault = 'target="woda_detail"';
        $SuffixDefault = "ToFrameD";
        $TargetParent  = 'target="_parent"';
        $SuffixSelf    = "ToFrameL";
        $TargetSelf    = 'target="_self"';
    }

    $time = time;

    if ( $CGI{'search'} =~ m/%20/ ) {
        $CGI{'search'} = &htUnescape( $CGI{'search'} );
    }

    if ( $CGI{'cookie'} eq 'NEWS' ) {
        $cookie = &htFormatCookie( 'NEWS', $time );
    }
    else {
        $cookie = "";
    }

    ( $isInHtmlFormat, $yh, $yf ) = &printSearchHttpHeader( $CGI{'format'}, $yh, $yf );

    $criteria = '';
    $op       = $CGI{'grouping'};
    for ( $i = 1 ; $i <= $nCriteria ; $i++ ) {
        $f = $CGI{"f:$i"};
        next if $f eq '';
        $e = $CGI{"e:$i"};
        $v = $CGI{"v:$i"};

        $e =~ s/x/$v/;
        $criteria .= " {$f} $e $op";
    }

    $criteria =~ m/$op$/;
    $criteria = $`;

    $CGI{'search'} .= $criteria;

    # remember the search expressions

    wbRememberSearch();

    if ($yh) {
        local ( $qs, $tt );
        $qs = $ENV{'QUERY_STRING'};
        $qs =~ s/&/&amp;/g;
        $qs =~ s/format=[^&]*//;
        $qs .= '&amp;format=RSS';
        $qs =~ s/&amp;&amp;/&amp;/;
        $MetaFields = <<EOM;
<link rel="alternate" title="This search as RSS feed" href="$SCRIPTBASE$SCRIPT/Search?$qs" type="application/rss+xml">
EOM
        printHead( "Search Results", $cookie );    #UK
        $MetaFields = '';
        print "<div class=SEARCHTITLES>\n";
    }

    # --- what fields to be shown, what fields to be edited

    @fields  = split( /\s|,/, $CGI{'fields'} );
    @efields = split( /\s|,/, $CGI{'efields'} );

    # --- which are editable

    foreach $f (@efields) {
        $f{$f} = '?';    # edit ... override display
    }

    # --- fields in given order ... if on command line order must be obeyed

    $fields = '';
    foreach $f (@fields) {
        if ( $f{$f} eq '?' ) {
            $fields .= "?$f,";
            delete $f{$f};
        }
        else {
            $fields .= "$f,";
        }
    }

    # --- any remaining fields in %f that were not set as visible

    foreach $f ( keys(%f) ) {
        $fields .= "?$f," if $f ne '_delete';
        $fields = "_delete,$fields" if $f eq '_delete';
    }

    chop($fields);
    $Log .= "Fields=$fields\n";
    $CGI{'fields'} = $fields;    # needed in FormFieldItems
    undef $f;

    # $fields now stores all fields to show, editable prepended with ?

    if (   $yh
        && $CGI{'search'}
        && $fields !~ /\b_delete\b/
        && $CGI{'noform'} eq '' )
    {
        $x = &wbSearchForm( $CGI{'search'} );
        print $x;
    }

    $CGI{'first'} = 0                if !$CGI{'first'};
    $CGI{'max'}   = $WBB{'dfltHits'} if !$CGI{'max'};
    $CGI{'max'}   = $WBB{'maxHits'}  if $CGI{'max'} > $WBB{'maxHits'};

    # --- default format

    $CGI{'format'} = 'DEFAULT' unless $CGI{'format'};

    # --- if not defined, just cram the fields

    if ( $CGI{'format'} eq 'DEFAULT' && !$WBB{'format;DEFAULT'} ) {

        wbSortFields();

        $f = '';
        $i = 0;
        foreach $x (@PFields) {
            next if $x =~ m/^_/;
            next if $WBF{ $x, 'type' } =~ m/BREAK|SUBMIT/;
            if ( $WBF{ $x, 'type' } !~ /AREA/ ) {
                $f .= "\$rec{'$x'} ;";
                $i++;
            }
            last if $i == 5;
        }
        $f = "\"$f\"";
        $WBB{'format;DEFAULT'} = $f;
    }

    $search = $CGI{'search'};
    $sort = join( "\n", $CGI{'sort'}, $CGI{'sort1'}, $CGI{'sort2'}, $CGI{'sort3'} );
    $sort =~ s/\n*$//g;

    $then = 0;
    $then = time - 24 * 3600 * $CGI{'days'} if $CGI{'days'};
    $then = $CGI{'since'} if $CGI{'since'};

    $search .= " {_at} > $then" if $then > 0;

    $niceThen = '';
    if ($then) {
        $niceThen = "<p>Changes since " . &formatYMDHMS($then);    #UK
    }

    # do the search

    ( $allHits, *list, *fields ) = &recSelect( '', $search, $sort, $CGI{'format'}, $fields, $CGI{'first'}, $CGI{'max'} );

    #	print "xxx";
    $dataHits = $#list + 1;

    # what did we get back

    $n = $CGI{'first'} + 1;
    $m = $CGI{'first'} + $dataHits;

    #print "must include: " . join(',',@aWords) . "<br>\n" if @aWords;
    #print "must not include: " . join(',',@nWords) . "<br>\n" if @nWords;
    #print "may include: " . join(',',@sWords) . "<br>\n" if @sWords;

    print $niceThen if $yh;
    print "<p>" . &filterText() if $yh;

    if ($dataHits) {
        if ($yh) {
            print "<p>Hits $n to $m of $allHits</p>";    #UK
        }
    }
    else {
        if ($yh) {
            print '<p>No hits.</p>', "\n", '</div><!-- SEARCHTITLES -->';    #UK
            wbHelpSearch();
            return;
        }
        else {
            if ( $CGI{'format'} =~ m/XML|RSS/ ) {

                # need output in any case
                formatFoundSeparators( $CGI{'format'} );
                print $t0;
                print $t3;
            }
            wbLog();
            wodaExit() unless $yf;
        }
    }

    $more     = 0;
    $newFirst = $CGI{'first'} + $CGI{'max'};
    undef @first;

    if ( $allHits != $dataHits ) {
        $more = 1 if $allHits > $CGI{'first'} + $CGI{'max'};
        $j = 0;
        for ( $i = 0 ; $i < $allHits ; $i += $CGI{'max'} ) {
            $first[$j] = $i;
            $j++;
        }
    }

    # reformat links

    if ( $dataHits > 0 && $yh && !$InFrame ) {    # disabled

        print '<div class="SEARCHREFORMAT">', "\n";
        print "<p>Reformat results as: \n";       #UK

        @altTexts = (
            "short",                              #UK
            "short into frame",                   #UK
            "detailed",                           #UK
            "detailed into frame"                 #UK
        );

        @formats = ( "DEFAULT", "DEFAULT", "LONG", "LONG" );
        @frames  = ( "NONE",    "V33",     "NONE", "V" );

        $qs = $ENV{'QUERY_STRING'};
        $qs =~ s/\&/&amp;/g;

        $qs =~ s/\&amp;format=\w+//;              # erase format info
        $qs =~ s/\&amp;frames=\w+//;              # erase frames info
        $i = 0;
        foreach $x (@altTexts) {
            $nqs = $qs . "&amp;format=$formats[$i]&amp;frames=$frames[$i]";

            print <<EOM;
<a class="BUTTON" href="$SCRIPT/Search?$nqs" $TargetParent> $x </a>&nbsp;
EOM
            $i++;
        }

        print "</p></div><!-- SEARCHREFORMAT -->\n";    # SEARCHREFORMAT

    }

    print "</div><!-- SEARCHTITLES -->\n" if $yh;       # SEARCHTITLES

    # prepare some output parameters

    $x    = $CGI{'first'} + 1;
    $cgif = $CGI{'format'};
    $head = &formatFoundHeader( $cgif, *fields );
    formatFoundSeparators( $cgif, $x );

    $hasBasketForm = !( $PrintingBasket || $HTMLFILE || !$WBB{'basketName'} );

    if ( $#efields >= 0 ) {
        $addAction = "AdmStoreMany";
        $addMethod = "POST";
        $addTarget = "";
        $addButton = "Store changes to found records";    #UK
        if ( $fields =~ m/^_delete/ ) {
            $instr = "Check the leftmost checkbox of each record to mark it for deletion";    #UK
            $js2   = &formatCheckAllJavaScript('_delete');
            if ($js2) {
                $js2 = "<p>$js2 Click here to check all/none for deletion.<br>";              #UK
            }
        }

    }
    elsif ($hasBasketForm) {
        $addAction = "BasketAdd";
        $addMethod = "GET";
        $addTarget = <<EOM;
target="woda_basket" 
onSubmit='
	a=window.open("","woda_basket","scrollbars=yes,resizable=yes,menubar=no,width=600,height=300");
	if (document.images) self.name = "woda_opener";
	a.focus();
'
EOM
        $addButton = "Add checked items to $WBB{'basketName'}";    #UK
        $js2       = '';

    }
    else {
        $addAction = '';

    }

    $js1 = &formatFormJavaScript();

    if ( $dataHits && $isInHtmlFormat ) {
        print <<EOM if $addAction;
<p>
$js1
$instr
<form name="main" action="$SCRIPT/$addAction" $addTarget METHOD=$addMethod>
EOM
        print '<div class="RECORDS">', $t0, $head;
        foreach $_ (@list) {
            print "$t1$_$t2";
        }
        print $t3;
        print '</div>';
        print <<EOM if $addAction;
$js2 <input class="BUTTON" type="submit" value="$addButton">
</form>
EOM

    }
    else {
        print $t0, $head;    # 4.190
        foreach $_ (@list) {
            print "$t1$_$t2";
        }
        print $t3;           # 4.190
    }

    return unless $yf;       # if footer is required

    @Toolbar = grep( /FormSave/, @Toolbar );

    #	undef @Toolbar;		# zakaj? subforma?
    #	do saveFormToolbar();	# search filed gets spoiled!

    if (@first) {

        print "<p>For more results click below:";    #UK
        $qs = $ENV{'QUERY_STRING'};
        unless ($qs) {                               # $qs may be empty because cgiHome sets CGI without passing data via URL
            $qs .= "&search=$CGI{'search'}" if $CGI{'search'};
            $qs .= "&sort=$CGI{'sort'}"     if $CGI{'sort'};
            $qs .= "&max=$CGI{'max'}"       if $CGI{'max'};
        }
        $qs =~ s/&/&amp;/g;
        $qs =~ s/\&amp;first=\d+//;

        $tale = -1;
        for ( $j = 0 ; $j <= $#first ; $j++ ) {
            if ( $first[$j] == $CGI{'first'} ) {
                $tale = $j;
                last;
            }
        }

        for ( $j = 0 ; $j <= $#first ; $j++ ) {

            $newFirst = $first[$j];

            if (   $j != 0
                && $j != $#first
                && $j != $tale
                && ( $tale - $j ) * ( $tale - $j ) > 25 )
            {
                if ( !$pikce ) {
                    push( @Toolbar, <<EOM);
...
EOM
                    $pikce = 1;
                }
                next;
            }

            $pikce = 0;
            $qsn   = $qs . "&amp;first=$newFirst";
            $pg    = "show page $j";                 #UK
            $pg0   = "this is page $j";              #UK

            if ( $j == $tale ) {
                push( @Toolbar, <<EOM);
<img $ICONPAR src="$ICONURL/up.gif" title="$pg0" alt="$pg0">
EOM
            }
            elsif ( $j == 0 ) {
                push( @Toolbar, <<EOM);
<a href="$SCRIPT/Search$SuffixSelf?$qsn" $TargetSelf><img  $ICONPAR  src="$ICONURL/first.gif" alt="$pg" title="$pg"></a>
EOM
            }
            elsif ( $j == $tale - 1 ) {
                push( @Toolbar, <<EOM);
<a href="$SCRIPT/Search$SuffixSelf?$qsn" $TargetSelf><img  $ICONPAR  src="$ICONURL/back.gif" alt="$pg" title="$pg"></a>
EOM
            }
            elsif ( $j == $tale + 1 ) {
                push( @Toolbar, <<EOM);
<a href="$SCRIPT/Search$SuffixSelf?$qsn" $TargetSelf><img  $ICONPAR  src="$ICONURL/forward.gif" alt="$pg" title="$pg"></a>
EOM

            }
            elsif ( $j == $#first ) {
                push( @Toolbar, <<EOM);
<a href="$SCRIPT/Search$SuffixSelf?$qsn" $TargetSelf><img  $ICONPAR  src="$ICONURL/last.gif" alt="$pg" title="$pg"></a>
EOM

            }
            elsif ( $j < $tale ) {
                push( @Toolbar, <<EOM);
<a href="$SCRIPT/Search$SuffixSelf?$qsn" $TargetSelf><img  $ICONPAR  src="$ICONURL/backg.gif" alt="$pg" title="$pg"></a>
EOM
            }
            elsif ( $j > $tale ) {
                push( @Toolbar, <<EOM);
<a href="$SCRIPT/Search$SuffixSelf?$qsn" $TargetSelf><img  $ICONPAR  src="$ICONURL/forwardg.gif" alt="$pg" title="$pg"></a>
EOM
            }
        }
    }
    else {
        if ($dataHits) {
            print "<p>No more hits.";    #UK
        }
    }

    # --- add stuff to toolbar

    local ( $s, $so, $fo );

    $s  = &htEscape($search);
    $so = &htEscape($sort);
    $fo = &htEscape( $CGI{'format'} );

    if ( &wbAllowed( $Group, 'AdmManyEditForm' ) && $#editable < 0 ) {

        $text = "EDIT/DELETE all these records.";    #UK

        push( @Toolbar, <<EOM);
<a href="$SCRIPT/AdmManyEditForm?$qs" $TargetParent>
<img $ICONPAR align="top" src="$ICONURL/editn.gif" alt="$text" title="$text"></a>
EOM
    }

    if ( &wbAllowed( $Group, 'AdmManySetForm' ) ) {

        $text = "Set one field in these records to one value.";    #UK

        push( @Toolbar, <<EOM);
<a href="$SCRIPT/AdmManySetForm?search=$s" $TargetParent>
<img $ICONPAR align="top" src="$ICONURL/editnf.gif" alt="$text" title="$text"></a>
EOM
    }

    if ( &wbAllowed( $Group, 'AdmMailForm' ) && $HasEmailField ) {

        $text = "Send serial email to them.";                      #UK

        push( @Toolbar, <<EOM);
<a href="$SCRIPT/AdmMailForm?search=$s" $TargetParent>
<img $ICONPAR align="top" src="$ICONURL/serialmail.gif" alt="$text" title="$text"></a>
EOM
    }

    if ( &wbAllowed( $Group, 'AgentAddForm' ) && !$agentFile ) {

        $text     = "I want to subscribe to searches like this.";    #UK
        $Formdata = &hashExport(*CGI);
        $Formdata = &htEscape($x);

        push( @Toolbar, <<EOM);
<a href="$SCRIPT/AgentAddForm?FORM=AgentAddForm&amp;formdata=$Formdata" $TargetParent>
<img $ICONPAR src="$ICONURL/angie.gif" alt="$text" title="$text"></a>
EOM
    }

}

# format separators for search results
# sets t0::above all
# t1::?
# t2::?
# t3::below all

sub formatFoundSeparators {

    local ( $format, $n ) = @_;

    if ( $format eq 'CSV' || $format eq 'CSVS' ) {
        $t0 = "";
        $t1 = "";
        $t2 = "\n";
        $t3 = "";
    }
    elsif ( $format eq 'TAB' ) {
        $t0 = "<XMP>\n";
        $t1 = "";
        $t2 = "\n";
        $t3 = "</XMP>\n";
    }
    elsif ( $format eq 'RAW' ) {
        $t0 = "";
        $t1 = "";
        $t2 = "\n";
        $t3 = "";
    }
    elsif ( $format eq 'XML' || $format eq 'XMLE' ) {
        $t0 = "<!DOCTYPE records>\n<records>\n";
        $t1 = "";
        $t2 = "";
        $t3 = "</records>\n";
    }
    elsif ( $format eq 'RSS' ) {
        $x  = &htUntag("$WBB{'apTitle'} - $WBB{'dbTitle'}");
        $x  = &htQuote($x);
        $y  = &htQuote( $WBB{'about'} );
        $t0 = <<EOM;
<rss version="0.91">
<channel>
<title>$x</title>
<link>$SCRIPTBASE$SCRIPT</link>
<description>$y</description>
<language>en-us</language>
EOM
        $t1 = "";
        $t2 = "";
        $t3 = <<EOM;
</channel>
</rss>
EOM
    }
    elsif ( $format eq 'TABLE' ) {
        $t0 = '<table class="DATATABLE" cellspacing="1">"\n"';
        $t1 = "";
        $t2 = "";
        $t3 = "</table><br>\n";
    }
    elsif ( $format eq 'LONG' || $format eq 'NAMEVAL' ) {
        $t0 = "";
        $t1 = "";
        $t2 = "<hr>\n";
        $t3 = "";
    }
    elsif ( $format eq 'DEFAULT' || $format =~ m/^\d$/ ) {
        $t0 = &evalMaybe( $WBB{"formatHead$format"} );
        $t1 = '';
        $t2 = '';
        $t3 = &evalMaybe( $WBB{"formatFoot$format"} );
    }
    elsif ( $format eq 'PARAGRAPH' ) {
        $t0 = '';
        $t1 = '';
        $t2 = '';
        $t3 = '<br><br>';
    }
}

# display form for search expression

sub cgiSearchForm {

    printHead("Simple search form");    #UK
    print &wbSearchForm( $CGI{'search'}, 'FORM,PROMPT,HISTORY' );
    if (   &wbAllowed( $Group, 'AdvancedSearchForm' )
        || &wbAllowed( 'friend', 'AdvancedSearchForm' ) )
    {
        print <<EOM;                    #UK
<br><a href="$SCRIPT/AdvancedSearchForm"><b>Advanced search form</b></a>
EOM
    }
    wbHelpSearch();
}

# add search to search history

sub wbSearchHistoryAdd {

    local (@s);

    @s = &wbSearchHistoryGet();

    foreach $s (@s) {
        return if ( $s eq $_[0] );
    }

    unshift( @s, $_[0] );

    if ( $#s > 9 ) {
        @s = $s[ 0 - 9 ];
    }

    return if $User =~ m/^_anon/;

    open( h, ">$DATADIR/_friend/$User.srch" );
    print h join( "\n", @s );
    print h "\n";
    close(h);
}

# returns array of all previous searches

sub wbSearchHistoryGet {

    local ($s);

    open( h, "$DATADIR/_friend/$User.srch" );
    @s = <h>;
    close(h);

    foreach $s (@s) {
        chop($s);
    }

    return @s;
}

# clears search history

sub wbSearchHistoryClear {

    unlink "$DATADIR/_friend/$User.srch";

}

# clears search history

sub cgiSearchHistoryClear {

    wbSearchHistoryClear();

    printHead("Search History Clear");    #UK

    print formatWindow( "Done", "Your search history has been erased" );    #UK
}

# displays many forms with all old search expressions

sub cgiSearchHistory {

    @s = &wbSearchHistoryGet();

    if ( $#s >= 0 ) {
        $n = $#s + 1;
        printHead("Search in your recent $n searches");                     #UK
        print "<ol>";
        foreach $s (@s) {
            print "<li>";
            print &wbSearchForm( $s, 'FORM' );
        }
        print '</li>';
        print <<EOM;                                                        #UK
<li>
<a href="$SCRIPT/SearchHistoryClear">CLEAR</a>
all of the search history above.
</li>
</ol>
EOM
    }
    else {
        cgiSearchForm();
    }
}

# Search by record ID

sub cgiSearchIDForm {

    printHead("Search by record ID");    #UK

    print <<EOM ;                        #UK
<form class="FORM" action="$SCRIPT/Show">
Show what record:
<br><input name="_id" size="40">
<br><span class="HINT">Enter record key.</span>
</form>
EOM

}

# formats an option list with values v, texts t
# where default value is default

sub wbFormatOptions {

    local ( *v, *t, $default ) = @_;
    local ( $nItems, $i );

    $nItems = $#v;
    $_      = '';
    foreach ( $i = 0 ; $i <= $nItems ; $i++ ) {
        if ( $default ne '' && $default eq $v[$i] ) {
            $_ .= "<option value=\"$v[$i]\" SELECTED>$t[$i]\n";
        }
        else {
            $_ .= "<option value=\"$v[$i]\">$t[$i]\n";
        }
    }

    return $_;
}

# returns the format as pulldown

sub formatFormats {

    local ($selected) = @_;
    local ( @vals, @texts, @v, @t );

    $selected = 'DEFAULT' unless $selected;

    @vals = ( "DEFAULT", "LONG", "PARAGRAPH", "TABLE", "CSV", "CSVS", "TAB", "XML", "XMLE", "RSS" );
    @texts = (
        "default short format",                 #UK
        "default detailed format",              #UK
        "paragraph of all selected fields",     #UK
        "table format - one record per row",    #UK
        "comma delimited format (.csv)",        #UK
        "semicolon delimited format (.csv)",    #UK
        "tab delimited format",                 #UK
        "XML format",                           #UK
        "easy to parse XML",                    #UK
        "RSS 0.91",                             #UK
    );

    if ( $WBB{'detail'} ) {
        push( @vals,  "NAMEVAL" );
        push( @texts, "custom detailed format" );    #UK
    }

    @x = keys %WBB;
    @x = grep ( /^formatName\d+/, @x );
    foreach $_ (@x) {
        $v = $WBB{$_};
        $_ =~ m/^formatName/;
        $text = $';
        $formats .= qq(<option value="$text">$v\n);
        unshift( @vals,  $text );
        unshift( @texts, $v );
    }

    # allowed works with parameters on the CGI

    for ( $vi = 0 ; $vi <= $#vals ; $vi++ ) {
        $v = $vals[$vi];
        if ( &wbAllowed( $Group, 'Search', "format=$v" ) ) {
            push( @v, $v );
            push( @t, $texts[$vi] );
        }
    }

    $formats = &wbFormatOptions( *v, *t, $selected );

    return $formats;
}

# Edit many fields at once

sub cgiAdmManyEditForm {

    return
      unless &saveFormCan('AdmManyEditForm');    # sets $SaveFormButton, sets %CGI

    printHead('Edit many fields at once');       #UK
    wbAdvancedSearchForm('EDIT');
}

# delete many records at once

sub cgiAdmManyDelForm {

    printHead("Deleting many records");

    if ( $CGI{'select'} eq 'basket' ) {

        readBasket();
        $ids = join( "\n", @BasketID );

        wbWarn( "Are you sure?", <<EOM);    #UK
This will delete all records in your basket from the database.
<form action="$SCRIPT/AdmManySet">
<textarea name="_delete">$ids</textarea>
<br>
<input class="BUTTON" type="submit" value="Delete!">
</form>
EOM
    }
    else {
        wbError("Nothing selected");
    }
}

# Change field in many records to one value

sub cgiAdmManySetForm {

    $field = $CGI{'field'};

    if ( $field ne '' ) {
        printHead("Change field $field in many records");

        if ( $field eq 'search' ) {
            wbError( 'Field with this name causes problems', 'Editing not allowed' );
        }

        if ( $CGI{'how'} eq 'input' ) {
            $name  = $field . ":";
            $input = qq(<input name="$field">);
        }
        else {
            $Help = 1;
            wbFormFieldItems($field);
            $name  = "Change $Form{$field,'name'}";
            $input = "$Form{$field,'prompt'}<br>$Form{$field,'input'}<br>$Form{$field,'help'}";
        }

        if ( $CGI{'select'} eq 'search' ) {
            $ps = "Search for records:";
            $fs = &wbSearchForm( $CGI{'search'}, 'HISTORY' );
        }
        elsif ( $CGI{'select'} eq 'records' ) {
            $ps = "Record IDs";
            $fs = qq(Enter one per row:<br><textarea wrap="PHYSICAL" name="_records" rows="10" cols="40">$CGI{'process'}</textarea>);
        }
        else {
            readBasket();
            $ps = "Records in the basket:";
            $fs = join( ', ', @BasketID )
              || '<em class="ERROR">Basket empty!</em>';
        }

        print <<EOM;

<p>This lets you set a field to <em>the same value</em> in many records.</p>
<form class="FORM" name="searchForm" action="$SCRIPT/AdmManySet" method="post">
EOM

        @prompts = ( $ps, "$name", 'Action:' );
        @flds = (
            $fs,
            $input,
            qq(
<input name="_field" type="hidden" value="$field">
<input class="BUTTON" type="submit" name="_action" value="Preview">
<input name="_action" class="BUTTON" type="submit" value="Modify them!">
<input class="BUTTON" type="reset" value="Reset">
)
        );

        print &formatNameVal( @prompts, @flds );
        print "</form>\n";

    }
    else {
        printHead("Change one field in many records");

        wbSortFields();
        $flds = '<select name="field">' . "\n";
        foreach $f (@IFields) {
            next if $f eq '_id';
            $flds .= "<option>$f\n";
        }
        $flds .= "</select>";

        $selects = 'SELECTED' if $CGI{'select'} eq 'search';
        $selectr = 'SELECTED' if $CGI{'select'} eq 'records';
        $selectb = 'SELECTED' if $CGI{'select'} eq 'basket';

        print <<EOM;
<p>Field will be set to the same one value in many records. You need to define:</p>
<form class="FORM" action="$SCRIPT/AdmManySetForm">
<input name="search" type="HIDDEN" value="$CGI{search}">
<table cellspacing="0" width="100%">
<tr>
<td>How to select the records:</td>
<td>
<select name="select">
<option value="search"  $selects >by using a search
<option value="records" $selectr >by entering several record IDs
<option value="basket"  $selectb >by using records from the basket 
</select>
</td>
</tr>
<tr>
<td>Which field to change:</td>
<td>$flds<td>
</tr>
<tr>
<td>How to edit the field:</td>
<td>
<select name="how">
<option value="nice">as in forms
<option value="input">as a plain input field
</select>
</td>
</tr>
<td>&nbsp;</td>
<td>
<input class="BUTTON" type="submit" value="Continue ...">
</td>
</tr>
</table>
</form>
EOM

    }
}

# Change field in many records to one value

sub cgiAdmManySet {

    printHead("Set field $CGI{'_field'} to new value");

    if ( $CGI{'_records'} ) {

        # get rid of LFs introduced by form
        $CGI{'_records'} =~ s/\r//g;
        @recToChange = split( /\n/, $CGI{'_records'} );
    }
    elsif ( defined $CGI{'search'} ) {
        wbRememberSearch();
        ( $allHits, *list, *fields ) = &recSelect( '', $CGI{'search'}, '', 'RAW', '_id' );
        @recToChange = @list;
    }
    else {
        readBasket();
        @recToChange = @BasketID;
    }

    if ( $#recToChange == -1 ) {
        wbError("No records found");
    }

    $fldToChange{ $CGI{'_field'} } = 1;
    foreach $rec (@recToChange) {
        $changeTo{ $rec, $CGI{'_field'} } = $CGI{ $CGI{'_field'} };
    }

    if ( $CGI{'_action'} eq 'Preview' ) {
        admStoreMany(1);
    }
    else {
        admStoreMany(0);
    }
}

# Edit many records

sub cgiAdmManyEdit {

    if ( $CGI{'format'} =~ m/^RAW|^TAB$|^TBL|^CSV|^RSS/ ) {    # why not XML
        wbError("Records cannot be edited in $CGI{'format'} format");
    }

    if ( $CGI{'efields'} eq '' ) {
        wbError("No fields were selected for editing");
    }

    cgiSearch();
}

# store changes to many records

sub cgiAdmStoreMany {

    local ( @recToChange, %fldToChange, %changeTo );
    local ( @datefields, $datefield );

    printHead("Stored changes to records");    #UK

    foreach $key ( keys(%CGI) ) {

        if ( $key =~ m/^\_id=(.+)\_field=(.+)/ ) {
            push( @recToChange, $1 ) if ( $recToChange{$1} eq "" );
            $recToChange{$1} = 1;
            $fldToChange{$2} = 1;
            $changeTo{ $1, $2 } = $CGI{$key};

        }
    }

    # identify date fields
    foreach $fld ( keys(%fldToChange) ) {

        if ( $fld =~ m/(.+)_year$/ ) {
            $datefield = $1;
            $mm        = $datefield . '_month';

            if ( $fldToChange{$datefield} && $fldToChange{$mm} ) {

                # see that there is a X_month and an X also
                # in case someone names fields like this: "next_year"
                push( @datefields, $datefield );
            }
        }
    }

    # merge date fields as y/m/d
    foreach $id ( keys(%recToChange) ) {    # records to change

        foreach $fld (@datefields) {        # are we editing a date field?
            $yy = $fld . '_year';
            $mm = $fld . '_month';

            if ( $changeTo{ $id, $fld } ) {

                # remove the _year and _month elements from %fldToChange and %changeTo
                my $date = "$changeTo{$id,$yy}/$changeTo{$id,$mm}/$changeTo{$id,$fld}";

                $changeTo{ $id, $fld } = $date;
                undef $changeTo{ $id, $yy };
                undef $changeTo{ $id, $mm };
                undef $fldToChange{$yy};
                undef $fldToChange{$mm};
            }
        }
    }

    admStoreMany();

}

# store many records

sub admStoreMany {

    # xxx problem ... not handling if ID of a record would change!

    local ($pretend) = @_;

    print '<p class="WARN">Preview only - nothing was actually written.</p>', "\n"
      if $pretend;

    # needs this set
    # @recToChange ... list of records to be changed
    # %fldToChange ... fields to be changed
    # %changeTo{record,field}

    foreach (@recToChange) {
        $oldID = $_;

        recRead($oldID);
        if ( $rec{'_id'} eq '' ) {
            print qq(<p><hr><em class="ERROR">Record $oldID not found!</em>\n);
            next;
        }
        $skipRec = 1;
        $x       = qq(<p><hr>Changes in record <a href="$SCRIPT/Show?_id=$oldID">$oldID</a>:);
        foreach $key ( keys(%fldToChange) ) {    # each fld to be changed
            if ( $rec{$key} ne $changeTo{ $oldID, $key } ) {

                # escape ???
                $x .= "<br><b>$key</b> from: $rec{$key}\n";
                $x .= "<br><b>$key</b> into: $changeTo{$oldID,$key}\n";
                $rec{$key} = $changeTo{ $oldID, $key };
                $skipRec = 0;                    # yes, there will be a change
            }
        }

        print $x;
        next if $skipRec;

        # assign a temporaly ID
        # this ID will be a random code always

        wbCheckDir($TMPDIR);
        $tmpID = &tmpID();
        recWriteFile("$TMPDIR/$tmpID.rec");

        $newID = $oldID;
        $newID = &evalWBBkey() || $newID;

        if ( $newID ne $oldID ) {
            print qq(<br><em class="ERROR">ERROR: Edits would cause an ID change to $newID.</em>\n);
            next;
        }

        # covert data from CGI variables into %rec
        # also saves attachment files

        #		   do recBackup($oldID);

        recReadFile("$TMPDIR/$tmpID.rec");

        if ( !%rec ) {
            wbError( "Data missing", "The submitted data in $TMPDIR/$tmpID.rec could not be read" );
            return;
        }

        #		do recMoveAttached($tmpID,$oldID) unless $pretend;
        recSetSystem();
        $rec{'_id'} = $oldID;
        recWrite() unless $pretend;
        recRead($oldID);

        if ( !%rec ) {
            wbError("Database Update Failed") unless $pretend;    #UK
        }

        unlink("$TMPDIR/$tmpID.rec");
        print "<br>stored!\n" unless $pretend;
        wbRecordChanged()     unless $pretend;
        wbDatabaseChanged()   unless $pretend;
    }

    print '<hr>';

    $CGI{'_delete'} =~ s/\r//g;
    @todelete = split( /\n/, $CGI{'_delete'} );
    foreach $id (@todelete) {
        recRead($id);    # we may need %rec
        recDelete($id) unless $pretend;
        wbDatabaseChanged();    # here !
        print "<br>$id deleted.\n";
    }

    print qq(<p class="WARN">Preview only - nothing was actually written.</p>\n)
      if $pretend;

}

# calls the advanced search form

sub cgiAdvancedSearchForm {

    return
      unless &saveFormCan('AdvancedSearchForm');    # sets $SaveFormButton, sets %CGI

    printHead('Advanced Search');                   #UK
    wbAdvancedSearchForm('SEARCH');
}

# generate and print the advanced search form

sub wbAdvancedSearchForm {

    local ($how) = @_;                              # EDIT | SEARCH

    # --- sorts

    @vals = ( "DEFAULT", "TIME" );
    @texts = ( "by number of matches", "new items first" );    #UK

    @x = keys(%WBB);
    @x = grep ( /^sortName;/, @x );
    foreach $_ (@x) {
        $n = $WBB{$_};
        ( $s, $text ) = split(/;/);
        push( @vals,  $text );
        push( @texts, $n );
    }

    push( @texts, "by fields selected below (A-Z)" );    #UK
    push( @vals,  "DYNAMICAZ" );

    push( @texts, "by fields selected below (Z-A)" );    #UK
    push( @vals,  "DYNAMICZA" );

    push( @texts, "by a field selected below (0-9)" );    #UK
    push( @vals,  "DYNAMIC09" );

    push( @texts, "by a field selected below (9-0)" );    #UK
    push( @vals,  "DYNAMIC90" );

    $sorts = &wbFormatOptions( *vals, *texts, $CGI{'sort'} );

    # --- formats

    if ( $CGI{'format'} ) {
        $df = $CGI{'format'};
    }
    elsif ( $how =~ m/EDIT/ ) {
        $df = 'TABLE';
    }
    else {
        $df = '';
    }

    $formats = &formatFormats($df);    # OPTION LIST READY

    # --- frames

    @vals = ( "NONE", "V", "V66", "V33", "H" );
    @texts = (
        "no frames",                   #UK
        "vertical",                    #UK
        "vertical - small detail",     #UK
        "vertical - big detail",       #UK
        "horizontal"                   #UK
    );

    $frames = &wbFormatOptions( *vals, *texts, $CGI{'frames'} );

    # what fields to show

    wbSortFields();
    undef @flds;
    foreach $x (@PFields) {
        next if $Hidden{$x};
        next if $x eq '_passswd';
        $ch = '';
        if ( $CGI{'fields'} eq '' ) {
            $ch = 'CHECKED' unless $x =~ m/^_/;
        }
        else {
            $ch = 'CHECKED' if $CGI{'fields'} =~ m/\b$x\b/;
        }
        push( @flds, qq(<input name="fields" value="$x" type="checkbox" $ch>&nbsp;$x \n) );
    }
    push( @flds, &formatCheckAllJavaScript('fields') );
    $flds = &formatColumns( 3, @flds );

    # what fields to edit :

    undef @flds;
    foreach $x (@IFields) {
        $ch = '';
        next if $x =~ m/^_/;
        $ch = '';
        if ( $CGI{'efields'} eq '' ) {

            # unchecked is default
        }
        else {
            $ch = 'CHECKED' if $CGI{'efields'} =~ m/\b$x\b/;
        }
        push( @flds, qq(<input name="efields" value="$x" type="checkbox" $ch>&nbsp;$x \n) );
    }

    push( @flds, &formatCheckAllJavaScript('efields') );
    $eflds = &formatColumns( 3, @flds );

    # the fields expression builder

    $fieldPull = "<option>\n";
    foreach $x (@IFields) {
        next if $Hidden{$x};
        $fieldPull .= "<option value=\"$x\">$x\n";
    }

    # default fields for sort

    $x           = $CGI{"sort1"};
    $fieldPullS1 = $fieldPull;
    $fieldPullS1 =~ s/\"$x\"/$& SELECTED/ if $x ne '';

    $x           = $CGI{"sort2"};
    $fieldPullS2 = $fieldPull;
    $fieldPullS2 =~ s/\"$x\"/$& SELECTED/ if $x ne '';

    $x           = $CGI{"sort3"};
    $fieldPullS3 = $fieldPull;
    $fieldPullS3 =~ s/\"$x\"/$& SELECTED/ if $x ne '';

    $opPull = <<EOM;
<option value="&amp;&amp;">and
<option value="||">or
<option value=")&amp;&amp;">)and
<option value=")&amp;&amp;(">)and(
<option value=")">)
EOM

    $brPull = <<EOM;
<option>
<option>(
EOM

    $grouping = <<EOM;    #UK
<select name="grouping">
<option value="and">satisfy all of the above
<option value="or">satisfy any of the above
</select>
EOM

    $grouping =~ s/\"$CGI{'grouping'}\"/$& SELECTED/;

    $exprPull = <<EOM;    #UK
<option value="">
<option value="=~ m/x/i">contains
<option value="!~ m/x/i">not contains
<option value="=~ m/^x/i">begins with
<option value="=~ m/x\$/i">ends with
<option value="=~ m/x/i">matches regexp.
<option value="!~ m/x/i">does not match regexp.
<option value=' eq ""'>is blank
<option value=' ne ""'>is not blank
<option value='eq "x"'>as string equal to
<option value='ne "x"'>as string not equal to
<option value='gt "x"'>as string greater than
<option value='lt "x"'>as string lesser than
<option value="=~ x">=~
<option value="== x">==
<option value="!= x">!= 
<option value="&gt; x">&gt; 
<option value="&lt; x">&lt;
<option value="&lt;= x">&lt;=
<option value="&gt;= x">&gt;= 
EOM

    $sex = '';
    for ( $i = 1 ; $i <= $nCriteria ; $i++ ) {
        $x = <<EOM;
<select name="f:$i">
$fieldPull
</select>
<select name="e:$i">
$exprPull
</select>
<input name="v:$i" size="15">
<br>
EOM

        # this does not work for the same selection value in both e and f

        $y = $CGI{"f:$i"};
        $x =~ s/\"$y\"/$& SELECTED/;

        $y = $CGI{"e:$i"};
        $x =~ s/\"$y\"/$& SELECTED/;

        $y = $CGI{"e:$i"};
        $x =~ s/\'$y\'/$& SELECTED/;

        $y = $CGI{"v:$i"};
        $x =~ s/\"v:$i\"/$& value=\"$y\"/;

        $x =~ s/SELECTED SELECTED/SELECTED/g;

        $sex .= $x;
    }

    $sex .= $grouping;

    @p = (
        "Search for",                                                                                 #UK
        "Search!",                                                                                    #UK
        "Enter space separated list of words.",                                                       #UK
        "Sorting",                                                                                    #UK
        "Formatting",                                                                                 #UK
        "Frames",                                                                                     #UK
        "Results per page",                                                                           #UK
        "Where applicable print the following fields",                                                #UK
        "Additional search criteria<br>They are applied on records satisfying the list of words.",    #UK
        "Fields to edit",                                                                             #UK
        "Password to unlock fields to edit",                                                          #UK
        "Enable deleting records",                                                                    #UK
        "Max age of data in days",                                                                    #UK
        "Paint&nbsp;found&nbsp;words"                                                                 #UK
    );

    $xby  = 'by';                                                                                     #UK
    $xand = 'then ';                                                                                  #UK

    if ( $how eq 'EDIT' ) {
        $action = "AdmManyEdit";
        $p[0]   = "Find and edit!";                                                                   #UK
        $p[1]   = $p[0];
        $pass   = qq(<input type="password" size="20" name="_password");
    }
    else {
        $action = "Search";
    }

    $s = $CGI{'search'};

    if ( $how =~ m/EDIT/ ) {
        $paint = '';
    }
    else {
        $paint = 'CHECKED';
    }

    $js = &formatFormJavaScript('searchForm');
    $sf = &wbSearchForm( $CGI{'search'}, 'HISTORY' );

    $max = $CGI{'max'} || 20;

    print <<EOM;
$SaveFormForm
$js
<form class="FORM" name="searchForm" action="$SCRIPT/$action">
<table cellspacing="0"><tr>
<td align="right" valign="top" width="30%">$p[0]:</td>
<td>$sf
<input class="BUTTON" type="submit" value="$p[1]">
</td></tr>

<tr>
<td align="right" valign="top">$p[13]:</td>
<td><input type="checkbox" name="paint" $paint></td>
</tr>

<tr>
<td align="right" valign="top">$p[8]:</td>
<td>$sex</td>
</tr>

<tr><td align="right" valign="top">$p[12]:</td>
<td><input name="days" size="8" value="$CGI{'days'}">
</td></tr>

<tr><td align="right" valign="top">$p[3]:</td>
<td><select name="sort">
$sorts
</select>
 
<br><select name="sort1">$fieldPullS1</select>
$xand <select name="sort2">$fieldPullS2</select>
$xand <select name="sort3">$fieldPullS3</select>
</td></tr>

<tr><td align="right" valign="top">$p[4]:</td>
<td><select name="format">
$formats
</select>
</td></tr>

<tr><td align="right" valign="top">$p[5]:</td>
<td><select name="frames">
$frames
</select>
</td></tr>

<tr><td align="right" valign="top">$p[6]:</td>
<td><input name="max" value="$max" size="8">
</td></tr>

<tr><td align="right" valign="top">$p[7]:</td>
<td>$flds</td>
</tr>
EOM

    if ( $how eq 'EDIT' ) {
        print <<EOM;
<tr><td align="right" valign="top">$p[11]:</td>
<td><input type="checkbox" name="efields" value="_delete" CHECKED></td>
</tr>

<tr><td align="right" valign="top">$p[9]:</td>
<td>$eflds</td>
</tr>

<tr><td align="right" valign="top">$p[10]:</td>
<td>$pass</td>
</tr>
EOM
    }

    print <<EOM;
<tr><td>&nbsp</td>
<td><input class="BUTTON" type="submit" value='$p[1]'> $SaveFormButton</td>
</tr>
</table>
</form>
EOM

}

# inserts data on the URL into the database
# name=Tom&surname=Jones&id=45

sub cgiXput {

    printHead('ok');

    %rec = %CGI;
    recWrite();

}

# put many records in csv format
# one should do a POST request to XputCSV
# CSV file should be POSTed
# content-type should be text/csv

sub cgiXputMany {

    cgiAdmImport();

}

# like show but returns XML

sub cgiXget {

    $id = $CGI{'_id'} || $CGI{'_isindex'};
    recRead($id);
    printXmlHead();    # in woda will be read by fetch!
    print "<!DOCTYPE rec>\n";
    print &rec2xmle( keys(%rec) );
}

# Show a record

sub cgiShow {

    &PIC() if 0;       # may need to have it compiled in, so call it, but don't use it

    $id = $CGI{'_id'} || $CGI{'_isindex'};

    $sort   = $CGI{'sort'};
    $format = $CGI{'format'};
    @fields = split( /,/, $CGI{'fields'} );

    recRead($id);

    if ( $rec{'_id'} eq '' ) {
        wbError( 'Error:No such record', "Record with id=$id does not exist or it has been recently deleted" );    #UK
    }

    if ( $format eq 'XML' ) {
        @fields = &visibleFields() unless $#fields >= 0;
        printXmlHead();
        print "<!DOCTYPE rec>\n";
        print &rec2xml(@fields);
    }
    elsif ( $format eq 'XMLE' ) {
        wbSortFields();

        # all stuff because GET is supposed to GET it all
        @fields = @IFields unless $#fields >= 0;
        printXmlHead();
        print "<!DOCTYPE rec>\n";
        print &rec2xmle(@fields);
    }
    else {

        #		$rec{'summary'}=~ s/\000/ /g;
        #wbFail($rec{'summary'});
        wbSetOwner();

        if (   $Group ne 'admin'
            && $WBB{'hideUnless'}
            && !&evalSafe( $WBB{'hideUnless'} ) )
        {    # setting
            wbError( "This record is not available", "User '$User' from group '$Group' is not allowed to see it" );    #UK
        }

        $MetaFields = '';
        @fields     = &visibleFields();
        foreach $field (@fields) {
            $meta = $WBF{ $field, 'meta' };
            if ( $meta ne '' ) {
                $val = &htQuote( $rec{$field} );
                $MetaFields .= qq(<meta name="$meta" content="$val">\n);
            }
        }

        $x = &evalSafe( $WBB{'detailTitle'} )
          || "$WBB{'recordTitle'} $id <!-- see WBB{'detailTitle'} and WBB{'recordTitle'} -->";
        printHead("$x");

        $html = &wbRecHtml();

        paintFoundRec(*html);
        print $html;

        $allHits = $CGI{'hits'};
        setShowToolbar();

        $idq = &htEscape($id);

        print <<EOM if $WBB{'similar'};    #UK
<ul>
<li>Users who put this record into basket also <a href="$SCRIPT/SimilarBasket?id=$idq">placed into basket these records</a>.</li>
<li>Users who displayed this record <a href="$SCRIPT/SimilarShow?id=$idq">also displayed these records</a>.</li>
</ul>
EOM

    }
}

# similar records based on basket

sub cgiSimilarBasket {

    local ( @list, $id, @ids );

    $id = $CGI{'id'};

    printHead("Records in basket with <a href=\"$SCRIPT/Show?_id=$id\">$id</a>");    # UK

    # all basket files

    @x = &glob('_friend/*.bskt');

    foreach $x (@x) {
        next if $x =~ m/$User/;
        open( h, "_friend/$x" );
        @y = <h>;
        close(h);
        next unless grep( /^$id/, @y );
        push( @list, @y );
    }

    # files in the basket with this one

    foreach $x (@list) {
        ( $did, $rest ) = split( /\t/, $x, 2 );
        $show{$did}++ unless $did eq $id;
    }

    @ids = sort { $show{$a} <=> $show{$b} } keys(%show);

    unless ( $#ids >= 0 ) {
        wbWarn('None');    #UK
        return;
    }

    printRecords( 'DEFAULT', *ids );

}

# similar records based on basket

sub cgiSimilarShow {

    local ( @list, $this, @ids );

    $this = $CGI{'id'};

    printHead("Records viewed by users who viewed <a href=\"$SCRIPT/Show?_id=$this\">$this</a>");    # UK

    @logs = &glob('_log/*.log');
    @logs = sort { $b cmp $a } (@logs);

    undef @rows;
    foreach $log (@logs) {

        # print "\n<br>Examined log file $log";
        open( h, "_log/$log" );
        push( @rows, grep( /\tShow\t/, <h> ) );
        close(h);

        # print "... $#rows relevant records";
        if ( $#rows > 10000 ) {

            # print "... enough!\n";
            last;
        }
    }

    # isolate interesting rows

    @thisShows = grep( /_id=$this/, @rows );
    $times = $#thisShows + 1;

    # find users who did the Show This

    foreach $row (@thisShows) {
        @x = split( /\t/, $row );
        $users{ $x[4] }++;
    }

    $since = &formatDate( $x[2] );

    # what else did these users Show

    @users  = keys(%users);
    $nusers = $#users + 1;

    $pattern = join( '\t|\t', @users );
    $pattern = "\\t$pattern\\t";

    @otherShows = grep( /$pattern/, @rows );

    foreach $rows (@otherShows) {
        $rows =~ m/_id=([^&\s]*)/;
        $show{$1}++;
    }

    @ids = sort { $show{$b} <=> $show{$a} } keys(%show);

    unless ( $#ids >= 0 ) {
        wbWarn('None');    #UK
        return;
    }

    print <<EOM;           #UK
<p>Examining $#rows /Show pages, record $this was shown <b>$times times</b>
since $since by <b>$nusers users</b>. These users also showed (more popular first):
EOM

    printRecords( 'DEFAULT', *ids, 10 );
}

# print records

sub printRecords {

    local ( $format, *ids, $max ) = @_;

    $max = $max | 10;

    undef @list;
    wbSortFields();

    $i = 0;
    foreach $id (@ids) {
        recRead($id);
        next unless $rec{'_id'};
        $i++;
        $ThisRecordNo = $i;
        $list[$i] = &formatFoundRec( $format, *PFields );    # Hidden ??
        last if $i eq $max;
    }

    $head = &formatFoundHeader( $format, *PFields );         # Hidden ??
    formatFoundSeparators( $format, 1 );

    print <<EOM;
<form action="$SCRIPT/BasketEdit" target="_self">	
<div class="RECORDS"> $t0 $head
EOM
    $i = 0;
    foreach $_ (@list) {
        print "$t1$_$t2";
        $i++;
        last if $i eq $max;
    }
    print $t3;

    $modify = "Add checked items to $WBB{'basketName'}";    #UK
    $x      = 'SUBMIT';
    $x      = 'HIDDEN' unless $WBB{'basketName'};
    print <<EOM;
</div>
<br>
<input $x class="BUTTON" type="$x" value="$modify">
</form>
EOM
}

# set record group to
#	 'owner' if the user is an owner
#	'public' if the record is not owned by anyone and user is ^_anon_
#	'other' in all other cases
#
# $RGroup::record group

sub wbSetOwner {

    undef $RGroup;

    if ( $WBB{'ownerField'} ) {
        if ( $rec{ $WBB{'ownerField'} } eq $User ) {
            $RGroup = 'owner';
        }
    }
    elsif (
        ( $User =~ m/^_anon_/ || $User eq '' )
        && (   $rec{ $WBB{'ownerField'} } eq '0'
            || $rec{ $WBB{'ownerField'} } eq '' )
      )
    {
        $RGroup = 'public';
    }
    else {
        $RGroup = 'other';
    }

    # may not be owner but belongs to right group
    if ( $WBB{'ownerGroup'} && $WBB{'recordLock'} =~ m/group/ ) {
        if ( $WBB{'ownerGroup'} eq $Group ) {
            $RGroup = 'owner';
        }
    }
}

# is the logged in user the owner of the opened record

sub isOwner {

    wbSetOwner();
    return $RGroup eq 'owner';
}

# is the record owned at all

sub isOwned {

    return $WBB{'ownerField'} && $rec{ $WBB{'ownerField'} } ne '';

}

# returns the /Show URL of the record of the currently logged user
# or blank, if not a registered user
# code shared with formatLoginRecordLink TODO code factoring
sub loginRecord {

    local ( $dir, $url, $u );

    # if ( $Group eq $WBB{'userTableGroup'} || $Group eq 'user' ) {
    if ( $WBB{'userTable'} && $Group !~ m/friend|guest|admin/ ) {
        $url = &wbTables( 'URL', $WBB{'userTable'} );
        $u = &wbFixId($User);

        # $u = &htEscape($u);
        return "$url/Show?_id=$u";
    }
    else {
        return '';
    }
}

# is the table used to verify usernames and passwords?
# returns 1 if this is the table used to
# verify usernames and passwords for
# logging in

sub isLoginTable {

    local ($ut);

    if ( $WBB{'userTable'} eq '' ) {

        return 0;

    }
    else {

        return &wbTables( 'URL', $WBB{'userTable'} ) eq &wbTables( 'URL', '' );

    }
}

# adds to @Toolbar the links to
# actions related to one record

sub setShowToolbar {

    @hlps = (
        "pick and add to $WBB{'basketName'}",        #UK
        'modify the information',                    #UK
        'delete entire record',                      #UK
        'delete password',                           #UK
        'explain the meaning of the fields',         #UK
        'view how record was changing over time',    #UK
        '',                                          #UK
        '',                                          #UK
        '',                                          #UK
        ''                                           #UK
    );
    @butts = ( 'PICK',         'EDIT', 'DELETE', 'DELETE PASSWORD', 'LEGEND',           'HISTORY',          'FIRST', 'PREVIOUS', 'NEXT', 'LAST' ); #UK
    @acts  = ( 'BasketAddOne', 'Edit', 'Delete', 'AdmDelPass',      'DisplayStructure', 'AdmRecordHistory', 'Go',    'Go',       'Go',   'Go' );
    @to    = ( '',             '',     '',       '',                '',                 '',                 'first', 'previous', 'next', 'last' );
    @icons = ( 'basket', 'edit', 'del', 'delpass', 'info', 'history', 'first', 'back', 'forward', 'last' );

    if ( !$WBB{'basketName'} ) {
        shift @hlps;
        shift @butts;
        shift @acts;
        shift @to;
        shift @icons;
    }

    $ret = "";

    $i = 0;
    foreach $act (@acts) {
        if ( &wbDenied( $Group, $act ) ) {
            $i++;
            next;
        }
        if ($InFrame) {
            $act .= "ToFrameD";
            $target = 'target="_self"';
        }
        else {
            $target = '';
        }
        $x = &htEscape($id);
        if ( $icons[$i] =~ m/edit|del|back|forward|basket|history/ ) {
            $params = "_id=$x";
        }
        else {
            $params = "";
        }
        $extra = '';
        if ( $act =~ m/^Go/ ) {
            $params = $params . '&' if $params;
            $params .= "sort=$sort&amp;to=$to[$i]";
            if ( $CGI{'search'} ne '' ) {
                $s = &htEscape( $CGI{'search'} );
                $params .= "&amp;search=$s";
            }
            next
              if $allHits > 100;    # only allow browsing if less than 100 results
            next unless $sort;
            next if $CGI{'search'} eq '';
        }

        if ( $act =~ m/^Basket/ ) {
            $target = 'target="woda_basket"';
        }

        $h = $hlps[$i] || $butts[$i];

        $params = "?" . $params if $params;

        push( @Toolbar, <<EOM);    #xxx
<a href="$SCRIPT/$act$params" $target><img 
$ICONPAR src="$ICONURL/$icons[$i].gif" alt="$h" title="$h"></a>
EOM
        $i++;
    }
}

# stores a record

sub cgiStore {

    $tmpID = $CGI{'tmpID'};

    #	do recBackup($newID);
    recReadFile("$TMPDIR/$tmpID.rec");

    if ( !%rec ) {
        wbError( "Data missing", "The submitted data in $TMPDIR/$tmpID.rec could not be read" );    #UK
        return;
    }

    # assign the new ID

    $newID = $CGI{'newID'} || &evalWBBkey();
    if ( !$newID ) {
        $newID = &recRandID();
    }

    # move attached

    recMoveAttached( $tmpID, $newID );
    recSetSystem();
    $rec{'_id'} = $newID;
    $Log .= "Write $newID\n";
    recWrite();
    $Log .= "Read $newID\n";
    recRead($newID);

    if ( !%rec ) {
        wbError("ERROR: Database Update Failed");    #UK
        return;
    }

    unlink("$TMPDIR/$tmpID.rec");
    wbTmpClean();

    $t = "Database Updated";                         #UK
    if ($HtmlHeadPrinted) {

        # &wbWarn("Success",$t); #UK
    }
    else {
        printHead($t);
    }

    wbRecordChanged();
    wbDatabaseChanged();

    $url = "$SCRIPTBASE$SCRIPT/Show?_id=$newID";

    $x = <<EOM ;    #UK
<p>Thank you for contributing to $WBB{'dbTitle'}. Your edits were saved at URL <p>
<center><p><a class="BUTTON" href="$url">$url</a>.</p></center>
<p>
This information may be useful to locate the record for editing or updating.
You may wish to <tt>[bookmark]</tt> it.
<p>
To add a record, click on the add icon in the toolbar.
EOM

    $y = "Success!";    #UK
    print &formatWindow( $y, $x );

    $t = "Add another record";    #UK
    push( @Toolbar, <<EOM);
<a href="$SCRIPT/Add">
<img $ICONPAR src="$ICONURL/add.gif" alt="$t" title="$t"></a>
EOM

    if ( &isLoginTable() ) {
        $t = 'WARNING: Login information changed!';    #UK
        $m = <<EOM;                                    #UK
<p>Please note that <b>$newID</b> is a username to
<a href="$SCRIPT/LoginForm?username=$newID">log into</a> this
application.
<p>The password to log in is the password that was used
to lock this record.
EOM
        if ( $WBB{'userAllow'} ) {
            $m .= <<EOM;                               #UK
<p>There are some special conditions set,
so that you might not be able to log-in
right away.
EOM
        }

        wbWarn( $t, $m, 'width="400"' );
    }

    if ( $WBB{'cacheExpire'} > 0 ) {

        $file = "_cache/default.tbl";
        $file =~ tr/[A-Z]/[a-z]/;
        open( h, "$file.time" );
        $secs = <h>;
        close(h);
        $secs += 2;

        $t = "WARNING: Data not indexed";    #UK
        $m = <<EOM;                          #UK
<p>Although your data has been entered, due to the setup of this database,
it may take up to <b>$WBB{'cacheExpire'} hours</b> before it can be found
by searches.
<p>If you want your edits to be visible
immediately, <a href="$SCRIPT/Add">add</a> or edit any other data you may have,
then click below and <b>be patient</b>. The processing may take about <b>$secs seconds</b>.
<form action="$SCRIPT/Reindex" method="post">
<input name="sort" type="hidden" value="default">
<input name="secs" type="hidden" value="$secs">
<input class="BUTTON" type="submit" value="OK, I'll wait">
</form>
EOM
        print &formatWindow( $t, $m );
    }

    print "<p>\n";
}

# reindex database

sub cgiReindex {

    printHead("Reindexing database");                                                                         #UK
    print "<p>Warning ... this may take about $CGI{'secs'} seconds! Do not disturb!\n";                       #UK
    wbReindex( $CGI{'sort'} );
    print "<p>Done. <a href=\"$SCRIPT/SearchForm\">Searching</a> should now find the most recent data!\n";    #UK
}

#
# display table of contents
#

sub cgiToc {

    printHead("Table of Contents");                                                                           #UK

    print <<EOM;                                                                                              #UK
This table of contents is intended for the
conveniance of Web spiders and crawlers. It is
not intended for humans!
EOM

    ( $allHits, *list, *fields ) = &recSelect();

    foreach $item (@list) {
        print "<br>$print{$item}\n";
    }
}

# undo record changes

sub cgiUndo {

    printHead("Undo is not available");    #UK
    return;

    $id = $CGI{'id'};

    #	$err = &recRestore($id);

    if ($err) {
        print <<EOM ;                      #UK
<h3 class="ERROR">Undo failed</h3>
$err
EOM
    }
    else {
        wbDatabaseChanged();
        print "<h3>Undo was successful</h3>\n";    #UK
    }
}

# Information about fields in database

sub cgiDisplayStructure {

    printHead("Information about fields in $WBB{'dbTitle'}");    #UK
    printStructure( '00000', '^p$|^head$|^type$|^help$|^options$', 1 );
}

# prepares $inarow and @fields parameters for keyword and contents browse

sub wbBrowseDefaults {

    wbSortFields();

    $inarow = $CGI{'inarow'} || 3;
    @fields = split( /,/, $CGI{'fields'} );

    if ( $#fields < 0 ) {
        foreach $field (@Fields) {
            $t = $WBF{ $field, 'type' };
            if ( $WBF{ $field, 'type' } =~ m/^OPTION$/ ) {
                push( @fields, $field );
            }
        }
    }
}

# Browse searching using keywords

sub cgiBrowseKeywords {

    wbBrowseDefaults();
    $f = join( ', ', @fields );

    printHead("Browse by keywords in fields $f");    #UK

    print <<EOM;                                     #UK
<p>Check any keywords you are interested in then press the
search button at the end of this page. The items with more matched
keywords will be listed on top <b>or</b></p>
<p>Click on keywords next to the checkboxes to search for records
belonging to that category only.</p>
EOM
    print <<EOM;
<form class="FORM" method="post" action="$SCRIPT/Search">
EOM

    @items   = '';
    $dItem   = 0;
    $dSearch = '';

    foreach $field (@fields) {

        if ( $WBF{ $field, typePar } =~ m/HEADING=(.{1,5})/i ) {    # check if the field contains HEADING
            $heading = $1;
        }
        else {
            $heading = '';
        }

        $dSearch = 0;                                               # reset position of label in @optionText;
        $head    = &formatFieldName($field);
        @kwords  = &wbParseOptions($field);
        my @optText = @optionText;                                  # prepare human readable display
        $print = "<b>$head</b><br />\n";

        foreach $kword (@kwords) {

            if ( $heading && $kword =~ m/^$heading/ ) {             # if it is a HEADING value print it
                $print .= "<p><i>$kword</i><br>";
                $dSearch++;                                         # skip HEADING for labels
                next;
            }

            $x = &wbOptionWordSearch( $field, $kword );
            $s = $optionText[$dSearch] || $kword;                   # makes human readable display
            $print .= <<EOM;
<input type="checkbox" name="search" value="&quot;$kword&quot;">
<a href="$SCRIPT/Search?search=$x">$s</a><br>
EOM
            $dSearch++;
        }

        $items[ $dItem++ ] = $print;
    }

    print '<table cellspacing="0">';

    for ( $i = 0 ; $i <= $#items ; $i += $inarow ) {
        print '<tr valign="top">', "\n";
        for ( $j = 0 ; $j < $inarow ; $j++ ) {
            $idx = $i + $j;
            print <<EOM;
<td>
$items[$idx]
</td>
EOM
        }
        print "</tr>\n";
    }

    $search = "S e a r c h";    #UK
    $reset  = "R e s e t";      #UK

    print <<EOM;
</table>
<p><br>
<input class="BUTTON" type="submit" value="   $search   ">
<input class="BUTTON" type="reset" value="   $reset   ">
</form>
EOM

}

# browse contents

sub cgiBrowseContents {

    wbBrowseDefaults();

    $icon = '<img src="' . $ICONURL . '/search.gif" ' . $ICONPAR . ' alt="search" title="search">';
    $join = "<br />";

    if ( $#fields == 0 ) {
        $f = &formatFieldName( $fields[0] );
        printHead("Browse by $f");    #UK
    }
    else {
        $f = join( ',', @fields );
        printHead("Browse by fields $f");    #UK
    }

    if ( $CGI{'comma'} ) {
        $join = ', ';
        $icon = '';
    }

    print <<EOM;                             #UK
<p>Follow the links to find items tagged with the keyword</p>
EOM

    @items   = '';
    $dItem   = 0;
    $dSearch = 0;

    foreach $field (@fields) {

        if ( $WBF{ $field, typePar } =~ m/HEADING=(.{1,5})/i ) {    # check if the field contains HEADING
            $heading = $1;
        }
        else {
            $heading = '';
        }

        $dSearch = 0;                                               # reset position of label in @optionText;

        $head   = &formatFieldName($field);
        @kwords = &wbParseOptions($field);
        $print  = "<b>$head</b><br>\n" if $#fields > 0;

        @list = ();
        foreach $kword (@kwords) {

            if ( $heading && $kword =~ m/^$heading/ ) {             # if it is a heading value print it
                push( @list, "<i>$kword</i>" );
                $dSearch++;                                         # skip HEADING for labels
                next;
            }

            $x = &wbOptionWordSearch( $field, $kword );
            $s = $optionText[$dSearch] || $kword;                   # makes human readable display
            push( @list, "<a href=\"$SCRIPT/Search?search=$x\">$icon $s</a>" );
            $dSearch++;
        }

        # $print .= &formatColumns(1,@list);
        $print .= join( $join, @list );
        $items[ $dItem++ ] = $print;
    }

    print "<table>";

    for ( $i = 0 ; $i <= $#items ; $i += $inarow ) {
        print '<tr valign="top">', "\n";
        for ( $j = 0 ; $j < $inarow ; $j++ ) {
            $idx = $i + $j;
            print <<EOM;
<td>
$items[$idx]
</td>
EOM
        }
        print "</tr>\n";
    }

    print "</table>";
}

# browse contents using a matrix

sub cgiBrowseMatrixList {

    printHead("Select Browse Matrix Combination");    #UK
    print "<p>\n";
    print <<EOM;                                      #UK
This page lets you define a browse matrix. Select 
combination of two fields. Values of the first will be used for rows and
values of the second for columns. Note that matrices which would include
fields marked with a * may take longer to generate.

EOM

    print <<EOM;                                      #UK
<p>You can also define the browse matrix using <a href="$SCRIPT/BrowseMatrixForm">a form</a>.
EOM

    wbSortFields();
    foreach $field (@Fields) {
        if ( $WBF{ $field, 'type' } =~ m/OPTION$/ ) {
            push( @fields, $field );
            $nice = &formatFieldName( $field, 1 );
            $nice .= " *" if $WBF{ $field, 'type' } eq 'LINKOPTION';

            # remove hyperlinks

            push( @niceFields, $nice );
        }
    }

    $opts = '';
    for ( $i = 0 ; $i <= $#fields ; $i++ ) {
        $rn = $fields[$i];
        $ru = &htEscape($rn);
        $rv = $niceFields[$i];
        print "<p><b>$rv</b><br>\n";
        for ( $j = 0 ; $j <= $#fields ; $j++ ) {
            next if $i == $j;
            $cn = $fields[$j];
            $cu = &htEscape($cn);
            $cv = $niceFields[$j];
            print <<EOM;    #UK
<a href="$SCRIPT/BrowseMatrix?row=$ru&amp;col=$cu">
$rv : $cv</a> 
(<a href="$SCRIPT/BrowseMatrix?row=$ru&amp;col=$cu&amp;count=on">count</a>)
<br>
EOM
        }
    }
}

# browse contents by matrix

sub cgiBrowseMatrixForm {

    printHead("Define Browse Matrix");    #UK

    print <<EOM;                          #UK
<p>This page lets you define a browse matrix. Select 
one field name to display in rows and one to display in colums.
Search matrix will be generated so that you will be able to search for items
with row and column value.<p>
Note that matrices which would include fields marked with a * may take longer to generate.

<p>You can also select a browse matrix using <a href="$SCRIPT/BrowseMatrixList">predefined lists</a>.

	<form class="FORM" action="$SCRIPT/BrowseMatrix">
EOM

    wbSortFields();
    foreach $field (@Fields) {
        if ( $WBF{ $field, 'type' } =~ m/OPTION$/ ) {
            push( @fields, $field );
            $nice = &formatFieldName( $field, 1 );
            $nice .= " *" if $WBF{ $field, 'type' } eq 'LINKOPTION';
            push( @niceFields, $nice );
        }
    }

    $opts = '';
    for ( $i = 0 ; $i <= $#fields ; $i++ ) {
        $n = $fields[$i];
        $v = $niceFields[$i];
        $opts .= "<option value=\"$n\">$v\n";
    }

    $c  = "Field to use for columns";    #UK
    $r  = "Field to use for rows";       #UK
    $cm = "Count matching records?";     #UK
    print <<EOM;
<table border="1">
<tr>
<td><br></td>
<td align="CENTER">
$c<br>
<select name="col">
$opts
</select>
</td>
</tr>

<tr>
<td align="CENTER">
$r<br>
<select name="row">
$opts
</select>
</td>
<td><br></td>
</tr></table>
<p>
<input name="count" type="checkbox"> $cm
</p>
<input class="BUTTON" type="submit" value="   OK   ">
</form>
EOM
}

# browse data using field matrix
# Utilise the values in @optionText to make human readable display

sub cgiBrowseMatrix {

    $field  = $CGI{'row'};
    $field2 = $CGI{'col'};

    $label  = $WBF{ $field,  'head' } || $field;
    $label2 = $WBF{ $field2, 'head' } || $field2;

    printHead("Browse Matrix by $label and $label2");    #UK
    print <<EOM;                                         #UK
<p>Click on row or column headers to search for records belonging to that
category <em>or</em>
Click on the search icons to search for
a combination of row and column values.
EOM

    if ( $CGI{'count'} ne '' ) {
        readCatCount( $field, $field2 );
    }

    @rowWords  = &wbParseOptions($field);
    @rowLabels = @optionText;
    @colWords  = &wbParseOptions($field2);
    @colLabels = @optionText;
    $rowTit    = &formatFieldName($field);
    $colTit    = &formatFieldName($field2);

    $nrows = $#rowWords + 1 + 1 + 1;    # legend,title
    $ncols = $#colWords + 1 + 1;
    $nc1   = $ncols - 1;

    print <<EOM;
<p><table border="1">
<tr>
<th><br></th>
<th COLSPAN=$nc1>
$colTit
</th>
</tr>
<tr>
<th align="right">$rowTit</th>
EOM

    # col headings
    $dSearch = 0;

    foreach $colWord (@colWords) {
        $x = &wbOptionWordSearch( $field2, $colWord );
        $s = $colLabels[$dSearch] || $colWord;
        print qq(<th><a href="$SCRIPT/Search?search=$x">$s</a></th>\n);
        $dSearch++;
    }

    print "</tr>\n";

    # rows

    $dSearch = 0;
    my $iconimage = 'search';
    foreach $rowWord (@rowWords) {
        $x = &wbOptionWordSearch( $field, $rowWord );
        $s = $rowLabels[$dSearch] || $rowWord;
        print qq(<tr><th align=RIGHT><a href="$SCRIPT/Search?search=$x">$s</a></th>\n);
        foreach $colWord (@colWords) {
            $search = "+$field2:\"$colWord\" +$field:\"$rowWord\"";
            $search = &htEscape($search);
            $surl   = "$SCRIPT/Search?search=$search";
            $count  = '';
            if ( $CGI{'count'} ) {
                $count = $catCount{"$rowWord,$colWord"};
                $count = '0' if $count eq '';
                $iconimage =
                  $count
                  ? qq(<img $ICONPAR src="$ICONURL/search.gif" alt="search" title=search>)
                  : '';
            }
            print <<EOM;
<td align="CENTER" valign=MIDDLE>
<a href="$surl">$iconimage</a> $count
</td>
EOM
        }
        print "</tr>\n";
        $dSearch++;
    }

    print "</table>\n";
}

# creates +field:"$word" and escapes it for url

sub wbOptionWordSearch {

    local ( $field, $word ) = @_;

    # $word =~ s/\(|\)|\//\\$&/g;	# escape ( ) and / in search words
    $word =~ s|[/()]|\\$&|g;    # escape ( ) and / in search words
    return &htEscape("+$field:\"$word\"");
}

########### # # # UTILITIES #######################################

# adds meta fields _time and _from to %rec

sub recSetSystem {

    # _from

    local ($from);
    $from         = $RH;
    $from         = $RA if !$from;
    $rec{'_from'} = $from;

    # _at

    $rec{'_at'} = time;

    # can be done with a hidden field!
    # $rec{'_user'}=$User;
}

# print @items in $num columns
# should be change to print in columns starting with HEADINGS defined fields

sub formatColumns {

    local ( $num, @items ) = @_;
    local ( $ret, $left );

    $num = 1 if $num == 0;    # parachute
    $left = 1 + int( ($#items) / $num );

    $ret = "<table><tr>\n";
    for ( $i = 0 ; $i <= $#items ; $i += $left ) {
        $j = $i + $left - 1;
        $x = join( "\n<br>", @items[ $i .. $j ] );
        $ret .= qq(<td valign="top" class="NOBORDER">$x</td>\n);
    }
    $ret .= "</tr></table>\n";
    return $ret;
}

# formats time as yyyy/mm/dd hh:mm

sub formatDate {

    local ($time) = @_;
    local ( @x, $i );

    @x = &fixDate($time);
    return "$x[5]/$x[4]/$x[3] $x[2]:$x[1]";
}

# formats date according to dateFormat

sub formatYMD {

    local ( $yyyy, $mm, $dd ) = @_;

    if ( $WBB{'dateFormat'} ne '' ) {
        return &evalSafe( $WBB{'dateFormat'} );
    }
    else {
        return "$yyyy/$mm/$dd";
    }
}

# formats date according to timeFormat

sub formatHMS {

    local ( $hh, $mm, $ss ) = @_;

    if ( $WBB{'timeFormat'} ne '' ) {
        return &evalSafe( $WBB{'timeFormat'} );
    }
    else {
        return "$hh:$mm";
    }
}

# returns Y M D H M S

sub formatYMDHMS {

    local ($time) = @_;
    local ( @x, $i );

    @x = &fixDate($time);
    return &formatYMD( $x[5], $x[4], $x[3] ) . ' ' . &formatHMS( $x[2], $x[1], $x[0] );
}

# returns today's date formatted in a way
# proper for Woda date fields

sub TODAY {

    local ( $d, $t );

    $d = &formatDate(time);
    ( $d, $t ) = split( / /, $d );
    return $d;
}

# pads the dat and time values with zeros

sub fixDate {

    local (@x) = localtime( $_[0] );

    $x[4]++;

    for ( $i = 0 ; $i <= $#x ; $i++ ) {
        $x[$i] = "0" . $x[$i] if length( $x[$i] ) == 1;
    }

    $x[5] += 1900;

    return @x;
}

# does nothing at all

sub NOOP {
}

# modifies $string so that it can be used
# as a record ID e.g. replaces does with underscores

sub ID {
    local ($string) = @_;

    return &wbFixId($string);
}

# attempts to disguise email address from screen scrapers

sub formatMailto {

    local ($addr) = @_;
    local ( $x, $y );
    $x = &htEscape($addr);
    $y = $addr;
    $y =~ s/\./&#46;/g;
    $y =~ s/\@/&#64;/g;
    if ( $JS >= 1 ) {
        return qq[<script LANGUAGE="JavaScript" type="text/javascript">document.write('<a HR'+'EF'+'="'+'mai'+'lto'+':'+'$x'+'">$y</a>');</script>];
    }
    else {
        return qq[<a href="mailto:$x">$y</a>];
    }

    #$x = int(1000*rand()) . '.';
    #$pad = "(without $x)"; #UK
    #return "<a href=mailto:$x$addr>$x$addr</a> $pad";
}

# formats field $field according to $WBB{field,'picture'}

sub formatField {

    local ($field) = @_;
    my $x;
    local $_;

    # available variables

    $_  = $rec{$field};
    $ID = $rec{'_id'};

    if ( $WBF{ $field, 'type' } eq 'BREAK' ) {
        return '';
    }

    if ( $WBF{ $field, 'type' } eq 'SUBMIT' ) {
        return '';
    }

    if ( $WBF{ $field, 'type' } eq 'COMPUTE' ) {
        $y = $WBF{ $field, 'picture' };
        $x = &evalSafe($y);

        # eval safe does it return "<p class="ERROR">$@</p><pre>$y</pre>" if $@;
        return "$x";
    }

    return '' if $_ eq '';

    # multiple options in OPTION or LINKOPTION

    $pict = $WBF{ $field, 'picture' };
    $pict = '' if $pict eq "\n";    # in forms this may be generated by itself

    if ( $_ =~ m/\n/ && $WBF{ $field, 'type' } =~ m/OPTION$|^LIST$/ ) {
        if ($pict) {
            local ($ret) = '';
            local ($_)   = $_;
            $_ =~ s/\r//g;          #scix
            local (@items) = split( /\n/, $_ );
            foreach $_ (@items) {
                $ret .= &evalSafe($pict);    # 30.12.2001 dodal ta BR
            }
            return $ret;
        }
        else {

            # ziga, ta nasuje podpicja !!!
            $_ =~ s/\n/, /g unless m/<br>/i;    # if we do our own linespacing ...
            return $_;
        }
    }

    # other options

    if ($pict) {

        return &evalSafe($pict);                # || "<em class="ERROR">$@</em>";

    }
    elsif ( $WBF{ $field, 'type' } eq "TEXTAREA" ) {

        if ( $_ =~ m/<p|<br/ ) {
            return $_;
        }
        else {
            local ( @rows, $max, $row );
            @rows = split( /\n/, $_ );
            $max = 0;
            foreach $row (@rows) {
                if ( length($row) > $max ) {
                    $max = length($row);
                }
            }
            if ( $max > 60 ) {
                $_ =~ s/\n/<br>\n/g;
                return "<tt>$_</tt>";
            }
            else {
                return "<pre>$_</pre>";
            }
        }

    }
    elsif ( $WBF{ $field, 'type' } eq "HTMLAREA" ) {

        if ( $_ =~ m/<p|<br|<b|<I/i ) {

            # nothing
        }
        else {
            s/\n\n/<p>/g;
        }

        return $_;

    }
    elsif ( $WBF{ $field, 'type' } eq "FILE" ) {

        return "<a href=\"$WBB{'htmlURL'}/att/$rec{$field}\">$rec{$field}</a>";

    }
    elsif ( $WBF{ $field, 'type' } eq "IMAGE" ) {

        return "<img alt=\"\" src=\"$WBB{'htmlURL'}/att/$rec{$field}\">";

    }
    elsif ( $WBF{ $field, 'type' } eq "USERFILE" ) {

        $unit = 'bytes';                             #UK
        @x    = stat("$HTMLDIR/att/$rec{$field}");
        $rec{$field} =~ m/([^\.]*)$/;
        $xt = $1;
        if ( $xt =~ m/$FILETYPES/ ) {
            $ic = "$ICONURL/$xt.gif";
        }
        else {
            $ic = "$ICONURL/binary.gif";
        }

        $s = $x[7];
        $s = '0' x ( 21 - length($s) ) . $s;
        $s =~ s/\d{3}/$&\,/g;
        chop($s);
        $s =~ s/^[0,]*//;

        return qq~<a href="$WBB{'htmlURL'}/att/$rec{$field}"><img alt="" $ICONPAR $xt align="top" src="$ic">file.$xt</a> ($s $unit)~;

    }
    elsif ( $WBF{ $field, 'type' } eq "URL" ) {

        return "<a href=\"$rec{$field}\">$rec{$field}</a>";

    }
    elsif ( $WBF{ $field, 'type' } eq "EMAIL" ) {

        return &formatMailto( $rec{$field} );

    }
    elsif ( $WBF{ $field, 'type' } eq "DATE" ) {

        ( $yyyy, $mm, $dd ) = split( /\//, $rec{$field}, 3 );
        return &formatYMD( $yyyy, $mm, $dd );

    }
    else {

        return $rec{$field};

    }
}

# is an email address

sub isEmail {

    local ($chr);

    return $_[0] =~ m/^[\w\-\+\.]+\@[\w\-\+\.]+\.[a-z]{2,4}$/i;    # something@something.com|si

}

# formats values in %rec hash into %pic hash
# using instructions in $WBB{afield,'picture'}
# only does it for comma separated list (no spaces) of dFields, if defined
# returns the PIC of the last field

sub PIC {

    local ($dFields) = @_;

    local ( $field, @f );
    undef %pic;

    if ($dFields) {
        @f = split( /,/, $dFields );
    }
    else {
        wbSortFields();
        @f = @Fields;
    }

    foreach $field (@f) {
        if ( $Hidden{$field} ) {
            $pic{$field} = '[hidden]';    #UK
        }
        else {
            $pic{$field} = &formatField($field);
        }
    }
}

# formats top/header row of search results
# should h

sub formatFoundHeader {

    local ( $format, *fields ) = @_;
    local ( $f, $hd, $ft );

    @fields = &visibleFields(@fields);

    if ( $format eq 'TABLE' ) {
        $hd = '<tr class="DATAH" ><th>&nbsp;</th>' . "\n";
    }
    else {
        $hd = '';
    }

    foreach $f (@fields) {
        if ( $format eq 'TABLE' ) {
            $x = &formatFieldName($f);
            $hd .= '<th align="left">' . $x . '</th>';
        }
        elsif ( $format eq 'TAB' ) {
            $hd .= "$f\t";
        }
        elsif ( $format eq 'CSV' ) {
            $hd .= "$f,";
        }
        elsif ( $format eq 'CSVS' ) {
            $hd .= "$f;";
        }
    }

    if (   $format eq 'TAB'
        || $format eq 'CSV'
        || $format eq 'CSVS' )
    {
        chop($hd);
    }

    if ( $format eq 'TABLE' ) {
        $hd .= "</tr>";
    }

    $hd .= "\n" if $hd ne '';

    return $hd;
}

# formats bottom / footer row of search results

sub formatFoundFooter {
    return '';
}

# format a found row

sub formatFoundRec {

    local ( $format, *fields, *editable ) = @_;
    local ( $this, $open, $s, $i );

    @ffields = &visibleFields(@fields);    # should preserve ... what of field hidden?

    if (   $format eq 'TAB'
        || $format eq 'RAW'
        || $format eq 'CSV'
        || $format eq 'XML'
        || $format eq 'XMLE' )
    {

        # --- nothing ... not editable!

    }
    else {

        # editable ?

        undef %pic;

        if ( $#editable >= 0 && &wbCanModifyRecord( $CGI{'_accessPassword'} ) ) {
            foreach $f (@editable) {
                wbFormFieldItems($f);
                $fld = $Form{ $f, 'input' };
                $fld =~ s|NAME="|NAME="_id=$rec{'_id'}_field=|gi;
                $fld =~ s/\n+/\n/g;
                $rec{$f} = $fld;
                $pic{$f} = $fld;
            }
        }

        # make all this shit ready

        $ese = &htEscape( $CGI{'search'} );
        $eso = &htEscape( $CGI{'sort'} || $sort );
        $id  = $ID = $rec{'_id'};
        $eid = &htEscape($id);

        # defined variables for use in the format definitions

        $this              = "$SCRIPT/Show$SuffixDetail?_id=$eid&amp;sort=$eso&amp;search=$ese&amp;hits=$allHits";
        $ThisRecordURL     = $this;
        $ThisRecordDate    = &formatYMDHMS( $rec{'_at'} );                                                           #4.151
        $ThisRecordMatches = $hits{$id};

        if ( $Group eq 'admin' && !$CMDLINE && $#editable < 0 ) {
            $z = 'HSPACE=0' if $format eq 'DEFAULT';

# code commented because obviously something is missing here
#             if (   $format eq 'PARAGRAPH'
#                 || $format eq 'DEFAULT'
#                 && $WBB{'format;DEFAULT'} =~ m/\$this/i )
#             {
#                 $ThisRecordIcons =
# "<a \nHREF=\"$SCRIPT/Delete$SuffixDetail?_id=$eid\"><img alt=\"\" \nSRC=\"$ICONURL/del.gif\" title=\"DEL\" $SMALLICONPAR $z></a> "
#                   . "<a \nHREF=\"$SCRIPT/EditForm$SuffixDetail?_id=$eid\"><img alt=\"\" \nSRC=\"$ICONURL/edit.gif\" title=\"EDIT\" $SMALLICONPAR $z></a> "
#
# #	. "<a \nHREF=\"$SCRIPT/AdmDelPass$SuffixDetail?_id=$eid\"><img alt=\"\" \nSRC=\"$ICONURL/delpass.gif\" title=\"DELPASS\" $SMALLICONPAR $z></a> "
#                   ;
#             }
#             else {
            $ThisRecordIcons =
                "<a \nHREF=\"$SCRIPT/Delete$SuffixDetail?_id=$eid\"><img alt=\"\" \nSRC=\"$ICONURL/del.gif\" title=\"DEL\" $SMALLICONPAR $z></a> "
              . "<a \nHREF=\"$SCRIPT/EditForm$SuffixDetail?_id=$eid\"><img alt=\"\" \nSRC=\"$ICONURL/edit.gif\" title=\"EDIT\" $SMALLICONPAR $z></a> "

     #	."<a \nHREF=\"$SCRIPT/AdmDelPass$SuffixDetail?_id=$eid\"><img alt=\"\" \nSRC=\"$ICONURL/delpass.gif\" title=\"DELPASS\" $SMALLICONPAR $z></a> "
              ;

            #             }
        }
        else {
            $ThisRecordIcons = '';
        }

        # --- delete checkbox

        if ( grep( /\b_delete\b/, @ffields ) ) {
            $alt          = "Check the checkbox to delete record";    #UK
            $onFocus      = qq(onClick="self.status='$alt'");
            $pic{_delete} = <<EOM;

<input
	type="checkbox"
	name="_delete"
	value="$rec{'_id'}" 
	$onFocus
>
EOM
            $rec{'_delete'} = "";
        }

        if ( $format ne 'TABLE' ) {
            $ThisRecordIcons .= $pic{_delete};
        }

        # --- provide the open icon

        if (   $format eq 'PARAGRAPH'
            || $format eq 'DEFAULT' && $WBB{'format;DEFAULT'} =~ m/\$this/i )
        {
            $ThisRecordIcons .= &formatOpenIcon( 'SMALL', $this );
        }
        else {
            $ThisRecordIcons .= &formatOpenIcon( '', $this );
        }

        # --- provide the basket checkbox unless
        #	 we are writing the BASKET
        #	 we are pushing the page out via command line
        #	we are formatting a record for editing

        if (   $PrintingBasket
            || $CMDLINE
            || !$WBB{'basketName'}
            || $#editable >= 0 )
        {

            # do nothing, $ThisRecordBasket may be set elsewhere, dont clear it !!!
        }
        else {

            local ($c);

            unless (%Basket) {
                readBasket();
            }
            $c                = '';
            $c                = 'CHECKED' if $Basket{$id};
            $ThisRecordBasket = <<EOM;
<input name="id$id" type="checkbox" $c>
EOM
        }
    }

    $print = '';

    if ( $WBB{"formatRow$format"} ne '' ) {
        $x = $WBB{"formatRow$format"};

        #do wbFail("$x");
        $print = &evalSafe($x);    # || "EVAL ERROR:" . &htQuote("$x:$@");
        paintFoundRec(*print);

    }
    elsif ( $WBB{"format;$format"} ) {
        $x     = $WBB{"format;$format"};
        $print = &evalSafe($x);                                              # || "EVAL ERROR:" . &htQuote("$x:$@");
        $print = "<li> $ThisRecordIcons $ThisRecordBasket $print </li>\n";
        paintFoundRec(*print);

    }
    elsif ( $format eq 'TAB' || $format eq 'RAW' ) {

        $print = &recToRow(@ffields);

    }
    elsif ( $format eq 'CSV' ) {
        foreach $x (@ffields) {                                              # kaj je to ???
            $v = $rec{$x};
            $v =~ s/\"/\"\"/g;
            $print = $print . '"' . $v . '"' . ',';
        }
        chop $print;

    }
    elsif ( $format eq 'XML' ) {
        $print = &rec2xml(@ffields);

    }
    elsif ( $format eq 'XMLE' ) {
        $print = &rec2xmle(@ffields);

    }
    elsif ( $format eq 'RSS' ) {

        local ($title);
        $title = &evalSafe( $WBB{'detailTitle'} ) || "$WBB{'recordTitle'} $id";
        $title = &htUntag($title);
        $title = &htQuote($title);
        $title = &htUTF8($title);

        $x     = $WBB{"format;DEFAULT"} || $WBB{"formatRowDEFAULT"};
        $print = &evalSafe($x);
        $print = &htUntag($print);
        $print = &htQuote($print);
        $print = &htUTF8($print);

        $print = <<EOM;
<item>
	<title>$title</title>
	<link>$SCRIPTBASE$SCRIPT/Show?_id=$ID</link>
	<description>$print</description>
</item>
EOM
    }
    elsif ( $format eq 'CSVS' ) {
        foreach $x (@ffields) {    # kaj je to ???
            $v = $rec{$x};
            $v =~ s/\"/\"\"/g;
            $print = $print . '"' . $v . '"' . ';';
        }
        chop $print;

    }
    elsif ( $format eq 'TABLE' ) {
        if ( $ffrt == 0 ) {
            $ffrt = 1;
        }
        else {
            $ffrt = 0;
        }
        $print = qq~\n<tr valign="top" class="DATA$ffrt"><td>$ThisRecordIcons $ThisRecordBasket</td>\n~;
        foreach $x (@ffields) {
            $value = $pic{$x}
              || &formatField($x);    # pic could be defined if this is a form
            $value = '&nbsp;' if $value eq '' || $value =~ m/^\s*$/;
            $print = $print . "<td>$value</td>\n";
        }
        $print .= "</tr>\n";
        paintFoundRec(*print);

    }
    elsif ( $format eq 'PARAGRAPH' ) {
        $print = "<p>$ThisRecordIcons $ThisRecordBasket";
        foreach $x (@ffields) {
            next if $x eq '_delete';
            $value = $pic{$x}
              || &formatField($x);    # pic could be defined if this is a form
            $value = '&nbsp;' if $value eq '' || $value =~ m/^\s*$/;
            $print = $print . "$value,\n";
        }
        chop($print);                 # comma
        paintFoundRec(*print);

    }
    elsif ( $format eq 'LONG' ) {
        $print = $ThisRecordBasket . &wbRecHtml( '', *ffields );
        paintFoundRec(*print);

    }
    elsif ( $format eq 'NAMEVAL' ) {
        $print = $ThisRecordBasket . &wbRecHtml( 'NAMEVAL', *ffields );

        #	} elsif ($WBB{"format;$format"}) {
        #		$x = $WBB{"format;$format"};
        #		$print = &evalSafe($x);

    }
    else {

        $print = &evalSafe($format);

    }

    return $print;
}

# specially paint words that were searched for
# carefull ... only paint outside HTML tags!

sub paintFoundRec {

    local (*print) = @_;

    # $Log .= "Painting for $SearchPatts\n";
    return if $SearchPatts eq '';
    return unless $CGI{'paint'} eq '1' || $CGI{'paint'} eq 'on';

    #			  /	  /				   /
    eval { $print =~ s/>[^<]+/&paintFoundText($&)/gei };
    return;
}

# specially paint words

sub paintFoundText {

    local ($x) = @_;

    #$Log .= "value='$x' patt='$SearchPatts'\n";

    $x =~ s/$SearchPatts/<span class="FOUND">$&<\/span>/gi;
    return $x;
}

# outputs name/value pairs in a table
# input: name1,name2,name3,name4,value1,value2,....

sub formatNameVal {

    local ($v0) = ( $#_ + 1 ) / 2;
    local ( $i, $out );

    return '&nbsp;' if $v0 == 0;    # empty table kills validation

    $out = '<table cellspacing="0">' . "\n";
    $v   = $v0;
    for ( $i = 0 ; $i < $v0 ; $i++ ) {
        $out .= qq(<tr valign="top"><th align="right">$_[$i]</th><td>$_[$v]</td></tr>\n);
        $v++;
    }

    $out .= "</table>\n";

    return $out;
}

# set variable to display the OPEN icon

sub formatOpenIcon {

    local ($size) = $_[0];
    local ($hits) = $ThisRecordMatches;
    local ($num);

    if ( $#sWords >= 0 ) {
        $num = ( $hits - $Xtra ) / ( $#sWords + 1 );
    }
    elsif ($then) {
        $num = ( $sort{$id} - $then ) / ( $now - $then );
    }
    else {
        $num = 1;
    }

    $iconpar = $ICONPAR;
    if ( $size eq 'SMALL' ) {
        $iconpar = $SMALLICONPAR;
    }

    $aa = 1;
    $aa = 2 if $num > .20;
    $aa = 3 if $num > .50;
    $aa = 4 if $num > .74;

    $proc = int( $num * 100 + .5 ) . '%';

    return <<EOM;    #UK
<a href="$this"><img src="$ICONURL/$ICONOPEN$aa.gif" $iconpar alt="$proc; open" title="$proc; open"></a> 
EOM

}

# converts html to ascii text

# was using $* for multiline matching.
# $* is obsolete in Perl 5.10
# quickly replaced by /m in regexp for multiline matching - to be checked

sub formatHtm2Txt {    # only used by cgiAdmAgentProcess ??

    local ($page) = @_;

    local ( $row, $tag, $pad, $content, $x, @rows );

    $asciiPage = '';

    @rows = split( /</m, $page );
    foreach $row (@rows) {
        $row =~ tr/\n\r/  /;          # can't do multiline here
        $row =~ m/^([^>]+)>(.*)$/;    # can't do multiline here
        $tag     = "<$1>";
        $content = $2;
        $content =~ s/\&nbsp\;/ /gm;
        $content =~ s/\&gt\;/>/gm;
        $content =~ s/\&lt\;/</gm;
        $content =~ s/\s+/ /gm;        # multiple space
        $content =~ s/^\s//gm;         # leading space
        $content =~ s/\s$/ /gm;        # trailing space

        if ( $tag =~ /<title>|<option>|<\!\-\-/im ) {
            next;
        }
        else {
        }
        if ( $tag =~ m/<p|<tr|<h1|<h2|<h3|<h4|<ol|<ul/im ) {
            $pad = "\n\n";             # these \n\n will be removed in cgiAdmAgentProcess
        }
        elsif ( $tag =~ m/<li/m ) {
            $pad = "\n\n[] ";          # these \n\n will be removed in cgiAdmAgentProcess
        }
        elsif ( $tag =~ m/<br/m ) {
            $pad = "\n";
        }
        else {
            $pad = "";
        }
        $asciiPage .= "$pad$content" if $pad || $content;
    }
    $asciiPage =~ s/ +/ /gm;
    return $asciiPage;
}

# prints XML header

sub printXmlHead {
    $e = $WBB{'intlCharset'} || 'UTF-8';
    print $HTTP200OK;
    print "Content-type: text/xml\n\n";
    print <<EOM;
<?xml version="1.0" encoding="$e" standalone="yes" ?>
EOM
}

# prints standard XML action reply

sub printXmlAction {

    $result = $_[0];

    print <<EOM;
<!DOCTYPE request>
<request name="$Page">
<in>
EOM

    foreach $key ( keys(%CGI) ) {
        next if $key eq '_isindex';
        print "<$key>$CGI{$key}</$key>\n";
    }

    print <<EOM;
</in>
<out>$result</out>
</request>
EOM
}

# gathers interface information

sub exportService {

    local ( $name, $in, $ret, $doc ) = @_;
    $ServiceIn{$name}  = $in;
    $ServiceDoc{$name} = &htQuote($doc);
    $ServiceRet{$name} = $ret || 'request';
}

# ----------------------------------------------

# nothing wrong, these calls to exportService are not in a sub
# used by cgiHtmlApi and cgiXmlApi

exportService( 'XmlApi', '', '', 'Returns XML API to this Web service.' );

exportService(
    'Show', 'format,_id,fields',
    'rec',  'Returns fields "fields" of record with id="_id". Format must be XML! "fields" is a comma separated list of field names.'
);

exportService(
    'Search',
    'search,format,fields',
    'records',
    'Returns "fields" of records matching the "search" criteria. 
    Format must be equal to XML. search parameter has the same syntax as in the interactive Search request.'
);

exportService( 'Xput', '', '',
'Inserts name value pairs into the database. For example ..../Xput?name=Tom&amp;surname=Jones&amp;_id=45 would add Tom Jones as record with id=45.'
);

exportService( 'XputMany', '', '', 'Inserts the POSTed .csv file of content-type text/csv into the database.' );

exportService( 'Xget', 'id', '', 'Gets a complete record with given id in XMLish format.' );

# return information for API in XML

sub cgiXmlApi {

    printXmlHead();
    $res = '';
    foreach $key ( sort( keys(%ServiceIn) ) ) {
        @ins = split( /,/, $ServiceIn{$key} );
        $url = "$SCRIPTBASE$SCRIPT/$key?";

        #		foreach $in (@ins) {
        #		$url .= "$in=value&";
        #		}
        chop($url);
        $url = &htQuote($url);
        $res .= <<EOM;
<requestApi name="$key">
<url>$url</url>
<method>GET|POST</method>
<parameterEncoding>URLencoded</parameterEncoding>
EOM
        foreach $in (@ins) {
            $res .= "<parameter>$in</parameter>\n";
        }
        $res .= <<EOM;
<result>$ServiceRet{$key}</result>
<documentation>$ServiceDoc{$key}</documentation>
<seeAlso>$SCRIPTBASE$SCRIPT/HtmlApi</seeAlso>
</requestApi>
EOM
    }
    printXmlAction($res);
}

# return information for API in HTML

sub cgiHtmlApi {

    printHead("Application interface of $WBB{'dbTitle'} service via HTTP");
    print <<EOM;
<p>See also the machine readable <a href="$SCRIPTBASE$SCRIPT/XmlApi">XmlApi</a>
</p>
EOM

    $res = '';
    foreach $key ( sort( keys(%ServiceIn) ) ) {
        @ins = split( /,/, $ServiceIn{$key} );
        $url = "$SCRIPTBASE$SCRIPT/$key?";

        #		foreach $in (@ins) {
        #		$url .= "$in=value&";
        #		}
        chop($url);
        $url = &htQuote($url);
        $res .= <<EOM;
<h3>Request $key</h3><pre>
url: $url
method: GET|POST
parameter encoding: URLencoded
EOM
        foreach $in (@ins) {
            $res .= "parameter: $in\n";
        }
        $res .= <<EOM;
result:$ServiceRet{$key}
documentation:</pre> $ServiceDoc{$key}
EOM
    }
    print $res;
}

# converts a record into an XML string

sub rec2xml {

    local (@fields) = @_;
    local ($ret);

    $id  = &htQuote( $rec{'_id'} );
    $ret = "<rec _id=\"$id\">\n";

    foreach $n (@fields) {
        $v = $rec{$n};
        $v = &htQuote($v);
        $ret .= "<$n>$v</$n>\n";
    }

    return "$ret</rec>\n";
}

# writes a record as an easy to parse XML

sub rec2xmle {

    local (@fields) = @_;
    local ($ret);

    $id  = &htQuote( $rec{'_id'} );
    $ret = "<rec _id=\"$id\">\n";
    $ret .= <<EOM;
<!-- Below are name=value pairs, one per line.						-->
<!-- To decode values tr char 034 with 012							-->
<!-- Ignore any lines that do not begin with letter,digit,_underscore -->
<!-- This is not in XML spirit but easy to parse					  -->
<!-- use format=XML if you need proper XML							-->
<!-- note that ]] is not escaped									  -->
<![CDATA[
EOM
    foreach $n (@fields) {

        $v = $rec{$n};
        $v =~ tr/\012/034/;
        $ret .= "$n=$v\n";
    }

    return "$ret]]></rec>\n";
}

# which of the fields entered as parameters are visible
# if no parameter is given, all visible fields are returned

sub visibleFields {

    local (@fields) = @_;

    wbSortFields();
    if ( $#fields < 0 ) {    # no parameters
        @fields = @PFields;
    }

    @ret = ();
    foreach $f (@fields) {
        push( @ret, $f ) unless $Hidden{$f};
    }

    return @ret;
}

# shortcut to httpAny, fetches a page

sub httpPost {

    httpAny( 'POST', @_ );

}

# shortcut to httpAny, fetches a page

sub httpGet {

    httpAny( 'GET', $_[0] );
}

# fetches the page under the URL
# returns status,header,page
# does not handle redirects etc.

sub httpAny {

    local ( $method, $url, $payload, $payloadType ) = @_;
    local ( $host, $dir, $them, $port );

    # --- parse URL

    #print "parameter=$_";

    $url =~ m|http://(.*)|;
    $url = $1;
    ( $host, $dir ) = split( /\//, $url, 2 );
    $dir = "/$dir";
    ( $them, $port ) = split( /:/, $host, 2 );
    $port = 80 unless $port;

    $cl = length($payload);
    $payloadType = $payloadType || 'text/plain';
    logTime("HTTP $method $cl bytes with http://$host:$port$dir");

    # --- setup sockets

    wbOpenS( $them, $port );

    print S <<EOM;
$method $dir HTTP/1.0
Host:$host
User-agent: WODA $VERSION
Woda-end-user: $RA
Content-type: $payloadType
Content-length: $cl

EOM
    print S $payload;

    @page = <S>;
    close(S);

    $header = '';
    $page   = '';
    while (@page) {
        $line = shift(@page);
        if ( $line eq "\n" || $line eq "\r\n" ) {
            $page = join( '', @page );
            @page = ();
        }
        else {
            $header .= $line;
        }
    }

    $header =~ m/\s(\d+)/;
    $status = $1;
    $page = '' unless $status eq '200';

    # wbError might have printed an error as well

    $page = '' if $header =~ m/\nWoda-error:/;
    $page =~ s/\r//g;    # get rid of carriage returns
    $Log .= $header if $page eq '';
    logTime("HTTP-DONE - status $status");
    return ( $status, $header, $page );
}

# determines paramaters how to use a table to run from
# command line
# returns executable file and definition to add to
# PATH_INFO part
# not to be used any longer!!!

sub wbPrepareCommandLine {

    local ($table) = @_;

    $cgiPrg = &wbTables( 'DEF', $table );
    unless ($cgiPrg) {
        return ( '', "ERROR - Table $table not known" );
    }

    # is it in fact a text file ?
    if ( $cgiPrg =~ m|$DATADIRDIR/([^/]+)/_datadict/dd| ) {
        $cgiDef = $1;
        $cgiPrg = $SCRIPTFILE;
    }
    elsif ( $cgiPrg =~ m|$TEMPLATEDIR/structures/| ) {
        $cgiDef = $';
        $cgiPrg = $SCRIPTFILE;
    }
    else {
        $cgiDef = '';
    }

    return ( $cgiPrg, $cgiDef );
}

# fetches $url and returns the part of page
# between A name=PAGEBODYN tags
# or whole page if that is not found
#
# this function may not work on some
# windows httpd servers (IIE, Xitami)

sub FETCH {

    local ($url) = @_;

    undef $ok;
    undef $headers;
    undef $page;

    ( $ok, $headers, $page ) = &httpGet($url);

    if ( $ok ne '200' ) {

        $headers =~ m/\n/;
        return $`;    # first row returned!

    }
    else {

        # local ($*); obsolete in Perl 5.10
        local ( $h, $b, $f ) = split( /\<a name="PAGEBODY".\>/m, $page );    # with multiline (/m)
        return $b || $page;
    }
}

# KEY ($id,$table)
# is there a record with id=$id in
# table $table ?

sub KEY {

    local ( $id, $table ) = @_;

    local (%t);

    if ( &wbTables( 'DIR', $table ) ) {
        return &recExists( $id, $table ) >= 0;
    }
    else {
        %t = &GET( $table, $id );
        return $t{_id} ne '';
    }
}

# IDS ($table)
# return array of all IDs in a given table
# does it through ROWS because ROWS can do http

sub IDS {

    local ($table) = @_;
    local ( @x, $x );

    $x = &ROWS( $table, '_id' );
    @x = split( /\n/, $x );

    return @x;
}

# get a value of field $name of record $id from table $table
# ($table,$id,$name)
# $table must be defined in $WBB{'tables'}

sub FLD {
    local ( $table, $id, $name ) = @_;
    local ( $dataDir, $dataFile );
    local $_;

    if ( $VALid eq "$table/$id" ) {

        # cache !

    }
    else {

        %VALrec = &GET( $table, $id );

    }

    return $VALrec{$name};
}

# like recRead but works with any table, even remote one
# returns hash
# includes ALL fields, including password

sub GET {
    local ( $table, $id ) = @_;
    local ( %rec, $url );

    if ( &wbTables( 'DIR', $table ) ) {

        recRead( &wbFixId($id), $table );

    }
    elsif ( $url = &wbTables( 'URL', $table ) ) {

        $id   = &wbFixId($id);
        $page = &FETCH("$url/Xget?_id=$id");
        @x    = split( /\n/, $page );
        @x    = grep( /^[^\<]/, @x );          # trash XML headers and footers

        # decode like rec2xmle

        foreach $x (@x) {
            ( $n, $v ) = split( /=/, $x, 2 );
            $v =~ tr/\034/\012/;
            $rec{$n} = $v;
        }

    }
    else {

        wbFail( "$table not known", "Table $table is neither local not defined in WBB{'tables'}" );

    }

    return %rec;
}

# puts several records from this table to another table
# $search and $then are same as in QRY

sub PUTN {

    local ( $othertable, $search, $then ) = @_;

    $data = &QRY( '', $search, $then, '', 'CSV' );
    $url = &wbTables( $othertable, 'URL' );
    $url = "$url/XputMany";
    ( $status, $headers, $page ) = &httpPost( $url, $data, 'text/csv' );
}

# puts %rec into table $table

sub PUT {

    local ($table) = @_;

    if ( &wbTables( 'DIR', $table ) ) {

        $Log .= "PUT into local $table\n";
        recWrite($table);

    }
    elsif ( $url = &wbTables( 'URL', $table ) ) {

        $Log .= "PUT into remote $table\n";

        local ($x);
        $x = '';
        foreach $field ( keys(%rec) ) {
            $x .= $field . '=' . &htEscape( $rec{$field} ) . '&';
        }

        chop($x) if $x;    # delete last &

        ( $status, $headers, $page ) = &httpPost( "$url/Xput", $x );
        if ( $status ne '200' ) {
            wbFail( "http POST returned $status", $headers );
        }

    }
    else {

        wbFail( "$table not known", "Table $table is neither local nor defined in WBB{'tables'}" );

    }

    return 1;
}

# decodes the value from an OPTION or LINKOPTION list
# so that full text is visible
#
# usage:
#	&DECODE() in the picture setting of a field
#	   it is assumned that the value is in $_ and the field is $field
#	$short ... returns decoded value only!

sub DECODE {

    local ( $fld, $val, $short ) = @_;

    if ( $#_ == -1 ) {
        $fld = $field;
        $val = $_;
    }

    if ( defined $Decode{ $fld, $val } ) {

        # ok ... we know it
    }
    else {
        @os = &wbParseOptions($fld);
        for ( $i = 0 ; $i <= $#os ; $i++ ) {
            $o = $os[$i];
            $t = $optionText[$i];
            $Decode{ $fld, $o } = $t;
        }
    }

    $d = $Decode{ $fld, $val };
    if ( $d ne '' ) {
        if ($short) {
            return $d;
        }
        else {
            return "$d ($val)";
        }
    }
    else {
        return $val;
    }
}

# return a link to a record with id=$id in table $table.
# if $text is supplied, that text is linking, otherwise
# the $id is linking.
#
# if $id is not specified, the value in $_ is
# assumed to hold the id
# typically used in 'picture' field definitions
# to make pointers to a related table
#
# usage:
#	WBB{$x,picture}='&LINK("people",$_,"lastName");

sub LINK {

    local ( $table, $id, $text ) = @_;
    local ($_) = $_;
    local ( $ddir, $durl );

    $id   = $id   || $_;
    $text = $text || $id;
    $id   = &wbFixId($id);
    $durl = &wbTables( 'URL', $table );

    return <<EOM;
<a href="$durl/Show?$id">$text</a>
EOM
}

# provides a decently formated related data from another table
# is fast - uses ROWS function
# may provide a small toolbar too
# WARNING - automatically adds $_id as first field
# only the table and search parameters need to be specified

sub RELATED {

    local ( $table, $flds, $search ) = @_;

    $Log .= "RELATED $table,$flds,$search\n";

    local ( $x, $out, $ddir, $durl );
    $durl = &wbTables( 'URL', $table );

    if ( $flds ne '' ) {
        $flds = "_id,$flds";
    }

    $x = &ROWS( $table, $flds, $search );

    @x = split( /\n/, $x );
    $out = '';

    $i = 0;
    foreach $x (@x) {
        $i++;
        $out .= "<tr>\n";
        @y = split( /\t/, $x );
        $c = $#y;
        if ( $flds eq '' ) {
            $c = 2;
        }
        $y[0] = qq(<a href="$durl/Show?$y[0]">$i</a>);
        $out .= '<td class="DATA1">' . join( '</td><td>', @y[ 0 .. $c ] ) . '</td></tr>';
    }

    if ($out) {
        $out = <<EOM;    #UK
<table>$out</table>
... <a href="$durl/Search?search=$search&amp;noform=1">details</a>
EOM
    }

    return $out;
}

# get raw data (tab separated) from $table
# similar to QRY function with $format=RAW
# $flds are comma separated field names
# does not take into account any setting of table $table
# unless table is empty and the function works on the
# local table
# ($table,$flds,$search,$then,$sort,$format,$first,$max)

sub ROWS {
    local ( $table, $flds, $search, $then, $sort, $format, $first, $max ) = @_;

    # make sure we are up to date!

    $ddir = &wbTables( 'DIR', $table );
    $format = $format || 'DEFAULT';
    $tblfile = "$format.tbl";
    $tblfile =~ tr/[A-Z]/[a-z]/;
    $tblfile = "$ddir/_cache/$tblfile";
    $max = 9999999999 || $max;

    # if local tbl file is outdated or non-existent it must be created
    # only local table can do this

    if ( $ddir ne '' && &wbExpired( $tblfile, $ddir ) ) {
        return &QRY( $table, $search, $then, $sort, 'RAW', $first, $max, $flds );
    }

    if ( $ddir eq '' ) {

        # QRY does the caching!
        return &QRY( $table, $search, $then, $sort, 'RAW', $first, $max, $flds );
    }

    # protect these variables

    local ( %rec, %data, %hits, $allHits, $id, $sex, @list, @fields, @fieldNames, @editable, @y );

    ( $allHits, *list, *fields ) = &recSelect( $table, $search, $sort, 'RAW', $flds );

    return ( join( "\n", @list ) ) . "\n" if @list;
    return '';
}

# This is obsolete!!!
#
# Performs a search in table $table
# and returns results formatted in HTML
# parameters correspond to CGI parameters
# of standard or advanced search
#
# Fails on some Windows Web servers such as IIE
# that don't let you call another Perl program.
#
# Run Debug info from Admin menu to check if it
# has a chance of working.

sub OLDQRY {
    local ( $table, $search, $then, $sort, $format, $first, $max, $flds ) = @_;
    local ( $ddir, $cgiPrg, $queryString, $dir, $prog, @x );

    ( $cgiPrg, $cgiDef ) = &wbPrepareCommandLine($table);

    if ( $cgiPrg eq '' ) {
        $Log .= "ERROR: QRY in $table - cgiPrg eq ''\n";
        return $cgiDef;    # error
    }

    $ddir = &wbTables( 'DIR', $table );

    # must have access3.txt
    # so that when command line is executed
    # the SCRIPT environment variable is known

    unless ( -s "$ddir/_cache/access3.txt" ) {
        $durl = &wbTables( 'URL', $table );
        wbFail( "$table not initialized", "Please <a href=\"$durl\">click here</a> to initialize table $table." );
    }

    if ( $search eq '' && $then eq '' ) {
        $cache = 1;    # cache it all ... always expensive
    }
    else {
        $cache = 1;
    }

    $_ = &htEscape($search);
    $queryString .= "search=$_&" if $_;

    $_ = &htEscape($then);
    $queryString .= "then=$_&" if $_;

    $_ = &htEscape($sort);
    $queryString .= "sort=$_&" if $_;

    $_ = &htEscape($format);
    $queryString .= "format=$_&" if $_;

    $_ = &htEscape($first);
    $queryString .= "first=$_&" if $_;

    $_ = &htEscape($max);
    $queryString .= "max=$_&" if $_;

    $_ = &htEscape($flds);
    $queryString .= "fields=$_&";

    chop($queryString);    # remove the &

    $cgiPrg =~ m|(.+)/([^/]+)$|;

    $dir  = $1;
    $prog = $2;

    ### cacheing comes here ###

    wbCheckDir("$DATADIR/_qry");

    ### clean the cache ###

    wbQryClean();

    ### this is the cache file ###

    $cacheFile = "$DATADIR/_qry/$table-$Group-$queryString.txt";

    ### now do it ###

    if ( &wbExpired( $cacheFile, $ddir ) ) {

        # go to data dir then run it
        $command = "$PERL $prog -x $cgiDef/Xsearch '$queryString'";
        if ($UNIX) {
        }
        else {
            $command =~ tr/'/"/;

            #$command =~ s/\%/\%\%/g;
        }
        logTime("Executing in $dir: $command");
        chdir($dir);
        $reply = `$command`;
        if ( $cache && length($reply) < 100000 ) {

            # need to write even if no reply ... meaning no hits!
            chdir($DATADIR);
            open( qry, ">$cacheFile" );
            binmode(qry);
            print qry $reply;
            close(qry);
        }
    }
    else {
        $Log .= "Using cached search";
        open( qry, $cacheFile );
        @x = <qry>;
        close(qry);
        $reply = join( '', @x );
    }

    logTime("Done");

    return $reply;
}

# returns URL of a given service

sub URL {

    return &wbTables( 'URL', @_ );

}

# Does the same as OLDQRY but uses HTTP to get the data

sub QRY {
    local ( $table, $search, $then, $sort, $format, $first, $max, $flds ) = @_;
    local ( $ddir, $cgiPrg, $queryString, $dir, $prog, @x );

    $ddir = &wbTables( 'DIR', $table );
    $url  = &wbTables( 'URL', $table );

    if ( $url eq '' ) {
        $Log .= "ERROR: QRY in $table - url unknown\n";
        return '';    # error
    }

    $cache = 0;
    $cache = 1 if $search ne '';
    $cache = 1 if $then ne '';
    $cache = 0 if $ddir eq '';     # table on remote server!

    $cache = 0 unless $WBB{searchCache};    # jan.2006

    $_ = &htEscape($search);
    $queryString .= "search=$_&" if $_;

    $_ = &htEscape($then);
    $queryString .= "then=$_&" if $_;

    $_ = &htEscape($sort);
    $queryString .= "sort=$_&" if $_;

    $_ = &htEscape($format);
    $queryString .= "format=$_&" if $_;

    $_ = &htEscape($first);
    $queryString .= "first=$_&" if $_;

    $_ = &htEscape($max);
    $queryString .= "max=$_&" if $_;

    $_ = &htEscape($flds);
    $queryString .= "fields=$_&";

    chop($queryString);    # remove the trailing &

    ### cacheing ###

    if ($cache) {
        wbCheckDir("$DATADIR/_qry");
        wbQryClean();
        $cacheFile = "$DATADIR/_qry/$table-$Group-$queryString.txt";
    }
    else {
        $cacheFile = '';
    }

    ### now do it ###

    if ( &wbExpired( $cacheFile, $ddir ) ) {
        $url = "$url/Xsearch?$queryString";
        ( $ok, $headers, $reply ) = &httpGet($url);
        if ( $cache && length($reply) < 100000 ) {

            # need to write even if no reply ... meaning no hits!
            chdir($DATADIR);
            open( qry, ">$cacheFile" );
            binmode(qry);
            print qry $reply;
            close(qry);

            # mf
            # in the real world too many things cause the file to be empty! #UK
            # however, it could be a valid search meaning no hits, so delete the file if $reply is empty #UK
            unlink $cacheFile unless $reply;
        }

    }
    else {
        $Log .= "Using cached search";
        open( qry, $cacheFile );
        @x = <qry>;
        close(qry);
        $reply = join( '', @x );
    }

    logTime("Done QRY");

    return $reply;
}

# return the URL of the /Home page

sub homeURL {

    if ( $WBB{'homePage'} ) {
        if ( $WBB{'homePage'} =~ m|/| ) {
            $homeURL = $WBB{'homePage'};
        }
        else {
            $homeURL = "$WBB{'htmlURL'}/$WBB{'homePage'}";
        }
    }
    else {
        $homeURL = "$SCRIPT/Home";
    }

    return $homeURL;
}

# print minimalistic header

sub printMiniHead {

    $MINIHEAD = 1;
    printHead(@_);
}

# print head for framed pages

sub printFrameHead {

    $MINIHEAD      = 1;
    $FRAMEHEAD     = 0;
    $TargetDefault = 'target="_parent"';

    printHead(@_);
}

# forwards request to another URL and exit

sub printLocation {

    local ( $location, $cookie ) = @_;

    print $cookie;
    print "Location: $location\n\n";
    wodaExit();
}

# show CSS

sub cgiColorSchemeShow {

    printHead("Show Color Scheme");
    printStyleSample();
}

# print style from WBB{'style'} # only works if 'style' contains CSS

sub printStyleSample {

    print <<EOM;
<table border="0" width="100%">
<tr>
<th>Style name</th>
<th>plain text</th>
<th>visited link</th>
<th>not visited link</th>
</tr>
EOM

    $r = rand(10000);

    @class = split( /\t\./, $WBB{'style'} );
    foreach $class (@class) {
        $class =~ m/^([A-Za-z0-9]*)/;
        $dclass = $1;
        print <<EOM;
<tr CLASS=$dclass>
<th>$dclass</th>
<td>some text</td>
<td><a href="http://www.yahoo.com/">visited link</a></td>
<td><a href="gopher://www.yahoo.com/$r">not visited link</a></td>
</tr>
EOM
    }

    print "</table>";
}

# print everything on top of page

sub printHead {

    local ( $pgTitle, $cookie ) = @_;
    local ( $homeURL, $dbTitle, $apTitle, $scriptURL, $p );

    if ( $Page eq 'Changed' && $WBB{'title'} ne '' ) {

        # title field would appear in header
        $pgTitle = $WBB{"title;$Page"} || $pgTitle;
    }
    else {
        $pgTitle = $CGI{'title'} || $WBB{"title;$Page"} || $pgTitle;
    }

    $dbTitle = $WBB{'dbTitle'};
    $apTitle = &evalMaybe( $WBB{'apTitle'} );

    # globals HTMLFILE ... don't print HTTP headers!
    # MINIHEAD ... just prints a simple title

    # make sure no funny stuff gets out on DOS

    binmode($STDOUT);

    unless ($HTMLFILE) {
        if ( $WBB{'intlCharset'} ne '' ) {
            $chs = "; charset=$WBB{'intlCharset'}";
        }
        else {
            $chs = '';
        }
        print $HTTP200OK . "Content-type: text/html$chs\n";
        print $cookie;
        print $DemandedCookies;
        print "\n";

        $Log .= "Cookies:\n";
        $Log .= $cookie;
        $Log .= $DemandedCookies;

        $HttpHeadPrinted = 1;
    }

    $tt = "$apTitle$dbTitle";
    $tt = &htUntag($tt);
    $tt =~ tr/\n/ /;

    $t = "$apTitle$dbTitle : $pgTitle";
    $t = &htUntag($t);
    $t =~ tr/\n/ /;

    $dt = $WBB{doctype} || '<!-- no doctype setting defined -->';

    print <<EOM;
$dt
<html>
<head>
<title>$t</title>
<meta name="GENERATOR" content="$POWERGEN">
EOM

    if ( &wbAllowed( $Group, 'Search' ) ) {
        print <<EOM;
<link rel="search" title="Search $tt" href="$SCRIPTBASE$SCRIPT/SearchForm">
<link rel="alternate" title="$tt RSS feed" href="$SCRIPTBASE$SCRIPT/Search?format=RSS&amp;sort=TIME&amp;max=10" type="application/rss+xml">
EOM
    }

    if ($chs) {
        print <<EOM;
<meta http-equiv="Content-Type" content="text/html$chs">
EOM
    }

    print qq(<base href="$SCRIPTURL">\n) if $HTMLFILE;
    print qq(<base $TargetDefault>\n)    if $TargetDefault;
    print $MetaFields;

    # print "<link rel=\"StyleSheet\" href=\"/woda/icons/woda.css\">
    # this is the default style sheet, there is no need to change it here
    # just write your own definition into $WBB{'style'} and it will
    # override it

    if ( $WBB{'style'} ) {
        $style = <<EOM;
<style type="text/css">
<!-- 
$WBB{'style'}
--></style>
EOM
    }
    else {

        $style = &defaultStyle();

    }

    if ( $WBB{'font'} ) {
        $style =~ s/font-family: [^;]*/font-family: $WBB{'font'}/g;
    }

    print $style;
    print $WBB{'script'} || "<!-- WBB{'script'} empty -->\n";
    print $WBB{'head'}   || "<!-- WBB{'head'} empty -->\n";
    print "</head>\n\n";

    $HtmlHeadPrinted = 1;

    # body

    $pb = $WBB{'pageBody'} || $styleBody;

    $pb = $WBB{'frameBody'}
      || 'STYLE="background-color: #CCC;" LEFTMARGIN=0 TOPMARGIN=0'
      if $FRAMEHEAD;

    print "<body onLoad=\"self.focus();\" $pb>\n<!-- body parameters came from WBB{'pageBody'} -->\n";

    if ($MINIHEAD) {
        if ($pgTitle) {
            print '<div class="MINITITLE">', $pgTitle, '</div>';
            printTrap();
        }

        print '<div class="CORE">', "\n";
        print &evalMaybe( $WBB{"head;$Page"} )
          || "<!-- WBB{head;$Page} would be here -->\n";

        return;    # ------------------------------>

    }

    $loginBox = &formatLoginBox() unless $Page =~ /^Login|^Pass/;
    print &evalMaybe( $WBB{'pageHead'} )
      || "<!-- WBB{'pageHead'} not defined -->";

    $x = time;
    printTrap();

    $homeURL     = &homeURL();
    $scriptURL   = $SCRIPT;
    $dbTitleLink = qq(<a href="$homeURL" $TargetParent>$dbTitle</a>);

    $phi = $WBB{'pageHeadIcon'} || "<!-- WBB{'pageHeadIcon'} -->";

    if ($pgTitle) {
        if ( $WBB{'pageTitle'} ) {
            $x = &evalSafe( $WBB{'pageTitle'} );
            print $x;
        }
        else {
            $x = "";
            print <<EOM;
<!-- WBB{'pageTitle'} not defined -->
<div class="DBTITLE">
$phi $apTitle $dbTitleLink
</div>
<div class="TITLE">
<h1>$pgTitle</h1>
</div>
EOM
        }

        print &evalMaybe( $WBB{"head;$Page"} )
          || "<!-- WBB{head;$Page} would be here -->\n";

    }

    print '<div class="CORE">';

}

# removes tags

sub htUntag {

    local ($_) = @_;

    s/<[^>]*//g;
    s/>//g;

    return $_;

}

# print the end of each page

sub printFoot {

    $PageCount = &wbAccessCount();
    wbLog();
    logTime("Done");

    return unless $HtmlHeadPrinted;

    # we did print HtmlHead ...

    print &evalMaybe( $WBB{"foot;$Page"} )
      || "\n<!-- WBB{foot;$Page} would be here -->";

    # Tracking device
    #	 if ($PageCount =~ m/^50*$/) {
    # 		print qq(<img alt="" width="1" height="1" src="http://www.ddatabase.com/register.png?c=$PageCount&amp;h=$SCRIPTBASE$SCRIPT">);
    #	 }
    print "\n</div><!-- closing div for CORE -->\n";

    #
    # $Page specific stuff for toolbar
    #

    if ( $WBB{'toolbar'} ne '' ) {
        @x = split( /\n/, $WBB{'toolbar'} );
        foreach $x (@x) {
            ( $patt, $string ) = split( /\s+/, $x, 2 );
            if ( $Page =~ m/$patt/ ) {
                my $x = &evalSafe($string) || $string;
                push( @Toolbar, $x );
            }
        }
    }

    #
    # prepare system stuff for toolbar
    #

    if ( !$MINIHEAD ) {

        $homeURL = &homeURL();

        if ( $Group eq 'admin' || $Group eq 'guest' || $Group eq 'friend' ) {
            $uicon = "u$Group";
        }
        else {
            $uicon = "uother";
        }
        local ( $hh, $w, $u );
        if ($User) {
            $us = "user $User from group $Group";    #UK
        }
        else {
            $us = "belong to group $Group";          #UK
        }

        $hh   = "Home page of this database";          #UK
        $hht  = "home";                                #UK
        $w    = "Web Oriented Database Home";
        $u    = "logout (you are $us)";                #UK
        $u    = "login" if $Group eq 'guest';          #UK
        $ah   = "ADMIN'S HOME";                        #UK
        $lurl = "$SCRIPT/LoginOrRegister?$Referer";    #UK

        push( @Toolbar, <<EOM) if &wbAllowed( $Group, 'Home' ) || $homeURL !~ m/Home$/;
<a href="$homeURL" $TargetParent><img $ICONPAR src="$ICONURL/home.gif" alt="$hh" title="$hht"></a>
EOM
        push( @Toolbar, <<EOM) if $Group eq 'admin';
<a href="$SCRIPT/AdmMenu" $TargetParent><img $ICONPAR src="$ICONURL/ahome.gif" alt="$ah" title="$ah"></a>
EOM

        push( @Toolbar, <<EOM) if ( !$NOLOGICON ) && &wbAllowed( $Group, 'LoginOrRegister' );
<a href="$lurl"><img $ICONPAR src="$ICONURL/$uicon.gif" alt="$u" title="$u"></a>
EOM

        $toolbarText = &evalMaybe( $WBB{"toolbarText"} )
          || "$WBB{'dbTitle'}<!-- WBB{'toolbarText'} -->";
        $woda = <<EOM;
<a href="$POWERURL" $TargetParent><img class="ICON" src="$POWERIMG" alt="$POWERTEXT" title="$POWERTEXT" $POWERSIZE></a>
EOM

    }    # MINIHEAD ?

    if ( !$MINIHEAD && $WBB{'pageEnd'} ) {

        #toolbarText
        #woda
        $toolbar = join( '', @Toolbar );

        print &evalMaybe( $WBB{'pageEnd'} );

    }
    elsif ( $#Toolbar >= 0 ) {

        # convert toolbar into a table ...

        $toolbar = join( qq(<img src="$ICONURL/pixel.gif" alt="">), @Toolbar );

        # get rid of whitespace

        $toolbar =~ s/>\n/>/g;
        $toolbar =~ s/>/\n>/g;

        # print the toolbar

        $cols = $toolbar =~ s/<td/<td/gi;
        $cols += 2;

        # prevent icons

        print <<EOM;
<!-- WBB{'pageEnd'} not defined -->
<div class="TOOLBAR">
<table width="100%"><tr>
<td align="left">$toolbar</td> 
<td align="center">$toolbarText</td>
<td align="right">$woda</td>
</tr></table>
</div>
EOM
    }

    if ($HtmlHeadPrinted) {
        print &evalMaybe( $WBB{'pageFoot'} )
          || "\n<!-- WBB{'pageFoot'} would be here -->";
        printTrap();
        print '</body></html>';
        print <<EOM if $PRINTLOG;
<!--  
++ ++ ++
WODA LOG
++ ++ ++
$Log
++ ++ ++
-->
EOM

    }

}

# formats the login box
#change: change class to LOGINBOX, force left hand alignment of table cells for input fields # mf
sub formatLoginBox {

    local ($al) = @_;

    local ( $a, $t, $c, $w );

    return '' unless $WBB{'userTable'};
    return '' if $User !~ m/^_anon/;
    return '' if $Group eq 'admin';
    return '' if $Group eq 'friend';

    if ( $User =~ m/^_anon/ ) {

        $a = $WBB{'userTable'};
        $url = &wbTables( 'URL', $a );

        $Referer =~ m/referer\=/;
        $r = $';
        $r = &htUnescape($r);

        $x1 = 'Username';    #UK
        $x2 = 'Password';    #UK
        $x3 = 'Login';       #UK
        $x4 = 'New user';    #UK

        return <<EOM;
<form action="$SCRIPT/Login" method="post">
<table class="BOX" cellspacing="0" border="0" $al>
<tr><td align="right">$x1:</td><td><input name="groupname" size="8"></td></tr>
<tr><td align="right">$x2:</td><td><input name="password" size="8" type="password"></td></tr>
<tr><td COLSPAN=2><input type="hidden" name="referer" value="$r">
<input class="BUTTON" type="submit" value="$x3" name="login">
<input class="BUTTON" type="submit" value="$x4" name="new">
</td></tr>
</table>
</form>
EOM
    }
    else {

        # not an anon user
        return '';
    }
}

# --- print the robot trap

sub printTrap {

    print qq(<a href="$SCRIPT/Trap"> </a>);

}

# --- returns default style

sub defaultStyle {

    local ( $bgcPage, $ink, $above, $border, $font, $lines, $error, $warning );

    # --- key parameters ---

    $fontsize = '10pt';
    $font     = 'Arial,Helvetica,sans-serif';

    $bodycolor      = 'black';
    $bodybackground = 'white';

    $headfootbackground = '#8DF';
    $headfootcolor      = '#004';
    $headfootborder     = "1px solid black";

    $abovebackground = '#EEE';
    $abovecolor      = $bodycolor;
    $aboveborder     = "1px solid grey";

    $buttonbackground = '#DDD';
    $buttoncolor      = '#440';
    $buttonborder     = "1px outset $buttonbackground";

    $linkcolor       = 'blue';
    $focusbackground = $headfootbackground;

    $modifycss = $WBB{'styleModify'} || '/* WBB{styleModify} not defined */';

    return <<EOM;
<!-- because WBB{'style'} not defined default sheet is used -->
<style type="text/css">
<!-- 
	/* '*' selector dangerous e.g. A on black background ... */
	/* note that uppercase is used for all tags and classes !!! */


	BODY,LI,P,DD,TD,TH {
		color: $bodycolor;
		font-size: $fontsize;
		font-family: $font;
	 }

	BODY {
		background-color: $bodybackground;
		margin:5px;
		padding:5px;
	}
	
	TT,PRE,CODE,SAMP {
			color: $bodycolor;
		font-size:$fontsize;
		font-family:monospace;
	}

	H1,H2,H3,H4,H5 { font-weight: bold }

	H1 { font-size: 20pt; }
	H2 { font-size: 16pt; }
	H3 { font-size: 14pt; }
	H4 { font-size: 12pt; }
	H5 { font-size: 10pt; }

	/* default anchors */

	A, A:active {color: $linkcolor;}
	A {xx-text-decoration: none}

	/* focus */

	A:hover, INPUT:focus, TEXTAREA:focus, SELECT:focus {
		background-color: $focusbackground;
	}

	A>IMG.ICON {
		-moz-opacity: 0.6;
		filter:alpha(opacity=60);
		border: 1px outset $buttonbackground;
		background-color:$buttonbackground;
	}

	A>IMG.ICON:hover {
		border: 1px inset $buttonbackground;
		filter:alpha(opacity=100);
		-moz-opacity: 1;
	}

/* elements at top and end of page */

	.DBTITLE, .TITLE, .MINITITLE, .TOOLBAR { 
		color: $headfootcolor;
		border: $headfootborder;
		background-color: $headfootbackground;
		width: 100%;
	}

	.TOOLBAR { 
		/* position:fixed; */
		/* bottom:0; left:0; */ 
	}


	.MENU {
		border: $aboveborder;
		background-color: $abovebackground;
		padding: 5px;
		margin: 0px;
		padding-bottom:2px;

	}

	DIV.DBTITLE A {
		text-decoration: none;
		color: $headfootcolor;
	}

	.TITLE {
		font-size: 20pt;
		margin-bottom: 10px;
		margin-top: 1px;
	}
	
	DIV.TITLE H1 {
		margin: 2px;
	}

	.MINITITLE {
		font-size: 14pt;
	}

	.CORE {
		/* in this div is the rest of the page */
		margin-top: 10px;
		margin-bottom: 10px;
	}
	
	.SEARCHTITLES {
		/* stuff on search results pages before results */
	}
	
	.SEARCHREFORMAT {
		/* links to reformat search results */
		display: none;
	}

/* some small stuff */

	.HIDDEN { display: none; }
	.NOBORDER { border: 0px none white !important; }
	.ERROR	{ color: red; }
	.WARN	 { color: orange; }
	.FOUND	 { background-color: $focusbackground; }
	.HINT   { font-size: 8pt; color: #666; }
	.SMALL	{ font-size: 8pt; }

/* menu */

	DIV.MENU BR {
		display: none;
	}

	DIV.MENU A {
		display:block;
	}

	DIV.MENU A, .BUTTON {
		border: $buttonborder;
		color: $buttoncolor;
		   background-color: $buttonbackground;
		text-align: left;
		text-decoration: none;
		padding: 1px;
		margin-top: 1px;
		margin-bottom: 1px;
	}

	DIV.MENU A:hover, .BUTTON:hover, A.BUTTON:hover {
		border: 1px inset $buttonbackground; 
		   background-color: $buttonbackground;
		text-decoration: none;
	}

	DIV.MENU A:active, .BUTTON:active, A.BUTTON:active {
		border: 1px inset $abovebackground; 
		   background-color:$abovebackground;
		text-decoration: none;
	}

	.MENUT {
		display:block;
		border: 0px solid $abovebackground;	/* same size as anchor */
		text-align: left;
		vertical-align: top;
		padding-top: 2px;
		margin-top: 0px;
	}

/* form elements */

	INPUT, TEXTAREA, SELECT {
		border: 1px inset $abovebackground;
	}

/* forms */

	.BOX, .SEARCHBOX, .FORM, DIV.RECORDS TABLE {
		background-color: $abovebackground;
		border: $aboveborder;
		vertical-align: middle;
		padding:4px;
	}

	FORM.FORM TD, FORM.FORM TH,
	DIV.RECORDS TD, DIV.RECORDS TH {	
		border-bottom: dotted 1px $bodycolor;
		padding:4px;
	}

	TABLE.BOX { /* cells need padding, TABLE not */
		padding:0px;
	}

/* boxes are similar */

	.BOX, .ERRBOX, .WARNBOX {
		margin-left:20%;
		margin-right:20%;
		width: 60%;
		text-align: center;
	}

	.ERRBOX {
		background-color: red;
		border: outset 1px red;
		color: black;
	}

	.WARNBOX {
		background-color: orange;
		border: outset 1px orange;
		color: black;
	}

/* tables for records */

	.DATATABLE {
		xx-background-color: $buttonbackground;
	}

	.DATAH  { 
		background-color: $abovebackground;
		color: $abovecolor;
	 }

	.DATA	 { background-color: $bodybackground }
	.DATA1	 { background-color: #DDD }
	.DATA0	  { background-color: #EEE }

/* end of WODA default */

	$modifycss	

-->
</style>
EOM
}

# --- items in cells ---

sub formatMenuItems {

    @stuff = @_;
    local ( $text, $url );
    local ($r);

    $r = '';

    #	$r = "<table><tr><td>\n";
    for ( $i = 0 ; $i < $#_ ; $i += 2 ) {
        $url  = $_[ $i + 1 ];
        $text = $_[$i];
        $url  = "$SCRIPT/$url" unless ( $url =~ m|^/| || $url =~ m|//| );
        next if $url =~ m/$SCRIPT/ && &wbDeniedURL( $Group, $url );
        next if $text eq '';

        #$text = 'all' if $text eq '';
        if ($url) {
            $r .= "<a href=\"$url\">$text</a><br>\n";
        }
        else {
            $r .= "$text<br>\n";
        }
    }

    #	$r .= "</td></tr></table>";
    $r .= "\n\n";
    return $r;
}

# --- title of a cell

sub formatMenuTitle {

    local ( $text, $url ) = @_;
    local ($r);

    $r = '';
    if ($url) {
        $r .= "<a href=\"$url\"><b>$text</b></a>:<br>\n";
    }
    else {
        $r .= '<span class="MENUT"><b>' . $text . ':</b></span><br>' . "\n";
    }
    return $r;

}

# print bad password warning

sub wbBadPass {

    local ($id) = @_;

    $id = &htEscape($id);

    local ( $t, $x );
    $t = "Bad password";    #UK

    $x = <<EOM;             #UK
Password does not match the password saved with the information.
You are not allowed to conclude this operation.
EOM

    $x .= <<EOM;            #UK
<p>
If an email address is stored with that information
it may be possible to
<a href="$SCRIPT/PassR1?_id=$id">
restore the password</a>.
EOM

    wbError( $t, $x );

}

# copy data from CGI variables to %rec

sub wbCgiToRec {

    local ($tmpID) = @_;

    undef %rec;

    foreach $field ( keys(%WBF) ) {

        # userFile-old stores the old file name in editing USERFILE typed fields

        if ( $CGI{"$field-old"} && !$CGI{$field} ) {
            $rec{$field} = $CGI{"$field-old"};
        }

        # skip those containing spaces only

        next if $CGI{$field} =~ m/^\s*$/;

        # the rest

        if ( $WBF{ $field, 'type' } eq 'DATE' ) {
            $y           = $field . "_year";
            $m           = $field . "_month";
            $rec{$field} = "$CGI{$y}/$CGI{$m}/$CGI{$field}";

        }
        elsif ($WBF{ $field, 'type' } eq 'USERFILE'
            || $WBF{ $field, 'type' } eq 'IMAGE' )
        {
            $rec{$field} = &wbUserFileSave( $field, $tmpID );

        }
        else {

            $rec{$field} = $CGI{$field};
        }
    }
}

#
# create cookie headers for sticky values
#

sub wbCookieCGI {

    local ( $out, $v, $stick );

    $out = '';

    foreach $x ( keys(%CGI) ) {
        next if ( $stick = $WBF{ $x, 'sticky' } ) eq '';

        if ( $stick =~ m|/| ) {
            $stick = "path=$stick;";
        }
        else {
            $stick = '';
        }
        $v = &htEscape( $CGI{$x} );
        next if $v eq '';    # make sure all cookies have values

        # $v =~ s|\"|\\\"|g;	# escape quotes, hope \n are ok

        $out .= "Set-Cookie: $x=$v; $stick expires=expires=Wed, 1-Jan-2020 23:59:59 GMT\n";
    }

    return $out;
}

#
# used for Windows because NT does not have crypt funtion
#

sub wbCryptPasswd {

    local ($what) = @_;
    local ($salt);

    unless ($NO_CRYPT) {

        # built in encryption
        $salt = '$1$wb' . substr( $what, 0, 2 );
        return crypt( $what, $salt );
    }
    else {

        # hide only
        $what =~ s/(.)/substr(log(ord($1)),2,5)/eg;
        $what =~ s/\.//g;
        return $what;
    }
}

# returns list of files in that dir
# does not prepend dir !!!
# &glob(*.pdf,timestamp)

sub glob {

    $_ = $_[0];
    s/\./\\\./g;
    s/\*/\.\*/g;

    # print "--- $_ ----\n";
    return &wbDir( $_, $_[1] );
}

# get files with pattern $patt younger than $then
# $patt is perl regexp

sub wbDir {

    local ( $patt, $then ) = @_;
    local ( @x, @y, @xx, $ddir );

    if ( $patt =~ m|/| ) {
        $patt =~ m|(.+)/([^/]+)$|;
        $dir  = $1;
        $patt = $2;
    }
    else {
        $dir = '.';
    }

    opendir( D, $dir );
    @x = readdir(D);
    closedir(D);
    @y = grep ( /$patt/, @x );

    if ( $then > 0 ) {
        undef @x;
        foreach $_ (@y) {
            @xx = stat("$dir/$_");
            next if $xx[9] < $then;
            push( @x, $_ );
        }
        return @x;
    }
    else {
        return @y;
    }
}

# did $file expire according to format
# definitions in SCRIPTFILE or elsewhere?

sub wbExpiredDD {

    local ($file) = @_;
    local ( @x, @y, @z );

    @x = stat($file);
    @y = stat($SCRIPTFILE);

    return 1 if $y[9] > $x[9];

    return 0;

}

# did $file expire according to data in $dataDir
# tolerance in hours is in $exp
# returns modification date of index file
# or 0 if it is up to date

sub wbExpired {

    local ( $file, $dataDir, $exp ) = @_;
    local ( $now, @x, @y, @z );

    return 1 if $file eq '';

    $exp     = $WBB{'cacheExpire'} if $exp     eq '';
    $exp     = 0                   if $exp     eq '';
    $dataDir = $DATADIR            if $dataDir eq '';

    return 1 if !-e $file;    # expired if it does not exist

    # any .rec files in the $DATADIR ?

    @x = stat($file);         # _cache file
    @y = stat($dataDir);      # _fat file

    if ( ( $y[9] - $x[9] ) > $exp * 3600 || !$UNIX ) {
        opendir( D, $dataDir );
        @z = readdir(D);
        closedir(D);
    }

    if ( grep( /[^_]\.rec$/, @z ) ) {
        appendRecFiles($dataDir);
    }

    # expired if empty database !
    if ( !-e "$dataDir/_data/fat.txt" ) {
        &evalSafe( $WBB{'onExpire'} );
        return 1;
    }

    # if onExpire is defined, we assume it is
    # expired regardless of the date on the
    # status of the fat.txt because fat.txt could be
    # out of date !

    if ( $WBB{'onExpire'} ne '' && time - $x[9] > $exp * 3600 ) {
        &evalSafe( $WBB{'onExpire'} );
        return $x[9];
    }

    # reread the fat.txt

    @y = stat("$dataDir/_data/fat.txt");

    # finally compare to the fat.txt

    if ( ( $y[9] - $x[9] ) > $exp * 3600 ) {
        &evalSafe( $WBB{'onExpire'} );
        return $x[9];
    }
    else {
        return 0;
    }
}

# do this after every change of the data in the table

sub wbDatabaseChanged {

    if ( $WBB{'afterTableModify'} ne '' ) {
        &evalSafe( $WBB{'afterTableModify'} );
    }

    #	open (hhh,">$DATADIR/_cache/modtime.dir") || &wbFail ("Can't open $x");
    #	print hhh "This file's date is set to when the data was last changed!";
    #	close(hhh);
}

# do this after every change of the data in a record

sub wbRecordChanged {

    if ( $WBB{'afterRecordModify'} ne '' ) {
        &evalSafe( $WBB{'afterRecordModify'} );
    }
}

# export the database

sub wbExport {

    local ( $sort, $format, $fields ) = @_;
    print "Content-type: text/plain\n\n";    ### unless $HTMLFILE;

    # BUG ... if called from command line $HTMLFILE will be
    # set and these lines will not be printed ... but they
    # are expected by the routine that uses the stuff

    ( $allHits, *records, *fields ) = &recSelect( '', '', $sort, $format, $fields );

    print &formatFoundHeader( $format, *fields );
    foreach $r (@records) {
        print $r;
        print "\n" if $format =~ m/^CSV/;

        # in searches, newline is added alsewhere
    }
    print &formatFoundFooter( $format, *fields );
}

# fatal error, print to screen and log error

sub wbFail {

    local ( $title, $more ) = @_;
    ( $package, $filename, $line ) = caller;

    $more = <<EOM;    # UK
<p>Cannot continue because of an error in $filename in line $line.
<p><b>$more</b>
<p>Please report this to the manager of this database: $WBB{'managerEmail'}
EOM

    $Err = 'FA';
    wbError( "FATAL ERROR: $title", $more, $caller );    # UK
}

# displays a window
# returns a table formatted by CSS .BOX

sub formatWindow {

    local ( $title, $body, $class ) = @_;

    $class = 'BOX'    if $class eq '';
    $body  = '&nbsp;' if $body  eq '';
    $title = '&nbsp;' if $title eq '';
    return <<EOM;
<p>
<table CLASS=$class cellspacing="2">
<tr><td CLASS=$class>
<h3>$title</h3>
</td></tr>
<tr><td CLASS=$class>
$body
</td></tr>
</table>
EOM
}

# terminates with error

sub wbError {

    local ( $title, $message, $caller, $ba ) = @_;

    $ba = -1 if $ba eq '';

    return if $IGNOREERRORS;

    select($STDOUT) if $STDOUT;

    if ( !$HttpHeadPrinted ) {

        # need header like Pragma:error=true
        # need to check this if woda is client!

        # do mainDefaults(); # so that we know where the icons are!
        printMiniHead( "", "Woda-error:error\n" );
        print "<br><br><br><br>";
    }

    $back = "Go back";    #UK
    $body = <<EOM;
<p>$message
<p>$caller
<p>
<form name="buttonbar" action="">
<input class="BUTTON" type=BUTTON value="$back" onClick="history.go($ba)">
</form>
EOM

    print &formatWindow( $title, $body, 'ERRBOX' );

    $ErrMsg = "$title:$message";
    $Err = $Err || 'ER';

    $ICONURL = $ICONURL || '/icons/default';

    printFoot();
    wodaExit();
}

# only prints a warning in a box

sub wbWarn {

    local ( $title, $message ) = @_;

    if ( !$HttpHeadPrinted ) {
        printHead('');
    }

    print &formatWindow( $title, $message, 'WARNBOX' );
}

# prints a warning in a box

sub wbFixId {
    local ($id) = @_;
    local ( $lcFrom, $lcTo, $asFrom, $asTo );

    ( $lcFrom, $lcTo ) = split( / /, $WBB{'intlLower'} );
    ( $asFrom, $asTo ) = split( / /, $WBB{'intlAscii'} );

    if ( $lcFrom ne '' ) {
        $_ = $id;
        eval "tr/$lcFrom/$lcTo/";
        eval "tr/$asFrom/$asTo/" if $asFrom;
        $id = $_;
    }
    else {

        #	    $id =~ tr|šðèæžŠÐÈÆŽ|sdcczSDCCZ|;
    }

    $id =~ tr/[A-Z]/[a-z]/;
    $id =~ tr| /\t\n|_|;
    $id =~ tr/\\()<>,;|"${}:\.\&\?/_/;    # does not allowed in ID!
    $id =~ tr/[\200-\377]/_/;

    return $id;
}

# sets HELP in cookie

sub cgiSetHelp {

    $help    = $CGI{'help'};
    $referer = $CGI{'referer'};

    $COOKIE{_conf} =~ s/:NOHELP//;
    $COOKIE{_conf} =~ s/:HELP//;

    if ( $help eq 'on' ) {
        $COOKIE{_conf} .= ':HELP';
    }
    else {
        $COOKIE{_conf} .= ':NOHELP';
    }

    $cookie = &htFormatCookie( '_conf', $COOKIE{_conf} );

    if ($referer) {
        printLocation( $referer, $cookie );
    }
    else {
        printMiniHead( "Set help $help", $cookie );
        if ( $JS >= 1 ) {
            print <<EOM;    #UK
<p> You can now
<form>
<input class="BUTTON" type=BUTTON value="go back" onClick="history.go(-1);">
</form>
<p>and reload that page.
EOM
        }
        else {
            print <<EOM;    #UK
<p> Please go back and reload for the changes to take effect.
EOM

        }
    }

    $NOLOGICON = 1;
}

# returns two JS functions handy in forms

sub formatFormJavaScript {

    local ($fname) = @_;    # name of the form

    return '' unless $JS;

    $fname = $fname || 'main';

    return <<EOM;
<script LANGUAGE="JavaScript" type="text/javascript"><!--

// check all or none of the checkboxes
function allBoxes(name,n) {
	for (var i=0; i<document.forms['$fname'].elements[name].length; i++) document.forms['$fname'].elements[name][i].checked=n;
}

// validate form element if it is a regular expression
function validText(name,expression,message) {

	myRe = expression;
	if (myRe.test(name.value)) {
	return
	} else {
	alert(message)
	}
}	


// --></script>
EOM
}

# generates javascript code the check all boxes in $field

sub formatCheckAllJavaScript {

    local ($field) = @_;

    return '' unless $JS;

    local ( $alltext, $nonetext );

    $alltext  = "check all";     #UK
    $nonetext = "check none";    #UK
    return <<EOM;
<a href="javascript:allBoxes('$field',1)"><img src="$ICONURL/check.gif" $ICONPAR alt="$alltext" title="$alltext"></a>&nbsp;
<a href="javascript:allBoxes('$field',0)"><img src="$ICONURL/uncheck.gif" $ICONPAR alt="$nonetext" title="$nonetext"></a>
EOM
}

# return all between <form> tags
# use data in %rec as default values

sub wbForm {

    my ( $out, $l );

    if ( $COOKIE{_conf} =~ m/:NOHELP/ ) {
        $Help = '';
    }
    else {
        $Help = 1;
    }

    $out = &formatFormJavaScript();

    $help    = "Help";              #UK
    $helpOff = "Show less help";    #UK
    $helpOn  = "Show more help";    #UK
    $helpAll = "Display manual";    #UK

    $out .= <<EOM;
<table cellspacing="0" width="100%">
<tr valign="top"><th align="right">&nbsp;</th>
EOM
    $referer = $Referer;
    if ($Help) {
        $out .= <<EOM;
<td><a class="BUTTON" href="$SCRIPT/SetHelp?help=off&$referer" target="_self">$helpOff</a>
EOM
    }
    else {
        $out .= <<EOM;
<td><a class="BUTTON" href="$SCRIPT/SetHelp?help=on&$referer" target="_self">$helpOn</a>
EOM
    }

    if ( &wbAllowed( $Group, 'DisplayStructure' ) ) {
        $out .= <<EOM;
<a class="BUTTON" href="$SCRIPT/DisplayStructure" target="woda_help">$helpAll</a></td></tr>
EOM
    }
    wbSortFields();
    foreach $field (@Fields) {
        wbFormFieldItems($field);
    }

    if ( $WBB{'form'} ne '' ) {
        $out .= "<!-- using WBB{'form'} -->\n";
        $out .= "<tr><td COLSPAN=2>\n";
        $out .= &evalSafe( $WBB{'form'} );
        $out .= "</td></tr>\n";
    }
    else {

        $out .= "<!-- not using WBB{'form'} -->\n";

        foreach $field (@Fields) {

            # OK for hidden polje ???

            next
              if ( $Form{ $field, 'name' } eq ''
                && $Form{ $field, 'input' } eq '' );

            if ( $Form{ $field, 'input' } eq '' ) {    # to izpisuje 'BREAK'
                $t = $Form{ $field, 'name' };
                $t =~ s/^"//;
                $t =~ s/"$//;

                $out .= <<EOM;
<tr valign="top">
<td COLSPAN=2>
<b>$t</b>
</td>
</tr>
EOM
            }
            else {

                $Form{ $field, 'prompt' } .= '<br>'
                  if $Form{ $field, 'prompt' };
                $Form{ $field, 'input' } .= '<br>' if $Form{ $field, 'help' };
                $out .= <<EOM;
<!-- ----------------------------- -->
<tr valign="top">
<th align="right">
$Form{$field,'name'}
$Form{$field,'err'}
</th>
<td align="left">
$Form{$field,'prompt'}
$Form{$field,'input'}
$Form{$field,'help'}
</td>
</tr>

EOM
            }
        }
    }

    $WBB{'recordLock'} = 'password|user' if $WBB{'recordLock'} eq '';

    @locks = split( /\|/, "$WBB{'recordLock'}|" );
    @locks = grep( /./, @locks );

    if ( $#locks == -1 || $#locks == 0 && $locks[0] eq 'none' ) {

        # no locking !

    }
    else {

        $x{password} = "with this password:";                #UK
        $x{user}     = "to user $User";                      #UK
        $x{group}    = "to group $Group";                    #UK
        $x{host}     = "to host $RA";                        #UK
        $x{none}     = "don't lock - anyone can edit it";    #UK

        $y{user}  = "u:";
        $y{group} = "g:";
        $y{host}  = "h:";

        if ( $WBB{recordLock} eq 'user' ) {                  # only one choice!

            $out .= qq(<input type="HIDDEN" name=_lock value="user">\n);

        }
        elsif ( $WBB{recordLock} eq 'group' ) {              # only one choice!

            $out .= qq(<input  type="HIDDEN" name=_lock value="group">\n);

        }
        else {
            $x = 'locking:';                                 #UK
            $out .= <<EOM;
<tr valign="top">
<th align="right">$x</th><td>
EOM

            $out .= <<EOM;                                   #UK
How would you like to lock this record?<br>
EOM

            $out .= '<select name="_lock">' . "\n";
            $na = 0;
            foreach $lock (@locks) {

                if ( $lock eq 'user' || $lock eq 'group' ) {
                    if ( $Group eq 'guest' ) {
                        $x = 'no locking - you must log in to lock this record';
                        $out .= qq(<option value="none" SELECTED>$x\n);
                        last;
                    }
                }
                $x = $x{$lock};

                $d = '';
                $d = 'SELECTED'
                  if $y{$lock} && $rec{'_password'} =~ m/^$y{$lock}/;
                $d = 'SELECTED'
                  if $lock eq 'password' && defined $CGI{'_accessPassword'};

                $na++;
                $out .= qq(<option value="$lock" $d>$x\n) if $x;

            }

            $out .= "</select>\n";

            if ( $na eq 0 ) {
                wbWarn( "Warning:", "Locking of this record will not be possible for an anonymous user. Please log in." );    #UK
            }

            if ( grep( /password/, @locks ) ) {

                if ( &isLoginTable() && &isOwner() ) {

                    # default
                    $CGI{'_accessPassword'} = $LogPass;
                }

                # keep value in _password field if new record
                $CGI{_accessPassword} = $CGI{_password} if ( !$CGI{_accessPassword} );

                $x = 'again';    #UK
                $out .= <<EOM;
<input type="password" name="_password" size="10" value="$CGI{'_accessPassword'}">
$x
<input type="password" name="_password1" size="10" value="$CGI{'_accessPassword'}">
EOM
                $out .= <<EOM;    #UK
<br><span class="HINT">Same password should be entered twice to avoid errors.</span>
EOM

                if ( &isLoginTable() ) {
                    $out .= <<EOM;    #UK
<br><span class="WARN"><b>
This password will also be used
for logging in into this application.
</b></span>
EOM
                }
            }
            $out .= "</td></tr>\n";
        }

    }

    $out .= "</table>\n";

    # print length($out);
    return $out;
}

# sets $Form{$field,'name'}
#	  $Form{$field,'prompt'}
#	  $Form{$field,'input'}
#	  $Form{$field,'help'}
#	  $Form{$field,'err'}

sub wbFormFieldItems {

    local ($field) = $_[0];
    local ( $onClick, $onFocus, $onBlur );

    $Form{ $field, 'name' }   = '';
    $Form{ $field, 'prompt' } = '';
    $Form{ $field, 'input' }  = '';
    $Form{ $field, 'help' }   = '';
    $Form{ $field, 'err' }    = '';

    #	 new record		adds set			 to us
    if (   $CGI{'_id'} eq ''
        && $WBF{ $field, 'adds' }
        && $Group =~ /admin|$WBF{$field,'adds'}/ )
    {

        # ok
    }
    else {
        return unless $Group =~ /admin|$WBF{$field,'modifies'}/;
    }

    if ( $CGI{'fields'} ne '' ) {

        # ne vem ce lahko dam to v klic, morda mora $Form postaviti kot se ssika
        return unless "$CGI{'fields'}" =~ m/\b$field\b/;
    }

    if ( $WBF{ $field, 'type' } eq 'BREAK' ) {
        $Form{ $field, 'name' } = $WBF{$field};
        return;
    }

    if ( $WBF{ $field, 'type' } eq 'SUBMIT' ) {
        return;
    }

    if ( $WBF{ $field, 'type' } eq 'COMPUTE' ) {
        return;
    }

    $hidden = 0;
    $hidden = 1 if $WBF{ $field, 'typePar' } =~ m/HIDDEN/i;

    $niceField = $WBF{ $field, 'head' };
    unless ($niceField) {
        $niceField = $field;
        $niceField =~ s/([A-Z]+)/ $1/g;
        $niceField =~ tr/[A-Z]/[a-z]/;
    }
    $niceField = "$niceField:";
    $niceField = '' if $hidden;

    $validIfText = "valid if";    #UK

    # begin ...

    $Form{ $field, 'err' } = '<img src="' . $ICONURL . '/del.gif" alt="Error" title="Error" ' . $SMALLICONPAR . '>'
      if $err{$field} == 1;

    $Form{ $field, 'name' } = $niceField;
    $Form{ $field, 'prompt' } = $WBF{ $field, 'p' } if $Help && !$hidden;

    $alert = '';

    $st = 'HINT';
    $st = 'ERROR' if $err{$field};
    if ( $WBF{ $field, 'cond' } ) {
        $alert = $WBF{ $field, 'cond' };
        $Form{ $field, 'prompt' } .= " <span CLASS=$st>($WBF{$field,'cond'})</span>"
          if $Help && !$continue && !$hidden;
    }
    else {
        $alert = "$validIfText: $WBF{$field}" if $WBF{$field} ne "1;";
        $Form{ $field, 'prompt' } .= " <span CLASS=$st>($validIfText: $WBF{$field})</span>"
          if $Help && $WBF{$field} ne "1;" && !$continue && !$hidden;
    }

    # is this field controlled?

####### 	if ($WBF{$field,'type'} eq 'EMAIL' || $WBF{$field} ne '1;'){ # && $WBF{$field} !~ m/\^\$/) {
    if ( $WBF{ $field, 'type' } eq 'EMAIL' || $WBF{ $field, 'required' } ) {    # MF necessary to show if admin says so
        $optional = '';
        $Form{ $field, 'name' } = '<span class="ERROR">*</span> ' . $Form{ $field, 'name' }
          unless $hidden;
    }
    else {
        $optional = 'class="DATA0" ';

        # does not work for checkboxes ...
    }

    if ( $JS >= 1.2 && $WBF{$field} =~ m|^m(/.*/[a-z]*$)| ) {
        $onChange = <<EOM;
onChange="validText(this,$1,'$field: $alert')"
EOM
    }
    else {
        $onChange = '';
    }

    if ( $JS && $status !~ m/<\(/ ) {    # weed out HTML !
        $status = "$field: $WBF{$field,'p'} ... $alert";
        $status =~ s/\n/ /g;
        $status =~ tr/\"\'\<\>/ /;

        #do wbFail($status);
        $status = $onClick = $onFocus = $onBlur = '';

        #		$onClick = qq(onClick = "self.status='$status'");
        #		$onFocus = qq(onFocus = "self.status='$status'");
        #		$onBlur =  qq(onBlur  = "self.status=''");
    }
    else {
        $status = $onClick = $onFocus = $onBlur = '';
    }

    #$Log .= "Default $field=$rec{$field}\n";

    $default = $rec{$field};
    $default = $CGI{$field} if $default eq '';
    $default = $COOKIE{$field} if $default eq '' && $WBF{ $field, 'sticky' };
    $default = $COOKIE{_email}
      if $default eq '' && $WBF{ $field, 'type' } eq 'EMAIL';

    # if ($default eq '') {
    if ( $default eq '' && $rec{_id} eq '' ) {    # claudio # mf
        $default = $WBF{ $field, 'd' };
        $default = &evalMaybe($default);
    }
    $default = &htQuote($default);
    $typePar = $WBF{ $field, 'typePar' };

    #	$typePar = "$typePar $optional";
    # $Log .= "$field - $typePar\n";

    if ( $typePar =~ s/CONTINUE// ) {
        $continue = 1;
    }
    else {
        $continue = 0;
    }

    if ( $WBF{ $field, 'type' } =~ m/^TEXTAREA|HTMLAREA|LIST$/ ) {
        $typePar = 'rows="8" cols="50"' if $WBF{ $field, 'typePar' } eq '';

        #		$typePar = "$typePar $optional";

        # width poglihati, heigth nastimati

        if (   $HtmlAreaWorks
            && $WBF{ $field, 'type' } eq 'HTMLAREA'
            && $WBF{ $field, 'typePar' } =~ m/RICH/ )
        {
            $typePar =~ s/COLS/XCOLS/;    # fixed width!
        }

        $Form{ $field, 'input' } = <<EOM;
<textarea name="$field" ID="$field" $typePar $optional wrap="virtual"
$onChange $onFocus $onBlur
>$default</textarea>
EOM

        if (   $HtmlAreaWorks
            && $WBF{ $field, 'type' } eq 'HTMLAREA'
            && $WBF{ $field, 'typePar' } =~ m/RICH/ )
        {
            $Form{ $field, 'input' } .= <<EOM;
<script type="text/javascript" DEFER="1">
	var config = new HTMLArea.Config();
	config.pageStyle = "body { background-color: #fff; font-family: serif; }";
	config.editorURL = '$WODAURL/htmlarea/';
	config.width = "604";
	config.height = "auto";
	config.statusBar = false;
	config.toolbar = [ [ 
		"formatblock", "separator", "bold", "italic", "underline", 
		"separator", "justifyleft", "justifycenter", "justifyright", "justifyfull", "separator",
		"insertorderedlist", "insertunorderedlist", "separator", "outdent", "indent", 
		"subscript", "superscript", "separator", "createlink",
		"htmlmode", "separator", "popupeditor", "space", "about" ]
	];
	HTMLArea.replace('$field',config);
</script>
EOM
        }

    }
    elsif ( $WBF{ $field, 'type' } =~ m/^OPTION$|^LINKOPTION$/ ) {

        @options = &wbParseOptions($field);

        # optionText also set
        $no = 0;

        $default = "\n$default\n";

        if ( $WBF{ $field, 'typePar' } =~ m/COLS=(\d+)/i ) {
            $cols = $1;
        }
        elsif ( $#options > 0 ) {
            $cols = int( 50. / ( length( join( '1234567', @options ) ) / ( $#options + 1 ) ) );
        }
        else {
            $cols = 1 if $#options < 3;
        }
        undef @x;

        if ( $WBF{ $field, 'typePar' } =~ m/HEADING=(.{1,5})/i ) {    # take the 5 first letters from HEADING
            $heading = $1;
        }
        else {
            $heading = '';
        }

        $typePar = 'RADIO'
          if ( $WBF{ $field, 'typePar' } eq '' && $#options < 2 );

        # as checkbox

        if ( $typePar =~ m/CHECKBOX/ ) {
            foreach $o (@options) {
                $okey = "($o)";
                $okey = '' if $WBF{ $field, 'typePar' } =~ m/NOID/;
                $optionText[$no] = "$o" if $optionText[$no] eq '';
                $t = $optionText[ $no++ ];
                $t .= ' <span class="HINT">' . $okey . '</span>' if $okey;

                $o   = &htQuote($o);     # default is htQuoted above
                $reo = &regexpEsc($o);

                if ( $heading && $o =~ m/^$heading/ ) {    # no checkbox ... prohibited
                    push( @x, "<b>$t</b>\n" );
                    next;
                }

                if ( $default =~ /\n$reo\n/ ) {
                    push( @x, qq(<input type="checkbox" name="$field" value="$o" checked $onClick>$t\n) );
                }
                else {
                    push( @x, qq(<input type="checkbox" name="$field" value="$o" $onClick>$t\n) );
                }
            }
            $alltext  = "check all";                       #UK
            $nonetext = "check none";                      #UK
            $allhtm   = <<EOM;
<a href="javascript:allBoxes('$field',1)"><img src="$ICONURL/check.gif" $ICONPAR alt="$alltext" title="$alltext"></a>
<a href="javascript:allBoxes('$field',0)"><img src="$ICONURL/uncheck.gif" $ICONPAR alt="$nonetext" title="$nonetext"></a>
EOM
            push( @x, $allhtm ) if $JS && $#options > 5;
            $Form{ $field, 'input' } .= &formatColumns( $cols, @x );

            # as radio

        }
        elsif ( $typePar =~ m/RADIO/ ) {
            foreach $o (@options) {
                $okey = "($o)";
                $okey = '' if $WBF{ $field, 'typePar' } =~ m/NOID/;
                $optionText[$no] = "$o" if $optionText[$no] eq '';
                $t = $optionText[ $no++ ];
                $t .= ' <span class="HINT">' . $okey . '</span>' if $okey;

                $o   = &htQuote($o);     # default is htQuoted above
                $reo = &regexpEsc($o);

                if ( $heading && $o =~ m/^$heading/ ) {    # no checkbox ... prohibited
                    push( @x, "<b>$t</b>\n" );
                    next;
                }

                if ( $default =~ /\n$reo\n/ ) {
                    push( @x, qq(<input type="radio" name="$field" value="$o" checked $onClick>$t\n) );
                }
                else {
                    push( @x, qq(<input type="radio" name="$field" value="$o" $onClick>$t\n) );
                }
            }
            $Form{ $field, 'input' } .= &formatColumns( $cols, @x );

            # as pulldowns

        }
        else {
            $Form{ $field, 'input' } .= "<select name=\"$field\" $typePar $onClick $optional>\n";
            foreach $o (@options) {
                $okey = "($o)";
                $okey = '' if $WBF{ $field, 'typePar' } =~ m/NOID/;
                $okey = "$o" if $optionText[$no] eq '';
                $t    = "$optionText[$no++] $okey";
                $o    = &htQuote($o);                                 # default is htQuoted above
                $reo  = &regexpEsc($o);

                if ( $default =~ /\n$reo\n/ ) {
                    $Form{ $field, 'input' } .= "<option selected value=\"$o\">$t\n";
                }
                else {
                    if ( $heading && $o =~ m/^$heading/ ) {           # no value ... prohibited - default still selected
                        $t = "* $t";
                        $o = '';
                        $Form{ $field, 'input' } .= "<option>$t\n";
                    }
                    else {

                        $Form{ $field, 'input' } .= "<option value=\"$o\">$t\n";
                    }
                }
            }
            $Form{ $field, 'input' } .= "</select>";
        }

    }
    elsif ( $WBF{ $field, 'type' } eq "FILE" ) {
        $typePar = "size=20";
        $Form{ $field, 'input' } .= qq(<input type="input" name="$field" value="$default" $typePar $onFocus $onBlur $optional>\n);

    }
    elsif ($WBF{ $field, 'type' } eq "USERFILE"
        || $WBF{ $field, 'type' } eq "IMAGE" )
    {

        $HasAttachmentField = 1;

        if ($default) {
            $keep = "$field exists; to delete uncheck the checkbox; to replace submit a file below";    #UK
            $keep = '' unless $Help;
            $Form{ $field, 'input' } .= qq(<input type="checkbox" name="$field-old" value="$default" CHECKED><span class="HINT">$keep</span><br>\n);
        }

        $typePar = "size=30" if !$typePar;
        $help = "select a file on <em>your</em> computer to upload to the Web (only works with Netscape Navigator 2+ or Microsoft Explorer 4+)";   #UK
        $help = '' unless $Help;
        $Form{ $field, 'input' } .= qq(<input $onFocus $onBlur type="file" name="$field" $typePar $optional><br><font class="HINT">$help</font>\n);

    }
    elsif ( $WBF{ $field, 'type' } eq "EMAIL" ) {
        $typePar = "size=40" if !$typePar;

        # $default = "$ENV{'REMOTE_USER'}\@$RH" if ! $default;
        $Form{ $field, 'input' } .= qq(<input $onFocus $onBlur type="input" name="$field" value="$default" $typePar $optional $onChange>\n);

    }
    elsif ( $WBF{ $field, 'type' } eq "DATE" ) {
        ( $y, $m, $d ) = split( /\//, $default );
        $ny   = $field . '_year';
        $nm   = $field . '_month';
        $help = "year/month/day";    #UK
        $help = '' unless $Help;
        $Form{ $field, 'input' } .= <<EOM;
<input $typePar $optional name="$ny" value="$y" size="4" $onFocus $onBlur
>/<input $typePar $optional name="$nm" value="$m" size="2" $onFocus $onBlur
>/<input $typePar $optional name="$field" value="$d" size="2" $onFocus $onBlur> $help
EOM

    }
    else {
        $typePar = "size=50" if $WBF{ $field, 'typePar' } eq '';
        if ( $typePar =~ m/HIDDEN/ ) {
            $t = 'hidden';
        }
        else {
            $t = 'text';
        }
        $Form{ $field, 'input' } .= <<EOM;
<input name="$field" type="$t" value="$default" $typePar $optional
	$onChange
	$onFocus
	$onBlur
>
EOM

    }

    # continue

    if ( !$continue ) {
        if ( $Help && $WBF{ $field, 'help' } && !$hidden ) {
            $Form{ $field, 'help' } .= '<span class="HINT">' . $WBF{ $field, 'help' } . '</span>';
        }
    }
}

# parses the $WBF{?,options} values
# returns list of options and sets @optionText to text values

sub wbParseOptions {

    local ($field) = @_;
    local ( @wbp, @x, $i, $o, $t, $alias, $fields, $options );
    undef @optionText;

    $options = $WBF{ $field, 'options' };

    if ( $WBF{ $field, 'type' } eq "LINKOPTION" ) {
        if ( $WBF{ $field, 'into' } =~ /^\&|^\#/ ) {
            $t = &evalSafe( $WBF{ $field, 'into' } );
            @wbp = split( /\n/, $t );
        }
        else {
            @x = &recAll( $WBF{ $field, 'into' } );
            return @x;
        }
    }
    elsif ( $options =~ /^\&|^\#/ ) {
        $t = &evalSafe($options);
        @wbp = split( /\n/, $t );
    }
    elsif ( !( $options =~ /\|/ ) ) {
        @wbp = split( /\n/, $options );    # /n delimited
    }
    else {
        return split( /\|/, $options );    # | delimited
    }

    for ( $i = 0 ; $i <= $#wbp ; $i++ ) {
        ( $o, $t ) = split( /\t/, $wbp[$i], 2 );
        $wbp[$i]        = $o;
        $optionText[$i] = $t;
        $wbp[$i] =~ s/\s+$//;
    }

    return @wbp;
}

# display other 'Friends' who have visited recently

sub cgiFriendsOnline {

    if ( $Group eq 'guest' ) {
        wbError( "Not registered", "We do not provide this information to unregistered users" );    #UK
        return;
    }

    printHead("Recent users");                                                                      #UK
    $now     = time;
    $minutes = $CGI{'minutes'} || 2;
    $secs    = $minutes * 60;
    $lf      = &logFile();
    open( h, $lf );
    while (<h>) {
        @x = split( /\t/, $_ );
        next if $now - $x[2] > $secs;
        $users{ $x[4] }++;
    }
    close(h);

    if ( &wbAllowed( $Group, 'AdmLogfilePrint' ) ) {
        $log = 1;
    }
    else {
        $log = 0;
    }

    $seconds = $minutes * 60;

    foreach $user ( keys(%users) ) {
        $u =~ s/^_//;
        ( $email{$user}, $x ) = &getRegisteredData($u) || 'email unknown';    #UK
        if ($log) {
            $email{$user} .= <<EOM;                                           #UK
 <a href="$SCRIPT/AdmLogfilePrint?last=$seconds&amp;format=short&amp;search=$user">logfile</a>
EOM
        }
    }

    $mform = <<EOM;                                                           #UK
<form action="$SCRIPT/FriendsOnline">
<br>Users in the last <input size="4" name="minutes" value="$minutes"> minutes.
<input class="BUTTON" type="submit" value="Refresh">
</form>
EOM
    print &formatHash( $mform, 0, %email );
}

# Print log file

sub cgiAdmLogfilePrint {

    $lf = $CGI{'logfile'} || &logFile();

    if ( $lf =~ m|$DATADIR/_log/| ) {
        $lf = $';
    }

    if ( $lf =~ m/\.\.|\/\// ) {
        wbError( "Bad logfile", $lf );
        return;
    }

    $search = $CGI{'search'};

    @hnos = ( 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 );
    @hdrs = ( 'stat', 'time', 'when', 'from', 'user', 'group', 'application', 'method', 'page', 'parameters', 'other' );

    if ( $CGI{'refresh'} && $CGI{'refresh'} > 5 ) {
        printHead( "Log file: $&", "Refresh: 5\n" );
    }
    else {
        printHead("Log file: $&");
    }

    @v      = ( 'normal', 'short', 'long', 'count', 'shows', 'searches' );
    @n      = @v;
    $format = $CGI{'format'};
    $field  = $CGI{'field'};
    $o      = &wbFormatOptions( *v, *n, $format );
    $f      = &wbFormatOptions( *hnos, *hdrs, $field );

    @names = ( 'Search:', 'In file:', 'Format:', 'Count field:', 'Refresh after:', 'Show last:', '' );
    @vals = (
        qq(<input name="search" value="$search" size="40">),
        qq(<input name="logfile" value="$lf" size="40"> ... ".*log" for all),
        qq(<select name="format">$o</select>),
        qq(<select name="field">$f</select> if Format is count),
        qq(<input name="refresh" value="$CGI{'refresh'}" size="3"> seconds.),
        qq(<input name="last" value="$CGI{'last'}" size="3"> seconds.),
        qq(<input class="BUTTON" type="submit" name="action" value="OK">)
    );

    $in = qq(<form action="$SCRIPT/AdmLogfilePrint">) . &formatNameVal( @names, @vals ) . qq(</form>);

    if ( $CGI{'refresh'} ) {
        print <<EOM;
<p>
Hits in the last $CGI{'last'} seconds. Screen will refresh every $CGI{'refresh'} seconds.
</p>
EOM
    }
    else {
        print &formatWindow( 'Define log printout settings', $in );
    }

    print "<p><br CLEAR=ALL><p><br>";

    return unless $CGI{'action'} ne '';

    # --- read the robots file

    open( h, "$DATADIR/_log/robots" );
    while ( $x = <h> ) {
        $x =~ m/([^\s]*)$/;
        $robot{$1} = '1';
    }
    close(h);

    opendir( D, "$DATADIR/_log" );
    @f = readdir(D);
    closedir(D);

    @f = grep( !/^\./, @f );
    @f = sort(@f);

    if ( $CGI{'format'} eq 'count' ) {
        admLogCount();
    }
    elsif ( $CGI{'format'} eq 'shows' ) {
        admLogCount('shows');
    }
    elsif ( $CGI{'format'} eq 'searches' ) {
        admLogCount('searches');
    }
    else {
        admLogShow();
    }
}

# show logs

sub admLogShow {

    print '<h3>Log data</h3>';
    print '<table class="DATATABLE" cellspacing="1"><tr class="DATAH" ><th>';

    @x = @hdrs;
    $x[10] = '' unless $format eq 'long';
    $x[1] = $x[5] = $x[6] = $x[7] = $x[9] = $x[10] = '' if $format eq 'short';

    print join( '</th><th align="left">', @x );
    print "</th></tr>\n";

    $i      = 0;
    $n      = 0;
    $search = '.' if $search eq '';

    foreach $f (@f) {
        next unless $f =~ m/$lf/;
        open( h, "$DATADIR/_log/$f" ) || wbFail("Could not open $f");
        while (<h>) {
            next unless $_ =~ m/$search/;
            @x = split( /\t/, $_ );

            # ziga, this is slooow
            if ( $CGI{'last'} && time - $CGI{'last'} > $x[2] ) {
                next;
            }

            $x[2] = &formatDate( $x[2] );
            $x[2] =~ s/ /&nbsp;/g;

            $x[10] = '' unless $format eq 'long';
            $x[1] = $x[5] = $x[6] = $x[7] = $x[9] = $x[10] = ''
              if $format eq 'short';

            print qq~<tr class="DATA$i" valign="top"><td>\n~;
            print join( "</td>\n<td>", @x );
            print "</td></tr>\n\n";
            $i = $i ? 0 : 1;
            $n++;
        }
    }
    print "</table>";
    print "<p>$n entries found\n";
}

# count lines from log

sub admLogCount {

    local ($what) = @_;

    $dPatt = $dPatt2 = '';

    if ( $what eq 'shows' ) {
        $dAction = 'Show';
        $dPatt   = 'id=([^&]*)';    # Show?id=1234
        $dPatt2  = '^([^&=]*)';     # Show?1234
        print "<h3>Records shown most often</h3>\n";
    }
    elsif ( $what eq 'searches' ) {
        $dAction = 'Search';
        $dPatt   = 'search=([^&]*)';
        print "<h3>Searches did most often</h3>\n";
    }
    else {
        print "<h3>Log counts</h3>\n";
    }

    $search = '.' if $search eq '';
    $field = $CGI{'field'};

    undef %count;

    print "<p>In log files:\n";

    foreach $f (@f) {
        next unless $f =~ m/$lf/;
        open( h, "$DATADIR/_log/$f" ) || wbFail("Could not open $f");
        print "$f\n";
        while (<h>) {
            next unless $_ =~ m/$search/;
            @x = split( /\t/, $_ );

            $rr++;
            $r++;
            next if $robot{ $x[3] } == 1;

            # next if $x[0] eq 'bot';
            $r--;

            if ( $what ne '' ) {

                # print $x[8] . $dAction . "\n";
                next unless $x[8] =~ m/^$dAction/;
                if ( $x[9] =~ m/$dPatt/ ) {
                    $count{$1}++;
                }
                if ( $dPatt2 && $x[9] =~ m/$dPatt2/ ) {
                    $count{$1}++;
                }
            }
            else {
                $count{ $x[$field] }++;
            }
        }
    }

    @a = keys(%count);
    @a = sort { $count{$b} <=> $count{$a} } @a;

    print "<br>Ignored $r robot accesses if $rr accesses from IPs in _LOG/robots <pre>";

    $i = 1;
    foreach $x (@a) {
        print "$i\t$count{$x}\t$x\n";
        $i++;
    }
    print "</pre>";
}

# returns name of current log file
# sets @x also

sub logFile {

    $lf   = $WBB{'logFile'} || '$yyyy-$mm.log';
    @x    = &fixDate(time);
    $yyyy = $x[5];
    $mm   = $x[4];
    $dd   = $x[3];
    $lf   = '"' . $lf . '"';
    $lf   = &evalSafe($lf);
    return $lf if $lf =~ m|^/|;
    return "$DATADIR/_log/$lf";
}

# only Robots can get here

sub cgiTrap {

    $time = time;

    open( h, ">$DATADIR/_tmp/robot.$RA" );
    print h $time;
    close(h);

    opendir( D, "$DATADIR/_tmp" );
    @files = readdir(D);
    closedir(D);

    # clean old robot files

    foreach $file (@files) {
        if ( $file =~ m/^robot/ ) {
            @x = stat("$DATADIR/_tmp/$file");
            if ( time > $x[9] + 3600 ) {
                unlink "$DATADIR/_tmp/$file";
            }
        }
    }

    redirectRobot();
}

# write to the log file

sub wbLog {

    local ( @x, $x, $yyyy, $mm, $dd, $lf, $t, $rh, $ll );

    $ll = $WBB{'logLevel'};
    $ll = 3 if $ll eq '';
    return unless $ll;

    $lf   = &logFile();
    $yyyy = $x[5];
    $mm   = $x[4];
    $dd   = $x[3];

    $t = &lapTime() || '?';

    $rh = $RH || $RA;
    if ($Err) {
        $e = $Err;
        return if $e eq 'ER' && $ll < 2;
    }
    else {
        $e = 'ok';
        return if $ll < 3;
    }

    $ErrMsg =~ tr/\n/ /;
    $ErrMsg = '' if $e eq 'NA';    # we don't want the not available error in the log
    open( h, ">>$lf" ) || return;
    flock( h, 2 );
    binmode(h);
    $tt = time;
    print h <<EOM;
$e	$t	$tt	$rh	$User	$Group	$SCRIPT	$ENV{'REQUEST_METHOD'}	$Page	$ENV{'QUERY_STRING'}	$ErrMsg
EOM

    if ( $ll > 3 ) {
        $t = $Log;
        $t =~ s/\n/\n-/g;
        chop($t);
        print h '-', $t;
    }
    close(h);
}

# increment the access count and return number of accesses
# access.txt:
#	 0 number of accesses
#	1 since
#	2 last time maintenance was run
#	3 at command is failing
#	4 UID=
#	5 URL=

sub wbAccessCount {

    local ( @data, $file, $job );

    $file = "$DATADIR/_cache/access3.txt";

    unless ( -e $file ) {
        open( access, ">$file" );
        binmode(access);
        print access "0\n";
        print access time() . "\n";
        close(access);
    }

    open( access, "+<$file" );
    binmode(access);
    @data = <access>;    # no locking !!!, newlines included !
    chop(@data);
    $data[0]++;
    if ( !$CMDLINE ) {
        $data[4] = "UID=$>";
        $data[5] = "URL=$SCRIPTBASE$SCRIPTNAME";
        $data[6] = "SCRIPTFILE=$SCRIPTFILE";
    }

    $freq = $WBB{'maintenancePeriod'} || 0;    # automatic maintenance off !!!
    $freq = $freq * 3600;
    $t    = time - $data[2];
    if ( !$CMDLINE && $freq && $data[3] ne '1' && ( time - $data[2] ) > $freq ) {
        $job = &wbAtdo();
        if ($?) {
            $data[3] = 1;
        }
        else {
            $data[2] = time;
        }
    }

    seek( access, 0, 0 );
    print access join( "\n", @data );
    print access "\n";
    close(access);

    return $data[0];
}

# return data in access.txt

sub wbAccessData {

    local ( @data, $file );

    $file = "$DATADIR/_cache/access3.txt";
    open( access, "$file" );
    @data = <access>;    # no locking !!!
    close(access);

    return (@data);
}

# greps @records into @list using @sWords, @aWords, @nWords, $then and $sex
# sets @list
# TODO remove $* used to call recFromRow and evalSafe with multiline
sub wbGrepRecords {

    local ( *records, *aWords, *nWords, *sWords, $then, $sex ) = @_;

    undef @list;
    undef $SearchPatts;    # search patterns to be used in formatting

    if ($then) {           # time zadnja zadeva !
        foreach $_ (@records) {
            m/([0-9]*)$/;
            push( @list, $_ ) if $1 > $then;
        }
        $Log .= "Processed time with $then ... $#list\n";
    }
    else {
        @list = @records;
        $Log .= "list=records ... $#list+1\n";
    }

    # remove nots
    foreach $word (@nWords) {
        next if $word eq '';
        $patt = &wordBound($word);
        @list = grep( !/$patt/i, @list );
        $Log .= "Perl grep NOT /$patt/i ... $#list+1\n";
    }

    # remove and
    foreach $word (@aWords) {
        next if $word eq '';
        $patt = &wordBound($word);
        $SearchPatts .= "$patt|";
        @list = grep( /$patt/i, @list );
        $Log .= "Perl grep AND /$patt/i ... $#list+1\n";
    }

    # extract Sex stuff
    if ($sex) {
        $* = 1;    # multiline matching
        undef @y;
        foreach $row (@list) {
            recFromRow( $row, *fieldNames );

            push( @y, $row ) if &evalSafe($sex);
        }
        $*    = 0;    # singleline matching
        @list = @y;

        #		$Log .=
        #print "<br>xyz Perl evaluation of $sex  ... $#list+1\n";
    }

    # count orwords
    undef @y;
    foreach $word (@sWords) {
        next if $word eq '';
        $patt = &wordBound($word);
        $SearchPatts .= "$patt|";
        @x = grep( /$patt/i, @list );
        push( @y, @x );
        $Log .= "Perl grep OR /$patt/i ... $#x/$#list\n";
    }

    # how much stuff was contributed by filter

    #	$nots = $#nWords - $filterNots + 1;
    #	$ands = $#aWords - $filterAnds + 1;

    $Xtra = 0;
    if ( $nots || $ands ) {

        # +aaa bbb ... must be aaa, maybe bbb
        # since bbb ... must be since, within that must be bbb
        push( @list, @y );
        $Xtra = 1;
        $Log .= "... OR words appended\n";
    }
    elsif (@sWords) {
        @list = @y;
        $Log .= "... using y because there are $#sWords\n";
    }
    else {
        $Log .= "... using list\n";

        # anything goes
    }

    # should return $Xtra too
    chop($SearchPatts);
    return *list;
}

#
# search by default means whole words
# car ... car* matches cardinal
#

sub wordBound {
    local ($word) = @_;

    $word =~ s/(\W)/\\$1/g;     # escape special characters '*' too!
    $word =~ s/\\\*/\.\*/g;     # lju\*ana into lju.*ana
    $word =~ s/\\\?/\./g;       # ljubl?ana into ljubl.ana
    $word =~ s/^(\w)/\\b$1/;    # prepend /b if it starts with a character
    $word =~ s/(\w)$/$1\\b/;    # append /b if it ends with a character

    return $word;
}

# wbTables() creates %TBL array based on WBB{tables??}
# wbTables($op,$alias) returns $TBL{$op,$alias}
# op = DIR|DEF|URL
# alias = /usr/local/x/y/somedatadir ... absolute
#	   = ../relative
#	   = people ... alias from 'tables'
#	= people213 ... dir from $DATADIRDIR

sub wbTables {

    local ( $op, $alias ) = @_;
    local ( @x, $x, $al, $dir, $def );
    local $_;

    # do we have it cached?

    if ( $TBL{ $op, $alias } =~ m|/| ) {
        my $s = $TBL{ $op, $alias };
        $s =~ s#/U=.*$##;
        return $s;
    }

    # is this about self

    if ( $alias eq '' ) {
        my $s;
        $s = $DATADIR              if $op eq 'DIR';
        $s = $DATADICT             if $op eq 'DEF';
        $s = $SCRIPTBASE . $SCRIPT if $op eq 'URL';
        $s =~ s#/U=.*$##;
        return $s;
    }

    # what if path, not alias is supplied
    # compatibility with old calls

    if ( $op eq 'DIR' && $alias =~ m|/| ) {

        # path supplied
        if ( $alias =~ m|^/| ) {

            # absolute
            my $s = $alias;
            $s =~ s#/U=.*$##;
            return $s;
        }
        else {

            # relative
            my $s = $DATADIR . '/' . $alias;
            $s =~ s#/U=.*$##;
            return $s;
        }
    }

    # parse WBB{tables} into hash
    # alias<spaces>definitionfile<spaces>datadir
    # do it once per session!

    if ( $WBB{'tables'} ne '' && !defined(%TBL) ) {
        @x = split( /\n/, $WBB{'tables'} );
        foreach $_ (@x) {
            ( $al, $def, $dir ) = split( /\s+/, $_, 3 );
            if ( $def =~ m/^http/ ) {

                # remote table
                $dir = '' if $dir eq '';
                $TBL{ 'URL', $al } = $def;
                $TBL{ 'DIR', $al } = $dir;    # likely to be empty
                $TBL{ 'DEF', $al } = '';
            }
            else {

                # local table
                if ( $def =~ m!/cgi|/db!i ) {
                    $TBL{'URL'} = $SCRIPTBASE . $& . $';
                }
                else {

                    # we could look into access3.txt!
                }
                $TBL{ 'DEF', $al } = $def;
                $TBL{ 'DIR', $al } = $dir;
            }
        }
    }

    # add values for self

    $TBL{ 'DIR', '' } = $DATADIR;
    $TBL{ 'DEF', '' } = $DATADICT;
    $TBL{ 'URL', '' } = $SCRIPTBASE . $SCRIPT;

    # we have a value

    if ( $TBL{ $op, $alias } =~ m|/| ) {
        my $s = $TBL{ $op, $alias };
        $s =~ s#/U=.*$##;
        return $s;
    }

    #printHead("$op $alias");
    #print &formatHash('','',%TBL);
    #do wbFail('bbb');
    # we have a defined but empty value

    if ( defined $TBL{ $op, $alias } ) {
        return '' if $TBL{ $op, $alias } eq '';
    }

    # still no results? ... lets see if it is laid out in parallel to self
    # e.g. we are /www/wdat/people ... we need to know where is 'companies'
    # perhaps it is at /www/wdat/companies
    # this is the only way to get to the URL

    $TBL{ 'DIR', '' } =~ m|[^/]*$|;
    $we = $&;
    $other = $TBL{ $op, '' };
    $other =~ s|/$we|/$alias|;

    $Log .= "table '$alias' not defined in WBB{'tables'} ... we are '$we' ... TBL{$op,$alias}=$other\n";

    $TBL{ $op, $alias } = $other;
    my $s = $TBL{ $op, $alias };
    $s =~ s#/U=.*$##;
    return $s;

}

# supports preferences set by WBB{'intlCollate'}

sub wbCollate {

    return ( '', '' ) unless $WBB{'intlCollate'};
    local ( $x, $c, $rest, $from, $to );

    @x = split( /\s/, $WBB{'intlCollate'} );
    foreach $x (@x) {
        $c = substr( $x, 0, 1 );
        $rest = substr( $x, 1 );
        $tr{$c} = $rest;
    }

    $from = $to = '';

    $acTo = 97;
    for ( $ac = 97 ; $ac <= 122 ; $ac++ ) {
        $c = pack( 'c', $ac );

        # print STDOUT "$ac -> $acTo <br>\n";

        $from .= $c;
        $to .= pack( 'c', $acTo++ );

        if ( $rest = $tr{$c} ) {
            for ( $i = 0 ; $i < length($rest) ; $i++ ) {
                $s = substr( $rest, $i, 1 );

                # print STDOUT "$s => $acTo <br>\n";
                $from .= $s;
                $to .= pack( 'c', $acTo++ );
            }
        }
    }

    return ( $from, $to );
}

# move file

sub wbMoveFile {
    local ( $from, $to ) = @_;

    return unless -e $from;
    return if $to eq '';

    rename( $from, $to ) || wbFail("Could not rename $from to $to");
}

#
#
# search section
#
#

# returns sWords[], aWords[], nWords[] arrays
# set Sex (search expression)

sub wbParseSearch {

    local ( $search, $table, $magic ) = @_;
    undef $then;

    if ($table) {
        $dontKnowFieldNames = 1;
    }
    else {
        wbSortFields();
    }

    ( $search, $sExpression ) = split( /\{/, $search, 2 );

    if ($sExpression) {
        $_ = "{$sExpression";

        # {_at}>123

        if ( $_ =~ m/\{_at\}\s*>\s*(\d+)/ ) {
            $then = $1;
            $_    = $` . $';
        }

        #		if ( $magic ne 'filter' && m/\w\s*\(|&|`|;|system|exec|spawn/ ) {
        #wbError("Invalid character in search expression '$_'"); #UK
        #		}
        s/\{/\$rec\{/g;
        s/ and / && /g;
        s/ or / || /g;
        s/ not / ! /g;

        #		{_basket} finds in basket only

        $x = '\$rec\{_basket\}';
        $y = '$Basket{$rec{\'_id\'}}>0';

        if ( $_ =~ m/$x/ ) {
            readBasket();
            $_ =~ s/$x/$y/;
        }

        $sExpression = $_;

        # print "\tsearch expression: $sExpression\n";

        # 4.152 ... no need to eval safe below?

        $x = &evalSafe($_);
        if ( $x =~ m/class="ERROR"/i ) {
            wbError( "Bad search expression", $x );    #UK
        }
    }

    @words = split( / /, $search );
    $s     = 0;
    $n     = 0;
    $a     = 0;
    $dWord = '';
    undef @sWords;
    undef @aWords;
    undef @nWords;

    while (@words) {
        $dWord .= shift(@words);
        if ( ( $dWord =~ tr/"/"/ ) % 2 ) {
            $dWord .= ' ';
            next;
        }

        # +email:ziga
        if ( $dWord =~ m/\:/ ) {
            ( $f, $w ) = split( /\:/, $dWord, 2 );
            if ( $f =~ m/"/ ) {

                #   "sdfsf:adasds" ... not special !
            }
            else {
                $we = $w;
                $we =~ s/"//g;
                $pwe = &wordBound($we);
                ( $s, $f ) =~ m/^([-+=\/]{0,1})(.*)$/;
                $s = $1;
                $f = $2;
                if ( $s ne '' && !$dontKnowFieldNames && !$IsField{$f} ) {

                    # +tyson:smith
                    # there is a plus or something there
                    # we know field names
                    # tyson is not a field name
                    wbError( "Warning: Bad search expression", "$f is not a field name. Perhaps you wanted to search for \"$dWord\"" );    #UK
                }
                else {
                    $s = $s || '+';
                    if ( $sExpression eq '' ) {
                        $sExpression = '1';
                    }
                    if ( $s eq '-' ) {
                        $sExpression .= " && ! (\$rec{'$f'} =~ m/$pwe/i)";
                        $dWord = '';
                        next;
                    }
                    if ( $s eq '+' ) {
                        $sExpression .= " && (\$rec{'$f'} =~ m/$pwe/i)";
                        $dWord = "+$w";

                        # process normally
                    }
                    if ( $s eq '=' ) {
                        $sExpression .= " && (\$rec{'$f'} eq '$we')";
                        $dWord = "+$w";

                        # process normally
                    }
                    if ( $s eq '/' ) {
                        $sExpression .= " && (\"\\n\$rec{'$f'}\\n\" =~ m/\\n$we\\n/i)";
                        $dWord = "+$w";

                        # process normally
                    }
                }
            }
        }

        $dWord =~ s/"//g;

        if ( $dWord =~ m/^\-(.+)/ ) {
            $nWords[ $n++ ] = $1;
        }
        elsif ( $dWord =~ m/^\+(.+)/ ) {
            $aWords[ $a++ ] = $1;
        }
        else {
            $sWords[ $s++ ] = $dWord if $dWord ne '';
        }
        $dWord = '';
    }

    if ( $dWord ne '' ) {
        wbError( "Error in search", "Search expression<br><b>$dWord</b><br>is invalid." );
    }

    # --- build a list of matching records

    $grepPatt = '';
    if ($GREP) {    # prepare search pattern for GREP
        if ( $aWords[0] ne '' ) {
            $grepPatt = $aWords[0];
        }
        elsif ( $#sWords >= 0 ) {
            $grepPatt = join( '|', @sWords );
        }
    }

    return ( *aWords, *nWords, *sWords, $then, $sExpression, $grepPatt );
}

# read records into @records array and field names into @fieldNames

sub wbReadRecords {

    local ( $table, $sort, $grepPatt ) = @_;
    local ($dd);

    $dd = &wbTables( 'DIR', $table );
    unless ($dd) {
        $Log .= "ERROR: No such table $table\n";
        return;
    }

    $sortf = $sort;
    $sortf =~ tr/[A-Z]/[a-z]/;

    if (   $GREP
        && $WBB{'grepBigger'}
        && $grepPatt ne ''
        && ( -s "$dd/_cache/$sortf.tbl" > $WBB{'grepBigger'} ) )
    {

        #print "USING GREP with '$grepPatt'\n";
        open( h, "$dd/_cache/$sortf.tbl" )
          || wbFail("Open of $sortf.tbl failed for $table");
        $records[0] = <h>;    # attn: newlines included !
        $records[1] = <h>;    # attn: newlines included !
        $records[2] = <h>;    # attn: newlines included !
        close(h);
        *fieldNames = &wbReadRecordsFnames();
        $grepPatt =~ s/"//g;
        $grepPatt =~ s/\*$//g;
        $grepPatt =~ s/^\*//g;
        $grepPatt =~ s/\*/\.\*/g;

        if ( $grepPatt =~ m/^\w*$/ ) {
            $sw = '-F';
        }
        elsif ( $grepPatt =~ m/\|/ ) {
            $sw = '-E';
        }
        else {
            $sw = '';
        }
        $g = "$GREP -a -i $sw \"$grepPatt\" $dd/_cache/$sortf.tbl |";
        open( h, $g );
        @records = <h>;
        $Log .= "$g ... $#records + 1\n";
        close(h);

    }
    else {

        #print "NOT USING GREP !\n";
        # brute force
        open( h, "$dd/_cache/$sortf.tbl" )
          || wbFail("Open of $sortf.tbl in $dd failed");
        @records = <h>;    # attn: newlines included !
        close(h);
        $searchProc = "Reading all records from $sortf.tbl\n";
        *fieldNames = &wbReadRecordsFnames();
    }

    return ( *records, *fieldNames );

}

# sets fieldNames based on first 3 rows

sub wbReadRecordsFnames {

    if ( $records[1] =~ m/^\n/ ) {
        $x = shift(@records);    # content type
        $x = shift(@records);    # blank
    }
    $x = shift(@records);
    chop($x);
    @fieldNames = split( /\t/, $x );

    return *fieldNames;

}

# format field names for display

sub formatFieldName {

    local ( $field, $nolink ) = @_;

    $_ = $field;

    if ( $WBF{ $_, 'head' } ) {
        return $WBF{ $_, 'head' };
    }

    s/([a-z])([A-Z])/$1 $2/g;
    tr/[A-Z]/[a-z]/;

    if ( !$nolink && $WBF{ $field, 'type' } =~ m/OPTION/ ) {
        $_ = qq(<a href="$SCRIPT/BrowseTree?field=$field&amp;order=AZ">$_</a>);
    }

    return $_;
}

# print rec in an HTML table or as defined in 'detail' format
# how = NAMEVAL ... standard name/value display

sub wbRecHtml {

    local (@fields);
    local ( $how, *fields ) = @_;

    local ( $out, $value, $x, $th1, $th2 );

    if ( $WBB{'detail'} ne '' && $how ne 'NAMEVAL' ) {
        $out = &evalSafe( $WBB{'detail'} );
        return $out;
    }

    # ($th1,$th2) = ("field","value") ; #UK

    $out .= <<EOM;
<table class="DATATABLE" cellspacing="1" cellpadding="2">
EOM

    #undef @fields;
    #$Log .= join(',',@fields) . "\n";

    if ( $#fields < 0 ) {
        wbSortFields();
        @f = @Fields;
        push( @f, '_from', '_at' );
    }
    else {
        @f = @fields;
    }

    foreach $field (@f) {

        # skip unseen

        next if $Hidden{$field};

        # banners

        $value = &formatField($field);

        if ( $WBF{ $field, 'type' } eq 'BREAK' ) {
            $t = $WBF{$field};
            $t =~ s/^"//;
            $t =~ s/"$//;
            $out .= '<tr class="DATAH" ><th>&nbsp;</th><th align="left">' . $t . '</th></tr>' . "\n";
            next;
        }

        if ( $WBF{ $field, 'type' } eq 'SUBMIT' ) {
            next;
        }

        next if $value eq '';

        $name = &formatFieldName($field);

        $c = '';
        $c = 'class="ERROR"' if $err{$field};

        $out .= qq~<tr class="DATA" valign="top">\n<th align="right">$name</th>\n<td $c>$value</td>\n</tr>\n\n~;
    }

    # admin details

    #	$niceDate = &formatYMDHMS($rec{'_at'});
    #	$from = "Entered" ; #UK
    #
    #	$out .= <<EOM unless $WBB{'hideUnderscore'} eq '1';
    #<tr>
    #<td COLSPAN=2 align="right">
    #<font size="2">
    #$from $rec{'_from'} $niceDate
    #</font></td>
    #</tr>
    #EOM

    $out .= "</table>\n";

    return $out;
}

# returns the text of the currently applied filter

sub filterText {

    @x = split( /\n/, $WBB{'filterText'} );    # only do it on self
    foreach $_ (@x) {
        ( $group, $text ) = split( /\s+/, $_, 2 );
        if ( $group eq $Group ) {
            return $text;
        }
    }
}

# find records, returns formatted @list of records
# fields in first row
# and @ids of found %data
# does not work with any table yet ! ... unless $table

sub recSelect {

    local ( $table, $search, $sort, $format, $fields, $first, $max ) = @_;

    logTime("---- recSelect in table '$table' for search '$search' with fields '$fields' started");

    if ( $table eq '' ) {

        #OK
    }
    elsif ( $sort eq 'TIME' || $sort eq 'DEFAULT' || $sort eq '' ) {

        #OK
    }
    else {
        wbFail( "Cannot select from $table", "Sort should be TIME or DEFAULT" );
    }

    # --- hide some records from the search with filter

    #	undef @x,@faWords,@fnWords,@fsWords,$fthen,$fsex,$fgrepPatt,$filter;
    undef @x;
    undef @faWords;
    undef @fnWords;
    undef @fsWords;
    undef $fthen;
    undef $fsex;
    undef $fgrepPatt;
    undef $filter;

    @x = split( /\n/, $WBB{'filter'} ) if $table eq '';    # only do it on self
    foreach $_ (@x) {
        ( $group, $filter ) = split( /\s+/, $_, 2 );
        if ( $group eq $Group ) {
            ( *aWords, *nWords, *sWords, $fthen, $fsex, $fgrepPatt ) = &wbParseSearch( $filter, $table, 'filter' );
            @faWords    = @aWords;
            @fnWords    = @nWords;
            @fsWords    = @sWords;
            $filterAnds = $#faWords + 1;                   # to determine relevance ranking
            $filterNots = $#fnWords + 1;                   # obsolete after 4.200
            last;
        }
    }

    # --- default format and sort

    $sort = 'DEFAULT' if $sort eq '';

    # --- do we need to update the static file

    if ( $WBB{"sort;$sort"} || $sort eq 'TIME' || $sort eq 'DEFAULT' ) {
        $dynamicSort = '';
    }
    else {    # comma separated list of fields
        $dynamicSort = $sort;
        $dynamicSort =~ tr/\n/,/;    # newline separated comes from forms!
        $sort = 'DEFAULT';           # this TBL will be used
        local ( @f, $f );
        @f = split( /,/, $dynamicSort );
        $dynamicSort = '';
        foreach $f (@f) {
            next unless $f =~ m/\w/;    # skip any blanks
            if ( $f =~ m/^DYNAMIC/ ) {
                $dynamic = $f;
                next;
            }
            $f =~ s/\s+//g;             # remove spaces
            $dynamicSort .= "\$rec{'$f'} ";    # convert to expression with $rec
        }
        $dynamicSort = '"' . $dynamicSort . '"';    # make a string
    }

    # --- reindex

    local ( $sort, $format ) = ( $sort, $format );
    wbReindex($sort) unless $table;

    # --- defaults

    $first = 0        if !$first;
    $max   = 99999999 if !$max;                     # adminExport !
    $last  = $first + $max - 1;

    # --- parse the search

    #$Log .= "A $fields\n";

    ( *aWords, *nWords, *sWords, $then, $sex, $grepPatt ) = &wbParseSearch( $search, $table );
    logTime("Search parameters parsed");

    #$Log .= "B $fields\n";

    # --- this is the actual search !!!

    $csearch = "$table-$search-$sort";
    $csearch = "_qry/" . &htEscape($csearch) . ".tbl";
    $csearch =~ s/%//g;
    if ( !$table && !&wbExpired($csearch) ) {
        $Log .= "\tUsing cached search\n";
        open( H, $csearch );
        $x = <H>;
        chop($x);
        @fieldNames = split( /\t/, $x );
        @list = <H>;
        close(H);
    }
    else {
        $Log .= "Working with table: $table\n";
        if ($search) {
            ( *records, *fieldNames ) = &wbReadRecords( $table, $sort, $grepPatt );    # read into RAM the presorted table
            logTime("Read with user's '$grepPatt': $#records+1 records in RAM");
        }
        else {
            ( *records, *fieldNames ) = &wbReadRecords( $table, $sort, $fgrepPatt );    # read into RAM the presorted table
            logTime("Read with filter's '$fgrepPatt': $#records+1 records in RAM");
        }

        *list = &wbGrepRecords( *records, *aWords, *nWords, *sWords, $then, $sex );     # create @list, may have duplicates
        logTime("Grep with user's search: $#list+1 records remain.");

        if ($filter) {
            $sp      = $SearchPatts;
            @records = @list;
            *list    = &wbGrepRecords( *records, *faWords, *fnWords, *fsWords, $fthen, $fsex );    # create @list, may have duplicates
            logTime("Grep to filter: $#list+1 records remain.");
            $SearchPatts = $sp;
        }

        # 	    force cache ? StoreSelect never defined ?!?!?!
        #	    if (($StoreSelect || $WBB{searchCache}) && $#list < 1000 && ($search ne '' || $then ne '')) {
        if ( $WBB{searchCache} && $#list < 1000 && ( $search ne '' || $then ne '' ) ) {

            wbQryClean();
            $Log .= "\tSearch cached - found $#list+1\n";
            open( H, ">$csearch" );
            binmode(H);
            print H join( "\t", @fieldNames ) . "\n";
            print H @list;
            close(H);
        }
        else {
            $Log .= "\tSearch not cached - found $#list+1\n";
        }
    }

    #$Log .= "C $fields\n";

    # --- trim the list

    undef $allHits;

    if ( $#sWords < 0 && $dynamicSort eq '' ) {    # 0 search words, trim now, no extra sorting will be needed

        $allHits = $#list + 1;
        wbTrimList( *list, $first, $last );
        $trimmed = 1;
        $Log .= "\tQuick trimmed to $#list+1\n";

    }
    else {

        $i = 0;
        foreach $_ (@list) {
            $i++;
            ( $id, $rest ) = split( /\t/, $_, 2 );
            $sort{$id} = $i unless $sort{$id};

            # why not simply store index into list?
            $data{$id} = $_ unless $data{$id};
            $hits{$id}++;
        }

        # sort on hits + original order

        foreach $_ ( keys(%data) ) {
            $sort{$_} += ( 100 - $hits{$_} ) * 10000;
        }

        # OK, dynamic sort is slow anyway
        # above we do some redundant things

        if ($dynamicSort) {
            logTime("Before dynamic sort with $dynamicSort");
            &evalSafe($dynamicSort);    # this will fail, below we can eval
            foreach $id ( keys(%data) ) {
                recFromRow( $data{$id}, *fieldNames );
                $sort{$id} = eval($dynamicSort);    # it is a $rec expression

                #$Log .= "$id ... $sort{$id}\n";
            }
        }

        #$Log .= join("\n\n",@list);

        if ( $dynamic eq 'DYNAMICZA' ) {
            @list = sort { $sort{$b} cmp $sort{$a} } keys(%hits);
        }
        elsif ( $dynamic eq 'DYNAMIC09' ) {
            @list = sort { $sort{$a} <=> $sort{$b} } keys(%hits);
        }
        elsif ( $dynamic eq 'DYNAMIC90' ) {
            @list = sort { $sort{$b} <=> $sort{$a} } keys(%hits);
        }
        else {
            @list = sort { $sort{$a} cmp $sort{$b} } keys(%hits);
        }

        #foreach $x (@list) {
        #	print "<br>sort($x)=$sort{$x}\n";
        #}

        #$Log .= join(',',values(%hits));

        #$Log .= join("\n\n",@list);

        logTime("$#list+1 sorted");
        $allHits = $#list + 1 unless $allHits;
        wbTrimList( *list, $first, $last );
        logTime("$#list+1 trimmed");
        @ids = @list;

        #$Log .= "	$#ids ids before format\n";
        undef @list;

        #$Log .= "	$#ids ids before format\n";
        foreach $id (@ids) {
            push( @list, $data{$id} );
        }
    }

    #$Log .= "D $fields\n";

    $Log .= "$#list+1 before format\n";

    # --- format the list

    if (   ( $format eq '' || $format eq 'TAB' || $format eq 'RAW' )
        && $fields eq ''
        && $Group  eq 'admin' )
    {

        # admin gets it all ... why??? ... bad if as admin one does A-Z index !!!
        @fields = @fieldNames;

    }
    else {

        if ( $fields eq '' ) {
            @fields = @fieldNames;
            undef @editable;
        }
        elsif ( $fields !~ m/,/ ) {
            @fields = ($fields);
        }
        else {
            @fields = split( /,/, $fields );
        }

        # any of them editable ? like ?name,$lastName

        foreach $f (@fields) {
            if ( $f =~ m/^\?/ ) {
                $f = $';
                push( @editable, $f );
            }
        }

        unless ( $Group eq 'admin' ) {
            @fields = grep( !/_password/, @fields );

            # @fields = grep(!/^_from$/,@fields) if $WBB{'hideUnderscore'};
            # @fields = grep(!/^_at$/,@fields) if $WBB{'hideUnderscore'};
        }

        $stevec = $first + 1;
        foreach $row (@list) {
            recFromRow( $row, *fieldNames );
            $ThisRecordNo = $stevec++;
            $row = &formatFoundRec( $format, *fields, *editable );    # format goes here !
        }
    }

    logTime("$#list+1 formated as $format ... recSelect done!");

    return ( $allHits, *list, *fields );
}

# return search form

sub wbSearchForm {

    local ( $sstring, $elements ) = @_;
    local ( $js, @s );
    if ( $elements eq '' ) {
        $elements = 'FORM,PROMPT,HELP,HISTORY';
    }

    local ( $prompt, $help, $button, $size );

    $prompt = "Search for:";    #UK
    $verb   = "Search";         #UK
    $help   = "Tip: ";          #UK

    $s = &htEscape( $CGI{'search'} );
    $help .= &formatSearchHelp();
    $help .= qq(<br>More: <a href="$SCRIPT/HelpSearch">all tips and help</a>);    #UK
    $help .= ' ';
    if (   &wbAllowed( $Group, 'AdvancedSearchForm' )
        || &wbAllowed( 'friend', 'AdvancedSearchForm' ) )
    {
        $help .= <<EOM;                                                           #UK
 <a href="$SCRIPT/AdvancedSearchForm?search=$s">advanced search</a>
EOM
    }

    if ( &wbAllowed( $Group, 'FullTextSearch' ) && &canFullTextSearch() ) {
        $help .= <<EOM;                                                           #UK
 <a href="$SCRIPT/FullTextSearch?search=$s">full text search</a>
EOM
    }

    $size = length($sstring) || 40;
    $size = 40 if $size < 40;

    # $x =~ s/"/%22/g;

    $help = '<span class="HINT">' . $help . '</span>';

    # $prompt = "$prompt<br>";

    $help   = '' unless $elements =~ m/HELP/;
    $prompt = '' unless $elements =~ m/PROMPT/;
    $histjs = '';
    $histbut = '';

    if ( $elements =~ m/HISTORY/ ) {

        # old searches

        @s  = &wbSearchHistoryGet();
        $ns = 0;
        $js = '';
        $n  = 1;
        foreach $s (@s) {
            next if $s eq '';
            $s =~ s/\"/\\\"/g;
            $js .= qq(\ts[$n] = "$s";\n);
            $ns++;
            $n++;
        }

        if ( $ns > 0 ) {
            $histjs = <<EOM;
<script LANGUAGE="JavaScript" type="text/javascript"><!--

ns = $ns;
cs = 0;
s = new Array();
$js

function next () {
	cs=cs+1;
	if (cs > ns) {
	cs = 1;
	}
	document.searchForm.search.value = s[cs];
}

function prev () {
	cs = cs - 1;
	if (cs <= 0) {
	cs = ns;
	}
	document.searchForm.search.value = s[cs];
}

//--></script>
EOM

            $p = "Previous search";
            $n = "Next search";
            $a = "All previous searches";
            $s = "Saved forms";

            $histbut = '';
            $histbut = <<EOM if $JS && $ns > 0;
<a href="javascript:prev()"><img src="$ICONURL/back.gif" $ICONPAR align="top" alt="$p" title="$p"></a>
<a href="javascript:next()"><img src="$ICONURL/forward.gif" $ICONPAR align="top" alt="$n" title="$n"></a>
EOM

            $histbut .= <<EOM;
<a href="$SCRIPT/SearchHistory"><img src="$ICONURL/last.gif" $ICONPAR align="top" alt="$a" title="$a"></a>
<a href="$SCRIPT/FormsList"><img src="$ICONURL/save.gif" $ICONPAR align="top" alt="$s" title="$s"></a>
EOM

        }
    }

    $ss = &htQuote($sstring);

    $x = <<EOM;
$histjs
$prompt
<input name="search" size="$size" value="$ss">
$histbut
EOM

    if ( $elements =~ m/FORM/ ) {
        return <<EOM;
<form action="$SCRIPT/Search" name="searchForm">
<table class="SEARCHBOX" cellspacing="0" width="100%">
<tr><td>
$x
<input type="hidden" name="paint" value="1">
<input type="hidden" name="_form" value="AdvancedSearchForm">
<input class="BUTTON" type="submit" value="$verb">
</td></tr><tr><td>
$help</td></tr>
</table>
</form>
EOM
    }
    else {
        return $x;
    }
}

# return search help

sub wbHelps {

    if ( $WBB{'searchHelp'} ) {
        $help = $WBB{'searchHelp'};
    }
    else {
        $help = <<EOM;    #UK
Type <b>car vehicle automobile</b> to search for any of the words.
Type <b>"New York" car</b> to search for 'New York' or 'car'.
Type <b>new +car</b> to find only records which do contain 'car' and may contain word 'new'.
Type <b>-used car vehicle</b> to find only records which contain words 'car' or 'vehicle' but not 'used'.
Type <b>n?w</b> to find 'now' and 'new'.
Type <b>car*</b> to find 'car' and 'cardinal'.
Type <b>*car</b> to find 'vicar' and 'car'.
Type <b>*car*</b> to find 'vicar', 'cardinal' and 'car'.
Type <b>" car "</b> or "<b>\\scar\\s</b>" to find 'car' and not 'vicar' or 'cardinal'.
Type <b>+email:uni-lj</b> to search for 'uni-lj' in the email field only.
Type <b>-email:edu</b> to find records which do not have word 'edu' in the email field only.
Type <b>=lastName:Turk</b> to find records with lastName field exactly equal to Turk. 
Type <b>/lastName:Turk</b> to find records with one line of lastName field exactly equal to Turk. 
Type <b>{price} < 2000</b> to search for items with price field less than 2000.
EOM
    }

    @helps = split( /\n/, $help );

}

# return random help tip

sub formatSearchHelp {

    local ( @helps, $help, $i );
    wbHelps();
    $i = int( rand(6.99) );
    return $helps[$i];
}

# display all search help

sub cgiHelpSearch {

    printHead("Search Tips");    #UK
    wbHelpSearch();
}

# format search help as html

sub wbHelpSearch {

    wbHelps();
    print "<ul>\n";
    foreach $help (@helps) {
        print "<li>$help</li>\n";
    }

    print <<EOM;    #UK
</ul>
<p>
<b>
<a href="$WodaHelpURL#search">Detailed search help.</a>
</b>
EOM
}

# remove files from TMPDIR

sub wbTmpClean {

    wbDirClean( $TMPDIR, 30, <<EOM);
This directory contains temporary files.
Files older than 30 days are deleted anyway.
EOM

}

# clean _qry directory according to searchCacheAge

sub wbQryClean {

    local ( $dir, $old, $f, $file );

    $dir = "$DATADIR/_qry";
    $old = $WBB{'searchCacheAge'} || 720;    # 30 days

    wbDirClean( $dir, $old, <<EOM);
Files here are results of searches.
You should delete them if you are testing
new scripts etc. Files older than $old hours
are deleted automatically.
EOM

}

# erases files older than $old days in $dir

sub wbDirClean {

    local ( $dir, $old, $readme ) = @_;

    local @x = stat("$dir/_readme");
    return if ( time - $x[9] ) < 3600 * 24;    # tries every 24 hrs

    # we need to do maintenance !

    open( h, ">$dir/_readme" );
    print h $readme;
    close(h);

    opendir( D, $dir );
    local @files = readdir(D);
    closedir(D);

    foreach $f (@files) {
        $file = "$dir/$f";
        next if -d $file;
        @x = stat($file);
        if ( time - $x[9] > $old * 3600 ) {
            unlink $file;
        }
    }
}

# sort fields - sets @Fields, %Hidden, @PFields, @IFields, %IsField
# sets @Fields - sorted array of all fields in the table, but not system fields
# sets %Hidden - lists hidden fields
# sets @PFields - array which also includes _id, _at and _from ... but not _passwd ... anything that can be printed
# sets @IFields - all fields to be in the default.tbl etc (without ones without value)
# sets %IsField to enable fast control of what is a field name and what is not
# does it only once

sub wbSortFields {

    return if defined @PFields;

    unless ( $WBF{'_at'} ) {
        $WBF{'_from'} = '';
        $WBF{ '_at', 'head' }    = 'last changed';      #UK
        $WBF{ '_at', 'picture' } = '&formatDate($_)';
    }

    unless ( $WBF{'_from'} ) {
        $WBF{'_from'} = '';
        $WBF{ '_from', 'head' } = 'changed from';       #UK
        $WBF{ '_from', 'sees' } = 'admin';
    }

    undef %Hidden;
    undef %IsField;
    undef @Fields;
    undef @IFields;
    undef @PFields;

    local ( $x, $y, $f );

    foreach $x ( keys(%WBF) ) {

        # print "$x<br>";
        next if $x =~ m/$;/;

        if ( $WBF{ $x, 'type' } eq 'EMAIL' ) {
            $HasEmailField = 1;
        }

        if ( $y = $WBF{ $x, 'sees' } ) {
            unless ( $Group =~ m/$y/ || $Group eq 'admin' || $RGroup =~ m/$y/ ) {
                $Hidden{$x} = 1;

                # print "$x is hidden";
            }
        }

        push( @Fields, $x )
          unless $x =~ m/^_/;    # do not add underscore named fields (hidden will be set!)
        $IsField{$x} = 1;
    }

    @Fields = sort { $WBF{ $b, 'srt' } <=> $WBF{ $a, 'srt' } } @Fields;
    @PFields = ( '_id', @Fields, '_from', '_at' );

    foreach $f (@Fields) {
        push( @IFields, $f )
          unless $WBF{ $f, 'type' } =~ m/BREAK|COMPUTE|SUBMIT/;
    }
    @IFields = ( '_id', @IFields, '_password', '_from', '_at' );

    return;
}

# makes sure a directory exists and returns it

sub wbCheckDir {

    unless ( -d $_[0] ) { mkdir( $_[0], $DIRMODE ); }
    unless ( -d $_[0] ) { wbFail("Cannot make/access directory $_[0]"); }
    return $_[0];
}

# makes sure a directory exists and returns it

sub wbCheckDirectories {

    local (@dirs) = (
        "$DATADIR",      "$DATADIR/_cache", "$DATADIR/_data", "$DATADIR/_friend", "$DATADIR/_tmp", "$DATADIR/_qry",
        "$DATADIR/_log", "$HTMLDIR",        "$HTMLDIR/att",   "$HTMLDIR/robots",  "$HTMLDIR/_doc"
    );

    foreach $dir (@dirs) {
        wbCheckDir($dir);
        if ( $dir eq $DATADIR || $dir eq $HTMLDIR ) {

            # see mainConfig
            open( h, ">$dir/$HtaccessFile" )
              || wbFail("Cannot create $dir/$HtaccessFile files");
            binmode h;
            print h $HtaccessDeny;
            close(h);
        }
        if ( $dir =~ m/att$|robots$/ ) {

            # see mainConfig
            open( h, ">$dir/$HtaccessFile" )
              || wbFail("Cannot create $dir/$HtaccessFile files");
            binmode h;
            print h $HtaccessAllow;
            close(h);
        }
        if ( $dir =~ m/att$/ ) {
            open( h, ">$dir/index.htm" );
            print h <<EOM;
Browsing denied!
EOM
            close(h);
        }
    }
    return @dirs;
}

# fixes the password data
# according to the authentification scheme
# also crypts it

sub wbFixPass {

    local ( $lock, $pass ) = @_;

    if ( $lock eq 'user' ) {
        return "u:$User";

    }
    elsif ( $lock eq 'group' ) {
        return "g:$Group";

    }
    elsif ( $lock eq 'none' ) {
        return "";

    }
    elsif ( $lock eq 'host' ) {
        return "h:$RA";

    }
    elsif ( $pass ne '' ) {
        return &wbCryptPasswd($pass);

    }
    else {
        return '';

    }
}

# test passwords
# returns 1 if OK

sub wbTestPass {

    local ( $new, $old ) = @_;
    local ($x);

    # the "passwords"

    $x = -1;

    if ( $old =~ m/^u:/ ) {
        $x = $' eq $User;
    }
    elsif ( $old =~ m/^g:/ ) {
        $x = $' eq $Group;
    }
    elsif ( $old =~ m/^h:/ ) {
        $x = $' eq $RH;
    }

    if ( $x > 0 ) {
        return $old;
    }

    if ( $Group eq 'admin' ) {
        $Log .= "Non password authentification OK, because you are admin\n";
        $ADMINPASS = 1;
        return 1;
    }
    elsif ( $Group ne '' && $Group eq $WBB{'ownerGroup'} ) {
        $Log .= "Non password authentification OK, because you are from group $WBB{'ownerGroup'}\n";
        $ADMINPASS = 1;
        return 1;
    }

    # the old way with passwords

    $x = &wbCryptPasswd($new);

    $Log .= "newpass='$x'\noldpass='$old'\n";

    if ( $old eq '' && $new eq '' || $x eq $old ) {
        return 1;
    }
    else {
        if ( $Group eq 'admin' ) {
            $Log .= "Password is OK, because you are admin\n";
            $ADMINPASS = 1;
            return 1;
        }
        else {
            return 0;
        }
    }
}

# validates %rec
# returns error string
# sets $err{$field} to 1 for all bad fields
# TODO remove $* used to call evalSafe with multiline
sub recValidate {

    local ($err);

    $err = "";

    foreach $field ( keys(%WBF) ) {
        next if $field =~ m/$;/;
        next if $field =~ m/^_/;
        $vex = $WBF{$field};
        $_ = $value = $rec{$field};
        $_ =~ s/\r//g;    # remove CR

        next if $WBF{ $field, 'type' } eq 'BREAK';
        next if $WBF{ $field, 'type' } eq 'SUBMIT';

        if (   $_
            && $WBF{ $field, 'type' } eq "FILE"
            && !-f "$HTMLDIR/att/$value" )
        {
            $err .= "WARNING: file $HTMLDIR/att/$value does not exist\n";    #UK
            $err{$field} = 1;
            next;
        }

        if (   $vex eq '1;'
            && $WBF{ $field, 'type' } eq 'EMAIL'
            && $value ne ''
            && !&isEmail($value) )
        {
            $err .= "WARNING: bad email address at $field\n";                #UK
            $err{$field} = 1;
            next;
        }

        if ( $vex eq '1;' && $WBF{ $field, 'type' } eq 'DATE' && $value ) {
            local ( $y, $m, $d ) = split( /\//, $value );
            if ( $y !~ /^\d\d\d\d$/ || $m > 12 || $m < 1 || $d < 1 || $d > 31 ) {
                $err .= "WARNING: bad date at $field\n";                     #UK
                $err{$field} = 1;
            }
            next;
        }

        if ( $WBF{ $field, 'type' } eq 'LIST' ) {
            @x = split( /\n/, $_ );
        }
        else {
            @x = ($_);
        }

        foreach $_ (@x) {

            $y = 1;

            $* = 1;                  # multiline matching
            $y = &evalSafe($vex);    # calculate
            $* = 0;                  # single line matching

            if ($y) {
                if ( $y =~ m/CLASS\=ERROR/i ) {
                    wbFail( "Error validation expression", "$y in $vex" );    #UK
                }
            }
            else {
                $m = $WBF{ $field, 'cond' };
                $m = $WBF{$field} if !$m;
                $err .= "WARNING: bad value for <b>$field</b>:\n\t'$value' does not satisfy '<b>$m</b>'\n";    #UK
                $err{$field} = 1;
            }
        }

    }

    if ( defined(&myRecVerify) ) {
        $err .= &myRecVerify();
    }

    $Log .= $err;

    return $err;
}

#
# return a good name for a temp file in current dir with ext. $_[0]
#

sub wbRandKey {

    local ($n) = $_[0];
    local (@a) = ( 0 .. 9, 'a' .. 'f' );
    local ( $x, $n, $c, $o, $base, @c, $i );
    $base = $#a + 1;

    $n = $n || 4;
    $x = $base**$n;
    $x = int( rand($x) );

    for ( $i = 1 ; $i <= $n ; $i++ ) {
        $o     = $x % $base;
        $c[$i] = $a[$o];
        $x     = int( $x / $base );
    }

    return ( join( '', @c ) );
}

# returns first URL field in a database

sub wbRingField {

    local ( $urlFld, $field );

    wbSortFields();
    $urlFld = $WBB{'ringField'};
    if ( $urlFld eq '' ) {
        foreach $field (@Fields) {
            if ( $WBF{ $field, 'type' } eq 'URL' ) {
                $urlFld = $field;
                last;
            }
        }
    }

    return $urlFld;
}

# loop through records, returning web sites when a URL field has a value

sub cgiBrowseRing {

    # sort
    # id ... or random
    # op=next,prev,last,first,this,nextu,index,home,close
    # fr=top|bot

    $urlFld = &wbRingField();

    if ( $urlFld eq '' ) {
        wbError("Database does not contain URL's");    #UK
        return;
    }

    # --- get URL ---

    $id   = $CGI{'id'};
    $to   = $CGI{'to'};
    $sort = $CGI{'sort'} || 'DEFAULT';
    $fr   = $CGI{'fr'};

    if ( $id ne '' && $to eq 'this' ) {

        recRead($id);
        $url   = $rec{$urlFld};
        $newid = $id;

    }
    else {
        ( *records, *fieldNames ) = &wbReadRecords( '', $sort );

        if ( $to eq 'first' ) {
            $i = 0;
        }
        elsif ( $to eq 'last' ) {
            $i = $#records;
        }
        elsif ( $to eq '' ) {
            $i = int( rand($#records) );
        }
        else {
            $i = 0;
            foreach $x (@records) {
                ( $did, $rest ) = split( /\t/, $x );
                last if $did eq $id;
                $i++;
            }

            $i++ if $to eq 'next' || $to eq 'nextu';
            $i-- if $to eq 'previous';

            if ( $i < 0 ) {
                $i = $#records;
            }

            if ( $i > $#records ) {
                $i = 0;
            }
        }

        $anchor = -3;

        $tried = 0;
        while (1) {

            ( $newid, $rest ) = split( /\t/, $records[$i] );
            recRead($newid);
            $url = $rec{$urlFld};

            if ( $url eq '' && $to eq 'nextu' && $i != $anchor ) {
                $anchor = $i;
                $i++;
                $i = 0 if $i > $#records;
                $tried++;
                if ( $tried > $#records ) {
                    wbError( "No URLs", "There are no URLs in this database" );
                    return;
                }
            }
            else {
                last;
            }
        }
    }

    if ( $url eq '' ) {
        $url = "$SCRIPT/Show?_id=$newid";
    }

    $noframes = "Your browser does not support frames. Sorry.";    #UK

    $eid  = &htEscape($newid);
    $eurl = &htEscape($url);

    if ( $CGI{'fr'} ne 'bottom' ) {
        print <<EOM;
Content-type: text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">
<head>
<title>$WBB{'dbTitle'} - RING</title>
</head>
<frameset rows="33,*" frameborder="0" framespacing=0 border="0">
	<frame name="woda_toolbar" src="$SCRIPT/BrowseRingToolbar?id=$eid&amp;sort=$sort&amp;url=$eurl&amp;fr=top"  scrolling=NO marginheight="1">
	<frame name="woda_page" src="$url" >
</frameset>		
<noframes><h1>$noframes</h1></noframes>
EOM
    }
    else {
        print <<EOM;
Content-type: text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">
<head>
<title>$WBB{'dbTitle'} - RING</title>
</head>
<frameset rows="*,28" frameborder="0" framespacing=0 border="0">
	<frame name="woda_page" src="$url" >
	<frame name="woda_toolbar" src="$SCRIPT/BrowseRingToolbar?id=$eid&amp;sort=$sort&amp;url=$eurl&amp;fr=bottom"  scrolling=NO marginheight="1">
</frameset>		
<noframes><h1>$noframes</h1></noframes>
EOM
    }

    wodaExit();
}

# generate toolbar frame for BrowseRing

sub cgiBrowseRingToolbar {

    $sort = $CGI{sort};
    $id   = $CGI{'id'};
    $url  = $CGI{'url'};
    $fr   = $CGI{'fr'};

    $eid  = &htEscape( $CGI{'id'} );
    $eurl = &htEscape( $CGI{'id'} );

    $ring = "Ring of Web pages";    #UK

    $homeURL = &homeURL();

    $texts = <<EOM;
First page
Previous page
Show database record
Next page
Next page with an URL
Last page
List of all
Home page
WODA Home Page
Navigation bar at page top
Navigation bar at page bottom
Close navigation bar
EOM

    @texts = split( /\n/, $texts );

    #gif	target-url
    @links = (
        'first',     "$SCRIPT/BrowseRing?fr=$fr&amp;sort=$sort&amp;to=first&amp;id=$eid",
        'back',      "$SCRIPT/BrowseRing?fr=$fr&amp;sort=$sort&amp;to=previous&amp;id=$eid",
        'open4',     "$SCRIPT/Show?_id=$eid",
        'forward',   "$SCRIPT/BrowseRing?fr=$fr&amp;sort=$sort&amp;to=next&amp;id=$eid",
        'fforward',  "$SCRIPT/BrowseRing?fr=$fr&amp;sort=$sort&amp;to=nextu&amp;id=$eid",
        'last',      "$SCRIPT/BrowseRing?fr=$fr&amp;sort=$sort&amp;to=last&amp;id=$eid",
        '',          '',
        'browse',    "$SCRIPT/Search",
        'home',      "$homeURL",
        'poweredby', "$POWERURL",
        'topframe',  "$SCRIPT/BrowseRing?fr=top&amp;sort=$sort&amp;to=this&amp;id=$eid",
        'botframe',  "$SCRIPT/BrowseRing?fr=bottom&amp;sort=$sort&amp;to=this&amp;id=$eid",
        'close',     "$CGI{'url'}"
    );

    print <<EOM;
Content-type: text/html

<head><title>Navigation Frame</title></head>
<body
	bgcolor="#000000"
	text="#ffffff"
	topmargin="0"
>
<table
	border="0"
	width="100%"
	cellpadding="0" 
	cellspacing="0"
>
<tr valign="top">
<td align="left">
EOM

    $j = 0;
    for ( $i = 0 ; $i < $#links ; $i += 2 ) {
        $gif = $links[$i];
        $url = $links[ $i + 1 ];
        if ( $gif eq '' ) {
            print "\n";
            next;
        }
        $text = $texts[ $j++ ];
        if ( $gif eq 'open4' && $CGI{'url'} =~ /$SCRIPT/ ) {
            $gif = 'open1';
            $url = '';
            $a   = 'AX';
        }
        else {
            $a = 'A';
        }
        next if ( $gif eq 'topframe' && $fr eq 'top' );
        next if ( $gif eq 'botframe' && $fr eq 'bottom' );

        if ( $j == 9 ) {
            print <<EOM;
<td align="CENTER">
$WBB{'dbTitle'} - $ring
</td>
<td align="right">
EOM
        }
        $iconpar = $ICONPAR;
        $iconpar = 'width="58" height="21"' if $gif eq 'poweredby';

        print qq(<$a href="$url" target="_parent"><img border="0" align="top" src="$ICONURL/$gif.gif" $iconpar title="$text"  alt="$text"></$a> );
    }

    print <<EOM;
</td></tr></table>
</body></html>
EOM
    wodaExit();

}

# enable browsing using A-Z index

sub cgiBrowseAZ {

    $name = $CGI{'name'};

    @indexes = split( /\n/, $WBB{'AZindex'} );
    @indexes = grep ( /^$name\s/, @indexes );
    ( $dbname, $fields, $title ) = split( /\s+/, $indexes[0], 3 );

    if ( $dbname ne $name || $name eq '' ) {
        wbFail("Index '$name' is not defined");    #UK
    }

    $splitMethod = '';
    @indexes     = split( /\n/, $WBB{'AZindexMethod'} );
    @indexes     = grep ( /^$name\s/, @indexes );
    ( $dbname, $splitMethod ) = split( /\s+/, $indexes[0], 2 );
    $splitMethod = $splitMethod || 'ASCII,LOWER,SPACE';

    $x = "A-Z Index: ";                            #UK
    $x .= $title;
    printHead($x);

    $x = "Index of words in field(s)";             #UK
    $x .= " <b>$fields</b>.";

    print "<p>$x\n";

    ( $lcFrom, $lcTo ) = split( / /, $WBB{'intlLower'} );
    undef $lcEval;
    $lcEval = "tr/$lcFrom/$lcTo/" if $lcFrom;

    $wordFname = "_cache/words-$name.txt";
    if ( &wbExpired($wordFname) ) {

        ( $allHits, *list, *fields ) = &recSelect( '', '', '', 'TAB', $fields );

        foreach $_ (@list) {

            #$Log .= "$_\n";
            tr/[\000-\040][!-@][\[-`][{-~]/ /s if $splitMethod =~ m/ASCII/;
            tr/[A-Z]/[a-z]/                    if $splitMethod =~ m/LOWER|TITLE/;
            tr/[a-z]/[A-Z]/                    if $splitMethod =~ m/UPPER/;
            @words = split( /\s+/,   $_ ) if $splitMethod =~ m/SPACE/;
            @words = split( /\,\s*/, $_ ) if $splitMethod =~ m/COMMA/;
            @words = split( /\;\s*/, $_ ) if $splitMethod =~ m/SEMICOLON/;
            @words = split( /\:\s*/, $_ ) if $splitMethod =~ m/COLON/;
            foreach $word (@words) {
                $word = &wbTitleCase($word) if $splitMethod =~ m/TITLE/;
                $word = &wbStemWord($word)  if $splitMethod =~ m/STEM/;
                $words{$word}++ if $word ne '';
            }
        }

        # do not index by some words like a, the, who, when
        undef @words;
        undef %sort;
        ( $coFrom, $coTo ) = &wbCollate();
        $coEval = "tr/$coFrom/$coTo/" if $coFrom;

        @stopWords = split( /\s+/, $WBB{'AZnot'} );
        foreach $word ( keys(%words) ) {
            next if length($word) == 1;
            foreach $stop (@stopWords) {
                if ( $word =~ m/^$stop$/i ) {
                    $words{$word} = 0;
                    last;
                }
            }
            $_ = $word;
            &evalSafe($coEval) if $coEval;
            $sort{$word} = $_;

            # print "<br>$word $sort{$word}\n";
        }

        # sort the words and print out
        @w = keys(%sort);
        @words = sort { $sort{$a} cmp $sort{$b} } @w;

        open( h, ">$wordFname" ) || wbFail("Cannot write to file $wordFname");
        binmode(h);
        foreach $word (@words) {
            next if $words{$word} == 0;
            print h "$word\t$words{$word}\n";
        }
        close(h);

    }

    open( h, $wordFname );
    @lines = <h>;
    close(h);

    if ( $CGI{'how'} eq 'frequent' ) {
        printAZfrequent( $CGI{'n'} );
    }
    else {
        printAZaz();
    }
}

# convert string to title case

sub wbTitleCase {

    local ($_) = @_;
    local ($x);

    tr/[A-Z]/[a-z]/;
    m/(.)(.*)/;
    $x = $1;
    $x =~ tr/[a-z]/[A-Z]/;
    return $x . $2;
}

# returns word stem for some suffixes

sub wbStemWord {

    # perhaps this http://www.tartarus.org/~martin/PorterStemmer/perl.txt
    # should be included!

    local ($_) = @_;

    return $_ if s/ing$/\*/;
    return $_ if s/ed$/\*/;
    return $_ if s/s$/\*/;

}

# transform to lowercase and format as html

sub printAZaz {

    $dletter = '';
    foreach $_ (@lines) {
        chop;
        ( $word, $count ) = split(/\t/);
        $_ = substr( $word, 0, 1 );
        unless (m/[a-z]/) {
            &evalSafe($lcEval)
              if $lcEval;    # do case lowering & other equivalences
            tr/[A-Z]/[a-z]/; # lower case
        }
        $sword = &htEscape($word);
        $entry{$_} .= <<EOM;
<a href="$SCRIPT/Search?search=$sword&amp;paint=1">$word</a> <em>($count)</em>
EOM
        if ( $_ ne $dletter ) {
            push( @letters, $_ );
            $dletter = $_;
        }
    }

    # --- print contents

    print "<h2><a name=\"az\">\n</a>\n";

    foreach $_ (@letters) {
        print "<a href=\"#$_\">$_</a>\n";
    }

    print "</h2>\n\n";

    $t = "Browse most frequent.";    #UK
    print <<EOM;
<p><a href="$SCRIPT/BrowseAZ?name=$CGI{'name'}&amp;how=frequent">$t</a>
<hr>
EOM

    foreach $_ (@letters) {
        @items = split( /\n/, $entry{$_} );
        $x = &formatColumns( 5, @items );
        print <<EOM;

<h2>
<a href=#az><img src="$ICONURL/up.gif" $ICONPAR align="BOTTOM" alt=up title=up></a>
<a name="$_">$_</a>
</h2>
$x
EOM

    }
}

# print list of most frequent words

sub printAZfrequent {

    $maxWords = $CGI{'n'} || 50;

    foreach $_ (@lines) {
        chop;
        ( $word, $count ) = split(/\t/);
        $_ = $word;

        # this should have been done at index building!
        #		&evalSafe($lcEval) if $lcEval;	# do case lowering & other equivalences
        #		tr/[A-Z]/[a-z]/;		# lower case
        $count{$_} += $count;
    }

    @words = keys(%count);
    @sortedWords = sort { ( ( 999999 - $count{$a} ) . $a ) cmp( ( 999999 - $count{$b} ) . $b ) } @words;

    if ( $#sortedWords + 1 < $maxWords ) {
        $maxWords = $#sortedWords + 1;
    }

    print "<h2>Most frequent $maxWords words</h2><ol>";

    undef @items;
    for ( $i = 0 ; $i < $maxWords ; $i++ ) {
        $word  = $sortedWords[$i];
        $sword = &htEscape($word);
        $count = $count{$word};
        print <<EOM;
<li><a href="$SCRIPT/Search?search=$sword">$word</a> <em>($count)</em></li>
EOM
    }

    print "\n</ol>\n";
}

# read counted number of categories

sub readCatCount {

    local ( $field, $field2 ) = @_;
    local ($file) = "_cache/count-$field-$field2.txt";
    local (@rows);
    if ( &wbExpired($file) ) {
        $Log .= "Rebuilding count file $file\n";
        $counts = &countCat( $field, $field2 );
        open( H, ">$file" ) || wbFail("Cannot open $file for writing\n");    #UK
        binmode(H);
        print H $counts;
        close(H);
    }

    open( H, $file );
    @rows = <H>;
    close(H);
    foreach $row (@rows) {
        chop($row);                                                          # newline
        ( $cat, $count ) = split( /\t/, $row );
        $catCount{$cat} = $count;
    }
}

# browse by fields in a tree

sub cgiBrowseTree {

    local ($how) = @_;

    $yh = 1 unless $how =~ /H/;
    $yf = 1 unless $how =~ /F/;

    $field     = $CGI{'field'};
    $separator = $CGI{'separator'} || ':';
    $value     = $CGI{'value'};
    $order     = $CGI{'order'};
    $recurse   = $CGI{'recurse'} || 0;

    readCatCount($field);

    $head = &formatFieldName($field);    # may return hyperlink ... results in two nested A links
                                         # --- the stuff above

    @subs = split( /$separator/, $value );
    if ( $#subs == -1 && $value ne '' ) {
        @subs = ($value);
    }

    $base = "$SCRIPT/BrowseTree?field=$field&amp;separator=$separator&amp;recurse=$recurse&amp;order=$order";

    $b    = "Browse by ";                         #UK
    $tit  = "$b <a href=\"$base\">$head</a> :";
    $last = pop(@subs);
    foreach $sub (@subs) {
        $esub = &htEscape($sub);
        $asub .= ":$sub";                         # dodano
        $asub =~ s|^$separator||o;                # pred ta prvim ni :
        $esub = &htEscape($asub);                               # dodano
        $text = "<a href=\"$base&amp;value=$esub\">$sub</a>";
        $tit   .= " $text $separator";
        $super .= "<b>$text</b><br>\n";
    }

    $tit   .= " $last";
    $super .= "<b>$last</b><br>\n";

    printHead($tit) if $yh;

    # --- subcategories

    $x   = 'Subcategories';        #UK
    $y1  = '(more)';               #UK
    $y2  = '(less)';               #UK
    $y   = $y1;
    $y   = $y2 if $recurse;
    $e   = &htEscape($value);
    $alt = "$base&amp;value=$e";

    # --- order

    @y = (
        'by count', 'by reverse count', 'a-z', 'predefined'    #UK
    );

    @x = ( 'COUNT', 'RCOUNT', 'AZ', 'ASIS' );

    $aalt = $alt;
    $aalt =~ s/&amp;order=[A-Z]*//;

    $order = $order || 'ASIS';

    for ( $i = 0 ; $i <= 3 ; $i++ ) {
        if ( $x[$i] eq $order ) {
            $ord .= $y[$i];
        }
        else {
            $ord .= qq(<a href="$aalt&amp;order=$x[$i]">$y[$i]</a>);
        }
        $ord .= ', ';
    }

    chop($ord);
    chop($ord);
    $ord .= '.';

    $xx = 're-order';    #UK

    if ( $recurse == 1 ) {
        $alt =~ s/recurse=1/recurse=0/;
    }
    else {
        $alt =~ s/recurse=0/recurse=1/;
    }

    print "<h3>$x <a href=\"$alt\">$y</a> $xx: $ord</h3>\n" if $yh;

    # these kwords are predefined

    @kwords = &wbParseOptions($field);

    # any other found by counting?

    foreach $x ( sort( keys(%catCount) ) ) {
        push( @kwords, $x ) unless grep ( /^$x$/, @kwords );
    }

    #@kwords = sort(@kwords);

    printTree( $value, $order );

    # --- now make a less demanding search

    $s2 = "\"$value\"";
    $x  = $value;
    $x =~ s/$separator/ /g;
    $x =~ s/\s+/ /g;
    $s2 = &htEscape("$s2 $x");

    if ( $value ne '' ) {

        $x = "Found items in $value";    #UK
        print "<h3>$x <a href=\"$SCRIPT/Search?search=$s2\">$y1</a>:</h3>\n";

        # exit through search

        $CGI{'search'} = "/$field:\"$value\"";
        cgiSearch('H');

        print "<ul>-</ul>" unless $dataHits;

        # toolbar should point to BrowseTree

        foreach $_ (@Toolbar) {
            s|/Search|/BrowseTree|g;
        }

    }
}

# prints Tree

sub printTree {

    # uses $field!

    local ( $value, $order ) = @_;

    local ( %seen, $yes, $subc, @kw, @kw1, $sep );

    # check if HEADING is used for OPTION fields

    local ($heading);

    if ( $WBF{ $field, typePar } =~ m/HEADING=(.{1,5})/i ) {    # take the 5 first letters from HEADING
        $heading = $1;
    }
    else {
        $heading = '';
    }

    $sep = $separator;
    $sep = '' if $value eq '';

    if ( $value ne '' ) {                                       # in perl4 // does not match anything
        @kw = grep( /^$value/, @kwords );
    }
    else {
        @kw = @kwords;
    }

    # --- find subcategories

    foreach $cat (@kw) {
        if ( $cat =~ m/^$value$sep([^$separator]+)/ ) {
            $sub = $1;                                          # subcategory
            push( @kw1, "$value$sep$sub" ) unless $seen{$sub};
            $seen{$sub} = 1;
        }
    }

    #print join('<br>',@kw1);

    # here they are !

    foreach $cat (@kw1) {

        next if ( $heading && $cat =~ m/^$heading/ );           # heading categories will not be counted

        $c1 = $catCount{$cat} || 0;                             # in this subcategory
        undef @x;

        @x  = grep( /^$cat/, @kw );
        $c3 = $#x + 1;                                          # num of categories
        $c2 = 0;                                                # in subcategories
        foreach $x (@x) {
            next if $x eq $cat;
            $c2 += $catCount{$x};
        }
        $c0       = $c1 + $c2;                                  # all
        $c0{$cat} = $c0;
        $c1{$cat} = $c1;
        $c2{$cat} = $c2;
        $c3{$cat} = $c3;
    }

    if ( $order eq 'COUNT' ) {
        @kw1 = sort { $c0{$b} <=> $c0{$a} } @kw1;
    }
    elsif ( $order eq 'RCOUNT' ) {
        @kw1 = sort { $c0{$a} <=> $c0{$b} } @kw1;
    }
    elsif ( $order eq 'AZ' ) {
        @kw1 = sort @kw1;
    }
    elsif ( $order eq 'ASIS' || $order eq '' ) {

        # keep as is
    }

    # --- print

    print "<ol>";
    foreach $cat (@kw1) {
        $nicecat = &DECODE( $field, $cat, 'SHORT' );

        # ampak poloviti je treba se vse ostale!
        $evlue = &htEscape($cat);

        # heading values - tested here because it can come from a related table
        if ( $heading && $cat =~ m/^$heading/ ) {    # $cat is not escaped, in case of alphanumeric values in HEADINGS
            print "<h5>$nicecat</h5>\n";
            next;
        }
        print "<li><a href=\"$base&amp;value=$evlue\"><b>$nicecat ($c0{$cat})</b></a></li>\n";
        print "($c2{$cat} in $c3{$cat} subcategories)\n" if $c2{$cat} > 0;    #UK
        printTree( $cat, $order ) if $CGI{'recurse'};
    }

    # print "-" unless $yes;
    print "</ol>\n";
}

# trims @list from $first to $last

sub wbTrimList {
    local ( *list, $first, $last ) = @_;

    $last = $#list if $#list < $last;
    @list = @list[ $first .. $last ];
    return '';
}

# if needed or $forced,
# reindexes the table and updates the $sort.tbl file
# ... 34" on olivetti

sub wbReindex {
    local ( $sort, $force ) = @_;
    local ( $now, $file, $id, $pos, $len, $time, $string );
    local ( %pos, %sort, %print );

    $now = time;

    $file = "_cache/$sort.tbl";
    $file =~ tr/[A-Z]/[a-z]/;

    unless ( $force || ( $tblm = &wbExpired($file) ) ) {

        # no need to do anything
        return;
    }

    if ( -e "$file.wrk" ) {
        if ( ( time - &wbMtime("$file.wrk") ) > 1200 ) {    # give up after 20 minutes
            unlink("$file.wrk");
            wbWarn( "Reindexing", "Please wait ..." );      #UK
        }
        elsif ( ( time - &wbMtime("$file.wrk") ) > 300 ) {
            wbWarn( "Reindexing", "Reindexing is taking longer than 5 minutes. Old index is being used." );
            return;
        }
        else {
            wbError( "Table is being reindexed.", "Please wait a minute. If the problem persists inform the manager" );    #UK
        }
    }

    logTime("Begin re-indexing table");

    open( hhh, ">$file.wrk" )
      || wbFail("Cannot reindex table and open $file.wrk");
    binmode(hhh);

    wbSortFields();
    print hhh join( "\t", @IFields );
    print hhh "\n" || wbFail("Cannot reindex table, disk full?");

    ( $coFrom, $coTo ) = &wbCollate();
    ( $lcFrom, $lcTo ) = split( / /, $WBB{'intlLower'} );
    undef $coEval;
    undef $lcEval;
    $coEval = "tr/$coFrom/$coTo/" if $coFrom;
    $lcEval = "tr/$lcFrom/$lcTo/" if $lcFrom;

    logTime("Before reading FAT");

    open( h, "$DATADIR/_data/fat.txt" );    # fat may not exist! ||wbFail("Cannot open fat.txt");
    while (<h>) {
        ( $id, $pos ) = split( /\t/, $_, 2 );
        $pos{$id} = $pos;
    }
    close(h);

    logTime("Parsed FAT");

    @keys = sort { $pos{"$a"} <=> $pos{"$b"} } keys(%pos);

    logTime("Optimised FAT - $#keys+1 records");

    open( h, "$DATADIR/_data/records.txt" );

    $sortexp = $WBB{"sort;$sort"};

    undef @badFatIDs;

    while ( ( $id, $pos ) = each(%pos) ) {

        next unless $pos >= 0;    # deleted record

        unless ( seek( h, $pos, 0 ) ) {
            push( @badFatPos, $id );
            next;
        }

        $meta = <h>;
        chop($meta);
        ( $len, $time ) = split( /\t/, $meta );

        unless ( read( h, $string, $len ) ) {
            push( @badFatPos, $id );
            next;
        }

        recFromVector($string);

        if ( $rec{'_at'} eq '' ) {

            #wbFail("$id","$rec{'_at'}");
            # so what if _at does not exist
        }

        if ( $sort eq 'TIME' ) {
            $sort{$id} = $rec{'_at'};
        }
        elsif ( $WBB{"sort;$sort"} ) {
            $_ = &evalSafe($sortexp);
            s/^\s*//;    # ignore leading spaces!
            &evalSafe($lcEval)
              if $lcEval;    # do case lowering & other equivalences
            tr/[A-Z]/[a-z]/; # lower case
            &evalSafe($coEval) if $coEval;    # different collate
            $sort{$id} = $_;

            # $Log .= "sort{$id}=$sort{$id}\n";
        }

        $print{$id} = &recToRow(@IFields);
    }

    close(h);

    # fat entries pointing somewhere strange ...
    # delete thoese records but with code -2
    # so that we know how they were corruped

    if ( $#badFatPos >= 0 ) {

        #locking ????
        open( h, ">>$DATADIR/_data/fat.txt" );
        flock( h, 2 );
        binmode(h);
        foreach $id (@badFatPos) {
            print h "$id\t-2\n";
        }
        close(h);
    }

    logTime("Have new data");

    if ( $sort eq 'TIME' ) {
        @list = sort { $sort{$b} <=> $sort{$a} } keys(%print);
    }
    else {
        @list = sort { $sort{$a} cmp $sort{$b} } keys(%print);
    }

    logTime("Sorted");

    foreach $id (@list) {
        print hhh $print{$id} . "\n";
    }
    close(hhh);

    unlink $file;
    rename( "$file.wrk", "$file" )
      || wbFail("Failed rename $file.wrk to $file");

    logTime("Table reindexed");

    open( h, ">$file.time" );
    $x = time - $now;
    print h "$x\n";
    print h "This is how many seconds the last reindexing took\n";
    close(h);
}

# write file to disk

sub wbUserFileSave {

    local ( $field, $tmpID ) = @_;
    local ( $fileName, $ext, $destDir );

    $destDir = &wbCheckDir("$DATADIR/_tmp");

    # find file's extension

    $ext = $EXT{$field};
    $ext = '.' . $ext if $ext;

    $fileName = "$tmpID.$field$ext";

    open( h, ">$destDir/$fileName" ) || wbFail("Cannot open $fileName\n");
    binmode(h);
    print h $CGI{$field};
    close h;

    if ( -e "$destDir/$fileName" ) {
        return $fileName;
    }
    else {
        return "";
    }
}

# escapes special regular expression characters in a string

sub regexpEsc {

    local ($_) = $_[0];
    s/(\W)/\\$1/g;
    return $_;
}

# ht ###################################################################

# demands that cookies @_ are written by the
# next printHead routine

sub htDemandCookie {

    $DemandedCookies .= join( '', @_ );
}

# format cookie

sub htFormatCookie {

    local ( $name, $value, $path, $time ) = @_;

    if ( $path ne '' ) {

        # $path = $path . '/' unless $path =~ m|/$|; # some explorers want cookies with a trailing slash ??
        $path = "path=$path;";
    }

    $path =~ s|:80/|/|;

    if ( $time eq '' ) {
        $time = 'expires=Wed, 1-Jan-2020 23:59:59 GMT';
    }
    elsif ( $time eq "RESET" ) {    # in the past
        $time = 'expires=Sun, 30-Nov-1997 23:59:59 GMT';
    }
    elsif ( $time eq "SESSION" ) {    # after session
        $time = '';
    }

    return '' if $name  eq '';
    return '' if $value eq '';

    $value = &htEscape($value);
    return "Set-Cookie: $name=$value; $path $time\n";
}

# parse key=value strings
# called only by htParseMulti which itself was called by sub run with multiline
# TODO: check against woda-max-pro

sub htParseSimple {
    local ($raw_data) = $_[0];
    local ( @items, $key, $value );

    # Split different "NAME" and their values.

    unless ( $raw_data =~ m/\=/m ) {    # multiline matching
        $_ = $raw_data;
        $_ =~ tr/+/ /;
        $_ =~ s/%(..)/pack("C", hex($1))/egm;    # multiline matching
        $CGI{'_isindex'} = $_;
        return;
    }

    # should we do multiline here ?
    @items = split( '&', $raw_data );

    # For each list of "NAME=its_value".

    for (@items) {

        $_ =~ tr/+/ /;

        if (m/\=/m) {    # multiline matching
            ( $key, $value ) = split( '=', $_, 2 );
        }
        else {
            ( $key, $value ) = ( '_isindex', $_ );
        }

        # The %xx hex numbers are converted to alphanumeric.
        $key   =~ s/%(..)/pack("C", hex($1))/eg;    # multiline matching
        $value =~ s/%(..)/pack("C", hex($1))/eg;    # multiline matching
        if ( defined $CGI{$key} ) {
            $CGI{$key} = $CGI{$key} . "\x0A" . $value;
        }
        else {
            $CGI{$key} = $value;
        }

    }

}

# get data passed from POST, GET or  STDIN

sub htGetData {
    local ( $buffer, $toread, $read );

    if ( $ENV{'REQUEST_METHOD'} eq "POST" ) {
        $toread = $ENV{'CONTENT_LENGTH'};
        $Log .= "post method - $toread bytes!\n";

        #if ($ENV{'SERVER_SOFTWARE'}=~ m/woda/i) {
        #	$read = read(server'NS,$buffer,$toread,0);
        #} else {
        binmode(STDIN);
        $read = read( STDIN, $buffer, $toread, 0 );

        #}
        if ( $read != $toread ) {
            wbFail("Read $read bytes from STDIN, should read $toread\n");
        }

        return $buffer;

    }
    elsif ( $ENV{'REQUEST_METHOD'} eq "GET" ) {
        return $ENV{'QUERY_STRING'};

    }
    else {
        return 0;

    }
}

# parse key=value strings
# called from sub run by  $*=1; do htParseMulti(); $*=0; (multiline matching in regexp)
# called only from sub run, same in old Woda 3505
# TODO: check against woda-max-pro

sub htParseMulti {

    local ( $boundary, @pairs, $position );
    local ( $raw_data, $value, $name, $part, $type );

    # do printHead("xxx");

    undef %CGI;
    $raw_data = &htGetData();    # QUERYSTRING if Get or STDIN if Post

    if ( $ENV{'CONTENT_TYPE'} =~ m|text/xml| ) {
        $CGI{'xml'} = $raw_data;
        return;
    }

    if ( $ENV{'CONTENT_TYPE'} =~ m|text/csv| ) {
        $CGI{'csv'} = $raw_data;
        return;
    }

    if ( $ENV{'CONTENT_TYPE'} =~ m|multipart/form-data| ) {

        # continue in this subroutine below

    }
    else {

        # all other cases but multipart/form-data
        &htParseSimple($raw_data);
        return;
    }

    # multipart/form-data
    # used for FILE|IMAGE|USERFILE fields, AdmUploadUserFile and AdmImport
    # a CreateFromData cgi function will use it

    # --- MIME style

    #do printHead("boundary stuff");
    #print length($raw_data),"---",$ENV{'CONTENT_LENGTH'},"---$nread";
    #do printENV();

    #	$Log .= $raw_data;

    #print "<pre>$raw_data</pre>";
    #$xxx = $raw_data;

    # "boundary=" is the parameter required for Content-Type field in multipart
    ( $boundary = $ENV{CONTENT_TYPE} ) =~ s/^.*boundary=(.*)$/$1/m;    # get the boundary value, seek in multilines
    @pairs = split( /--$boundary/m, $raw_data );                       # boundaries begins with --, seek in multilines
    @pairs = splice( @pairs, 1, $#pairs );

    #print "<pre>$raw_data";
    #print join('<br>',@pairs);
    #print "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";

    foreach $part (@pairs) {

        if ( $part =~ m/\x0D\x0A$/m ) {
            $CRLF = "\x0D\x0A";
        }
        else {
            $CRLF = "\x0A";
        }
        ( $head, $content ) = split( /$CRLF$CRLF/m, $part, 2 );
        chop $content;
        chop $content if length($CRLF) == 2;

        $head =~ s/\r//gm;
        $head =~ s/\n/;/gm;    # replace new lines with ";"
        (@headers) = split( /\s*;\s*/m, $head );    # do we still need multiline (/m) here and bottom ?
        undef %hdr;
        foreach $item (@headers) {
            next if ( $item eq '' || $item eq '--' );
            ( $name, $value ) = split( /=|:/m, $item, 2 );
            $value =~ m/\s*"*([^"]*)/m;
            $value = $1;
            $hdr{$name} = $value;

            #print "<br>hdr($name)=($value)\n";
        }

        #print "<br>",%hdr;

        # store into variables

        $name = $hdr{'name'};
        next if $name eq '';

        #print "<br>CGI($name)\n";
        if ( defined $CGI{$name} ) {
            $CGI{$name} = $CGI{$name} . "\x0A" . $content;
        }
        else {
            $CGI{$name} = $content;
        }

        if ( $hdr{filename} ) {
            $hdr{filename} =~ m/([^\.]*)$/m;
            if ($`) {
                $EXT{$name} = $1 if $`;
                $EXT{$name} =~ tr/[A-Z]/[a-z]/;
            }
            else {
                $EXT{$name} = '';
            }
            $C_TYPE{$name} = $hdr{"Content-Type"};
        }

        #print "<br>",length($content),"content=($ddcontent)";

    }

}

# escapes charaters in strings that are > 127 into UTF8 encoding

sub htUTF8 {

    local ($_) = @_;
    $_ =~ s/([\200-\377])/&htUTF8help(ord($1))/ge;
    return $_;

}

# escapes characters >127 into like &#x00FF for XML
# this is correct for ISO Latin 1 ... not correct for Windows 1250
# but should travel well

sub htUTF8help {
    local ($_) = @_;
    return '&#x00' . ( 0 .. 9, 'a' .. 'f' )[ $_ / 16 ] . ( 0 .. 9, 'a' .. 'f' )[ $_ & 15 ] . ';';
}

# escapes non alphanums into %hex ... as for URLs

sub htEscape {
    local ($x) = @_;
    $x =~ s/(\W)/&htToHex(ord($1))/ge;
    return $x;
}

# quote special html characters

sub htQuote {
    local ($_) = $_[0];

    s/\&/\&amp;/g;
    s/\>/\&gt;/g;
    s/\</\&lt;/g;
    s/\"/\&quot;/g;
    s/\'/\&#39;/g;    # &apos; not generally used by browsers !!!

    return $_;
}

#
# convert cookies into Assoc array
#

sub htReadCookies {

    local ( $n, $v, $x, @x );

    @x = split( /;/, $ENV{'HTTP_COOKIE'} );
    foreach $x (@x) {

        # do printHead ("$x");
        ( $n, $v ) = split( /\=/, $x );

        # do printHead("$n=$v"); to ga nekaj lomi na win32
        $n =~ s/\s//g;
        $v =~ s/\+/ /g;
        $v =~ s/%([\da-f]{1,2})/pack(C,hex($1))/eig;
        $Log .= "cookie $n = $v\n";
        $COOKIE{$n} = $v;

    }
}

# convert ascii value to hex

sub htToHex {
    local ($_) = @_;
    return '%' . ( 0 .. 9, 'a' .. 'f' )[ $_ / 16 ] . ( 0 .. 9, 'a' .. 'f' )[ $_ & 15 ];
}

# convert hex value to ascii

sub htUnescape {

    $_[0] =~ s/\+/ /g;
    $_[0] =~ s/%([\da-f]{1,2})/pack(C,hex($1))/eig;
    return $_[0];
}

# prints settings of the database
# WBBs matching bpatt
# WBFs matching fpatt
# legends in legends set to 1

sub printStructure {

    local ( $bpatt, $fpatt, $legends, $validate ) = @_;

    admMeta();

    print '<h2>1. Database settings</h2>';    #UK
    print "\n", '<table class="DATATABLE" cellspacing="1">';

    $j = 0;
    foreach $wbase (@MetaB) {

        # headers

        if ( $MetaB{ $wbase, 'what' } eq 'H' ) {
            next unless $wbase =~ m/$bpatt/;
            $hdr = <<EOM;
<tr class="DATAH" >
<td align="CENTER" COLSPAN=2 class="MENUT">
<b>$wbase</b>
</td>
</tr>
EOM
        }
        else {

            next unless $wbase =~ m/$bpatt/;
            next unless defined $WBB{$wbase};

            if ($hdr) {
                print $hdr;
                $hdr = '';
            }

            if ($validate) {
                ( $error, $x ) = &admMetaValidate( $WBB{$wbase}, $MetaB{ $wbase, 'type' } );
                $error = &htQuote($error);
                $error = '<p class="ERROR">' . $error . '</p>';
            }
            else {
                $error = '';
            }

            $value = &htQuote( $WBB{$wbase} );
            $about = $MetaB{ $wbase, 'uk' };
            if ($legends) {
                $about = "<td>$about aaa </td>\n";
            }
            else {
                $about = '';
            }
            print <<EOM;
<tr class="DATA1"  valign="top">
<td align="right">
<a name="_b$wbase">
<b>$wbase</b>
</a>
</td>$about<td valign="top">
<pre>$value</pre>
$error
</td>
</tr>
EOM
        }
    }

    print "</table>\n";

    print "<h2>2. Field settings</h2>\n";    #UK

    wbSortFields();

    $nf = 0;
    foreach $field (@Fields) {
        $nf++;
        print <<EOM;
<table class="DATATABLE" cellspacing="1" width="80%">
<tr class="DATAH" ><th COLSPAN=2 align="left"><a name="$field">$field</a></th></tr>
EOM
        $j = 1;
        foreach $attr (@MetaF) {
            next unless $attr =~ m/$fpatt/;
            next unless defined $WBF{ $field, $attr } || $attr eq "";

            $x = $WBF{ $field, $attr };
            $x = $WBF{$field} if $attr eq '';

            $x = &htQuote($x);

            $a = $attr;
            $a = "<em>valid-if</em>" if $attr eq "";

            if ( $a eq 'options' ) {
                @x = &wbParseOptions($field);
                $x = join( ',<br>', @x );
            }

            $j = !$j + 0;
            print qq(<tr class="DATA1" ><td align="right" valign="top" width="20%"><b>$a</b>\n</td>\n<td valign="top" width="80%">$x</td></tr>\n);

        }
        print "</table><br>\n";

    }

    print <<EOM;    #UK
<p>The <a href="$SCRIPT/AdmConfAttrs">legend</a> of the above values is available as well.
EOM
}

# agent ################################################################

# display form to add agent search

sub cgiAgentAddForm {

    #	shodi cgi
    #	do saveFormCan('AgentAddForm');	# sets $SaveFormButton, sets %CGI

    printHead("Subscribe to a search");    #UK

    if ( $CGI{'select'} ) {
        wbError( "Error", "This order cannot be modified" );    #UK
    }

    print <<EOM;                                                #UK
<p>This page lets you save a search, so that you will be receiving results of this search periodically to the specified email address.
EOM

    hashImport( *x, $CGI{'formdata'} );

    undef @v;
    undef @p;

    # name

    push( @p, "Name of this search:" );                                   #UK
    push( @v, qq(<input name="_formname" value="$CGI{'_formname'}">) );

    # email address unverified !

    $email = $COOKIE{_email};
    push( @p, "Your email address:" );                                    #UK
    push( @v, qq(<input name="_email" value="$email">) );

    # when to do it

    @daysVal = ( 'x',     '0',   '1',   '2',   '3',   '4',   '5',   '6',       'm1',  'm8',  'm16',  'm24',  'm28' );
    @days    = ( 'Never', 'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat<br>', '1st', '8th', '16th', '24th', '28th in the month' );    #UK
    $i       = 0;
    $days    = '';
    foreach $day (@days) {
        $d    = '';
        $name = "day:$daysVal[$i]";
        $d    = 'CHECKED' if $CGI{$name} eq 'on';
        $days .= qq(<input name="$name" type="checkbox" $d>$day \n);
        $i++;
    }
    push( @p, "When to search:" );                                                                                                     #UK
    push( @v, $days );

    $df = &htEscape( $CGI{'formdata'} );
    $t  = time;

    #	$pf = $CGI{'formdata'};
    #	$pf =~ s/\037/, /g;
    push( @p, '' );
    push(
        @v, qq(
<input type="hidden" name="_form" value="AgentAddForm">
<input type="hidden" name="since" value="$t">
<input type="hidden" name="formdata" value="$df">
)
    );

    push( @p, '' );
    push( @v, qq(<input class="BUTTON" type="submit" value="OK">) );

    $x = &formatNameVal( @p, @v );

    print <<EOM;
<form class="FORM" action="$SCRIPT/AgentSave" method="post">
$x
</form>
EOM
}

# Saving agent searches

sub cgiAgentSave {

    printHead("Saving order for searches");    #UK

    hashImport( *x, &htUnescape( $CGI{'formdata'} ) );

    foreach $key ( keys(%CGI) ) {
        next if $key eq 'formdata';
        $x{$key} = $CGI{$key};
    }
    unless ( &isEmail( $x{_email} ) ) {
        wbError( "Error", "$x{_email} is not a valid email address" );    #UK
        return;
    }

    $x{_confirm} = &wbCryptPasswd( $x{_email} );
    $dform       = $x{_form};
    $formdata    = &hashExport(*x);

    if ( $COOKIE{_email} && $COOKIE{_email} eq $x{_email} ) {

        # no need to check email address

        saveFormAppend($formdata);
        saveFormSavedMessage($dform);

    }
    else {

        $r1 = int( rand(100000000) );

        open( h, ">>$DATADIR/_tmp/$r1.form" );
        binmode(h);
        print h $formdata;
        print h "\n";
        close(h);

        $message = <<EOM;    #UK
From: $WBB{'managerEmail'}
To: $x{_email}
Reply-to: $WBB{'managerEmail'}
Subject: Your request in $WBB{'dbTitle'}

Hi,

someone (you) order an agent at
$WBB{'dbTitle'} at $homeURL
to periodically search the database
and send you the results by email to
the address $x{_email}.

To confirm this order, please open this
URL:

$SCRIPTBASE$SCRIPT/AgentConfirm?key=$r1

or enter this value:

$r1

into the form shown at

$SCRIPTBASE$SCRIPT/AgentConfirm

If you don't want to subscribe to the
search just ignore this message.

Automatically,

$WBB{manager},
manager.
EOM

        if ( &sendMail($message) ) {
            wbError( "Error sending email", "Error was $message" );
        }
        else {
            print <<EOM;    #UK
You will need to <a href="$SCRIPT/AgentConfirm">confirm</a> your request by email.
A message with instructions was sent to $x{_email}.
EOM
        }

    }
}

# Agent Search Confirm

sub cgiAgentConfirm {

    $key = $CGI{'key'};

    if ( $key eq '' ) {

        printHead("Search order confirmation");    #UK
        print '<p>';

        $tit  = "Confirmation key needed";         #UK
        $form = <<EOM;                             #UK
<form action="$SCRIPT/AgentConfirm">
Key that you received by email: <input name="key" size="10">
<input class="BUTTON" type="submit" value="OK">
</form>
EOM
        print &formatWindow( $tit, $form );        #UK
        return;
    }

    unless ( open( h, "$DATADIR/_tmp/$key.form" ) ) {
        wbError( "Bad key", "The key supplied is invalid, used or outdated" );    #UK
        return;
    }

    $formdata = <h>;
    close(h);
    unlink("$DATADIR/_tmp/$key.form");

    chop($formdata);
    hashImport( *x, $formdata );

    # demand a cookie

    $cscope = $WBB{'userScope'} || $COOKIESCOPE;
    $ec = &htFormatCookie( '_email', $x{_email}, $cscope );
    &htDemandCookie($ec);

    printHead("Search order confirmed");    #UK

    $dform = $x{_form};
    saveFormAppend($formdata);
    saveFormSavedMessage($dform);
}

# send mail

sub sendMail {

    if ( $SENDMAIL eq '' ) {
        $Log .= "Sent mail via Sockets\n";
        return &sendMailSock(@_);
    }
    else {
        $Log .= "Sent mail via $SENDMAIL\n";
        return &sendMailProg(@_);
    }
}

# send mail using program defined in $SENDMAIL

sub sendMailProg {

    local ($message) = @_;

    $| = 1;

    findSendmail();

    # -t ... to,cc,bcc specified in mail body
    # -i ... ignore dots alone on lines

    open( MAIL, "|$SENDMAIL -t -i $SENDMAILFLAGS 2>&1" ) || return 1;
    print MAIL $message || return 1;
    close(MAIL) || return 1;

    $Log .= "$SENDMAIL -t -i $SENDMAILFLAGS 2>&1\n";

    return '';
}

# send mail using sockets
# returns error string or null string

sub sendMailSock {

    local ($message) = @_;

    $message =~ m/From:\s*([^\n]*)/;
    $from = $1;
    $from =~ s/\s//g;

    $message =~ m/To:\s*([^\n]*)/;
    $to = $1;
    $to =~ s/\s//g;

    unless ( &isEmail($to) ) {
        $to = &htQuote($to);
        return "ERROR: '$to' is not a valid email address";
    }

    $message =~ s/^\./\.\./gm;    # . as first character m=treat string as multiple lines
    $message =~ s/\r//g;          # remove CR
    $message =~ s/\n/\r\n/gm;     # CR-LF always

    # set port, open the sockets

    $port = ( getservbyname( 'smtp', 'tcp' ) )[2];
    $e = &wbOpenS( $SMTPHOST, $port );
    if ( $e ne '' ) {
        return "ERROR: connecting to $SMTPHOST port $port; $e.";
    }

    # start conversation

    print S "helo woda\r\n";
    $i = <S>;

    if ( $i !~ m/^[45]/ ) {
        $i = &smtpSend( $from, $to, $message );
    }

    if ( $i =~ m/^[45]/ ) {
        close(S);
        open( h, ">>$DATADIR/_log/sockMail" );
        print h time, "\tError to $to\n";
        close(h);
        return ("Sending mail to $to failed Error was: $i");
    }
    else {
        print S "quit\r\n";
        $_ = <S>;
        close(S);
        open( h, ">>$DATADIR/_log/sockMail" );
        print h time, "\tSent to $to\n";
        close(h);
        return '';
    }

}

# print mail message

sub smtpSend {

    local ( $from, $to, $message ) = @_;

    print S "mail from: $from\r\n";
    $_ = <S>;
    if (/^[45]/) { return $_; }

    print S "rcpt to: $to\r\n";
    $_ = <S>;
    if (/^[45]/) { return $_; }

    print S "data\r\n";
    $_ = <S>;
    if (/^[45]/) { return $_; }

    print S "X-Mailer: Woda http://woda.sourceforge.net/\r\n";
    print S "$message";
    print S "\r\n.\r\n";

    $_ = <S>;
    if (/^[45]/) { return $_; }

    return '';
}

# connects stream S to $port of $host
# returns 1 on success, 0 or less on failure

sub wbOpenS {

    local ( $them, $port ) = @_;

    return "Host not defined" if $them eq '';

    local ( $sockaddr, $aliases, $proto, $type, $len, $thisaddr, $thataddr, $x );

    $sockaddr = 'S n a4 x8';
    ( $name, $aliases, $proto ) = getprotobyname('tcp');
    ( $name, $aliases, $type, $len, $thisaddr ) = gethostbyname('localhost');
    ( $name, $aliases, $type, $len, $thisaddr ) = gethostbyname($name);

    # funny but without the above line it crashes on win95

    ( $name, $aliases, $type, $len, $thataddr ) = gethostbyname($them);

    $this = pack( $sockaddr, $AF_INET, 0,     $thisaddr );
    $that = pack( $sockaddr, $AF_INET, $port, $thataddr );

    socket( S, $AF_INET, $SOCK_STREAM, $proto ) || return "socket failed";
    bind( S, $this ) || return "bind failed";
    connect( S, $that ) || return "connect to $name failed";
    binmode(S);

    $x = select(S);
    $| = 1;
    select($x);

    return '';
}

# locate sendmail program

sub findSendmail {

    return if -e $SENDMAIL;

    $SENDMAIL = '/usr/libexec/sendmail/sendmail';
    return if -e $SENDMAIL;

    $SENDMAIL = '/usr/sbin/sendmail';
    return if -e $SENDMAIL;

    $SENDMAIL = '/usr/lib/sendmail';
    return if -e $SENDMAIL;

    $SENDMAIL = '/usr/bin/sendmail';
    return if -e $SENDMAIL;

    wbFail( '$SENDMAIL not configured properly', "Cannot execute $SENDMAIL" );
}

# admin ################################################################

# process agent requests

sub cgiAdmAgentProcessForm {

    printHead("Processing agent requests");

    $pattUser  = $CGI{'users'}  || '.';
    $pattEmail = $CGI{'emails'} || '.';
    $force     = $CGI{'force'};
    $where     = $CGI{'where'};

    print "<pre>";

    wbCheckDir('_friend');
    chdir("$DATADIR/_friend")
      || wbFail( "ERROR", "Cannot chdir to _friend" );

    @agentRecords = &wbDir('.*\.forms$');

    @x     = localtime(time);
    $dDay  = $x[6];
    $dDate = 'm' . $x[3];

    foreach $_ (@agentRecords) {

        m/(.*).forms$/;
        $friend = $1;

        if ( $friend =~ m/$pattUser/ ) {
            print "\nPROCESSING $friend\n";
        }
        else {
            print "\nSkipping $friend\n";
            next;
        }

        @rows = &hashRead("$DATADIR/_friend/$friend.forms");
        @rows = grep( /_form\037AgentAddForm/, @rows );

        foreach $row (@rows) {

            hashImport( *row, $row );

            $email = $row{_email};
            next if $email !~ m/$pattEmail/;
            next unless &isEmail($email);
            next unless &wbCryptPasswd($email) eq $row{'_confirm'};

            print "handling $row{'_formid'}\n";

            unless ( $row{"day:$dDay"} eq 'on' || $row{"day:$dDate"} eq 'on' ) {
                print "... today no need\n";
                next unless $force;
                print "... force\n";
            }

            %CGI = %row;
            if ( $force == 2 ) {
                $CGI{'since'} = 1;
            }
            else {
                $CGI{'since'} = $row{'since'};
            }

            $xx = $$;

            # --- now we do the search

            wbCheckDir($TMPDIR);
            $agentFile = "$TMPDIR/agent-process.$xx.htm";
            open( AGENTFILE, ">$agentFile" );
            binmode(AGENTFILE);
            local ($selected);
            $selected = select(AGENTFILE);
            chdir $DATADIR;
            $HTMLFILE = 1;
            wbDo("cgiSearch");
            print "\n</body></html>";
            $HTMLFILE = 0;
            chdir '_friend';
            select($selected);
            close AGENTFILE;

            open( h, $agentFile );
            $html = join( '', <H> );
            $ascii = &formatHtm2Txt($html);
            $ascii =~ s/\n\n/\n/g;
            close(h);
            unlink($agentFile);

            $html = "Content-type:text/html\n\n" . $html;

            # restore

            $Group = $agentGroup;

            print "... $allHits hits\n";

            if ( $where eq 'MAIL' || $where eq 'SCREEN' ) {
                $to = $email;
            }
            else {
                $to = $WBB{'managerEmail'};
            }

            if ( $allHits > 0 ) {
                $bo      = rand(10000000);
                $bo      = "==========$bo";
                $subject = "$allHits new hits in $ap[0]";    #UK
                $ascii   = <<EOM;                            #UK
Attached is an HTML file with $allHits new records
from $WBB{dbTitle} database.

Here's an ASCII transcript:

$ascii

Regards,

Angie
EOM
                $headers = <<EOM;
From: "WODA Aganet" <$WBB{managerEmail}>
Reply-to: $WBB{managerEmail}
To: $to
Subject: $subject
Mime-Version: 1.0
Content-Type: multipart/mixed; boundary="$bo"
EOM

                $message = <<EOM;
$headers

This is a multi-part message in MIME format.

--$bo
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit

$ascii
--$bo
$html
--$bo--
EOM

                if ( $where eq 'SCREEN' ) {
                    $h = &htQuote($headers);
                    print <<EOM;
</pre>
<table class="DATATABLE" cellspacing="0"><tr><td>
<pre>$h
</pre></td></tr><tr><td>
<pre>$ascii</pre>
</td></tr></table>
<pre>
EOM
                }
                elsif ( $where ne 'NULL' ) {

                    if ( !&sendMail($message) ) {
                        push( @done, "$row{_formid} " . time );
                        print "... message sent to $to\n";
                    }
                    else {
                        print qq(... <em class="ERROR">ERROR: Sendmail to $to failed</em>\n);
                    }
                }

            }    # hits

        }    # rows of one user

        saveFormUpdateSince( $friend, @done );
        print "Updated .forms file for $friend\n";

    }    # all users

    print "</pre>\n";
}

# How to process agent requests

sub cgiAdmAgentProcess {

    printHead("Processing agent requests");

    $pattUser  = $CGI{'users'}  || '.';
    $pattEmail = $CGI{'emails'} || '.';
    $force     = $CGI{'force'};
    $where     = $CGI{'where'};

    print "<pre>";

    wbCheckDir('_friend');
    chdir("$DATADIR/_friend") || wbFail( "ERROR", "Cannot chdir to _friend" );

    @agentRecords = &wbDir('.*\.forms$');

    @x     = localtime(time);
    $dDay  = $x[6];
    $dDate = 'm' . $x[3];

    foreach $_ (@agentRecords) {

        m/(.*).forms$/;
        $friend = $1;

        if ( $friend =~ m/$pattUser/ ) {
            print "\nPROCESSING $friend\n";
        }
        else {
            print "\nSkipping $friend\n";
            next;
        }

        @rows = &hashRead("$DATADIR/_friend/$friend.forms");
        @rows = grep( /_form\037AgentAddForm/, @rows );

        foreach $row (@rows) {

            hashImport( *row, $row );

            $email = $row{_email};
            next if $email !~ m/$pattEmail/;
            next unless &isEmail($email);
            next unless &wbCryptPasswd($email) eq $row{_confirm};

            print "handling $row{_formid}\n";

            unless ( $row{"day:$dday"} eq 'on' || $row{"day:$ddate"} eq 'on' ) {
                print "... today no need\n";
                next unless $force;
                print "... force\n";
            }

            %CGI = %row;
            if ( $force == 2 ) {
                $CGI{'since'} = 1;
            }
            else {
                $CGI{'since'} = $row{'since'};
            }

            $xx = $$;

            # --- now we do the search

            wbCheckDir($TMPDIR);
            $agentFile = "$TMPDIR/agent-process.$xx.htm";
            open( AGENTFILE, ">$agentFile" );
            binmode(AGENTFILE);
            local ($selected);
            $selected = select(AGENTFILE);
            chdir $DATADIR;
            $HTMLFILE = 1;
            wbDo("cgiSearch");
            print "\n</body></html>";
            $HTMLFILE = 0;
            chdir '_friend';
            select($selected);
            close AGENTFILE;

            open( h, $agentFile );
            $html = join( '', <h> );
            $ascii = &formatHtm2Txt($html);
            $ascii =~ s/\n\n/\n/g;
            close(h);
            unlink($agentFile);

            $html = "Content-type:text/html\n\n" . $html;

            # restore

            $Group = $agentGroup;

            print "... $allHits hits\n";

            if ( $where eq 'MAIL' || $where eq 'SCREEN' ) {
                $to = $email;
            }
            else {
                $to = $WBB{'managerEmail'};
            }

            if ( $allHits > 0 ) {
                $bo      = rand(10000000);
                $bo      = "==========$bo";
                $subject = "$allHits new hits in $ap[0]";    #UK
                $ascii   = <<EOM;                            #UK
Attached is an HTML file with $allHits new records
from $WBB{'dbTitle'} database.

Here's an ASCII transcript:

$ascii

Regards,

Angie
EOM
                $headers = <<EOM;
From: "WODA Aganet" <$WBB{'managerEmail'}>
Reply-to: $WBB{'managerEmail'}
To: $to
Subject: $subject
Mime-Version: 1.0
Content-Type: multipart/mixed; boundary="$bo"
EOM

                $message = <<EOM;
$headers

This is a multi-part message in MIME format.

--$bo
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit

$ascii
--$bo
$html
--$bo--
EOM

                if ( $where eq 'SCREEN' ) {
                    $h = &htQuote($headers);
                    print <<EOM;
</pre>
<table class="DATATABLE" cellspacing="0"><tr><td>
<pre>$h
</pre></td></tr><tr><td>
<pre>$ascii</pre>
</td></tr></table>
<pre>
EOM
                }
                elsif ( $where ne 'NULL' ) {

                    if ( !&sendMail($message) ) {
                        push( @done, "$row{_formid} " . time );
                        print "... message sent to $to\n";
                    }
                    else {
                        print qq(... <em class="ERROR">ERROR: Sendmail to $to failed</em>\n);
                    }
                }

            }    # hits

        }    # rows of one user

        saveFormUpdateSince( $friend, @done );
        print "Updated .forms file for $friend\n";

    }    # all users

    print "</pre>\n";
}

# print WODA status report

sub cgiAdmWodaStatus {

    printHead('WODA Status');

    print "<h3>Can we execute another Perl program?</h3>";

    $!    = '';
    $perl = "$PERL -v";

    $out = `$perl`;

    if ( $out eq '' ) {
        print <<EOM;
<p class="ERROR">
Looks like it will not be possible to run external Perl programs.
Relational features such as &QRY function will not work!
Maype just <br>\$PERL=$PERL<br>was not set up right!
This is the error reported: <b>$!</b>
</p>
EOM
    }
    else {
        print <<EOM;
<p>
Command <b>$perl</b> produced:
<pre>
$out
</pre>
<p>
EOM
    }

    print <<EOM;
<h3>Directories related to this application</h3>
<p>Note that only <b>$TMPDIR</b> and <b>$DATADIR</b> must exist.
<p>
EOM
    @x = &wbCheckDirectories();
    push( @x, $TESTCGIDIR );

    $testFile = "_test.xxx";
    $textText = "This is some test text";
    $close    = "</em></td></tr>\n";
    $et       = '<em class="ERROR">';
    print '<table border="1">';

    foreach $x (@x) {
        $x = $DISK . $x unless $x =~ m/^[A-Z]\:/;
        print '<tr align="left" valign="top"><th>', $x, '</th><td><em class="ERROR">', "\n";

        unless ( -e "$x" ) {
            print "Does not exist!\n";
            print $close;
            next;
        }

        unless ( chdir($x) ) {
            print "Cannot chdir to $x!\n";
        }

        unless ( open( h, ">$x/$testFile" ) ) {
            print "Cannot open file $testFile for writing. Check permissions!\n";
            print $close;
            next;
        }
        else {
            print h $testText;
            close(h);
        }

        unless ( open( h, "$x/$testFile" ) ) {
            print "Cannot open file $testFile for reading!\n";
        }
        else {
            $xx = <h>;
            if ( $xx ne $testText ) {
                print "Read text not equal to written text!\n";
            }
            close(h);
        }

        if ($MV) {
            `$MV $testFile $testFile.zzz`;
            unless ( -e "$testFile.zzz" ) {
                print "Could not MV $testFile\n";
            }
            else {
                `$MV $testFile.zzz $testFile`;
            }
        }

        unless ( unlink("$x/$testFile") ) {
            print "Cannot delete file $testFile\n";
            print $close;
            next;
        }

        print "</em>OK<em>\n";
        print $close;
    }
    print "</table>\n";

    print "<h3>External programs</h3> Essential functions work without them! <table>";

    @progs = ( '$SENDMAIL', '$AT', '$GREP', '$DUMPABLEPERL', '$UNDUMP' );
    foreach $prog (@progs) {
        $xprog = eval($prog);
        print "<tr><td>$prog</td><td>$xprog</td><td>\n";

        if ( !$xprog ) {
            print '<em class="WARN">not defined</em>';
        }
        elsif ( -x $xprog ) {
            print "executable";
            next;
        }
        elsif ( -e $xprog && !UNIX ) {
            print "exists";
            next;
        }
        else {
            print '<em class="WARN">does not exist</em>';
        }

    }

    print "</table>";

    if ( $SCRIPTPACKAGE ne $WodaPackage ) {
        print "<h3>Subroutines defined in package '$SCRIPTPACKAGE'</h3>\n";

        @subs = @MYSubs;
        foreach $sub (@MYSubs) {
            print "$SCRIPTPACKAGE'$sub()<br>\n";
        }
    }

    print <<EOM;
<h3>Can we use FETCH to get a Webpage into application?</h3>
EOM

    $pg   = 'http://woda.sourceforge.net/index.html';
    $back = &FETCH($pg);

    print <<EOM;
<p>Tried &FETCH(<a href="$pg">$pg</a>). This is what I got back:
<table border="1"><tr><td>
$back
</td></tr></table>
EOM
}

# Database statistics

sub cgiAdmStatistics {

    printHead('Database statistics');
    chdir($DATADIR);
    print "<h3>Number of Records</h3>\n";
    recAll();
    $numFiles = ( scalar keys %IDS );
    print "<p>$numFiles records</p>\n";
    print "<h3>Age of Records</h3>\n";
    $now = time;
    while ( $id = &recNext() ) {
        recRead($id);
        $week[ int( ( time - $rec{'_at'} ) / ( 7 * 24 * 3600 ) ) ]++;
    }
    print '<table border="1"><tr  class="DATAH"><th>week</th><td># records</td></tr>';
    for ( $i = 0 ; $i <= $#week ; $i++ ) {
        print '<tr><th>', $i + 1, '</th><td>', $week[$i], '</td></tr>';
    }
    print '<tr><th>total</th><td>' . $numFiles . '</td></tr></table>';
    print '<h3>Capacity Used by Attachments</h3>';
    $text = $WBB{'quotaText'} || 'not defined';
    $bin  = $WBB{'quotaBin'}  || 'not defined';
    ( $ptext, $pbin ) = &ddQuota();
    print <<EOM;
<table border="1">
<tr class="DATAH">
<th>class</th><th>quota</th><th>used</th></tr>
<th>text</th><td>$text</td><td>$ptext</td></tr>
<th>binary</th><td>$bin</td><td>$pbin</td></tr>
</tr>
</table>
EOM

}

# compile and dump perl

sub cgiAdmCompile {

    printHead('Making Compiled Version');

    if ( !$DUMPABLEPERL ) {
        print "DUMPABLEPERL not defined!\n";
        return;
    }

    $source = $0;
    if ( $source =~ m|\.pl([^/]*)$| ) {
        $exe = "$`$1";
    }
    else {
        print "Source $0 does not contain '.pl' string!\n";
        return;
    }

    if ( $source eq $exe ) {
        print "Source and exe are identical!";
        return;
    }

    print "Compiling <b>$0</b> into <b>$exe</b>\n<pre>The results of compilation:\n";

    print `$DUMPABLEPERL $source -dump 2>&1` if $DATADIR;    # compile application
    print `$DUMPABLEPERL -u $source 2>&1`
      unless $DATADIR;                                       # compile wbxx.pl only
    print `$UNDUMP $DUMPABLEPERL core -o $exe 2>&1`;
    print `$CHMOD +x $exe 2>&1` if $CHMOD;
    print "\n";
    print `ls -l $exe 2>&1`;

}

# defragment the data files

sub cgiAdmDefragment {

    if ( $CGI{'go'} ) {

        printHead("Defragmenting ...");
        print "<pre>";
        print &defragment();
        print "</pre>";

    }
    else {

        printHead("Defragment table");

        ( $n, $r, $d, $m ) = &checkIntegrity();

        if ($m) {
            wbFail( "Cannot defragment", "$m" );
        }

        $text = <<EOM;
You are about to defragment this table. It has $r records,
$n are old versions that can be removed, $d need to be deleted.
<p>
THE GOOD: This will create a new and more compact versions of
fat.txt and data.txt files in the _data directory.
Less space will be used and the application will work faster.
<p>
THE BAD: Old versions of data will be erased. This will prevent you from undeleting
records or reseting changes.

<form action="$SCRIPT/AdmDefragment">
<input class="BUTTON" type="submit" value="Yes I am">
<input type="hidden" value="1" name="go">
</form>
EOM

        $sure = "Are you sure?";    #UK
        print &formatWindow( $sure, $text );

    }

}

# rebuild FAT.TXT

sub cgiAdmRebuildFat {

    if ( $CGI{'go'} ) {

        printHead("Rebuilding fat.txt ...");
        print "<pre>";
        print &rebuildFat();
        print "</pre>";

    }
    else {

        printHead("Rebuild fat");

        ( $n, $r, $d, $m ) = &checkIntegrity();

        if ($m) {
            wbWarn( "Fat is corrupt:", "<pre>$m</pre>" );
        }

        $text = <<EOM;
Shall I try to rebuild the fat.txt? <b>Some deleted records will
re-appear.</b>
<p>If the records.txt is corrupt as well, this will fail.
However,you can still try to
rebuild the table by deleting *.txt in _data directory and
then using the 'rebuild from tbl' command.
<form action="$SCRIPT/AdmRebuildFat">
<input class="BUTTON" type="submit" value="YES">
<input type="hidden" value="1" name="go">
</form>
EOM

        print &formatWindow( "Continue?", $text );

    }

}

# print Configurable Attributes of a Database

sub cgiAdmConfAttrs {

    printHead("Configurable Attributes of a Database");

    admMeta();
    admConfAttrs();
}

# writes the definition as a
# plain text perl script file

sub cgiAdmScript {

    print "Content-type: text/plain\n\n";
    admPrintScript();
}

# writes the definition as a plain text perl script file

sub admPrintScript {

    if ( $CGI{'filename'} ne '' ) {

        if ( open( h, ">$TESTCGIDIR/$CGI{'filename'}" ) ) {
            select(h);
        }
        else {
            wbError( "Open failed", "Could not open $TESTCGIDIR/$CGI{'filename'} for writing\n" );
        }
    }

    admMeta();

    $CGI{'language'} =~ tr/[A-Z]/[a-z]/;

    $lib = "$WodaFileDir/$WodaFile";
    $now = &formatDate(time);

    $top = <<EOM;
#!$PERL

require "$lib" unless defined &main;

# Data dictionary definition for WODA $Version engine
# stored as a Perl script file.
#
# generated by $lib version $VERSION at $now
#
# INSTRUCTIONS:
#	 save as *unix* file into cgi directory of your server
#	 Save ... in MSDOS/Windows often creates pairs of CRLF as line terminators
#	 unix needs CR only!

#
# Database definition:
#

EOM

    $end = <<EOM;

# insert any add-in functions here:

# call the real thing

main();

EOM

    # add to string $definition

    $definition = '';

    foreach $a (@MetaB) {
        next if $MetaB{ $a, 'what' } eq 'H';
        if ( $WBB{$a} ne '' ) {
            $name = '$WBB{' . "'" . $a . "'" . '}';
            $definition .= &admMetaOutput( $name, $WBB{$a} );
        }
    }

    $definition .= <<EOM;

#
# Fields:
#

\$i=1000;		# counter

EOM

    wbSortFields();
    foreach $field (@Fields) {
        $value = $WBF{$field};
        $value =~ s/'/'."'".'/g;    # make sure ' does not cause
        $value =~ s/\r\n/\n/g;      # CR-LF into LF only !
        $definition .= <<EOM;

\$x = '$field'; ############################
\$WBF{\$x,'srt'}=\$i--;
\$WBF{\$x}='$value';
EOM

        foreach $attr (@MetaF) {
            next if $attr eq 'srt';
            $value = $WBF{ $field, $attr };
            next if $value eq '';
            $value =~ s/'/'."'".'/g;    # make sure ' does not cause
            $value =~ s/\r\n/\n/g;      # CR-LF into LF only !
            $definition .= <<EOM;
\$WBF{\$x,'$attr'}='$value';
EOM
        }
    }

    print $top;
    print $definition;
    print $end;
    select STDOUT;
}

# prints reassuring message

sub cgiAdmSetEnv {

    printHead("Storing values of some ENV variables");

    print <<EOM;
<p>The environment is being saved with each use of WODA through the
Web.
EOM

}

# Modifying Crontab Entries

sub cgiAdmCronModify {

    printHead("Modifying Crontab Entries");

    $cron = `$CRONTAB -l`;

    @cron = split( /\n/, $cron );
    @cron = grep ( !m|$0|, @cron );
    $cron = join( "\n", @cron );

    if ( $CGI{'action'} eq 'add' ) {
        $cron .= &wbCronjob();
    }

    wbCheckDir($TMPDIR);
    open( h, ">$TMPDIR/$$.cro" ) || wbFail("Cannot open crontab file");
    binmode(h);
    print h $cron;
    close(h);
    print "<p>Results from crontab:<pre>\n";
    print `$CRONTAB $TMPDIR/$$.cro 2>&1`;
    print "</pre>\n";

    # unlink ($TMPDIR/$$.cro);

    $crontab = `$CRONTAB -l`;
    print <<EOM;
<p>The new crontab file is:
<pre>
$cron
</pre>
EOM
}

# process cron job using AT

sub cgiAdmAtNow {

    printHead("Processing of Cron Job Now Using At");

    print &wbAtdo();
    if ($?) {
        print <<EOM;
<hr>
<p>Looks like the at command failed.
EOM
    }
}

# process cron job

sub wbCronjob {

    $a = &wbAtjob();
    return <<EOM;
#
# WODA job for $0
0 1 * * * $a
EOM
}

# do job with AT

sub wbAtdo {

    local ($job);

    $job = &wbAtjob();

    wbCheckDir($TMPDIR);
    open( at, ">$TMPDIR/$$.at" );
    binmode(at);
    print at $job;
    close(at);

    return `$AT -f$TMPDIR/$$.at now 2>&1`;

}

# reports on AT job

sub wbAtjob {

    findSendmail();
    ( $prg, $prefix ) = &wbPrepareCommandLine();

    return "$prg -x $prefix/AdmDaily 2>&1 | $SENDMAIL $WBB{'managerEmail'}";

}

# generate cron tasks

sub cgiAdmCrons {

    printHead("Required crontab entries");

    print <<EOM;
<p><b>WARNING:</b> If you called
this function form a test database, then results are not valid.
You should run it from the real database!
EOM

    print <<EOM;
<p>These tasks could be part of your crontab file:
EOM
    $prog = &wbCronjob();

    print "<pre><b>$prog</b></pre>";

    print <<EOM;
<p>On systems that unlike UNIX do not support cron,
you should execute the above command once a day or
have "system agent" run it from time to time.
EOM

    $crons = `$CRONTAB -l 2>&1`;
    $x     = getpwuid($>);
    if ( $crons =~ /allowed/ ) {
        print <<EOM;
Cron jobs cannot be set automatically on this system.
You should ask the systm administrator to allow user
$x to run cron and at jobs by adding
<tt>$x</tt> to (usually)
<tt>/usr/lib/cron/cron.allow</tt> and to 
<tt>/usr/lib/cron/at.allow</tt>.
EOM
    }
    else {
        print <<EOM;
<p> More options:
<ul>
<li><a href="$SCRIPT/AdmCronModify?action=add">
Add this to the Cron table now.
</a>
</li><li>
<a href="$SCRIPT/AdmCronModify?action=remove">
Remove this from the Cron table.
</a>
</li><li>
<a href="$SCRIPT/AdmAtNow">
Run this now using command '$AT'.
</a>
</li></ul>
EOM
    }
}

# delete password from a record

sub cgiAdmDelPass {

    $id = $CGI{'_id'};
    if ( $id eq '' ) {

        printHead("Delete password for a record");

        print <<EOM;
<form action="$SCRIPT/AdmDelPass">
<br>Id of the record <input name="_id">
<br><input type="submit" value="Reset password">
</form>
EOM
        return;
    }

    printHead("Deleting password for record with id=$id");

    recRead($id);
    if ( $rec{'_id'} ) {
        undef $rec{'_password'};
        $rec{'_id'} = $id;
        recWrite();
        print "<h1>Done!\n</h1>";
    }
    else {
        wbError( "No such record", "Record $id not known here" );
    }
}

# print database structure

sub cgiAdmDisplayStructure {

    printHead("Structure of database $WBB{'dbTitle'}");
    printStructure( '.', '.', !$CGI{'nolegend'}, $CGI{'errors'} );
}

# delete records older than 'days' or 'deleteAfter'

sub cgiAdmExpire {

    $CGI{'days'} = $CGI{'days'} || $WBB{'deleteAfter'} || 0;
    $CGI{'days'} *= 1;
    if ( $CGI{'days'} <= 0 ) {
        printHead("Nothing expired");
        return;
    }

    printHead("DELETING records older than $CGI{'days'} days");

    $then = time - $CGI{'days'} * 24 * 3600;

    recAll();
    $allN = 0;
    $delN = 0;

    while ( $id = &recNext() ) {

        $allN++;

        recRead($id);

        next if $rec{'_at'} > $then;

        $delN++;

        recDelete($id);
        print "<br>Deleted: $id\n";
    }

    wbDatabaseChanged();
    print "<p>All files $allN. Deleted $delN.\n";
}

# print form to delete expired records

sub cgiAdmExpireForm {
    printHead("Delete old records from the database");
    $d = $WBB{'deleteAfter'};

    print <<EOM;

<form action="$SCRIPT/AdmExpire" method="get">
<p>
Delete files older than <input name="days" size="5" value="$d"> days.
<p>
<input class="BUTTON" type="submit"  value="DELETE!">
</form>
EOM
}

# Synchronise userfiles and table data

sub cgiAdmSyncUserfilesForm {

    printHead("Synchronise userfiles and table data");

    print <<EOM;

<p>Binary files attached to records (images, sounds, documents) are stored as 
a file with name ID.FIELD.ext in the ./att directory. The name that file
also appears as the value of the FIELD. For many reasons there may be a
contradicion in these two pieces of information. This option lets you
handle it.</p>

<form class="FORM" action="$SCRIPT/AdmSyncUserfiles">
<table cellspacing="0">
<tr>
<td>
	What data is valid:
</td>
<td>
	<input type="radio" name="valid" value="files">Files. Insert file names to fields.<br>
	<input type="radio" name="valid" value="fields">Fields. Delete files not mentioned in fields.
</td>
</tr>
<tr>
<td>
	Action:
</td>
<td>
	<input class="BUTTON" type="submit" name="action" value="sync">
	<input class="BUTTON" type="submit" name="action" value="preview">
</td>
</tr>
</table>
EOM
}

# Synchronise userfiles and table data

sub cgiAdmSyncUserfiles {

    printHead("Syncronise userfiles");
    $| = 1;

    opendir( D, "$DATADIR/att" );
    @fsx = readdir(D);
    closedir(D);
    print "<p>$#fsx files<br>";
    $i = 0;
    foreach $file (@fsx) {
        next if $file =~ m/^\./;
        next if $file eq 'index.htm';
        $file =~ m/^([^\.]*)\.([^\.]*)/;
        $id    = $1;
        $field = $2;
        $_id   = '';
        $fw    = '';    # field value
        if ( &recRead($id) eq '' ) {
            $fw = $rec{$field};
        }

        next if $fw eq $file;

        print "\n<br>";

        $i++;

        if ( $CGI{'valid'} eq 'files' ) {
            print "$i - In record $id value of field $field should be $file.";
            if ( $CGI{'action'} eq 'sync' ) {
                $rec{$field} = $file;
                $rec{'_id'} = $id;
                recWrite();
                print " CORRECTED!";
            }
            else {
                print " IGNORED!";
            }
        }
        else {
            print "$i - $file is not mentioned in a record.";
            if ( $CGI{'action'} eq 'sync' ) {
                unlink "$DATADIR/att/$file";
                print " DELETED!";
            }
            else {
                print " IGNORED!";
            }
        }
    }
}

# export records

sub cgiAdmExportCSV {

    wbExport( 'DEFAULT', 'CSV', $CGI{'fields'} );
    wodaExit();
}

# export records

sub cgiAdmExportCSVS {

    wbExport( 'DEFAULT', 'CSVS', $CGI{'fields'} );
    wodaExit();
}

# export record IDs

sub cgiAdmExportID {

    print "Content-type: text/plain\n\n" unless $HTMLFILE;

    ( $allHits, *list, *fields ) = &recSelect( '', '', '', '$rec{\'_id\'}' );

    print '$tempList = (';
    print "\n";

    foreach $x (@list) {
        print "'" . $x . "'";
        print ",\n";
    }

    print ');';
    print "\n";

    wodaExit();
}

# counts categories in $field and returns them in
# an ascii string ...
# category and count are TAB separated

sub countCat {

    local ( $field, $field2 ) = @_;
    local ($ret);

    if ( $field2 eq '' ) {
        $all = &ROWS( '', "$field" );
    }
    else {
        $all = &ROWS( '', "$field,$field2" );
    }

    chop($all);
    $all =~ tr/\034/\n/;    # multiple value
    @list = split( /\n/, $all );

    undef %cat;

    foreach $x (@list) {
        $x =~ tr/\t/,/;
        $cat{$x}++;
    }

    foreach $cat ( sort( keys(%cat) ) ) {
        next if $cat eq '';
        $ret .= "$cat\t$cat{$cat}\n";
    }

    return $ret;
}

# export records

sub cgiAdmExportTAB {

    wbExport( 'DEFAULT', 'TAB', $CGI{'fields'} );
    wodaExit();
}

# converts data in CGI{'csv'}
# into arrays:
#	$err[$line] ... errors
#	$rec[$line] ... record in feature vector format
#	$at[$line] ... original date stamp
#	$id[$line] ... id

sub admReadCSV {

    local ($how) = @_;

    # get rid of CRs

    $CGI{'csv'} =~ s/\r//g;
    @csv = split( /\n/, $CGI{'csv'} );
    $line = 0;

    # print "<pre>$#csv == $CGI{'csv'}</pre>";

    while ( $dLine = shift(@csv) ) {

        $line++;
        undef %rec;

        if ( $line == 1 ) {
            if ( $dLine =~ m/\,/ ) {
                $delim = ',';
            }
            elsif ( $dLine =~ m/\;/ ) {
                $delim = ';';
            }
            else {
                wbError( 'Bad delimiter', 'Delimiter of the .csv file should be comma (,) or semicolon (;)' );
                return;
            }
        }

        while ( ( $dLine =~ tr/"/"/ ) % 2 ) {

            if (@csv) {
                $dLine .= "\n";           # removed by split
                $dLine .= shift(@csv);    # next line
            }
            else {
                print "<h3>Bad .csv file at line $line; \" did not close</h3>\n";
                return;
            }
        }

        next
          if $dLine =~ m/^$delim*$/;      # ignore lines with nothing but blank cells

        @flds = split( /$delim/, $dLine );

        if ( $line == 1 ) {
            @csvf = @flds;
            print "<p>The fields are:<ol><li>";
            print join( "</li>\n<li>", @flds );
            print "</li></ol><p>\n";
        }

        $field = 0;

        while (1) {
            last unless (@flds);

            $f = shift(@flds);

            $field++;

            while ( ( $f =~ tr/"/"/ ) % 2 ) {
                if (@flds) {
                    $f .= $delim . shift(@flds);
                }
                else {
                    print "<h3>Bad .csv file at line $line; \" did not close</h3>\n";
                    return;
                }
            }

            $f =~ s/^"//;
            $f =~ s/"$//;
            $f =~ s/""/"/g;

            if ( $line == 1 ) {
                if ( $f eq '' ) {
                    wbError( 'Column without name', 'Aborted' );    #UK
                }
                $key[$field] = $f;
                $CGI{"_f$field"} = $f;

            }
            else {
                $rec{ $key[$field] } = $f;
                admGuessFieldType( $field, $f ) if $how eq 'CREATE';
            }

        }

        if ( $line > 1 ) {

            $e = &recValidate();

            if ( $rec{'_id'} ne '' ) {    # _id is in the file ... keep original ID

                $id[$line] = $rec{'_id'};

            }
            elsif ( ( $newId = &evalWBBkey() ) ne '' ) {    # compute new ID via WBBkey

                $id[$line] = $newId;

            }
            else {

                # random key will be assigned at some point

            }

            if (   $id[$line] ne ''
                && $CGI{'how'} eq 'addonly'
                && &recExists( $id[$line] ) >= 0 )
            {
                $e = $e . "\n<b>Record exists!</b>\n";
            }

            $err[$line] = $e;
            $keepUnderscores = 1
              if $rec{'_from'} || $rec{'_at'} || $rec{'_password'};
            $rec[$line] = &recToVector();
            $at[$line] = $rec{'_at'} if $rec{'_at'};

            unless ( $CGI{'quiet'} ) {
                local $line;
                print "<hr>";
                print &wbRecHtml( 'NAMEVAL', *csvf );
            }
        }
    }

    print "<h3>$line rows read from file</h3>";

}

# tries to guess what type a field should be

sub admGuessFieldType {

    local ( $field, $data ) = @_;

    return if $CGI{"type$field"} =~ m/EMAIL|URL|HTMLAREA/;
    if ( $data =~ m/\@/ ) {
        $CGI{"type$field"} = 'EMAIL';
        print "<br>I guess column $key[$field] is of type email!\n";
    }
    elsif ( $data =~ m/^http:\/\/|^ftp:\/\/|^news:|^mailto:/ ) {
        $CGI{"type$field"} = 'URL';
        print "<br>I guess column $key[$field] is of type URL!\n";
    }
    elsif ( $data =~ m/\034/ || length $data > 100 ) {
        $CGI{"type$field"} = 'HTMLAREA';
        print "<br>I guess column $key[$field] is of type HTMLAREA!\n";
    }
}

# upload user files

sub cgiAdmUploadUserfiles {

    if ( $CGI{'userfiles'} ) {
        admUploadUserfiles();
    }
    else {
        admUploadUserfilesForm();
    }
}

# upload user files

sub admUploadUserfilesForm {

    printHead("Upload userfiles");

    print <<EOM;
<p>This option allows you to upload several userfiles
(binary attachement to the records) at once.
<p>Instructions
<ol>
<li>Prepare the files on your computer, so that they have correct names and extensions. These names are like
id.field.ext where id is the id of the record, field is the name of the field and the ext
is the extension of the file.
</li><li>Pack the files into an archive. Archive should be made with the program called 'tar'.
</li><li>Compress the archive with a program called gzip. The resulting file should have the extension tgz.
</li><li>Fill the form below to upload the .tgz file to the server.
</li></ol>
<form ENCTYPE="multipart/form-data" method="post" action="$SCRIPT/AdmUploadUserfiles">
<input name="userfiles" type=FILE>
<input class="BUTTON" type="submit" value="Upload">
</form>
EOM

}

# upload multiple files compressed into a tgz archive

sub admUploadUserfiles {

    printHead("Userfiles Uploaded");

    if ( $EXT{userfiles} ne 'tgz' ) {
        wbError( "Bad file format", "The uploaded file must be a gzip-ed tar file with .tgz extension" );
    }

    $bytes = length( $CGI{'userfiles'} );
    print "<p>$bytes uploaded\n";

    &wbCheckDir("$HTMLDIR/att");
    chdir("$HTMLDIR/att");
    $tmpfile = time . 'upload' . '.tgz';
    open( h, ">$tmpfile" ) || wbError("Cannot open file $tmpfile");
    binmode(h);
    print h $CGI{'userfiles'} || wbError("Failed to write $tmpfile");
    close(h);

    $command = <<EOM;
tar -xvzf $tmpfile
rm $tmpfile
EOM

    print "<p>Executing this:<p><pre>$command</pre>";
    print `$command`;

    print <<EOM;
<p>If the above information make you believe
that the files were properly uploaded,
then you should link them into the database
using the
<a href="$SCRIPT/AdmSyncUserfilesForms">Sync Userfiles</a>
command
EOM
}

# import records from CSV file

sub cgiAdmImport {

    $makeRecordChanges = ''
      || $CGI{doOnRecordChange};    # must be requested explicitly

    printHead("Import results");

    if ( $CGI{'_password'} && $CGI{'_password'} ne $CGI{'_password1'} ) {
        wbFail("Passwords do not match!");
    }

    if ( $CGI{'_password'} ) {
        $cPassword = &wbCryptPasswd( $CGI{'_password'} );
    }
    else {
        $cPassword = '';
    }

    $delim = $CGI{'fsep'} || ';';

    if ( $CGI{'csv'} eq '' ) {
        print '<h3 class="ERROR">No csv file supplied</h3>', "\n";
        return;
    }
    else {
        admReadCSV();
    }

    # --- write to disk ?

    $x = join( "", @err );
    if ( $x && !defined( $CGI{'eignore'} ) ) {
        print '<h3 class="ERROR">ERRORS IN INPUT DATA:</h3>', "\n";

        for ( $i = 2 ; $i <= $#err ; $i++ ) {
            if ( $err[$i] ) {
                $id = $id[$i];
                $err[$i] =~ s/WARN/<br>WARN/g;
                print <<EOM;
<h3 class="ERROR">Errors in row $i ... _id=$id</h3>
<p>$err[$i]
EOM
            }
        }
        wbError( "Errors were found", <<EOM);
Nothing was saved.
EOM

    }
    else {
        print "<h2>No errors in input data</h2>\n" unless $x;
        print "<h2>Errors in input data ignored</h2>\n" if $x;

        $| = 1;

        for ( $inLine = 2 ; $inLine <= $#rec ; $inLine++ ) {

            $home = $DATADIR;

            if ( $id[$inLine] ) {    # before 24.11.1999 was ($WBB{'key'} || $id[$inLine])
                $id = $id[$inLine];
            }
            else {
                $id = &recRandID();
            }

            # yeah, this is slow
            # I could have dumped it out to a file as is

            recFromVector( $rec[$inLine] );

            # 27.5.2002 13:53 this was checked in the admReadCSV
            #
            #		if (&recExists($id)>=0 && $CGI{'how'} eq 'addonly') {
            #			print "<br>Line $inLine NOT SAVED (duplicate id=$id)\n";
            #			next;
            #		}

            # set _at to value from file

            if ( $at[$inLine] ) {
                $rec{'_at'} = $at[$inLine];
            }

            # _at data is quite essential!

            if ( $rec{'_at'} eq '' ) {
                $rec{'_at'} = time;
            }

            # merge in values from existing record

            if ( $CGI{'how'} eq 'merge' ) {
                %newrec = %rec;
                recRead($id);

                foreach $field ( keys(%newrec) ) {

                    # import only non-empty fields i.e. blank fields doesn't erase data
                    $rec{$field} = $newrec{$field} if ( $newrec{$field} );
                }
            }

            # passwords

            if ( $CGI{'crypt'} && $rec{'_password'} ) {
                $rec{'_password'} = &wbCryptPasswd( $rec{'_password'} );
            }

            if ( $cPassword && $rec{'_password'} eq '' ) {
                $rec{'_password'} = $cPassword;
            }

            $rec{'_id'} = $id;
            recWrite();
            wbRecordChanged() if $makeRecordChanges;
            print "<br />Line $inLine was saved to <a href=\"$SCRIPT/Show?_id=$id\">$id</a>\n";
        }
        wbDatabaseChanged();    #moved out of loop, so we call it once at the end, not every loop
    }
}

# print form to manage Importing records

sub cgiAdmImportForm {

    printHead("Import / merge .CSV data into database");

    print <<EOM;
<p>You can make CSV with office tools such as Excel or Access.</p>

<form class="FORM" ENCTYPE="multipart/form-data" action="$SCRIPT/AdmImport" method="post">

<p>
<b>CSV file on your disk</b>, which contains a title line with field names and
.CSV records (as saved, for example, by MS Excel or MS Access):
<br><input type=FILE size="40" name="csv">
<!--<br><select name="fsep"><option value=",">comma (,)<option value=";">semicolon (;)</select> Field separator in the above file.
-->
<br>
<br> How to merge old with new data:
<br><input type="radio" name="how" value="addonly" CHECKED> The imported record may only be added as new record.
<br><input type="radio" name="how" value="overwrite"> The imported record replaces the whole existing record.
<br><input type="radio" name="how" value="merge"> Values from imported records are merged with values in existing records.
<br>
<br><input type="checkbox" name="eignore"> Ignore errors in the imported data.
<br><input type="checkbox" name="quiet"> Don't display all records.
<p class="warn">
<br><input type="checkbox" name="doOnRecordChange"> Perform commands in WBB{afterRecordmodify} when records change?
<br> afterRecordModify contains this command:<pre>$WBB{'afterRecordModify'}</pre></p>
<br>
<br>How to handle passwords:
<br><input type="checkbox" name="crypt"> Encrypt the data in the _password column. 
Use only if you are adding new records and the passwords are in plain text!
If you are moving some existing records back into database and the
password column is encrypted already, don't encrypt again!
<br>Password (optional) for <em>all</em> records that do not have a password.
<br><input type="password" name=\"_password\" size="10">
again
<input type="password" name=\"_password1\" size="10">
<p>
<input class="BUTTON" type="submit"  value="IMPORT!">
</p>
</form>
<p>
EOM

    wbSortFields();

    print '<p>Sample table:</p>
<table border="1"><tr>';
    foreach $f (@PFields) {
        print "<th>$f</th>";
    }

    print "</tr><tr>";
    foreach $f (@Fields) {
        print "<th>&nbsp</th>";
    }

    print "</tr><tr>";
    foreach $f (@Fields) {
        print "<th>&nbsp</th>";
    }

    print "</tr><tr>";
    foreach $f (@Fields) {
        print "<th>&nbsp</th>";
    }

    print "</tr></table>";

}

# create records from CSV data

sub admCreateFromData {

    local ($DATADIR) = @_;

    # --- parse CSV file

    print "<p>Parsing the .CSV file\n\n";

    $delim = $CGI{'fsep'} || ';';

    if ( $CGI{'csv'} eq '' ) {
        print "<h3>No csv file supplied</h3>\n";
        return;
    }
    else {
        $CGI{'quiet'} = 1;
        admReadCSV('CREATE');
    }

    # --- write records

    print "<p>";

    for ( $inLine = 2 ; $inLine <= $#rec ; $inLine++ ) {

        $home = $CGI{'dataDir'};

        if ( $id[$inLine] ) {
            $id = $id[$inLine];
        }
        else {
            $id = &recRandID();
        }

        recFromVector( $rec[$inLine] );
        $rec{'_id'} = $id;
        recWrite();

        print "<br>Wrote row $inLine to $id\n";
    }
}

# display admin menu

sub cgiAdmMenu {

    printHead("Administration menu");

    wbCheckDirectories();

    print '<div class="MENU"><table><tr><td width="33%" valign="top">';

    print &formatMenuTitle("Close/open application");
    print &formatMenuItems( "Close application for maintenance", "$SCRIPT/AdmClose", "Open application after maintenance", "$SCRIPT/AdmOpen", );

    print &formatMenuTitle("Table definition");
    print &formatMenuItems(
        "Modify database definition",  "$SCRIPT/AdmModifySettingFrames",
        "Apply pageTemplate setting",  "$SCRIPT/AdmApplyPageTemplate",
        "Show possible page template", "$SCRIPT/AdmPageTemplate",

        #HIDE"
        "Upgrade CGI definition to dd.txt definition",
        "$SCRIPT/AdmSaveDataDict",
        "Display definition",         "$SCRIPT/AdmDisplayStructure",
        "Display definition (terse)", "$SCRIPT/AdmDisplayStructure?nolegend=1",
        "Display errors in definition",
        "$SCRIPT/AdmDisplayStructure?nolegend=1&amp;errors=1",
        "Display definition as Perl script file", "$SCRIPT/AdmScript",
        "Show color scheme",                      "$SCRIPT/ColorSchemeShow",
        "Refresh cached HTML",                    "$SCRIPT/AdmTouch",
    );

    print &formatMenuTitle("Documentation");
    print &formatMenuItems(

        #		"Getting started","/started/",
        #		"Administrator's manual","/man/",
        #		"FAQ","/use/faq/",
        "Configurable parameters",  "$SCRIPT/AdmConfAttrs",
        "Display global variables", "$SCRIPT/AdmPrintVars",
        "CGI GUI functions",        "$SCRIPT/AdmPrintCGIAPI",
        "Perl API functions",       "$SCRIPT/AdmPrintPerlAPI",
        "HTTP API in HTML",         "$SCRIPT/HtmlApi",
        "General WSDL",             "$SCRIPT/WSDL",
        "Office Research WSDL",     "$SCRIPT/OfficeWSDL",
    );

    print '</td><td width="33%" valign="top">';

    print &formatMenuTitle("Manage users");
    print &formatMenuItems(
        "Add/delete user",                                "$SCRIPT/AdmModifySetting?type=b&amp;name=groups",
        "Delete record's password",                       "$SCRIPT/AdmDelPass",
        "Edit rights",                                    "$SCRIPT/AdmModifySetting?type=b&amp;name=rights",
        "List registered friends",                        "$SCRIPT/AdmRegisteredPrint",
        "Send serial email to addresses of friends",      "$SCRIPT/AdmRegisteredMailForm",
        "Send serial email to addresses in the database", "$SCRIPT/AdmMailForm",
    );

    print &formatMenuTitle("Periodic tasks");
    print &formatMenuItems(
        "Rebuild fat.txt ...",                                           "$SCRIPT/AdmRebuildFat",
        "Defragment data.txt ...",                                       "$SCRIPT/AdmDefragment",
        "Clean cache",                                                   "$SCRIPT/AdmClean",
        "Reindex",                                                       "$SCRIPT/AdmReindex",
        "Index PDF files (slow)(<b>new</b>)",                            "$SCRIPT/AdmIndexPDF",
        "Delete expired records",                                        "$SCRIPT/AdmExpireForm",
        "Make compiled script",                                          "$SCRIPT/AdmCompile",
        "Make TOC for robots",                                           "$SCRIPT/AdmStatic",
        "View what should be part of crontab file and save environment", "$SCRIPT/AdmCrons",
        "Clean, expire and update static ... all in one go",             "$SCRIPT/AdmDaily",
        "Process Agent Requests ...",                                    "$SCRIPT/AdmAgentProcessForm"
    );

    print &formatMenuTitle("Status");
    print &formatMenuItems(
        "Log file",        "$SCRIPT/AdmLogfilePrint", "Statistics",      "$SCRIPT/AdmStatistics",
        "Database status", "$SCRIPT/AdmWodaStatus",   "Show debug info", "$SCRIPT/AdmDebug",
    );

    print '</td><td width="33%" valign="top">';

    print &formatMenuTitle("Export table");
    print &formatMenuItems(
        "as CSV comma delimited", "$SCRIPT/AdmExportCSV", "as CSV semicolon delimited", "$SCRIPT/AdmExportCSVS",
        "as TAB delimited",       "$SCRIPT/AdmExportTAB", "as OPTION array",            "$SCRIPT/AdmExportID",
        "as text files",          "$SCRIPT/AdmExportTextfiles",
    );

    print &formatMenuTitle("Import data, edit table");
    print &formatMenuItems(
        "Import from CSV file",                             "$SCRIPT/AdmImportForm",
        "Upload userfiles",                                 "$SCRIPT/AdmUploadUserfiles",
        "Syncronise userfiles ...",                         "$SCRIPT/AdmSyncUserfilesForm",
        "Change a field in many records",                   "$SCRIPT/AdmManySetForm",
        "Edit fields in many records",                      "$SCRIPT/AdmManyEditForm",
        "Rebuild from default.tbl file (with processing!)", "$SCRIPT/AdmRebuildFromTabForm"
    );

    print &formatMenuTitle("Backup");
    print &formatMenuItems( "Backup table", "$SCRIPT/AdmBackupData", "Restore table ...", "$SCRIPT/AdmRestoreData", );

    print &formatMenuTitle("History");
    print &formatMenuItems(
        "Rollback table ...",                        "$SCRIPT/AdmRollbackForm",
        "Rollback, undelete, unchange record ...",   "$SCRIPT/AdmRecordHistoryForm",
        "see also history icon on record's toolbar", ''
    );

    print &formatMenuTitle("Other");
    print &formatMenuItems(
        "The /<b>Home</b> page of this table", "$SCRIPT/Home", "Try this database as guest", "$SCRIPT/U=guest/",
        "Manage raw files (dangerous)",        "$SCRIPT/AdmDir",
    );

    print "</td></tr></table></div>\n";

    print "<H3>Main menu (in case it is hiddent in the app)</H3>\n";

    print &formatHomeMenu(4);    # columns

}

# display form to show record history

sub cgiAdmRecordHistoryForm {

    printHead("Select record to examine");

    print <<EOM;
<form action="$SCRIPT/AdmRecordHistory">
<input class="BUTTON" type="submit" value="Show"> history of record <input name="_id" size="20">
</form>
EOM
}

# show record history

sub cgiAdmRecordHistory {

    $id = $CGI{'_id'};
    printHead("History of record $id");
    print <<EOM;
<p>This is only the history from the current records.txt and fat.txt files!
<p>
EOM

    ( $HData, $HFat ) = &recOpen();
    @x = <$HFat>;
    @x = grep( /^$id\t/, @x );

    # this one depends on the format of fat.txt
    undef %rec;
    undef %orec;

    $text = 'na';
    foreach $x (@x) {
        chop($x);
        ( $id, $pos ) = split( /\t/, $x );
        if ( $pos < 0 ) {
            $text = "deleted";
            $date = "?";
        }
        else {
            if ( $text eq 'na' ) {
                $text = 'created';
            }
            elsif ( $text eq 'deleted' ) {
                $text = 'restored';
            }
            else {
                $text = 'edited';
            }

            %orec = %rec;
            undef %rec;
            $date = '?';

            recReadAt($pos);

            $date  = &formatYMDHMS( $rec{'_at'} );
            $date1 = &formatYMDHMS($time);

            $diff = '';
            foreach $f ( sort( keys(%rec) ) ) {
                if ( $rec{$f} ne $orec{$f} ) {
                    $diff .= "<br>changed $f\n";
                }
            }

            $text = <<EOM;
$text (edited _at $date saved $date1)
<a href="$SCRIPT/AdmRestoreRecTo?id=$id&amp;pos=$pos">
restore to this version
</a>
$diff
EOM
        }
        print "<hr>$text<br>\n";
    }
    print qq(<a href="$SCRIPT/Show?_id=$id"><b>This is the current version</b></a><p>\n);
    &recClose();
}

# make sure ID and POS are valid

sub wbCheckIdPos {

    local ( $id, $pos ) = @_;

    if ( $id =~ m/\t|\n/ ) {
        wbFail("ID contains newlines or tabs\n");
    }

    if ( $pos !~ m/^[0-9-]+$/ ) {
        wbFail("Position is not a number\n");
    }

}

# restore a record

sub cgiAdmRestoreRecTo {

    $id  = $CGI{'id'};
    $pos = $CGI{'pos'};

    printHead("Restoring record $id");

    ( $HData, $HFat ) = &recOpen();
    @x = grep ( /$id\t$pos\n/, <$HFat> );
    recClose();

    if ( $#x >= 0 ) {

        wbCheckIdPos( $id, $pos );

        open( h, ">>$DATADIR/_data/fat.txt" ) || &wbFail("Cannot open fat");
        flock( h, 2 );
        binmode(h);
        print h "$id\t$pos\n";
        close(h);
        wbWarn( "Restored", "The record is <a href=\"$SCRIPT/Show?$id\">here</a>" );
    }
    else {
        wbError( "No such record", "Record with id=$id cannot be resored" );
    }
}

# print blank template page

sub cgiAdmPageTemplate {

    $WBB{'pageTitle'} =~ s/\$/\\\$/g;
    $WBB{'pageEnd'}   =~ s/\$/\\\$/g;

    printHead('whatever');
    print '$body';
}

# Use template file to generate pageBody, pageTitle and pageEnd

sub cgiAdmApplyPageTemplate {

    $t = $WBB{'pageTemplate'};

    if ( $t eq '' ) {
        wbError( "pageTemplate not defined", "Setting \$WBB{'pageTemplate'} is empty!" );
        return;
    }

    if ( $t =~ m/^http/ ) {
        $text = &FETCH($t);
    }
    else {
        chdir($DATADIR);
        open( h, $t ) || wbError( "Cannot open file", "File $t not found" );
        @t = <h>;
        close(h);
        $text = join( '', @t );
    }

    if ( $text !~ m/\<body([^>]*)\>/i ) {
        wbError( "Bad template", "$t does not include BODY tag" );
    }

    $CGI{'pageBody'} = $1;
    $h               = $`;
    $b               = $';

    if ( $b !~ m/\$body/ ) {
        wbError( "Bad structure", "\$body string not found in template body" );
    }

    $CGI{'pageTitle'} = $`;
    $CGI{'pageEnd'}   = $';

    # quote and escape

    $CGI{'pageTitle'} =~ s/\"/\\\"/g;    # quote strings
    $CGI{'pageTitle'} =~ tr/\r//;        # CRs
    $CGI{'pageTitle'} = '"' . $CGI{'pageTitle'} . '"';

    $CGI{'pageEnd'} =~ s/\"/\\\"/g;      # quote strings
    $CGI{'pageEnd'} =~ tr/\r//;          # CRs
    $CGI{'pageEnd'} = '"' . $CGI{'pageEnd'} . '"';

    print "Content-type: text/plain\n\n";
    print <<EOM;
# insert this code into $SCRIPTFILE
#

\$WBB{'pageBody'}='$CGI{'pageBody'}';

\$WBB{'pageTitle'}=<<'EOT';
$CGI{'pageTitle'}
EOT

\$WBB{'pageEnd'}=<<'EOT';
$CGI{'pageEnd'}
EOT

#
# end of code to insert
EOM
}

# change Modification dates on data files

sub cgiAdmTouch {

    printHead("Touching Data");

    print <<EOM;
<p>Modification dates on data files were changed so that
most cached files will be regenerated and all modifications
due to the modification of the database definition will
show on screen.
EOM

    recTouch();
}

# Close application to enable maintenance

sub cgiAdmClose {

    printHead("Closing application to enable maintenance");
    $message = <<EOM;    #UK
This application is closed for maintenance.
Please try again at 14:00 GMT.
EOM

    if ( $CGI{'message'} ) {
        open( H, ">_cache/closed.ip" )
          || wbFail("Cannot open file closed.ip\n");
        binmode(H);
        print H $RA;
        close(H);
        open( H, ">_cache/closed.txt" )
          || wbFail("Cannot open file closed.ip\n");
        binmode(H);
        print H $CGI{'message'};
        close(H);
        print "The database is now closed for all but users from $RA\n";
    }
    else {
        print <<EOM;
<form action="$SCRIPT/AdmClose">
<br>Enter a message to show to users:
<br><textarea name="message" rows="5" cols="50">$message</textarea>
<br><input class="BUTTON" type="submit" value="Close for all but yourself">
</form>
EOM
    }
}

# open database to general public

sub cgiAdmOpen {

    unlink "_cache/closed.ip";
    unlink "_cache/closed.txt";

    printHead("Opened the database for general public");

    print "<p>The database is now open for all.\n";
}

# Rebuild database from TBL file

sub cgiAdmRebuildFromTabForm {

    return unless &saveFormCan('AdmRebuildFromTabForm');

    printHead("Rebuild database from TBL file");

    if ( &wbExpired( "_cache/default.tbl", "", 0 ) ) {
        wbWarn( "Warning!", <<EOM);
The index file, default.tbl, which would be used
to update the raw data is older than the raw data!
If you continue, you may loose recent edits,
but perhaps this is what you want.
EOM
    }

    print <<EOM;
<p class="WARN">
This will first erase all records and then it will
rebuild them from the _cache/default.tbl file.</p>
$SaveFormForm
<form class="FORM" action="$SCRIPT/AdmRebuildFromTab" method="post">

<p>Processing to apply to each record while rewriting. You should manipulate \$rec{} values.
You can use Perl's <em>next</em> statement to delete some records!
</p>

<p><textarea name="action" cols="60" rows="20">$CGI{'action'}</textarea></p>
<p>$SaveFormButton
<p><input class="BUTTON" type="submit" value="Do it"> It may take time, don't panic!

</p>
</form>
<br>
EOM
}

# Rebuild database from default.tbl

sub cgiAdmRebuildFromTab {

    return unless &saveFormToolbar();

    printHead("Rebuilding database from tab file");

    open( h, "_cache/default.tbl" ) || wbFail("Open of default.tbl failed");
    @records = <h>;    # attn: newlines included !
    close(h);

    open( h, ">_cache/default.old" );
    binmode(h);
    print h @records;
    close(h);

    print "<br>default.tbl backed up to default.old\n";

    admRebuild();
}

# rebuild data from default.tbl

sub admRebuild {

    *fieldNames = &wbReadRecordsFnames();

    # @records contains our database
    # fieldNames are the field names

    # delete the *.rec

    # do recAll();
    # $count = 0;
    # while($id = &recNext()) {
    #	&recDelete($id,1) && $count++;
    #}

    #print "<br>$count files deleted\n";

    wbBackupData();

    print "<br>Old versions of data backed up!\n";

    $todo = $CGI{'action'};
    $todo =~ tr/\r//d;

    open( H, ">>_tmp/rebuild-actions.txt" );
    flock( H, 2 );
    binmode(H);
    $date = &formatDate(time);
    binmode(H);
    print H "# --- $date from $RA\n";
    print H "$todo\n\n";
    close(H);

    $| = 1;

    print "<br>Actions will be logged to _tmp/rebuild-actions.txt\n";
    print "<br>Will do this to each record:<br><pre>$todo</pre>\n\n\n";

    $count = 0;
    foreach $rec (@records) {
        recFromRow( $rec, *fieldNames );
        &evalSafe($todo);
        if ($@) {
            print qq(<p class="ERROR">$id: $@<br>Failed and aborted!\n);
            last;
        }
        $id = $rec{'_id'};
        delete $rec{'_id'};

        if ( ( $newID = &evalWBBkey() ) ne '' ) {

            # newID was computed
        }
        else {

            # automatic ID ... no new ID
            $newID = $id;
        }

        $rec{'_id'} = $newID;

        recWrite() && $count++;
        wbRecordChanged();
    }

    print "<br>$count records written\n";

    wbDatabaseChanged();
}

# print frame set for modifying settings

sub cgiAdmModifySettingFrames {

    $type = $CGI{'type'};
    $type =~ tr/[a-z]/[A-Z]/;
    $t = $type . 'ToFrameX';

    print <<EOM;
Content-type: text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">
<head>
<title>Modify Setting</title>
</head>
<frameset rows="100,*" border="1" bordercolor=#000000>
	<frame src="$SCRIPT/AdmModifySettingMenu" name="woda_menu" marginwidth="0" marginheight="0">
	<frameset cols="25%,*" border="1">
		<frame src="$ICONURL/pixel.gif" name="woda_sidebar" marginwidth="0" marginheight="0">
		<frame src="$ICONURL/pixel.gif" name="woda_right">
	</frameset>
</frameset>
EOM
}

# print menu for modifying settings

sub cgiAdmModifySettingMenu {

    printFrameHead('Modify Settings');

    print <<EOM;
<table cellspacing="0" width="100%"><tr><td align="left">
<nobr>
<form action="$SCRIPT/AdmModifySettingBToFrameX" target="woda_sidebar">
<b><span class="SMALL">
Modify:
<a class="BUTTON" href="$SCRIPT/AdmModifyFieldsForm" target="woda_right">Structure</a>
<a class="BUTTON" href="$SCRIPT/AdmModifySettingFToFrameX" target="woda_sidebar">Fields' settings</a>
<a class="BUTTON" href="$SCRIPT/AdmModifySettingBToFrameX?titles=1" target="woda_sidebar">Table settings</a>
<a class="BUTTON" href="$SCRIPT/AdmModifySettingBToFrameX" target="woda_sidebar">(long)</a>
<a class="BUTTON" href="$SCRIPT/AdmModifySettingBToFrameX?sort=1&amp;titles=0" target="woda_sidebar">(AZ)</a>
<input name="search" size="10">
EOM

    print <<EOM;
</span></b>
</form>
</nobr>
</td>
<td align="right" width="50">
<nobr>
<a href="$SCRIPT/Home" target="_parent"><img src="$ICONURL/home.gif" $ICONPAR title="Main Menu" alt="Main Menu"></a>
<a href="$SCRIPT/AdmMenu" target="_parent"><img src="$ICONURL/ahome.gif" $ICONPAR alt="Admin Menu" title="Admin Menu"></a>
</nobr>
</td></tr></table>
EOM
}

#  print frame with links to settings

sub cgiAdmModifySettingB {

    printFrameHead("Table settings:");
    print "<br>";
    admMeta();
    admPrintWBBs();
}

# print frame with links to settings

sub cgiAdmModifySettingFind {

    printFrameHead("Find settings:");
    print "<br>";
    admMeta();
    admPrintWBBs();
}

# print frame with links to field definitions

sub cgiAdmModifySettingF {

    printFrameHead("Modify Fields:");
    print "<br>";
    admMeta();
    wbSortFields();
    $i = 1;
    foreach $f (@Fields) {
        print qq(<br><a href="$SCRIPT/AdmModifySettingToFrameX?type=f&amp;name=$f" target="woda_right">$f</a>\n);
        $i++;
    }
}

# print form to modify a setting

sub cgiAdmModifySetting {

    $stt     = $CGI{'type'};
    $name    = $CGI{'name'};
    $feature = $CGI{'feature'};

    printFrameHead("Edit setting '$name'");

    if ( $stt eq '' ) {
        print 'Select a setting in the left frame and edit its value here.';    #UK
        return;
    }

    admMeta();
    wbSortFields();

    $act = 'AdmCreateCGIToFrameX';

    if ( $COOKIE{_conf} !~ m/:NOHELP/ ) {
        print qq(<div style="margin:3px;"><a href="$SCRIPT/SetHelp?help=off&$Referer" class="BUTTON" $TargetSelf>Turn Help Off</a></div>);    #UK
        $Help = 1;
    }
    else {
        print qq(<div style="margin:3px;"><a href="$SCRIPT/SetHelp?help=on&$Referer" class="BUTTON" $TargetSelf>Turn Help On</a></div>);      #UK
        $Help = 0;
    }

    print <<EOM;
<form class="FORM" action="$SCRIPT/$act" method="post" target="_self">


<table border="0" cellspacing="0">
EOM

    if ( $stt eq 'b' && $MetaB{ $name, 'what' } ne 'H' ) {

        print &admMetaInput( $name, '', $WBB{$name} );

    }
    elsif ( $stt eq 'b' && $MetaB{ $name, 'what' } eq 'H' ) {

        $yes = 0;
        foreach $fld (@MetaB) {
            if ( $MetaB{ $fld, 'what' } eq 'H' ) {
                if ( $fld eq $name ) {
                    $yes = 1;
                }
                else {
                    $yes = 0;
                }
            }
            else {
                print &admMetaInput( $fld, '', $WBB{$fld} ) if $yes;
            }
        }

    }
    else {

        print &admMetaInputField( 1, $name, $feature );

    }

    print "</table>";

    if ($DBDEF) {
        $fn = "$DBDEF.pl";
    }
    else {
        $SCRIPTNAME =~ m|/([^/]*$)|;
        $fn = "$1";
        $fn = $fn . '.pl' unless $fn =~ m/pl$|cgi$/;
    }

    admChangeDefinitionButton();

    print "</form>\n";
}

# prints WBB parameters in $ncols columns
# links them to a huge form in the frame 'woda_right'
# or to a form that displays only one of them

sub admPrintWBBs {

    print '<table width="100%" class="DATATABLE"><tr><td>';

    undef @y;
    undef @x;
    if ( $CGI{'sort'} eq '1' ) {
        @MetaB = sort(@MetaB);
    }

    foreach $fld (@MetaB) {

        if ( $MetaB{ $fld, 'what' } eq 'H' ) {
            if (@x) {
                print &formatMenuTitle(@y) unless $CGI{'titles'} eq '0';
                print '<span class="SMALL">', "\n";
                while (@x) {
                    $text = shift(@x);
                    $url  = shift(@x);
                    if ( $CGI{'titles'} eq '1' ) {
                        print ".";
                    }
                    else {
                        print qq(<a href="$url">$text</a><br>\n);
                    }
                }
                print "<br></span>";
                undef @y;
                undef @x;
            }
            $y[0] = $fld;
            $x    = &htEscape($fld);
            $y[1] = qq($SCRIPT/AdmModifySettingToFrameX?type=b&amp;name=$x" class="BUTTON" target="woda_right);    # there is no missing quote here
            next;
        }

        next unless &ddAdminCanModify( 'B', $fld );
        if ( $CGI{'search'} ne '' ) {
            $s = $CGI{'search'};
            next unless $fld =~ m/$s/i;
        }

        if ( substr( $MetaB{ $fld, 'uk' }, 0, 1 ) eq '*' ) {
            $mand = "<b>*</b>";
        }
        else {
            $mand = "";
        }

        if ( $WBB{$fld} ne '' ) {
            $set = '<tt>[X]</tt>';
        }
        else {
            $set = '<tt>[&nbsp;]</tt>';
        }

        push( @x, "$set$fld" );
        push( @x, "$SCRIPT/AdmModifySettingToFrameX?type=b&amp;name=$fld\" target=\"woda_right" );

    }

    print "</td></tr></table>";

}

# shortcut to run Expire, Static and Agent Process

sub cgiAdmDaily {

    if ($CMDLINE) {
        $x = $WBB{'dbTitle'};
        print <<EOM;
Subject: $x: Daily Administration 
MIME-Version: 1.0
Content-type: text/html; charset=US-ASCII
Content-transfer-encoding: 7BIT

EOM
    }
    else {
        printHead("Daily Administration");
    }

    $HTMLFILE = 1;
    wbDo("cgiAdmExpire");
    $HTMLFILE = 1;
    wbDo("cgiAdmStatic");
    $HTMLFILE = 1;
    wbDo("cgiAdmAgentProcess");    # if $WBB{'agentGroups'};

}

# remove cached files etc.
# remove static files from _cache.
# Be Warned: You probably don't want to Use AdmClean?all=true to remove lastID!

sub cgiAdmClean {

    my @remove = ( '_readme', 'access', 'closed', 'register' );
    push( @remove, 'lastid' ) if $CGI{'all'};    # option to remove last ID
    my $fileList = join( '|', @remove );

    printHead("Cleaning old data");

    # --- qry dir

    $dir = "$DATADIR/_qry";
    opendir( D, $dir );
    @files = readdir(D);
    closedir(D);
    foreach $f ( sort(@files) ) {
        next if $f =~ m/_readme/;
        $file = "$dir/$f";
        next if -d $file;
        print "<br />unlink $f\n";
        unlink $file;
    }

    # --- cache dir
    $dir = "$DATADIR/_cache";
    opendir( D, $dir );
    @files = readdir(D);
    closedir(D);

    foreach $f ( sort(@files) ) {
        next if $f =~ m/$fileList/i;    #MF
        $file = "$dir/$f";
        next if -d $file;
        print "<br />unlink $f\n";
        unlink $file;
    }
}

# generate static pages

sub cgiAdmStatic {

    # static page is what is after /cgi-bin/application
    # extension is htm
    # ?/|()/ --> _

    local ($selected);

    printHead("Making table of contents for robots");

    #	print <<EOM;
    #<b>Warning</b>: generated static pages will have links relative to
    #<a href=$SCRIPT>$SCRIPT</a>. If you are testing some new script
    #on a database which is served by a different script, the generated
    #pages are all wrong!
    #</p>
    #<p class="WARN">There seem to be some problems when doing this under WINDOS. The page looks bad,
    #but the right things seem to get done anyway. Looks like Perl's select() function is not ported
    #100% onto WIN-DOS.
    #</p>
    #EOM

    # --- main title page

    # --- write static html pages for each record into htmlDir

    #	if ($WBB{'robots'} ne '0') {
    admStatic();
    print "<h3>Table of contents</h3> <p> At <a href=\"$WBB{'htmlURL'}/robots/index.htm\">index.htm</a>\n";

    #		}

    # this is done regardless of the robots setting

    # write the static home page (maybe with a link to the static)

    if ( $WBB{'generatedHomePage'} ) {
        $hp = "$HTMLDIR/$WBB{'generatedHomePage'}";
        open( cache, ">$hp" ) || wbFail("Cannot open file $hp");
        binmode(cache);
        $selected = select(cache);
        $HTMLFILE = 1;
        wbDo("cgiHome");
        printFoot();
        $HTMLFILE = 0;
        close(cache);
        select($selected);
        print "<br>Wrote home page into $hp\n";
    }
}

# generate Table of contents for robots

sub cgiRobotToc {

    printHead("TOC for Robots");

    recAll();
    while ( $id = &recNext() ) {
        print "<a href=\"$SCRIPT/Show?$id\">$id</a>, \n";
    }
}

# generates static pages for robots
# should not make any output

sub admStatic {

    wbCheckDirectories();

    if (1) {
        $toc = "";
        recAll();
        while ( $id = &recNext() ) {
            $toc .= "<a href=\"$SCRIPT/Show?$id\">$id</a>, \n";

            #			$hp = "$HTMLDIR/robots/$id.htm";
            # is hp newer than $rec{'_time'} ...
            #		$toc .= "<li> <a href=$WBB{'htmlURL'}/robots/$id.htm>$id</a>\n";
            #			open (cache,">$hp") ||wbFail("Cannot open file $hp");
            #		binmode(cache);
            #		print "<br>Page $WBB{'htmlURL'}/robots/$id.htm\n";
            #			   $selected = select(cache);
            #			$HTMLFILE=1;
            #		  $CGI{'_id'}=$id; $group=$Group; $Group='guest';
            #		  $IGNOREERRORS=1;
            #wbDo ("cgiShow");
            #		  undef @Toolbar;
            #		  delete $CGI{'_id'};
            #		  $Group=$group;
            #			$HTMLFILE=0;
            #			close (cache);
            #		select($selected);
        }

        # write table of contents

        $hp = "$HTMLDIR/robots/index.htm";
        open( cache, ">$hp" ) || wbFail("Cannot open file $hp");
        binmode(cache);
        $selected = select(cache);
        $homeURL  = &homeURL();
        $HTMLFILE = 1;
        printHead("Database Contents");    #UK
        print <<EOM;
<p>
$WBB{'about'}
<h3>Index</h3>
<ul>
$toc
</ul>
EOM
        close(cache);
        select($selected);
    }
}

# export fields as text files

sub cgiAdmExportTextfiles {

    printHead("Export as many text files");

    if ( $CGI{'fields'} eq '' ) {
        print <<EOM;
This will export some fields as text files (which can then be used, for example, for indexing
with a third party software. The files will be created in $DATADIR/_tmp/txt/.
<form action="$SCRIPT/AdmExportTextfiles">
Fields:<input name="fields">
<br><input class="BUTTON" type="submit" value="Export">
</form>
EOM
        return;
    }

    @fields = split( /\,/, $CGI{'fields'} );

    $dir = "$DATADIR/_tmp/txt";
    mkdir( $dir, 0777 );

    $count = 0;
    recAll();
    while ( $id = &recNext() ) {
        recRead($id);
        open( h, ">$dir/$id.txt" );
        binmode(h);
        foreach $field (@fields) {
            print h $rec{$field};
            print h "\n";
        }
        close(h);
        $count++;
    }

    print "<p>$count records printed into $dir\n";
}

# reindex database

sub cgiAdmReindex {

    printHead("Reindexing database");

    # touch the cached files here, generate above

    foreach $sort ( DEFAULT, TIME, U1, U2, U3 ) {
        next if !$WBB{"sort;$sort"} && $sort ne 'TIME';
        wbReindex( $sort, 1 );
        print "<br>Updated $sort.tbl\n";
    }

    return;
}

# reads the environment; SCRIPTNAME not used!

sub admGetEnv {

    local ($otherDataDir) = @_;
    local ( $name, $value, $x );

    $otherDataDir = &wbTables( 'DIR', $otherDataDir );

    open( h, "$otherDataDir/_cache/access3.txt" ) || return ();

    while (<h>) {
        chop;
        $i++;
        ( $name, $value ) = split(/=/);
        if ( $name eq 'UID' ) {
            $> = $value if $UNIX;
        }
        elsif ( $name eq 'URL' ) {
            $x = $value;
        }
        elsif ( $name eq 'SCRIPTNAME' ) {
            $sn = $value;
        }
    }
    close h;

    return ( $1, $2, $3, $4 ) if $x =~ m|(.*)://([^/]*):(\d*)/(.*)|;
    return ( $1, $2, 80, $3 ) if $x =~ m|(.*)://([^/]*)/(.*)|;
}

# print global variables

sub cgiAdmPrintVars {

    printHead("Global Variables in WODA");
    $var = <<'EOM';
$AT::(W) what is UNIX at program
$CHMOD::(W) what is UNIX chmod program
$CRONTAB::(W) where is crontab program
$GREP::(W) where is grep program
$DUMPABLEPERL::(W) where is a dumpable version of Perl
$MV::(W) where is the command to move files
$UNDUMP::where is undump program

$DATADIR::base directory for non HTML related files of this table
$HTMLDIR::base directory for HTML related files of this table

$ADMINPASS::password was ignored, because we are admin
$ADMUSER::reserved
$AF_INET::reserved
$Action::action currently executed
$BIGICONPAR::(W) IMG tags for big icons
$CMDLINE::are we running from command line?
$CRLF::reserved
$CookieGroup::reserved
$CookiePass::reserved
$DIRMODE::(W) in what mode to create directories
$DISK::(W) on WINDOS, on which drive letter is WODA related stuff 
$FILETYPES::for which file types we have extensions
$Group::current group the user belongs to
$HTMLFILE::is output going to a file
$HTTP200OK::(W) add extra HTTPD messages to please some NT servers
$Help::(i)does the user requier a lot of help in forms
$HtmlHeadPrinted::was top of HTML pages printed
$HttpHeadPrinted::(i)was HTTP header printed
$ICONOPEN::reserved
$ICONPAR::(W) standard parameters for IMG tags of icons
$ICONURL::(W) URL of the icons
$WODAURL::(W) URL of various woda related support files
$InFrame::is output going into a frame
$JS::current JavaScript version supported by this browser
$MINIHEAD::print a miniature head only
$MyAction::what would be the name of the overloaded function for this Page
$NOCACHE::reserved
$NOLOGICON::don't display log icon in toolbar
$NO_CRYPT::Perl does not have crypt function
$NO_DIRTIMES::operating system does not show directory modification times
$NO_TIMES::Perl does not have times function
$PDFTOHTML::where is PDFTOHTML program (for swish-e indexing)
$SWISH::where is swish-e (for indexing and searching)
$PERL::where is Perl
$PERLV::version of Perl
$Page::name of the Page requested
$PageCount::number of pages served
$PrintingBasket::are we printing shopping basket
$RGroup::record group
$Referer::the full URL of this page
$ROBOTIP::IP of where robots come from
$ROBOTAGENT::user agent pattern for robots
$SCRIPT::URL of this script, relative to host, including all login data etc. passed on in PATH_INFO
$SCRIPTNAME::like above but without password etc.
$SCRIPTBASE::base host for this script
$SCRIPTURL::full length URL of this script
$SCRIPTFILE::file executed to run this application, can be the code with WBB and WBF definitions or /use$ or /tt$ file that loads a definition
$SCRIPTDIR::directory of SCRIPTFILE
$SCRIPTPACKAGE::in which package is the script
$SOCK_STREAM::reserved
$SENDMAIL::where is sendmail program
$SMTPHOST::address of a machine that can relay email for us 
$SuffixDefault::reserved
$SuffixDetail::reserved
$SuffixSelf::reserved
$TESTCGIDIR::(W)directory to write generated definitions into
$TargetDefault::reserved
$TargetDetail::reserved
$TargetParent::reserved
$TargetSelf::reserved
$ThisRecordNo::record's serial number in search results
$ThisRecordBasket::URL to add this record to basket
$ThisRecordDate::data when this record was modified
$ThisRecordIcons::icons that go with this record
$ThisRecordMatches::how this record matches
$ThisRecordURL::URL of current record
$UNIX::is this a UNIX system ?
$URL::URL of the current page including all login and frams junk
$URLGroup::(before 4.0) group of current user as set in the URL
$LogPass::password of current user as set in the URL or cookies
$User::username of current user
$VALid::reserved
$VERSION::current version of WODA
$WodaPackage::in which Perl package is WODA code 
$WodaFile::name of this script
$WodaFileDir::in which directory is this script
$WodaHelpURL::URL of the WODA help files
$WodaHomeURL::URL of home of WODA
$WodaLanguage::language of messages in this WODA
$WodaLanguageAuthor::who translated WODA to this language
$WodaLibDir::directory with library of subroutines
EOM

    @var = split( /\n/, $var );
    print <<EOM;
<p>This table lists the global variables that are used
within the WODA engine. Variables marked with (W) are static
setings defined in the WODA code. Some of these variables are
defined at all times, others only during some operations.
<p><b>Within the database definition file, all these variables should
be considered read-only! Information about them is povided for power users
and for debugging purposes.
They are not cast in concrete and their names and values might change
in the future</b></p>
 
<table border="1">
<tr>
<th>Variable</th>
<th>Description</th>
<th>Current Value</th>
</tr>
EOM

    foreach $var (@var) {
        ( $name, $desc ) = split( /::/, $var, 3 );
        $value = &evalSafe("$name");
        $value = &htQuote($value);
        print <<EOM;
<tr>
<td><b>$name</b></td>
<td>$desc</td>
<td>$value&nbsp;</td>
</tr>
EOM
    }
    print "</table>\n";
}

# print database attributes

sub admConfAttrs {

    print "<h2>Legend - database attributes </h2><p>These attributes apply to the database as a whole.";    #UK
    print '<table border="0">';

    foreach $wbase (@MetaB) {

        next unless &ddAdminCanModify( 'B', $wbase );

        if ( $MetaB{ $wbase, 'what' } eq 'H' ) {
            print <<EOM;
<tr  class="MENUT" valign="top"><td>&nbsp;</td>
<td align="left">
<b><u>$wbase</u></b>

</td>
</tr>

EOM
            next;
        }

        $value = $MetaB{ $wbase, uk };
        $type = &admMetaType( $MetaB{ $wbase, 'type' } );
        $type = <<EOM;
(<a href="$SCRIPT/AdmMetaTypes" target="woda_help">$type</a>)
EOM
        print <<EOM;
<tr valign="top">
<td align="right">
<b><a name="_b$wbase">$wbase</a></b>
</td>

<td>
$type $value
</td>
</tr>
EOM
    }
    print "</table>\n";

    print '<h2>Legend - field attributes </h2><p>The attributes apply to individual fields of a table.';    #UK
    print '<table cellspacing="0">';

    foreach $wbase (@MetaF) {

        next unless &ddAdminCanModify( 'F', $wbase );

        $value = $MetaF{ $wbase, uk };
        $type = &admMetaType( $MetaF{ $wbase, 'type' } );
        $type = <<EOM;
(<a href="$SCRIPT/AdmMetaTypes" target="woda_help">$type</a>)
EOM

        print <<EOM;
<tr valign="top">
<td align="right">
<b><a name="_f$wbase">$wbase</a></b>
</td>

<td>
$type $value
</td>
</tr>

EOM
    }

    print "</table>\n";

    print "<h2>Field Types Supported by WODA</h2>";

    setTypes();
    @types = sort( keys(%FieldType) );
    undef @val;
    foreach $type (@types) {
        push( @val, $FieldType{$type} );
    }

    print &formatNameVal( @types, @val );
}

# populate %MetaF and %MetaB

sub admMA {

    local ( $what, $name, $uk, $size, $deflt ) = @_;

    "$what " =~ m/^(.)(.)/;
    $what = $1;
    $type = $2;

    if ( $uk =~ s/^\*// ) {
        $mandatory = 1;
    }
    else {
        $mandatory = 0;
    }

    if ( $what eq 'B' || $what eq 'H' ) {

        $uk = &formatBaseDocs( $uk, $name );

        push( @MetaB, $name );
        $MetaB{ $name, 'what' } = $what;
        $MetaB{ $name, 'uk' }   = $uk;
        $MetaB{ $name, 'size' } = $size;
        $MetaB{ $name, 'dflt' } = $deflt;
        $MetaB{ $name, 'type' } = $type;
        $MetaB{ $name, 'mand' } = $mandatory;

    }
    elsif ( $what eq 'F' ) {

        $uk = &formatFldDocs( $uk, $name );

        push( @MetaF, $name );
        $MetaF{ $name, 'what' } = $what;
        $MetaF{ $name, 'uk' }   = $uk;
        $MetaF{ $name, 'size' } = $size;
        $MetaF{ $name, 'dflt' } = $deflt;
        $MetaF{ $name, 'type' } = $type;
        $MetaF{ $name, 'mand' } = $mandatory;
    }
}

# return link to WODA manual

sub formatBaseDocs {

    local ( $uk, $wbb ) = @_;

    #	$uk = &htQuote($uk);
    $uk =~ s|manual\b|<a href="$WodaHomeURL/man.htm">manual</a>|g;			# \b: word limit because we are greedy (g)
    $uk =~ s/\^(b|f)([a-zA-Z;]+)/<a href="#_$1$2" $TargetSelf><em>$2<\/em><\/a>/g;

    $x = '<table width="62%" class="DATATABLE" ><tr><td><pre>';
    $y = '</pre></td></tr></table>';

    $uk =~ s/\[([^\]]*)\]/$x . &htQuote($1) . $y/ge;

    return $uk;
}

# generate links to WODA help within pre-formatted text

sub formatFldDocs {

    local ( $uk, $wbb ) = @_;

    #	$uk = &htQuote($uk);
    $uk =~ s|manual\b|<a href="$WodaHomeURL/man.htm">manual</a>|g;			# \b: word limit because we are greedy (g)
    $uk =~ s/\^(b|f)([a-zA-Z;]+)/<a href="#_$1$2" $TargetSelf><em>$2<\/em><\/a>/g;

    $x = '<table width="62%" class="DATA1" ><tr><td><pre>';
    $y = '</pre></td></tr></table>';

    $uk =~ s/\[/<br>Example:$x/g;
    $uk =~ s/\]/$y/g;

    return $uk;
}

# populate %FieldType

sub setTypes {

    return if defined %FieldType;

    $FieldType{'INPUT'}    = "The defaul  field type. A line  of text.";
    $FieldType{'TEXTAREA'} = "Several rows of preformatted text. Adjust  the
 size with rows=n cols=m. Will be tagged as &lt;PRE&gt; unless P or BR tags  are
 present";

    $FieldType{'HTMLAREA'} = "A paragraph  of text  or a  whole tagged HTML
 document. Any formatting should be done  in HTML. If HTML tags are  not present
 two newlines  will be  replaced by  P tag.  Adjust size  with rows=n  cols=m in
 typePar. Enable rich editing with RICH in typePar";

    $FieldType{'OPTION'} = "An option list. Define options in the  'options'
 attribute. 'typePar' may include RADIO, SIZE, MULTIPLE and TREE. In the  second
 case field will be offered for hierarchical browsing.";

    $FieldType{'LINKOPTION'} = "(obsolete) An option list where the options are keys from another table.
Define table's alias or dataDir in the 'into' attribute";

    $FieldType{'LIST'} = "A list of values. Values should be entered one per row.
Formatting and verification applies one per each item";

    $FieldType{'DATE'} = "Date type. Internally represented as yyyy/mm/dd.";

    $FieldType{'EMAIL'} = "An email address. By default rendered as mailto: HREF, checked to include '\@a'.";

    $FieldType{'URL'} = "A URL. By default rendered as A HREF";

    $FieldType{'FILE'} = "A file somewhere on the server, relative to htmlURL.
E.g. picture, full text document ... By default rendered as link to that file.";

    $FieldType{'USERFILE'} = "Like FILE but database users can upload it to the server using the Netscape 2+ or Explorer 4+.";

    $FieldType{'IMAGE'} = "Like USERFILE but will be rendered on screen as an IMG.";

    $FieldType{'COMPUTE'} = "A field which is not stored in the database but is computed on the fly when record is read.";

    $FieldType{'BREAK'} = "A horizontal line to separate fields in forms and tabular output";

    $FieldType{'SUBMIT'} = "Link to submit buttons. Use if you want an extra section in the middle of the form.";
}

# return all directories within $ICONDIR

sub getIconSets {

    local (@all);

    return () if $ICONDIR eq '';
    opendir( D, $ICONDIR )
      || wbWarn( "Cannot open icon directory",
        "Setting \$ICONDIR=$ICONDIR sucks! This error will prevent you from choosing alternate sets of icons!" );
    @all = readdir(D);
    closedir(D);

    @all = grep( !/\./, @all );
    @all = sort(@all);

    return @all;
}

# print out explanation of all variable types

sub cgiAdmMetaTypes {

    printHead("Setting types");
    foreach $t ( 'c', 's', 'l', 'r', 'x' ) {
        push( @n, &admMetaType($t) );
        push( @v, &admMetaType( $t, 'LONG' ) );
    }

    print &formatNameVal( @n, @v );
}

# return explanation of meta variable type

sub admMetaType {

    local ( $what, $how ) = @_;

    if ( $how ne 'LONG' ) {
        return "constant" if $what eq '' || $what eq ' ' || $what eq 'c';
        return "string expression"  if $what eq 's';
        return "string or constant" if $what eq 'x';
        return "logical expression" if $what eq 'l';
        return "regular expression" if $what eq 'r';
        return "unknown type '$type'";
    }
    else {
        return <<EOM if $what eq '' || $what eq ' ' || $what eq 'c';
A fixed, constant value. Will be printed as is.
EOM
        return <<'EOM' if $what eq 's';
A little Perl program, that returns a string:<pre>
"The value of variable thisvariable = $thisvariable"
</pre>
EOM

        return <<'EOM' if $what eq 'x';
If it looks like Perl code it will be evaluated,
if not it is taken to be a constant.
EOM

        return <<'EOM' if $what eq 'l';
A little Perl program, that returns true (no zero) or false (zero, empty):<pre>
if ($variable > 1000) {
	1;
} else {
	0;
}
</pre>
EOM
        return <<'EOM' if $what eq 'r';
This matches anything starting with a capital letter:<pre>
[A-Z].*
</pre>
EOM

    }
}

#
# admMeta ... defines database's meta information into MetaB and MetaF
#
# s = string expression
# l = logical expression
# r = regular expression
# c = constant
# x = constant or string expression, subject to evalMaybe

sub admMeta {

    return if @MetaB;

    setTypes();
    $types = join( '|', sort( keys(%FieldType) ) );
    $types = '|' . $types;

    &admMA( 'H', 'Essentials' );

    &admMA( 'B', 'dbTitle', <<EOM, 40 );
* database title. Is displayed linked to this table's main menu.
[members]
EOM

    &admMA( 'Bx', 'apTitle', <<EOM, 40 );
application title. Constant or string that will be printed to the left of the ^bdbTitle 
Should be linked to a main menu page of an application dealing with several tables.
Example below (combined with the setting of ^bdbTitle) would create a page supertitle beach club : members 
[beach club : ]
EOM

    &admMA( 'Bc', 'about', <<EOM, 100 );
short descriptive text about the table
[This database stores the members of the
Beach Club]
EOM

    &admMA( 'B', 'recordTitle', <<EOM);
title of one record. See also ^bdetailTitle.
[Club member] 
EOM

    &admMA( 'B', 'manager', <<EOM);
* name of database's manager (see ^bmanagerEmail)
[Joe Smith]
EOM

    &admMA( 'B', 'managerEmail', <<'EOM');
* ^bmanager's email
[jsmith@beachclub.com]
EOM

    &admMA( 'B', 'dataDir', <<EOM);
* directory where data of the database is stored (without trailing slash!)
[/usr/local/www/wdat/beachclub/members]
EOM

    &admMA( 'B', 'htmlDir', <<EOM);
directory within the document path of the server,
where the attachments and other html documents
will be stored. If undefined, it will be set to ^bdataDir
[/usr/local/www/htdocs/beachclub/members]
EOM

    &admMA( 'B', 'htmlURL', <<EOM);
A URL which points to the same directory as the ^bhtmlDir.
Usually it can be determined automatically if the server
reports the DOCUMENT_ROOT (e.g. /usr/local/www/htdocs)
variable (try with /AdmDebug) and
it matches ^bhtmlDir. The unmatched part is
what should be written here.
[/beachclub/members]
EOM

    &admMA( 'Bs', "sort;DEFAULT", <<'EOM');
string to use for DEFAULT sorting of records.
If undefined records will be sorted
by the first field.
["$rec{'lastName'} $rec{'firstName'}"]
EOM

    &admMA( 'Bs', "format;DEFAULT", <<'EOM', 100 );
REPLACED WITH ^bformatRowDEFAULT in WODA 3.4 and higher.
String, which formats DEFAULT
printout of one record in a listing. It will be placed between
&lt;OL&gt; tags. If undefined a table will be printed. 
["$rec{'firstName'} $rec{'initial'} $rec{'lastName'}"]
EOM

    &admMA( 'Bs', 'key', <<'EOM');
string using $rec{''} variables, which generates record key.
Default is a random number 4 hex digits long. Set to &NEXT(f,i)
to set the ID of first record to number f and then increment by i.
String &NEXT or NEXT() must be seen in this setting.
["$rec{'lastName'} $rec{'firstName'}"]
EOM

    &admMA( 'H', 'Parameters describing Web Services' );

    &admMA( 'Bc', 'provider', <<EOM, 40 );
provider of this service; needed only in services such as /Office calls.
[Santa Monica Beach Club Ltd.]
EOM

    &admMA( 'Bc', 'license', <<EOM, 100 );
usage license for this service; needed only in services such as /Office calls.
[some legal text]
EOM

    &admMA( 'Bc', 'officeCategory', <<EOM, 40 );
Service category for /Office calls.
[RESEARCH_GENERAL]
EOM

    &admMA( 'H', 'Change the appearance of every page' );

    &admMA( 'B', 'style', <<EOM, 100 );
Stylesheet to apply on pages generated.
Value of this field will
be inserted between &lt;STYLE&gt; tags.
View source to see which CLASSes are used.
If empty, the default style sheet will be applied.
Set to '/* */' to insert an empty style sheet.
EOM

    &admMA( 'B', 'styleModify', <<EOM, 100 );
Include content of this style AFTER the standard WODA
style sheet. This gives a possibility to modify some WODA
styles while keeping most of them intact.
EOM

    @x = split( /\n/, <<EOM);

Times New Roman,Times,serif
Helvetica,Arial,sans-serif
Albertus,sans-serif
Arial,sans-serif
Balthazar,serif
BankGothis,sans-serif
Bodoni,serif
Book Antigua,serif
Bookman,serif
Century Gothic,sans-serif
Clarendon,serif
Comic,Comic Sans MS,sans-serif
Garamond,serif
Graphos,serif
Letter Gothic,Courier,Courier New
Mistral AV,Comic,Comic Sans MS,sans-serif
Simplex,sans-serif
Stylus BT,Comic,Comic Sans MS,sans-serif
Tahoma,Arial,sans-serif
Technic,Arial,sans-serif
Univers,Universe Condensed,sans-serif
VAG Round,Arial,sans-serif
Verdana,Arial,sans-serif
EOM

    &admMA( 'B', 'font', <<EOM, join( '|', @x ) );
Font to be used in pages. This setting override any
font settings in the ^bstyle sheet.
EOM

    &admMA( 'B', 'script', <<EOM, 100 );
Reserved.
EOM

    &admMA( 'B', 'head', <<EOM, 100 );
Tags to insert into the HEAD section of all generated pages
*after* any HEAD elements that are inserted automatically.
EOM

    &admMA( 'B', 'pageBody', <<EOM);
parameters for the HTML's BODY tag;
since 4.3 the use of this setting is discouraged.
[BGCOLOR=#000000 text=#000000 link=#0000FF]
EOM

    &admMA( 'B', 'frameBody', <<EOM);
parameters for the BODY tag for 
title and system frames on the few framed pages
[BGCOLOR=#808080]
EOM

    &admMA( 'Bx', 'pageHead', <<EOM, 100 );
text to display on every page top above ^bpageTitle and below BODY tag
EOM

    &admMA( 'B', 'pageHeadIcon', <<EOM, 100 );
text to display on every page inside the
default generated table for the title.
Good place to put LEFT or RIGHT aligned IMG tags.
EOM

    &admMA( 'Bx', 'pageFoot', <<EOM, 100 );
text to display at the end of every page, below
the default generated toolbar.
EOM

    &admMA( 'Bs', 'dateFormat', <<'EOM', 30 );
String evaluating to how to display dates.
$yyyy contains year, $mm contains month and $dd the day.
The default behaviour is
["$yyyy/$mm/$dd"]
EOM

    &admMA( 'Bs', 'timeFormat', <<'EOM', 30 );
String evaluating to how to display times.
$hh contains hour in 24 hour format,
$mm contains minute and $ss second.
The default behaviour is
["$hh:$mm"]
EOM

    &admMA( 'B', 'JavaScript', <<EOM, 4 );
Maximum JavaScript version that should be used
to enhance the generated pages. Set to 0 to disable JavaScript.
Keep empty to let the script determine the JavaScript version
that a browser can support.
EOM

    &admMA( 'Bs', 'doctype', <<'EOM', 80 );
The !doctype of pages generated. Leave empty for default. See http://www.w3.org/QA/2002/04/valid-dtd-list.html for a list of valid doctypes.
[<!DOCTYPE HTML>]
[<!-- DOCTYPE HTML this is like not having it -->]
EOM

    &admMA( 'H', 'Icons' );

    &admMA( 'B', 'iconURL', <<EOM);
URL of a directory where icons are kept. No trailing slash!
Keep empty if a system set of icons is used.
EOM

    @x = &getIconSets();
    $x = join( '|', @x );

    &admMA( 'B', 'iconSet', <<EOM, $x );
name of the set of predefined icons to be used by this application.
This setting will not be used if ^biconURL is defined! You may
<a href="$SCRIPT/IconDocs">display sample icons from the icon sets</a>.
EOM

    &admMA( 'B', 'iconOpen', <<EOM);
Base name for the icons that denote found item in a database. The default is 
'open' and icons open1.gif ... open4.gif are used in search listings etc.
You can also set it to floppy, paper, mail, face or any other graphics
where you created a set of four icons.
[face]
[floppy] 
EOM

    &admMA( 'H', 'Make big changes to header and footer of every page' );

    &admMA( 'B', 'pageTemplate', <<'EOM', 100 );
If it starts with http, an URL, otherwise a file (absolute or relative to
^bdataDir) that contains a template how a page should look like.
This template can be created with any HTML design application
(Frontpage, DreamWeaver ...). When you use the 'Apply Page Template'
command on the Admin menu, this template is copied to database definition
settings ^bhead, ^bpageBody, ^bpageTitle and ^bpageEnd like this:
<ul>
<li> everything above string $body is considered ^bpageTitle.
Use $dbTitleLink and $pgTitle.
<li> $body is replaced with the main text of the page
<li> everything below $body is considered to be ^bpageEnd. Use $toolbar
to place icons and $woda to place Woda icon.
<li> BODY parameters are copied to ^bpageBody
<li> The entire HEAD section is
copied to ^bhead. TITLE and BASE are renamed to XTITLE and XBASE!
</ul>
WARNING: Without calling 'Apply Page Template' this setting
has no effect. Also, changes in the template are sensed by this
application only after you manually apply them.
[template.dwt
http://www.what.is/Templates/niceone]
EOM

    &admMA( 'Bs', 'pageTitle', <<'EOM', 100 );
string expression to generate on every page top below pageHead
You can use the following variables:
<ul>
<li>$phi - content of ^pageHeadIcon
<li>$pgTitle - the title of the page
<li>$dbTitle - the value in ^bdbTitle
<li>$homeURL - URL of the database's home page
<li>$scriptURL - URL of this script,
<li>$loginBox - a box offering users to log in - only displayed if ^buserTable is set
</ul>
If not set, something like this is printed
as default:
["
<div class="DBTITLE">
<a href="$homeURL">$dbTitle</a>
</div>
<div class="PGTITLE">
$pgTitle
</div>
"]
EOM

    &admMA( 'Bx', 'pageEnd', <<'EOM', 100 );
string expression which generates whatever
is printed at the end of every page, just above ^bpageFoot. Variables:
<ul><li>$PageCount contains page count of this application.
<li>@Toolbar contains the elements that should
be placed into the toolbar, 
<li>$toolbar contains the toolbar icons
<li>$woda contains the icon that points to Woda. THIS ONE MUST BE PRESENT.
</ul>
If not set, something
like this is printed as default:
[
$x = join("</td><td>",@Toolbar);
"<table><tr><td>$x</td></table>";
]
EOM

    &admMA( 'H', 'Add to toolbars' );

    &admMA( 'Bx', 'toolbarText', <<EOM, 60 );
The default toolbar has some system icons on the
left and right.
In between is a wide empty cell. This text will be
printed in it.
[(c) Blinky Bill 1993]
EOM

    &admMA( 'B', 'toolbar', <<EOM, 100 );
Contains a table with space-separated 
columns and several rows.
The left column contains a pattern to match against
the action part of the URL (Search, Add, Home ...).
The right contains a string expression (variables can be used!)
to add into one cell of the toolbar table. In the below example, 
the first item is added to Add pages only, the second one to all pages.
Third item calls a subroutine to generate the string.
[Add	"<img src=help.gif> help"
.	"(c) 2001 The Oddisey Group"
Edit	&myEditToolbar()]
EOM

    &admMA( 'H', 'Database home page' );

    &admMA( 'B', 'homePage', <<EOM);
Name of a HTML file which is the homepage of the database
(in ^bhtmlDir) unless contains a / in which case it is
treated as a URL. If empty, the database's home will
be the /Home main menu page automatically
generated by this script and not a static HTML file.
[home.htm]
[http://www.somesite.com/index.htm]
EOM

    &admMA( 'B', 'generatedHomePage', <<EOM);
WODA can automatically generate a .htm file for the database
home page. Define here into which file it should create it (relative to ^bhtmlDir.
Can be the same as ^bhomePage in which case the generated .htm will replace
what you wrote in it. Mainly used to speed up the delivery of the database's
home page.
[home.htm]
EOM

    &admMA( 'Bx', 'homePageFormat', <<'EOM', 100 );
HTML that defines the home page / main menu of the
application. Predefined variables are $MENU, $SEARCH, $ABOUT, $LOGIN and $DATA.
$DATA depends on ^bhomePageData. $MENU can be changed by ^buserMenuBrowse and
related settings.
["
<table width="100%">
<tr valign="top">
<td>
$SEARCH
$ABOUT
$DATA
$LOGIN
</td>
<td>
<td bgcolor=#000000>&nbsp;</td>
<td>$MENU</td>
</tr>
</table>
"]
EOM

    &admMA( 'B', 'homePageData', <<EOM);
What data from the database should be inserted into the home
or main menu page. Possible values are:
<ul>
<li> (blank) ... last 5 items are displayed
<li> N ... last N items are displayed
<li> fieldName ... browse menu of that field is displayed
</ul>
[8]
EOM

    &admMA( 'H', 'Add to menus on the Home page' );

    &admMA( 'B', 'userMenuBrowse', <<EOM, 100 );
comma separated list of text,url to add to Browse menu. One per row!
If it does not include // or start with a / assumed to be relative to current application.
EOM

    &admMA( 'B', 'userMenuFind', <<EOM, 100 );
comma separated list of text,url to add to Find menu. One per row!
If it does not include // or start with a / assumed to be relative to current application.
[Search websites,http://www.x.com/search
Search here,Search?search=foobar]
EOM

    &admMA( 'B', 'userMenuAdd', <<EOM, 100 );
comma separated list of text,url to add to Add menu. One per row!
If it does not include // or start with a / assumed to be relative to current application.
EOM

    &admMA( 'B', 'userMenuOther', <<EOM, 100 );
comma separated list of text,url to add to other menu. One per row!
If it does not include // or start with a / assumed to be relative to current application.
EOM

    &admMA( 'H', 'Change the look of the data entry form' );

    &admMA( 'Bx', 'formFoot', <<EOM, 100 );
text to display at the end of every entry form
EOM

    &admMA( 'B', 'verifyEdits', <<EOM, 1 );
If set to non zero, users will need to verify
the records that they enter, before it is submitted.
EOM

    &admMA( 'Bs', 'form', <<'EOM', 100 );
What does the main part of the input form look like. Should return a string.
The defined variables for each field are:
<br>$Form{field,name}
<br>$Form{field,prompt}
<br>$Form{field,input}
<br>$Form{field,help}
<br>$Form{field,err}
<br>$Form{field,continue}
EOM

    &admMA( 'H', 'Change the look of the Detail page' );

    &admMA( 'Bs', "detail", <<EOM, 100 );
how to format a page with one record. If undefined the fields will
be printed in default format - a two column table. See also ^bdetailTitle.
EOM

    &admMA( 'Bs', 'detailTitle', <<'EOM', 100 );
string expression to generate as title on the default detail page
Use $phi, $pgTitle, $dbTitle, $homeURL, $scriptURL, which contain site icon (^bpageHeadIcon),
page title (^bpageTitle), database title (^bdbTitle), database's home URL (^bhomePage) 
and script's home URL.
["Full details of $rec{'firstName'} $rec{'lastName'}"]
EOM

    &admMA( 'H', 'Change the Look of the Search Results Page' );

    &admMA( 'B', 'searchTextLinks', <<'EOM', 2 );
Use text for links to other pages of search results. By default, the links are created with images and placed in the toolbar.
[1]
EOM

    &admMA( 'B', 'searchNoHitCount', <<'EOM', 2 );
Do not print hit count or text links. By default, the number of records found and the current record range is printed. <em>Do not use this option with searchTextLinks (^bsearchTextLinks) as it prevents text links from being printed.</em>
[1]
EOM

    &admMA( 'B', 'searchNoTools', <<'EOM', 2 );
Turn user tools off. By default, all user tools (format, sort, show, export) are displayed on the search results page. It is a much easier way to say no format tools (^bsearchNoFormat),  no sort tools (^bsearchNoSort),  no show tools (^bsearchNoShow),  no export tools (^searchNoExport).
[1]
EOM
    &admMA( 'B', 'searchNoFormat', <<'EOM', 2 );
Turn off format tools. By default, users can choose to display the results in different formats. 
[1]
EOM

    &admMA( 'B', 'searchNoSort', <<'EOM', 2 );
Turn off sort tools. By default, users can change the sort order.
[1]
EOM

    &admMA( 'B', 'searchNoShow', <<'EOM', 2 );
Turn off Show tools. By default, users can choose to show more or less records per page
[1]
EOM

    &admMA( 'B', 'searchNoExport', <<'EOM', 2 );
Turn off export tools. By default, privileged users can export records.
[1]
EOM

    &admMA( 'H', 'Speed up searches' );

    &admMA( 'B', 'grepBigger', <<EOM, 300000 );
If nonzero and if \$GREP WODA variable is defined,
UNIX's grep will be used to speed up searches if
the size of data is larger than this number of bytes. Set to
500000 or more. See also ^bsearchCache if you want to
speed up searches.
[300000]
EOM

    &admMA( 'B', 'cacheExpire', <<EOM, '0|1|2|3|4|6|8|12|24|48|72|96|168' );
after how many hours after the changes in the database should
the cached files and indexes be rebuilt. Default - 0 (zero)
means 'after each change in the database'.
Since indexing takes some time a value of 0 may
slow down very large databases. This value in fact means
that it may take at most this many hours before an entered record
is found.
EOM

    &admMA( 'B', 'onExpire', <<EOM);
Perl code to execute when cache would expire.
This routine can be used to occasionally bring into the
database some data from external sources.
[do myOnExpire]
EOM

    &admMA( 'B', 'searchCache', <<EOM);
If nonzero, WODA will cache search results in the _qry
directory. Files will be named *.tbl. This speeds up searches.
EOM

    &admMA( 'B', 'searchCacheAge', <<EOM);
How many hours after a cached file was created
it should be deleted. Setting this to 24
would make sure that cached files that have not been used
a day or more will be deleted automatically.
Default is one month (720 hours). Cleaning takes place
at most once an hour.
EOM

    &admMA( 'H', 'More ways to sort search results. Searches from U1 to U9' );

    foreach $p ( 1 .. 3 ) {    # U1 .. U9

        &admMA( 'B', "sortName;U$p", <<EOM, 20 );
name of $p sorting
[by state and town]
EOM

        &admMA( 'Bs', "sort;U$p", <<EOM);
string to use for $p sorting of records. See ^bsort;DEFAULT for details.
["\$rec{'state'} \$rec{'town'}"]
EOM
    }

    &admMA( 'H', 'User defined search result formats from one to nine' );

    foreach $p ( 'DEFAULT', 1 .. 9 ) {

        &admMA( 'B', "formatName$p", <<EOM, 20 );
This and next settings (ending in digits 1-9)
define how found records are printed out.
Here, the name of this format is
defined and will be used in menus and pull-downs.
[a very detailed format]
EOM

        &admMA( 'Bx', "formatHttpHead$p", <<EOM, 100 );
Http Header to be printed for this particular format.
Specify only if the setting is not text/html!
WARNING: Currently it can not provide correct Content-length!
[Content-type: text/xml]
EOM

        &admMA( 'Bx', "formatHead$p", <<EOM, 100 );
Next settings define how a list of found
records is laid out. First the formatHead
is printed, then, for each record, formatRow,
and finally, ^bformatFoot.
[<table>]
EOM

        &admMA( 'Bx', "formatRow$p", <<'EOM', 100 );
See ^bformatHead for details. The following variables
can be used:
<ul>
<li>$ID ... ID of this record
<li>$ThisRecordNo ... serial number of this found record
<li>$ThisRecordURL ... URL of the found record.
<li>$ThisRecordDate ... date when the record was last modified.
<li>$ThisRecordIcons ... system action actions for this record.
<li>$ThisRecordBasket ... checkbox for the shopping basket.
<li>$ThisRecordMatches ... numeric value of the relevance of this record.
</ul>
["<tr>
<td><a href="$ThisRecordURL">$rec{'firstName'}</a></td>
<td>$rec{'lastName'}</td>
</tr>
"]
EOM

        &admMA( 'Bx', "formatFoot$p", <<EOM, 100 );
What is printed after all found records.
[</table>]
EOM
    }

    &admMA( 'H', 'Custom browse options' );

    &admMA( 'Bs', 'tree', <<'EOM');
expression which creates database subtitles for browsing. It takes
some time to create this index therefore all 'trees' are cached. Changing this value
will not require a cache update. After change manually clear the *.3.* files from _cache!
For most applications, ^btree is an overkill. See TREE as ^ftypePar if
tree data comes from one field only.
WARNING: this tag is not obsolete, but it is no more implemented since Woda 4.000. Might be in future.
["$rec{'category'};$rec{'subcategory'}"]
EOM

    &admMA( 'Br', 'treeSplit', <<EOM);
pattern which splits '^btree' into category and subcategory.
WARNING: this tag is not obsolete, but it is no more implemented since Woda 4.000. Might be in future.
[;]
EOM

    &admMA( 'B', 'AZindex', <<EOM, 100 );
Table, which defines how A-Z indexes of the database are defined.
Each row defines one A-Z index. The columns are (1) ID (a word which defines
a name of the A-Z index, (2) the comma separated list of field names which are used to
provide the words for the A-Z index and (3) the rest is a nice textual description of
the A-Z index.
[byname	lastName,firstName	Index of member names.
bystate	state	Index of members by state]
EOM

    &admMA( 'B', 'AZindexMethod', <<EOM, 100 );
Table, which defines how A-Z indexes of the database are computed.
The columns are (1) ID (a word which defines
a name of the A-Z index, as in ^bAZindex. (2) the comma separated list of methods to
apply. Methods ASCII,UPPER,LOWER,TITLE,STEM translate found text. STEM is rather primitive. Any combination may be
applied. Methods to split text into words are SPACE,COMMA,SEMICOLON and COLON. Only one can be used.
[bystate	UPPER,SPACE
byname	LOWER]
EOM

    &admMA( 'B', 'AZnot', <<EOM, 100 );
List of words which should not be included in the ^bAZindex. 
The words should be white-space separated. Regular expressions may be used.
[a the is at]
EOM

    &admMA( 'B', 'ringField', <<EOM, 100 );
The name of a field to be used by BrowseRing. 
If more than one field has been defined as a URL then ringField can be used to specify which field to use. 
If ringField is empty the first field that is defined as URL is used.
[publicUrl]
EOM

    &admMA( 'H', 'Changing headers and footers on some pages' );

    foreach $p ( 'Home', 'Search', 'Add', 'Changed', 'Show', 'Store', 'Edit', 'NewForm', 'AdvancedSearchForm', 'SearchIDForm', 'AdmMenu' ) {

        &admMA( 'Bx', "title;$p", <<EOM, 100 );
text to display on top of page $p
EOM

        &admMA( 'Bx', "head;$p", <<EOM, 100 );
text to display on top of page $p
EOM

        &admMA( 'Bx', "foot;$p", <<EOM, 100 );
text to display at the end of page $p
EOM
    }
    &admMA( 'H', 'Shopping basket' );

    &admMA( 'B', 'basketName', <<EOM, 100 );
If defined, WODA will allow adding to a shopping basket in all
search and similar pages. This is not only useful for shopping but
for compiling a list of relevant found records. These can later be 
printed out together etc. The value entered should fit nicely after
"add to " ... e.g. "favourites","shopping list","your personal selection"
etc. Baskets related to user's ID. Files named like Userid.bskt are in the
_friend directory. Expired baskets are deleted as a part of maintenance procedure.
[local favourite members]
EOM

    &admMA( 'B', 'basketExpire', <<EOM, 5 );
Baskets expire in after 24 hours. This tag is currently not implemented.
EOM

    &admMA( 'H', 'Automated maintenance' );

    &admMA( 'Bs', 'afterRecordModify', <<EOM);
Expression (Perl statement or program) that is
executed after a record has been modified and written to disk.
%rec still contains field values.
EOM

    &admMA( 'Bs', 'afterTableModify', <<EOM);
Expression (Perl statement or program) that is
executed after a table is modified.
[`/usr/local/bin/execHtm whatsnew.htm`]
EOM

    &admMA( 'B', 'maintenancePeriod', <<EOM);
Every how many hours should the automatic
maintenance of the database take place (or 0 to disable). HTTPD must be
allowed to use UNIX's 'at' command. 
If you set up Angie, this should be once a day. If the database 
is little used, use cron instead.
[24]
EOM

    &admMA( 'Bs', 'deleteAfter', <<EOM, 2 );
delete records older than this many days (0=never)
EOM

    &admMA( 'H', 'Relational features' );

    &admMA( 'B', 'tables', <<EOM, 100 );
<p>In functions such as ROWS, QRY etc. and
some filed settings, one needs to refer to
another table. This setting defines where that
table is. The definition is only needed if
it is not possible to guess, where the table is.
<p>
If the related table is laid out like this one, there is no need to
define it in this field. If table 'products' is handled by 
/cgi/products with data in /www/wdat/products and if 'people' are handled by
/cgi/people with data in /www/wdat/people there is no need
to mention them here!
<p>
(since version 4.400) If the related table is on another server
it is defined by a row consisting of an ALIAS and URL. Some operations on such
tables will be much slower than on local tables.
<ol>
<li>ALIAS is the logical name to be used in the program.
<li>URL is the URL of the script handling this table. Starts with http.
<li>DIR (optional). If defined it is assumed this is the data directory 
on this server. Could speed things up. However, it the table is really on another server,
this field should be empty!
</ol>
<p>
(for backward compatibility) If the related table is on the same server and its location
is different then a row in this setting defines it. The row has three whitespace
separated fields:ALIAS PROGRAM DATADIR.
<ol>
<li>ALIAS is the logical name to be used in the program.
<li>PROGRAM program is the name of the database definition file. This will be used
to guess the URL so that any part of the path before /cgi or /db will be
discarded and the remaining part used after server URL.
<li>DATADIR is the name of the database's data directory.
</ol>
[members	/www/cgi-bin/members /www/wdat/members
states	/www/cgi-bin/states /www/wdat/states
users	http://users.scix.net/cgi-bin/users]
EOM

    &admMA( 'B', 'requireFile', <<EOM);
file with add-in subroutines and settings to be require-ed by Perl. If
path
<ol>
<li>starts with / it is absolute
<li>starts with letter of digit it should be stored in \$TEMPLATEDIR/lib
<li>otherwise it is relative to ^bdataDir of this table.
</ol>
Typically used to store the ^btables and other settings to ensure consistent
formatting of several tables of a relational application. First the original table definition is loaded,
then the require file, then the common.pl file at ^bdataDir/../common.pl
[/usr/www/cgi-bin/common.pl]
EOM

    &admMA( 'H', 'Group based security' );

    &admMA( 'B', 'groups', <<EOM, 100, $ADMUSER );
New in 4.105! What groups of users does this application support. Setting contains
a table with three tab or whitespace separated columns: (1)group-name, (2)host-pattern, (3)password.
During login, this table will be scanned so that for the given login parameters
both password AND host pattern would match. 
See below: 'ziga' can be only a person coming from 193.2.92.14 AND supplying the password 'turk'.
Ziga can be also a person coming from other places but knowing password 'secret'.
Slovenians are all that come from 193.2.92. To become Italian or English or German
(# are translated to . for compatibility reasons) one must supply a password. You cant become french,
no such IP is possible! System defined groups are guest, admin, friend, owner and client.
[
ziga	193.2.92.14	turk
ziga	.	secret
slovene	193.2.92.*
italian	.	milano
english	.	london
french  !	paris
german	#	berlin
]	
EOM

    &admMA( 'B', 'rights', <<EOM, 100 );
Who is allowed to do what. Also see manual.
Table with with four space
separated columns (1) allow/deny, (2) group (from ^bgroups) or group pattern, (3) action-pattern, (4-optional) parameter-pattern.
By default, guests cannot use Basket and SearchHistory and nobody is
allowed to use Agent* and Admin* actions. Any other actions are
allowed by default. See manual. See ^bownerField for definition of
two magic groups. Below, we deny guests to search - they will need
to register as friend first! Friends can Add or Edit stuff.
Special can subscribe to Agent searches. Enemies are denied everything.
We also deny anyone search for 'sex!' (note that there is no need to escape the '!')
and guest to set max parameter to start
with 3.  
[deny	guest	Search
allow	friend	Add|Edit
allow	special Agent.*
deny	enemies	.*
deny	.	Search	sex!
deny	guest	Search	max=3]
EOM

    &admMA( 'Bs', 'userScope', <<EOM, 10 );
To pass usernames and passwords, cookies are used. By default,
cookies are only set for one table at a time and as a result
a user must log in to several related tables separately. This
string should be set to the local part of the URL for which
cookies should be valid, starting with a slash. E.g. /cgi below will
pass username and password to all cgi programs on the same server. 
[/cgi]
EOM

    &admMA( 'H', 'User based security' );

    &admMA( 'B', 'userTable', <<EOM, 10 );
An alias of a table from ^btables field that
contains information about the users of this table.
Username will be the ^bkey as defined in the ^buserTable.
Password of the user will be the same as the password
used lock a record in the ^buserTable. The user will
belong to a group given in ^buserTable's field called 'group', or
to the default group given in the ^buserTableGroup field
here or to group 'guest' (in that order).
See manual.
[members]
EOM

    &admMA( 'B', 'userTableGroup', <<EOM, 10 );
Group (from ^bgroups) to which all users
defined in table ^buserTable belong, unless
the ^buserTable contains field named 'group' which
overrides this default. If undefined, the users defined
in ^buserTable default to group 'user'.
[friends]
EOM

    &admMA( 'B', 'ownerGroup', <<EOM, 10 );
Name of group that gets record owner status
(can edit, modify,delete record) even if
does not know the password of the record.
Users from this group in effect get admin's
rights in the record.
[editor]
EOM

    &admMA( 'Bl', 'userAllow', <<EOM);
Logical expression, based on data in table pointed to by ^buserTable
that evaluates to true, if a given user is allowed to log into
this database or to false, if not. If empty or equal to 1
anyone is allowed to login. Otherwise admin can check applicants
before letting them use the database for example (below),
by setting the status field to OK. It would make sense to set 
^fmodifies of status to 'admin'.
[\$rec{'status'} eq "OK"]
EOM

    &admMA( 'B', 'recordLock', <<EOM, 60, 'password|user' );
How a person entering some data may lock it. The options are:
<ul>
<li>password ... record will be locked with a password
<li>user ... record will be locked to a user (option is not avalaible to guests)
<li>group ... record is locked to a group from ^bgroups
<li>host ... record is locked to an IP number
<li>none ... record can not be locked at all
</ul>
These words most be entered using | as a delimiter. The default is:
[password|user]
EOM

    &admMA( 'B', 'ownerField', <<EOM, 10 );
Name of field that contains the
username of the owner of this record.
If the value in this field is the same as
the username of the logged in user (one from ^buserTable)
the ^frights, ^fsees and ^fmodifies for a magic group 
'owner' will apply in the context of this record.
If the value of this field is '0' or ''
the rights for magic group 'public' will apply. This setting lets you
build databases where records are protected by user,
not by password, and a combination of both.
[recordOwner]
EOM

    &admMA( 'H', 'Special protection of records' );

    &admMA( 'B', 'filter', <<EOM, 100 );
Hide some records from some ^bgroups. Table with two tab separated columns. In the first is
a name of a group as defined in the ^bgroups attribute. In the second is the
search condition which will be silently added to any search any user from this group 
performs. The example below only records that contain the word "abracadabra" will
be found. WARNING: without the + the search below would behave exactly the same way!
See also ^bfilterText
[guest	+abracadabra]
EOM

    &admMA( 'B', 'filterText', <<EOM, 100 );
Used to define explanations of filters defined in ^ffilter.
[guest	Filtered to abracadabra]
EOM

    &admMA( 'Bl', 'hideUnless', <<'EOM', 100 );
^ffilter prevents record to be found. hideUnless prevents them
to be shown if someone happens to know the id and keys in
the URL. Only records for which this expression evaluates
to TRUE will be displayed.
[$Group eq "user" && $rec{'_id'} eq $User]
EOM

    &admMA( 'B', 'hideUnderscore', <<EOM, 1 );
Obsolete in versions above 4.300
EOM

    &admMA( 'H', 'Limiting number of found records' );

    &admMA( 'B', 'dfltHits', <<EOM, 5, $WBB{'dfltHits'} );
default number hits per page in searches
EOM

    &admMA( 'B', 'maxHits', <<EOM, 5, $WBB{'maxHits'} );
maximum allowed hits per page in searches. On the advanced search page,
users may ask for up to this many hits per page.
EOM

    &admMA( 'H', 'Support for non-english languages' );

    &admMA( 'B', 'language', <<EOM, 2, 'uk' );
* User interface language. Enter a two letter language code (English=UK, Slovenian=SI ...).
You must have the appropriate woda-xx.pl library installed, to support your language.
[UK]
EOM

    &admMA( 'B', 'intlCharset', <<EOM, 10 );
character set used in pages of the database
[ISO-8859-1]
[WINDOWS-1250]
EOM

    &admMA( 'B', 'intlCollate', <<EOM);
how to sort international characters. Many space separated strings like
e&#234;&#233; o&#246;&#244; which mean that &#234;&#233; should be sorted after
e, &#246;&#244; after o ... Define on lower case only!
EOM

    &admMA( 'B', 'intlLower', <<EOM);
how to convert uppercase international letters into lower case
or other character equivalencies in sorting.
Two space separated strings must be defined - the first are upper case letters, 
the second corresponding lower case ones: like &#201;&#202;&#218;&#219; eeuu 
In this case they will be sorted as if they were equivalent to e-s and u-s
EOM

    &admMA( 'B', 'intlAscii', <<EOM);
how to convert lower case international letters into corresponding ASCII ones
(eg. &egrave; into e).
Two space separated strings must be defined - the first are the non-ascii
characters, then are the corresponding ascii ones. Both strings must be equal size!
EOM

    &admMA( 'H', 'Other options' );

    &admMA( 'B', 'logLevel', <<EOM, 1, 3 );
Log level. The default is 3.
<ul>
<li>0 - no logging
<li>1 - logging fatal errors only
<li>2 - logging all errors
<li>3 - logging all actions
<li>4 - logging all actions with comments
</ul>
[1]
EOM

    &admMA( 'B', 'logFile', <<EOM, 10 );
Name of logfile. Defaults to $yyyy$mm.log 
File will be in the ^bdataDir/_log unless name starts with a /, assuming
an absolute path. Name may include variables $yyyy
for year, $mm for month and $dd for day.
Note that this is not the server's log or error log file.
Errors which prevented WODA to start will not show up here.
The columns in the log file
are separated with tabs and are (left to right):
<ol>
<li>ok|ER|FA to denote ERror or FAtal error
<li>time in seconds used to process the request
<li>date and time when request was made in ISO format
<li>host or IP number from which request was made
<li>username of the user
<li>group to which the user belonged
<li>request method
<li>script executed
<li>page requested
<li>query string (encoded)
<li>any WODA error messages
</ol>
This will create a new log file every day:
[$yyyy-$mm-$dd.log]
EOM

    &admMA( 'B', 'robots', <<EOM, 1 );
Setting is obsolete since version 4.596 (automatic handling of robots). This setting is no more implemented.
EOM

    &admMA( 'B', 'adFile', <<EOM);
Obsolete. This setting is no more implemented.
EOM

    #	do wbHelps();
    #	do htQuote($help); ... too long
    &admMA( 'B', 'searchHelp', <<EOM, 100 );
several rows of help messages that will be displayed below the
search form. Each row should contain one, self contained help message.
In the example below are the built-in help messages. Note that some rows may be wrapped in your browser - in fact there is one
sentence per row!
[Sorry, no help example available.]
EOM

    &admMA( 'B', 'quotaText', <<EOM, 10, 0 );
Maximum allowed size (in megabytes) for the data in the text part of the table.
After that size is reached, the modifications of the table will not be allowed.
Unlimited if 0.
[0.5]
EOM

    &admMA( 'B', 'quotaBin', <<EOM, 10, 0 );
Maximum allowed size (in megabytes) for all binary attachments the data in the table.
After that size is reached, the modifications of the table will not be allowed.
Unlimited if 0.
[2]
EOM

    &admMA( 'B', 'quotaLoad', <<EOM, 100, 0 );
What operations are allowed at what server load.
A table with two tab separated columns. Left=load
above which operation is denied. Right=pattern
of actions, like in the ^brights setting.
[10	Search
5	Show]
EOM

    &admMA( 'H', 'z End' );

    ### --- field level data --- ###

    &admMA( 'F', 'type', <<EOM, $types );
* Field type.
[OPTION]
EOM

    &admMA( 'F', 'typePar', <<EOM);
Additional field parameters. By default these are added to the INPUT (e.g. size="n"),
TEXTAREA (e.g. rows="x" cols="y") or SELECT tags (MULTIPLE).
Special values are:
<ul>
<li> size="n" ... as in HTML
<li> CONTINUE (does not work since 3.5),
<li> HIDDEN (makes an INPUT field hidden),
<li> OWNER says that if this field is one of the many EMAIL TYPE fields, it should be used to reset the password.
<li> RICH enables rich text editing of HTMLAREA fields.
</ul>
These apply to OPTION or LINKOPTION fields where a selection
of many values is possible:
<ul>
<li> cols="n" puts CHECKBOXES in n columns
<li> RADIO - options should be presented as radio buttons.
<li> CHECKBOX - options should be presented as checkboxes (the default is a pulldown list)
<li> TREE - field should be used for hierarchical browsing of
the database. ^foptions must be like europe:italy:rome, that is,
categories and subcategories separated by colons.
<li> NOID hides IDs in OPTION fields, where display and stored values are different.
<li> HEADING="xxxxx" will display items starting with x as headings of OPTION fields, not as values to be selected. "xxxxx" can be 1 to 5 characters.
</ul>
[CHECKBOX MULTIPLE]
EOM

    &admMA( 'F', 'p', <<EOM);
Short prompt for field, shown above field.
[Check all operating systems on which your software works]
EOM

    &admMA( 'F', 'help', <<EOM, 100 );
Help text on field input, shown below field. Usually much longer than prompt text.
[Do not check all systems if
it is a Java program.
Just click next to Java,
you dummy.]
EOM

    &admMA( 'Fl', '', <<EOM, '', '1;' );
* Unnamed attribute. Expression which validates the field value in \$_. 
Keep 1; if anything goes. In BREAK fields text of the break. In USERFILE fields, \$_
is the name of the temp version of the file. Good enough to check for
extension or existence. length(\$CGI{'FieldName'}) gives the length of the
uploaded file. Note that you can write your own subroutine
to validate record as a whole, the subroutine must be named myRecVerify.
[length($_)>20;]
EOM

    &admMA( 'F', 'cond', <<EOM);
The above validation expression but written in plain English (or another language).
[Text should be at least 20 characters!]
EOM

    &admMA( 'F', 'head', <<EOM);
Name of the field for table headings. If undefined, field name is used.
[Supported operating system]
EOM

    &admMA( 'F', 'options', <<EOM, 100 );
* Possible options for OPTION typed fields in the following formats:
<ul>
<li> one the same line separated by |
<li> one per line. If the options are in one-per row format, then the row 
can have several columns separated with TABs. The first column is used as
a value to be stored in the database. All others are shown on screen to make
the selection clearer. 
<li>Value starts with & or # it will be evaluated as a Perl program
which should return rows as described above.
</ul>
Several ^ftypePar options provide details.

[yes|no|maybe]
[&ROWS('people','_id,givenName,lastName')]
[	(unknown)
WIN3.1
WIN95
WIN98	(also includes WIN2000)
WINNT
MAC
LINUX
FreeBSD
HP-UX
Java
Excel
Word]
EOM

    &admMA( 'Fx', 'd', <<EOM);
Default value. Constant string or a perl expression returning a default value. 
[WIN98]
[
$User;
]
EOM

    &admMA( 'F', 'into', <<EOM);
DO NOT USE ... WILL BE OBSOLETE SOON! Defines that the data in this field point(s) to record(s) in table defined as parameter here.
(1) dataDir or alias of the related database.
Keys of this database will be extracted in forms if field type is LINKOPTION.
(2) If this field starts with an & character it will be evaluated.
The result of the program should be formatted as the ^foptions setting above.
This lets you add more information about the selection. Typically you would use
&ROWS function.
[platforms] # (1)
[&ROWS("platforms","_id,name")] # (2)
EOM

    &admMA( 'Fs', 'picture', <<'EOM', 100 );
expression which nicely formats field value. Expression may use variables
$_ (raw value of field) and $_id (ID of the current record).
Use &ROWS(), &QRY() and &FLD() functions to relate to other databases.
Below we display an icon instead of text.
["&lt;IMG src=/imgs/platforms/$_.gif&gt;"]
["&lt;B&gt;$_&lt/B&gt;"]
EOM

    &admMA( 'Fs', 'change', <<'EOM', 100 );
expression which changes the value after it was entered in the form and before
it is validated or written into the database. Make sure that repeated 'changes'
not keep changing the value. The value to be changes is in $_. The field is changed to
whatever this code returns. The example removes any newlines.
[$_=~s/\n/ /g;
return $_;]
EOM

    &admMA( 'F', 'sticky', <<EOM);
Leave empty (default) or set to 1 to make the value entered here
default next time the same user enters this form.
E.g. email addresses could be made sticky.
Set to any value including a / to set the path= variable in HTTPD-COOKIE header.
[1]
[/]	
[/cgi/data]
EOM

    &admMA( 'F', 'meta', <<EOM);
Value of the field should map to a meta tag in the html head
section on appropriate pages (e.g. /Show).
[DC.TITLE]
EOM

    &admMA( 'Fr', 'modifies', <<EOM);
| separated list of ^bgroups allowed to modify this field (or empty for all)
[friend|special]
EOM

    &admMA( 'Fr', 'adds', <<EOM);
| separated list of ^bgroups allowed to enter data into this field on the Add page (or empty for all)
[friend|special]
EOM

    &admMA( 'Fr', 'sees', <<EOM);
| separated list of ^bgroups allowed to see the value of the field (or empty for all)
[user|friend]
EOM

    &admMA( 'F', 'srt', <<EOM, 4 );
order of field in forms. The higher the number, the more at the top!
EOM

}

# return HTML code for changing settings
# name of setting
# number of field
# current value

sub admMetaInput {

    # css class .MENUT has caused problems in some browsers #UK
    # redefine MENUT if problems occur #UK

    local ( $name, $number, $value ) = @_;
    local ($out);

    if ($number) {
        $bf = 'F';
    }
    else {
        $bf = 'B';
    }

    if ( !&ddAdminCanModify( $bf, $name ) ) {
        return;
    }

    $nav = '';

    # $nav = "<a href=#_end>[!]</a>\n";

    if ( $number eq '' ) {    # database

        if ( $MetaB{ $name, 'what' } eq 'H' ) {
            return '' if $CGI{'quick'};
            $out = <<EOM;
<tr valign="top" class="MENUT">
<td>$name</td>
</tr>
EOM
            return $out;
        }

        $type = &admMetaType( $MetaB{ $name, 'type' } );
        $prompt = $MetaB{ $name, 'uk' };
        $size   = $MetaB{ $name, 'size' };
        $dflt   = $MetaB{ $name, 'dflt' };
        $bookmark = "<a name=\"_b$name\">&nbsp;</a>";
        $spc      = "<hr>";
        $nameplus = '';
    }
    else {    # field
        $type = &admMetaType( $MetaF{ $name, 'type' } );
        $prompt = $MetaF{ $name, 'uk' };
        $size   = $MetaF{ $name, 'size' };
        $dflt   = $MetaF{ $name, 'dflt' };
        $bookmark = '';
        $spc      = '';

        #$nameplus="[$number]";
        $nameplus = '';
    }

    $dflt = $value if $value ne '';
    $dflt = &htQuote($dflt);

    # $prompt =~ s/\|/, /g;

    $size = 50 if $size eq '';

    if ( $prompt !~ m/^\*/ && $CGI{'quick'} ) {
        return '';
    }

    $out = <<EOM;

<tr valign="top" class="MENUT">
<td align="left">$bookmark<b>$name $nameplus</b>
[<a href="$SCRIPT/AdmMetaTypes" target="woda_help">$type</a>] $nav </td>
</tr>
<tr>
<td align="left">
EOM

    $nn = $name . $number;

    if ( $size =~ m/\|/ ) {    # optionlist
        @opts = split( /\|/, $size );
        $out .= qq(<select name="$nn" size="1">\n);
        foreach $opt (@opts) {
            if ( $opt eq $dflt ) {
                $out .= "<option SELECTED>$opt\n";
            }
            else {
                $out .= "<option>$opt\n";
            }
        }
        $out .= "</select>\n";

    }
    elsif ( $size eq '0' ) {    # reserved
        return '';

    }
    elsif ( $size > 80 ) {
        $out .= qq(<textarea rows="5" cols="50" name="$nn">$dflt</textarea>\n);

    }
    else {                      # one liner

        $out .= qq(<input size="$size" name="$nn" value="$dflt">\n);

    }

    $prompt = '' unless $Help;
    $out .= <<EOM;
<br>
<font class="SMALL">
$prompt
</font>
$spc
</td>
</tr>
EOM
    return $out;
}

# returns name/value pairs as expression

sub admMetaOutput {

    local ( $name, $value ) = @_;

    $value =~ s/'/'."'".'/g;    # make sure ' does not cause
                                # sytax errors
    $value =~ s/\r\n/\n/g;      # CR-LF into LF only !

    return "$name = '$value';\n";
}

# returns HTML code (table) of all @MetaF data

sub admMetaInputField {

    local ( $i, $f, $feature ) = @_;
    local ($out);

    $out = <<EOM;
<br><br>
<a name="$i">
</a>
<table class="FORM" cellspacing="0">
<tr class="TITLE">
<td align="CENTER"><b>Field $f</b>
<input size="20" value="$f" name="_f$i" type="hidden"></td>
</tr>
EOM

    foreach $mf (@MetaF) {
        if ( $feature ne '' ) {
            next unless $mf =~ m/$feature/;
        }

        $value = $WBF{ $f, $mf };
        $value = $WBF{$f} if $mf eq '';

        $out .= &admMetaInput( $mf, $i, $value );
        $out .= "\n";
    }

    $out .= "</table>\n";
}

# print button for action to create new CGI definition in Test directory

sub admChangeDefinitionButton {

    if ( $DDTYPE eq 'CGI' ) {
        print <<EOM;
<p><a name="_end"></a>
<input class="BUTTON" type="submit"  value="Create database definition"> into file
<br><tt>$TESTCGIDIR/</tt><input name="filename" value="$fn" size="20">
<p>or (if empty) print the definition into the browser.
<em class="WARN">Warning!</em> Make sure that the new definition
does not replace the definition currently used by the end-users of the database.
New definition is written if it is found to be syntactically correct and if the
web server software has the right to write it to disk.
<p><b>If there are errors, the will not start with #! but with some HTML garbage!</b> 
EOM
    }
}

# print all CGI functions (web browser actions)

sub cgiAdmPrintCGIAPI {

    printHead("CGI functions");
    print <<EOM;
Below are all cgi functions defined in WODA.
They can be called as $SCRIPT/Name where name is
a listed function. They are deliberately not linked, because
most should not be called out of context and without
proper parameters. The parameters are listed in braces.
The list is machine generated and therefore usually incomplete.
EOM

    "$WodaFileDir/include/api.pl";
    print $cgis;
}

# print all API functions (BIG letter functions)

sub cgiAdmPrintPerlAPI {

    printHead("Perl functions");
    print <<EOM;
<p>These are Perl functions that can be called in
Woda parameters that allow expressions. In Perl you call functions
like this:<pre>
	&myFunction(parameter1,parameter2)
</pre>
EOM
    "$WodaFileDir/include/api.pl";
    print $xxxs;
}

# functions dealing with rows of hashes ... the preferred woda format
# does the reverse of hashExport

sub hashImport {

    local ( *hash, $value ) = @_;
    local (@x);

    @x = split( /\037/, $value );
    foreach (@x) {
        tr|\034|\012|;
    }
    %hash = @x;
}

# translates a hash to a string so that
# there is US unit separator (ASCII 31, octal 037)
# between names and values; newlines are escaped to octal 034

sub hashExport {

    local (*hash) = @_;
    local ( $key, $val, $out );

    $out = '';
    foreach $key ( sort( keys(%hash) ) ) {
        $val = $hash{$key};

        # next if $val eq '';
        $key =~ tr|\012|\034|;
        $val =~ tr|\034| |;      # not 100% simetric!
        $val =~ tr|\012|\034|;
        $out .= "$key\037$val\037";
    }
    chop($out);
    return $out;
}

# reads a file and returns @rows

sub hashRead {

    local ($file) = @_;
    local (@rows);
    open( h, $file );
    while (<h>) {
        chop;
        push( @rows, $_ );
    }
    close(h);

    return @rows;
}

# writes @rows to file

sub hashWrite {

    local ( $file, *rows ) = @_;
    open( h, ">$file" );
    binmode(h);
    foreach $_ (@rows) {
        print h $_ . "\n";
    }
    close(h);
}

# delete  a row from @rows

sub hashDelete {

    local (@rows) = @_;
    local (@out);

    foreach $row (@rows) {
        $row =~ m/_formid.(\d+)/;
        next if $CGI{'delete'} eq $1;
        push( @out, $row );
    }
    return @out;
}

# delete a row from file

sub hashProcess {

    local ($file) = @_;

    @rows = &hashRead($file);
    @rows = &hashDelete(@rows);
    hashWrite( $file, *rows );

    return @rows;
}

# display list of user's forms (searches, etc)

sub cgiFormsList {

    $dform = $CGI{'form'};

    if ($dform) {
        printHead("Saved forms of user $User related to $dform");    #UK
    }
    else {
        printHead("All saved forms of user $User");                  #UK
    }

    @rows = &hashProcess("$DATADIR/_friend/$User.forms");

    print <<EOM;
<p>Click on delete icons to delete archived form.
<p>
EOM

    # list the forms

    print '<table class="DATATABLE" cellspacing="1">', "\n";

    foreach $row (@rows) {
        undef %hash;
        hashImport( *hash, $row );
        next if $dform && $hash{_form} ne $dform;
        $id   = $hash{_formid};
        $form = $hash{_form};
        $name = $hash{_formname};
        print <<EOM;
<tr class="DATA0"  valign="top">
<td>
<a href="$SCRIPT/FormsList?delete=$id">
<img $SMALLICONPAR src="$ICONURL/del.gif" alt="Delete this form" title="Delete this form"></a>
</td>
<td align="right">
$form
</td>
<td>
<a href="$SCRIPT/$form?select=$id">$name</a>
</td>
<td>
EOM

        foreach $key ( sort( keys(%hash) ) ) {
            next if $key =~ m/^_formid|^_form|^_formname|^_confirm/;
            $val = $hash{$key};
            $val = &htQuote($val);
            print "<b>$key</b>=$val\n" if $val;
        }

        print "</td></tr>\n";

    }

    print "</table>";
}

# Save user's forms

sub cgiFormSave {

    printHead("Saving form data");    #UK

    # escaped form data ... CGI unescaped!
    $eFormdata = &htQuote( $CGI{'formdata'} );

    # convert to hash
    hashImport( *x, $CGI{'formdata'} );

    # override the formname
    $x{_formname} = $CGI{'_formname'};

    if ( $x{_formname} eq '' ) {

        %y = %x;
        foreach $key ( keys(%y) ) {
            delete $y{$key} if $y{$key} eq '';
        }

        $h = &formatHash( "Parameters to be saved are:", 1, %y );    #UK

        $a = "Define name";                                          #UK
        $p = "Save";                                                 #UK
        $b = <<EOM;                                                  #UK
<form action="$SCRIPT/FormSave">
Before a form can be saved, it must be named:<br>
<input name="_formname" size="30">
<input type="hidden" name="formdata" value="$eFormdata">
&nbsp;
<input class="BUTTON" type="submit"  value="Save">
</form>
$h
EOM
        print &formatWindow( $a, $b );
        return;
    }
    else {
        $formdata = &hashExport(*x);

        #print "formdata = $CGI{'formdata'}<br>\n";
        #print "name = $x{_formname}\n";
        #do wbFail("$formdata");

        saveFormAppend($formdata);
        $dform = $x{_form};

        saveFormSavedMessage($dform);
    }
}

# called by pages showing forms
# to display save form button *and* restore selections

sub saveFormCan {

    local ($caller) = @_;

    # assing CGI

    if ( $CGI{'formdata'} ne '' ) {
        hashImport( *CGI, $CGI{'formdata'} );
        delete $CGI{'_formid'};
        delete $CGI{'_formname'};
        delete $CGI{'_form'};
    }

    if ( $CGI{'listall'} ne '' ) {
        cgiFormsList($caller);
        return 0;
    }

    $sfa            = 'Save form as';    #UK
    $SaveFormButton = <<EOM;
<input type="hidden" name="_form" value="$caller"><input class="BUTTON" type="submit"  name="saveform" value="$sfa">
<input name="_formname" size="15">
EOM

    # restore forms ...

    @rows = &hashRead("$DATADIR/_friend/$User.forms");
    @rows = grep( /_form.$caller/, @rows );

    if ( $#rows >= 0 ) {

        $select = '<option value="">';
        foreach $row (@rows) {
            hashImport( *x, $row );
            if ( $x{_formid} eq $CGI{'select'} ) {
                %CGI = %x;
                $d   = 'SELECTED';
            }
            else {
                $d = '';
            }
            $x = $x{_formname};
            $select .= qq(<option value="$x{_formid}" $d>$x);
        }

        $SaveFormForm = <<EOM;
<form class="FORM" action="$SCRIPT/$caller">
Archived forms: <select name="select">
$select
</select>
<input class="BUTTON" type="submit"  value="Load">
<input class="BUTTON" type="submit"  name="listall" value="List all">
</form>
EOM

    }
    else {

        $SaveFormForm = '';

    }

    return 1;
}

# called in pages executing forms
# so that save button appears in the toolbar
# also sets $Formdata
# anomaly: cgiSearch resets toolbar so this is called twice!

sub saveFormToolbar {

    # if user selected SAVE in the form

    if ( $CGI{'saveform'} ne '' ) {
        delete $CGI{'saveform'};
        $CGI{'formdata'} = &hashExport(*CGI);
        cgiFormSave();
        return 0;    # caller must return too
    }

    # if user clicked SAVE button in results

    if ( $CGI{'_form'} ) {
        $x        = &hashExport(*CGI);
        $x        = &htEscape($x);
        $Formdata = $x;
        push( @Toolbar, <<EOM) if &wbAllowed( $Group, 'FormSave' ) && !$agentFile;
<a href="$SCRIPT/FormSave?formdata=$Formdata"><img $ICONPAR src="$ICONURL/formsave.gif" alt="Save form data" title="Save form data" $TargetParent></a>
EOM
    }

    return 1;

}

# save recently updated forms

sub saveFormUpdateSince {

    local ( $friend, @records ) = @_;
    local ( @rows, $row, $r, $id, $time, %r );

    open( h, "$DATADIR/_friend/$friend.forms" )
      || wbFail("Cannot open $DATADIR/_friend/$friend.forms");
    @rows = <h>;
    close(h);
    foreach $row (@rows) {
        chop($row);
        hashImport( *r, $row );
        @d = grep( /^$r{_formid} /, @records );
        if (@d) {
            ( $x, $time ) = split( / /, $d[0] );
            $r{since} = $time;
            $row = &hashExport(*r);
        }
    }

    open( h, ">$DATADIR/_friend/$friend.forms" );
    binmode(h);
    print h join( "\n", @rows );
    print h "\n";
    close(h);
}

# append form to user's forms

sub saveFormAppend {

    local ($formdata) = @_;

    open( h, ">>$DATADIR/_friend/$User.forms" );
    flock( h, 2 );
    binmode(h);
    print h "_formid\037" . time . "\037";
    print h $formdata;
    print h "\n";
    close(h);
}

# print message with links to forms

sub saveFormSavedMessage {

    local ($dform) = @_;

    print <<EOM;
<p>Form data saved.
<ul>
<li>Here are <a href="$SCRIPT/FormsList?form=$dform">forms related to $dform</a>.
</li><li>Here are <a href="$SCRIPT/FormsList">all forms</a>
</li></ul>
EOM
}

# send serial email

sub cgiAdmMailForm {

    return unless &saveFormCan('AdmMailForm');    # sets $SaveFormButton, sets %CGI

    printHead("Send serial email");

    wbSortFields();
    $f = join( '}, $rec{', @IFields );
    $f = "\$rec{$f}";

    print <<EOM;
$SaveFormForm

<p>HELP: Fill in the fields. Use \$rec{'fieldname'} notation
to denote field data e.g. if the email address is in the 'email'
field, write \$rec{'email'} into the To field. Use normal search syntax
in the search field. Use preprocessing field to define any processing
to be done before a record is composed (e.g. &PIC() to get nicely formatted
values of fields.
<p><b>The fields in this table are:</b>
$f
<form class="FORM" name="searchForm" action="$SCRIPT/AdmMailSend" method="post">
EOM

    $me = $CGI{'from'} || $WBB{'managerEmail'};
    $em = $CGI{'to'}   || '$rec{\'email\'}';

    @prompts = ( 'Search:', 'Preprocessing:', 'From:', 'To:', 'Cc:', 'Subject:', 'Message:', 'Send to:', 'Action:' );
    @flds = (
        &wbSearchForm( $CGI{'search'}, 'HISTORY' ),
        qq(<textarea wrap="PHYSICAL" name="process" rows="5" cols="50">$CGI{'process'}</textarea>),
        qq(<input name="from" size="40" value="$me">),
        qq(<input name="to" size="40" value="$em">),
        qq(<input name="cc" size="40" value="$me">),
        qq(<input name="subject" size="40" value="$CGI{'subject'}">),
        qq(<textarea wrap=\"PHYSICAL\" name="message" rows="15" cols="50">$CGI{'message'}</textarea>),
qq(<select name="where"><option>SCREEN<option>MAIL<option value="PERL">SCREEN AS PERL CODE</select> Try 'SCREEN' first to check that the right content is sent!),
        qq(<input class="BUTTON" type="submit"  value="Send"> <input class="BUTTON" type="reset"  value="reset" > $SaveFormButton )
    );

    print &formatNameVal( @prompts, @flds );
    print "</form>\n";
}

# send serial email

sub cgiAdmMailSend {

    return unless &saveFormToolbar();

    if ( $CGI{'where'} eq 'PERL' ) {
        print "Content-type: text/plain\n\n";
        print <<EOM;
# this perl code will let you send
# mails like that some other time

EOM
        foreach $name ( keys(%CGI) ) {
            next if $name =~ m/name|where|paint/;
            print '$CGI{' . $name . "} = <<'EOM';\n$CGI{$name}\nEOM\n\n";
        }
        print "do cgiAdmMailForm();";
        return;
    }

    printHead("Sending Serial Mail");

    open( h, ">>$DATADIR/_log/serialmails" );
    flock( h, 2 );
    print h &hashExport(*CGI);
    print h "\n";
    close(h);

    $CGI{'to'}   =~ s/\@/\\\@/;
    $CGI{'cc'}   =~ s/\@/\\\@/;
    $CGI{'from'} =~ s/\@/\\\@/;

    ( $x, *smids ) = &recSelect( '', $CGI{'search'}, '', 'TAB', '_id', 0, 99999999 );

    print "<p>Sending $x messages.<ol>\n";

    foreach $id (@smids) {

        #		print "<hr>id=$id\n";

        recRead($id);

        #	print "<pre>";
        #		print join (',',@fieldNames);
        #		$item =~ tr/\t/,/;
        #		print "\n$item";
        #
        #		print join ("<br>\n",%rec);
        #
        #		last;

        &evalSafe( $CGI{'process'} );
        wbError($@) if $@;

        $cc      = &evalSafe( '"' . "$CGI{'cc'}" . '"' );
        $to      = &evalSafe( '"' . "$CGI{'to'}" . '"' );
        $from    = &evalSafe( '"' . "$CGI{'from'}" . '"' );
        $subject = &evalSafe( '"' . "$CGI{'subject'}" . '"' );
        $message = &evalSafe( '"' . "$CGI{'message'}" . '"' );
        $mess    = <<EOM;
From: $from
Reply-to: $from
Cc: $cc
To: $to
Subject: $subject

$message
EOM

        if ( $to =~ m/\w\@\w/ ) {

            # OK
        }
        else {
            print qq(<li><em class="ERROR">Bad destination address 'To:$to'</em></li>\n);
            next;
        }

        if ( $CGI{'where'} eq 'MAIL' ) {
            if ( !&sendMail($mess) ) {
                print "<li>To: $to\n";
            }
            else {
                print '<li><em class="ERROR">Error sending mesage</em></li>', "\n";
            }
        }
        else {
            $mess = &htQuote($mess);
            print "<li><pre>$mess</pre><hr></li>\n";
        }
    }

    print "</ol>\n";
}

# return HTML link to a url

sub formatTmpWinOpen {
    local ( $url, $name ) = @_;

    $name = $name || 'woda_other';

    if ( $url =~ m/LoginForm$/ && $url =~ m/:/ ) {
        $name = '_self';
        $url =~ s|/[^/:]+:[^/]*||;
        $url .= '?win=self';
        return <<EOM;
<a href="$url" target="$name">
EOM
    }

    if ( $JS >= 1.2 ) {
        return <<EOM;
<a href="#"
onClick='
window.open("$url","$name","scrollbars=yes,resizable=yes,toolbar=no,width="400",height="400"");'
>
EOM
    }
    else {
        return <<EOM;
<a href="$url" target="$name">
EOM
    }
}

# append lapTime to $Log

sub logTime {

    $Log .= &lapTime() . ":\t$_[0] \n";
}

# measure time to perform tasks

sub lapTime {

    local (@x) = eval 'times;' || ( time - $StartTime );
    $x[0] =~ m/^.{0,5}/;
    return $&;
}

# return HTML code to close a window

sub formatTmpWinClose {

    local ( $msg1, $msg2 );
    $msg1 = "Done. Click here to close.";                                              #UK
    $msg2 = "Close this window and <b>reload</b> the other one to see the change.";    #UK
    if ( $JS >= 1.2 ) {
        return <<EOM;
<form>
<input class="BUTTON" type="BUTTON" value="$msg1" size="20"
OnClick='window.opener.document.location.reload();window.close()'
></form>
EOM

    }
    else {
        return "<p>$msg2";
    }
}

#
# # # # # CODE for BASKET SUPPORT
#
# - BasketAdd?id=ID&num=2
# - BasketList
# - BasketEdit?id1=id1&no1=no1...
# - BasketSearch?searchParameters
# - basket file per IP number at /_friend/
# - id<TAB>number

# add one item to the basket

sub cgiBasketAddOne {

    readBasket();
    push( @BasketID, $CGI{'_id'} );
    push( @BasketNO, $CGI{'num'} || 1 );
    writeBasket();
    readBasket();
    printBasket();
}

# add an item to the basket

sub cgiBasketAdd {

    readBasket();
    foreach $key ( keys(%CGI) ) {
        if ( $key =~ m/^id(.*)/ ) {
            if ( $CGI{$key} eq 'on' && $Basket{$key} eq '' ) {
                push( @BasketID, $1 );
                push( @BasketNO, 1 );
            }
        }
    }

    writeBasket();
    printBasket();
}

# fills @BasketID and @BasketNO and %Basket and returns numbers of items in basket

sub readBasket {

    local ( @x, $id, $i, $no, $line, $file );

    undef @BasketID;
    undef @BasketNO;
    undef @lines;

    $file = &basketFile();

    @x = stat($file);
    if ( $x[9] < ( time - 24 * 3600 ) ) {

        # selection expired
        unlink $file;
    }
    else {
        open( b, $file );
        @lines = <b>;
        close(b);
    }

    $i = 0;
    foreach $line (@lines) {
        chop($line);
        ( $id, $no ) = split( /\t/, $line );
        $BasketID[$i] = $id || $line;
        $BasketNO[$i] = $no || 1;
        $Basket{$id}  = $no || 1;
        $i++;
    }

    return $#BasketID;
}

# returns mtime of a file

sub wbMtime {

    local ($file) = @_;
    local (@x);

    @x = stat($file);
    return $x[9];
}

# writes Basket to disk

sub writeBasket {
    local ( %printed, $file, $i, $ii );

    $file = &basketFile();
    if ( $#BasketID < 0 ) {
        unlink($file);
        return 0;
    }
    else {
        open( h, ">$file" );
        binmode(h);
        for ( $i = 0 ; $i <= $#BasketID ; $i++ ) {
            $ii = $BasketID[$i];
            next unless &recExists($ii) >= 0;
            if ( !$printed{$ii} ) {    # remove duplicates
                print h $ii . "\t" . $BasketNO[$i] . "\n" if $BasketNO[$i] > 0;
                $printed{$ii} = 1;
            }
        }
        close(h);
    }
    return $i + 1;
}

# returns name of basket file

sub basketFile {
    local ( $ip, $x );
    $x = wbCheckDir("$DATADIR/_friend");
    return ("$x/$User.bskt");
}

# prints basket on screen

sub cgiBasketShow {
    if ( $CGI{'editable'} eq '' ) {
        $ed = 0;
    }
    else {
        $ed = 1;
    }
    printBasket( $CGI{'format'}, $ed );
}

# prints basket as HTML ... does the form as well
# by default all output goes to the caller window

sub printBasket {

    # by default all output goes to the caller window
    $TargetDefault = 'target="woda_opener"';

    local ($format) = $_[0] || 'DEFAULT';
    local ($editable) = $_[1];

    $editable = 1 if $editable ne '0';

    $n = &readBasket() + 1;

    if ( $n <= 0 ) {

        wbError( "Nothing is in $WBB{'basketName'}", "Nothing selected" );    #UK

    }
    else {

        ( $html, $yh, $yf ) = &printSearchHttpHeader( $format, 1, 1 );

        $modify = "Keep only checked items in the basket";                    #UK
        $go     = "Reformat list as:";                                        #UK
        $eable  = "in editable format";                                       #UK

        printMiniHead("$n items in $WBB{'basketName'}") if $html;             #UK
        $editable = 0 unless $html;

        $formats = &formatFormats( $CGI{'format'} );

        $ch = '';
        $ch = 'CHECKED' if $editable;

        print <<EOM if $editable;
<table><tr><td>
<form class="FORM" action="$SCRIPT/BasketShow" target="_self">
$go
<select name="format">
$formats
</select> $eable
<input name="editable" type="checkbox" $ch>
<input class="BUTTON" type="submit"  value="OK">
</form>
</td></tr>

<tr><td>
EOM

        print <<EOM if $editable;
<form action="$SCRIPT/BasketEdit" target="_self">
EOM

        wbSortFields();
        $PrintingBasket = 1;

        for ( $i = 0 ; $i <= $#BasketID ; $i++ ) {
            $id = $BasketID[$i];
            recRead($id);
            if ( $editable && $FORMAT ne 'TABLE' ) {
                $ThisRecordBasket = '<input type="checkbox" name="r' . $i . '" CHECKED>';
            }
            else {
                $ThisRecordBasket = '';
            }
            $ThisRecordNo = $i;
            $list[$i] = &formatFoundRec( $format, *PFields );    # Hidden ??
        }

        $head = &formatFoundHeader( $format, *PFields );         # Hidden ??
        formatFoundSeparators( $format, 1 );

        print '<div class="RECORDS">' . $t0 . $head if $html;
        $i = 0;
        foreach $_ (@list) {
            print "$t1$_$t2";
            $i++;
        }
        print $t3;
        print '</div>' if $html;

        $formats = &formatFormats();

        print <<EOM if $editable;
<br>
<input class="BUTTON" type="submit"  value="$modify">
</form>
</td></tr>
</table>

EOM
        $PrintingBasket = 0;
    }

    # edit

    $ic = 'edit';
    $tx = 'edit basket';                      #UK
    $a  = "$SCRIPT/BasketShow?editable=on";

    push( @Toolbar, <<EOM);
<a href="$a" target="_self"><img title="$tx" alt="$tx" $ICONPAR src="$ICONURL/$ic.gif"></a>
EOM

    # delete

    $ic = 'del';
    $tx = 'delete from database all records that are in the basket';    #UK
    $a  = "$SCRIPT/AdmManyDelForm?select=basket";

    push( @Toolbar, <<EOM) if &wbAllowed( $Group, 'AdmManyDelForm' );
<a href="$a" ><img title="$tx" alt="$tx" $ICONPAR src="$ICONURL/$ic.gif"></a>
EOM

    # many set form

    $ic = 'editnf';
    $tx = 'edit a field in these records';                              #UK
    $a  = "$SCRIPT/AdmManySetForm?select=basket";

    push( @Toolbar, <<EOM) if &wbAllowed( $Group, 'AdmManySetForm' );
<a href="$a" ><img title="$tx" alt="$tx" $ICONPAR src="$ICONURL/$ic.gif"></a>
EOM

    # search in basket

    $ic = 'search';
    $tx = 'search in basket';                                           #UK
    $a  = "$SCRIPT/Search?search=%7B_basket%7D";

    push( @Toolbar, <<EOM) if &wbAllowed( $Group, 'Search' );
<a href="$a"><img title="$tx" alt="$tx" $ICONPAR src="$ICONURL/$ic.gif"></a>
EOM

    # serial email to basket

    $ic = 'serialmail';
    $tx = 'serial mail to people in basket';                            #UK
    $a  = "$SCRIPT/AdmMailForm?search=%7B_basket%7D";

    push( @Toolbar, <<EOM) if &wbAllowed( $Group, 'AdmMailForm' );
<a href="$a"><img title="$tx" alt="$tx" $ICONPAR src="$ICONURL/$ic.gif"></a>
EOM

}

# edit content of basket

sub cgiBasketEdit {

    readBasket();

    for ( $i = 0 ; $i <= $#BasketID ; $i++ ) {
        unless ( $CGI{"r$i"} ne '' ) {
            $BasketNO[$i] = 0;
        }
    }

    writeBasket();
    printBasket();
}

# prints specified HTML file with current databases livery
# shows the file with standard WODA defined headers, footers,
# styles etc. Ignores any settings in the <head> except title.
# only shows files within $WBB{'htmlDir'}/_doc
# TODO: check the multiline matching
sub cgiHtml {

    $file = $CGI{'file'};
    $dd   = $HTMLDIR . "/_doc";

    if ( $file =~ m/\.\./ || $file !~ m/^[\w\d\.\-\_ ]*$/ ) {
        wbFail("It is not possible to show this file $dd\n");    #UK
    }

    open( h, "$dd/$file" ) || wbFail("$dd/$file could not be opened.\n");    #UK
    @rows = <h>;
    close(h);
    $text = join( '', @rows );
    ( $head, $body ) = split( m|</head\s*>|im, $text, 2 );                   # multiline matching

    if ( $body eq '' ) {
        $body  = $text;
        $title = $file;
    }
    else {

        # extract title
        if ( $head =~ m|<title>([^<]+)<|im ) {                               # multiline matching
            $title = $1;
        }
        else {
            $title = $file;
        }
    }

    $body =~ s|<body|<xxbody|im;                                             # multiline matching
    $body =~ s|</body|</xxbody|im;                                           # multiline matching

    # body could use relinking
    # how about gif files?

    printHead($title);
    print $body;
    return;
}

# moves records and data form old style directories to new style ones
# (from woda 3.x store format to woda 4.x store format)

sub cgiUpgrade24 {

    printHead("Maintenance: Updating database to WODA 4.x format");    #UK
    print <<EOM;
<p>Please wait.
EOM

    wbCheckDirectories();
    $moved = &appendRecFiles();

    print <<EOM;                                                       #UK
<br>Moved $moved files from $DATADIR to $DATADIR/_data/records.txt.
<br>Attachments are in $HTMLDIR/att.
<p>
<p>You can now go back and retry.
EOM
}

# recover backed up data

sub cgiAdmRestoreData {

    # problem spot ... fat and data may not have the same datestamp!

    $d = $CGI{'date'};

    if (   $d
        && -s "$DATADIR/_data/fat.txt.at.$d"
        && -s "$DATADIR/_data/records.txt.at.$d" )
    {
        printHead("Recovered data to time $d");

        unlink("$DATADIR/_data/fat.txt");
        unlink("$DATADIR/_data/records.txt");

        print "<p>Current data deleted.\n";

        rename( "$DATADIR/_data/fat.txt.at.$d", "$DATADIR/_data/fat.txt" )
          || wbFail( "Recovery failed", "You have no valid data now" );
        rename( "$DATADIR/_data/records.txt.at.$d", "$DATADIR/_data/records.txt" )
          || wbFail( "Recovery failed", "You have no valid data now" );

        print "<p>Data form date $d restored as working data; backup from date $d does not exist any more.\n";

    }
    else {

        printHead("Recover data $d");

        chdir("$DATADIR/_data");
        @files = &wbDir('records.txt.at.\d+');

        undef @x;

        foreach $file (@files) {
            next unless -s $file;
            $file =~ m/(\d+)$/;
            push( @x, $1 );
        }

        @x = sort { $b <=> $a } @x;

        if ( $#x < 0 ) {
            wbError( "Restore impossible", "No suitable backup files available" );
        }

        $opts = &wbFormatOptions( *x, *x );

        # get all the records.at.files!

        print <<EOM;
<p>This fill recover the data in this table to the selected backup file.
Prior to recovery attempt current data will be erased.
Proceed with care!
<form action="$SCRIPT/AdmRestoreData" method="get">
<p>Select backup file:
<select name="date">
$opts
</select>
<input class="BUTTON" type="submit"  value="Replace current with selection">
</form>
EOM
    }
}

# print file to screen

sub cgiAdmDirGet {

    $file = $CGI{'file'};

    if ( $file =~ m/\.\./ ) {
        wbFail( "Off limits", "You may not get file $file" );
    }

    open( h, "$DATADIR/$file" );
    @x = <h>;
    close(h);

    print <<EOM;
Content-type: text/plain
Content-Disposition: inline; filename="$file"

EOM
    print @x;
}

# print directory to screen

sub cgiAdmDir {

    $dir = $CGI{'dir'} || '';

    printHead("Raw files in ./$dir");

    if ( $dir =~ m/\.\./ || $dir =~ m/^\// ) {
        wbFail( "Off limits", "You are not allowed to view $dir" );
    }

    if ( $CGI{'action'} eq 'delete' ) {
        @f = split( /\n/, $CGI{'file'} );
        foreach $f (@f) {
            unlink "$DATADIR/$dir/$f";
        }
    }

    opendir( D, "$DATADIR/$dir" );
    @files = readdir(D);
    closedir(D);

    print <<EOM;
<form action="$SCRIPT/AdmDir">
<input type="hidden" name="dir" value="$dir">
<table cellspacing="0">
EOM

    @files = sort(@files);

    if ( $dir ne '' ) {
        $dir =~ m|/[^/]*$|;
        $up = $`;
        print <<EOM;
<tr>
<td>[DIR]</td>
<td><a href="$SCRIPT/AdmDir?dir=$up">up one level</a></td>
</tr>
EOM
    }

    foreach $f (@files) {
        next if $f =~ m/^\./;
        if ( -d "$DATADIR/$f" ) {
            print <<EOM;
<tr>
<td>[DIR]</td>
<td><a href="$SCRIPT/AdmDir?dir=$f">$f</a></td>
</tr>
EOM
        }
        else {
            $a = "<a href=\"$SCRIPT/AdmDirGet?file=$dir/$f\">$f</a>";
            print <<EOM;
<tr>
<td><input type="checkbox" name="file" value="$f"></td>
<td>$a</td>
</tr>
EOM
        }
    }

    print <<EOM;
</table>
<p>
<input class="BUTTON" type="submit"  name="action" value="delete"> checked files (dangerous!).
</p>
</form>
EOM
}

# make a copy of fat.txt and records.txt
# date in second is append to file's name

sub cgiAdmBackupData {

    printHead("Backup up current snapshot of the database");

    foreach $f ( "fat.txt", "records.txt" ) {
        $in  = "$DATADIR/_data/$f";
        $out = &wbBackupFile($in);
        open( in, $in );
        @x = <in>;
        close(in);
        open( out, ">$out" );
        print out @x;
        close(out);
        print "<br>Copied $in to $out\n";
    }
}

# appends .rec files in dataDir to records.txt

sub appendRecFiles {

    local ($dataDir) = @_;
    $dataDir = $dataDir || $DATADIR;

    chdir($dataDir);
    wbCheckDir('_rec');

    if ( !-d "_data" ) {
        wbFail("_data directory does not exist in $dataDir");
    }

    @files = &glob("*");
    foreach $file (@files) {
        next if -d $file;
        if ( $file =~ m|([^/]*)\.rec$| ) {
            $id = $1;
            recReadFile($file);
            $rec{'_id'} = $id;
            recWrite($dataDir);
            wbMoveFile( $file, "$dataDir/_rec/$file" );
            $rec++;
        }
        elsif ( $file =~ m/\.bak$/ || $file =~ m/^\./ ) {

            # nothing
        }
        else {
            $att++;
            wbMoveFile( $file, "$dataDir/att/$file" );
        }
    }
    chdir($DATADIR);
    return $rec;
}

# returns a valid, random ID for a .rec
# file in directory $DATADIR/_tmp

sub tmpID {

    local ($id);
    while (1) {
        $id = &wbRandKey();
        last unless -e "$TMPDIR/$id.rec";
    }
    return $id;
}

#############################################

# returns a time stamped file name for backups

sub wbBackupFile {

    local ($file) = @_;
    local ($add);
    @x   = &fixDate(time);
    $add = $x[5] . $x[4] . $x[3] . $x[2] . $x[1] . $x[0];
    return "$file.at.$add";
}

# rebuilds the fat.txt file, based on data.txt
# deleted records will reappear

sub rebuildFat {

    $file = "$DATADIR/_data/fat.txt";
    rename( $file, &wbBackupFile($file) );

    openFile( H, "$DATADIR/_data/records.txt" );
    openFile( F, ">$DATADIR/_data/fat.txt" );

    $pos = 0;
    while (<H>) {
        chop;
        ( $len, $time ) = split( /\t/, $_ );
        if ( $len !~ m/\d+/ || $time !~ m/\d+/ ) {
            wbFail( "records.txt is corrupt", "Try rebuilding the table from .tbl files. Contents of _data is corrupt." );
        }
        read( H, $string, $len )
          || wbFail( "Cannot read", "$l bytes from $file" );
        recFromVector($string);
        $id = $rec{'_id'};

        print F $rec{'_id'} . "\t" . $pos . "\n";

        $pos = tell(H);
    }
    close(H);
    close(F);

    return;
}

# openFile(handle,name,canFail)

sub openFile {
    open( $_[0], $_[1] ) || $_[2] || wbFail("Cannot open $_[1]");
    binmode( $_[0] );
}

# check integrity of fat.txt and records.txt

sub checkIntegrity {

    open( h, "$DATADIR/_data/fat.txt" )
      || return ( '?', '?', '?', "fat.txt does not exist" );
    $r = $n = $d = '0';
    while (<h>) {
        $r++;
        chop;
        ( $id, $pos ) = split( /\t/, $_ );
        if ( $pos{$id} ne '' ) {
            $n++;
        }
        $pos{$id} = $pos;
    }
    close(h);

    @keys = sort { $pos{"$a"} <=> $pos{"$b"} } keys(%pos);

    openFile( h, "$DATADIR/_data/records.txt" );

    foreach $id (@keys) {
        $pos = $pos{$id};

        if ( $pos < 0 ) {
            $d++;
            next;
        }

        seek( h, $pos, 0 ) || ( $ret .= "Cannot seek $pos for $id\n" );
        $meta = <h>;
        chop($meta);
        ( $len, $time ) = split( /\t/, $meta, 2 );
        if ( $len =~ /^\d+$/ && $time =~ /^\d+$/ ) {

            # ok
        }
        else {
            $ret .= "Bad fat entry for $id\n";
        }
    }
    close(h);

    return ( $n, $r, $d, $ret );
}

# defragments table

sub defragment {

    openFile( h, "$DATADIR/_data/fat.txt" ) || return "fat.txt does not exist";
    flock( h, 2 );
    $r = $n = 0;
    while (<h>) {
        $r++;
        chop;
        ( $id, $pos ) = split( /\t/, $_ );
        if ( $pos{$id} ne '' ) {
            $n++;
        }
        $pos{$id} = $pos;
    }
    close(h);

    $ret .= "$r all records, $n obsolete.\n";
    if ( $n == 0 ) {
        $ret .= "Nothing to do.\n";
        return $ret;
    }

    @keys = sort { $pos{"$a"} <=> $pos{"$b"} } keys(%pos);

    openFile( h, "$DATADIR/_data/records.txt" );
    flock( h, 2 );
    openFile( d, ">$DATADIR/_data/records.wrk" );
    openFile( f, ">$DATADIR/_data/fat.wrk" );

    foreach $id (@keys) {
        $pos = $pos{$id};

        next if $pos < 0;    # deleted record

        seek( h, $pos, 0 ) || wbFail("Cannot seek $pos for $id");
        $meta = <h>;
        chop($meta);
        ( $len, $time ) = split( /\t/, $meta );
        read( h, $string, $len ) || wbFail("Cannot read $len from $pos");

        $newpos = tell(d);
        print d "$meta\n";
        print d $string;
        print f "$id\t$newpos\n";
    }

    close(d);
    close(f);
    close(h);

    wbBackupData();

    rename( "$DATADIR/_data/fat.wrk", "$DATADIR/_data/fat.txt" )
      || &wbFail("Cannot rename fat.wrk");
    rename( "$DATADIR/_data/records.wrk", "$DATADIR/_data/records.txt" )
      || &wbFail("Cannot rename records.wrk");

    $ret .= "Defragmented!\n";
    return $ret;
}

# backs up fat.txt and data.txt

sub wbBackupData {
    local ($file);

    $file = "$DATADIR/_data/fat.txt";
    if ( -e $file ) {
        rename( $file, &wbBackupFile($file) ) || &wbFail("Cannot rename $file");
    }
    $file = "$DATADIR/_data/records.txt";
    if ( -e $file ) {
        rename( $file, &wbBackupFile($file) ) || &wbFail("Cannot rename $file");
    }
}

#############################################
# conversions to and from %rec
# handling records in one per file format
#
# recReadFile
# recWriteFile
# recToVector
# recToRow
# recFromVector
# recFromRow
# recMoveAttached
#
#############################################

# fill %rec array with values from file

sub recReadFile {
    local ($inFile) = @_;
    local (@rows);

    open( h, "$inFile" ) || return "Cannot open $inFile";
    @rows = <h>;
    close(h);
    recFromVector( join( '', @rows ) );
}

# writes entire _rec to file.rec provided as parameter

sub recWriteFile {

    local ($file) = @_;

    open( h, ">$file" ) || return 0;
    binmode(h);
    print h &recToVector();
    close(h);
    return 1;
}

# converts a record to a string
# does it in a binary format
# converts newlines to \034 =
#
# was wbRecPrint

sub recToVector {

    local ( $out, $x, $v );

    foreach $x ( sort( keys(%rec) ) ) {
        next if $x eq '';    # skip values with blank keys!
        $v = $rec{$x};
        $v =~ tr/\012/\034/;
        $out .= "$x\n$v\n";
    }

    return $out;
}

# converts a string, created by recToVector
# into a %rec array

sub recFromVector {

    local (@x);
    undef %rec;

    #$Log.=&htEscape($_[0]);

    $_[0] =~ s/\x0D//g;    # get rid of any CRs

    if ( $_[0] !~ m/^\%/ ) {

        # --- binary format

        #		chop($_[0]);		# split ("a\nb\n") should be 3 elements!
        #					# on some perls it is not
        #		odd number of elements bug

        @x = split( /\n/, $_[0] );
        shift(@x) while $x[0] eq '';    # remove all blank rows up front

        foreach $x (@x) {
            $x =~ tr/\034/\012/;        # escaped newlines replaced in-place
        }

        unless ( $#x % 2 ) {            # not a pair
            pop(@x);
        }

        %rec = @x;

    }
    else {

        # --- readable format

        @x = split( /\n\%/, $_[0] );
        foreach $_ (@x) {
            next if m/^\%/;    # comment %%
            ( $tag, $value ) = split( /\:/, $_, 2 );
            $value =~ s/\n\s+/\n/g;    # delete space after newline
            $value =~ s/^\s*//;        # delete leading space
            $value =~ s/\s*$//;        # delete trailing space
            $rec{$tag} = $value;
        }
    }
}

# converts %rec to a row of fields passed as parameter
# does not include newline

sub recToRow {

    local ( $out, $x );

    foreach $x (@_) {
        $v = $rec{$x};
        $v =~ tr/\t/\033/;
        $v =~ tr/\012/\034/;
        $out .= $v . "\t";
    }

    chop $out;
    return $out;
}

# converts a row of tab separated fields to %rec
# fields and row are a parameter
# was: wbTab2Rec

sub recFromRow {
    local ( $row, *fields ) = @_;
    local ( $field, $i, $x, @x );

    undef %rec;

    $x = chop($row);    # newline ?
    if ( $x ne "\n" ) {
        $row .= $x;
    }

    @x = split( /\t/, "$row\t." );    # because split(/\t/,"\t\t\t") does not return array of 4 empty fields
    pop(@x);
    if ( $#x != $#fields ) {
        $x = '<table border="1"><tr><th>field</th><th>20 chars of value</th></tr>';
        $i = 0;
        while (1) {
            local ( $n, $v );
            $n = $fields[$i];
            $v = substr( $x[$i], 0, 20 );
            $x .= "<tr><tr><td>$n\n</td><td>$v\n</td></tr>\n";
            last if $n eq '' && $v eq '';
            $i++;
        }
        $x .= '</table>';
        wbFail( "TBL file inconsistent", "$#x data items for $#fields fields!\n<hr> $x" );
    }

    $i = 0;
    foreach $field (@fields) {
        $x[$i] =~ s/\033/\t/g;
        $x[$i] =~ s/\034/\012/g;
        $rec{$field} = $x[$i];
        $i++;
    }
}

# %rec includes pointers to attached files
# if a pointer is empty, any such file is deleted
# if a pointer is non-empty, any such file is properly stored as
# id.field.random.ext and file id.field.* is deleted

sub recMoveAttached {

    local ( $tmpID, $newID ) = @_;
    local ( $destDir, $srcDir, $srcFile, $destFile, $d, @toDelete );

    $destDir = &wbCheckDir("$HTMLDIR/att");
    $srcDir  = "$DATADIR/_tmp";

    wbSortFields();
    foreach $field (@Fields) {
        next
          unless ( $WBF{ $field, 'type' } eq "USERFILE"
            || $WBF{ $field, 'type' } eq "IMAGE" );
        $srcFile = $rec{$field};

        if ( $srcFile eq '' ) {

            # srcfile is blank ... we need to delete it

            @toDelete = &glob("$destDir/$newID.$field.*");
            foreach $d (@toDelete) {
                unlink "$destDir/$d";
            }

        }
        elsif ( -e "$srcDir/$srcFile" ) {

            # there is a new file just uploaded

            $srcFile =~ m/([^\.]*)$/;    # extract .extension
            $ext = $1;
            $x   = rand();
            $x =~ s/\D//g;
            $x = substr( $x, 0, 5 );
            $destFile = "$newID.$field.$x.$ext";

            # delete old version

            @toDelete = &glob("$destDir/$newID.$field.*");
            foreach $d (@toDelete) {
                unlink "$destDir/$d";
            }

            # move from tmp to att

            wbMoveFile( "$srcDir/$srcFile", "$destDir/$destFile" );
            $rec{$field} = $destFile;

        }
        elsif ( $srcFile !~ m/^$newID\./ ) {

            # the new id is different than the ID stored in the %rec

            $srcFile =~ m/\.$field\./;
            $destFile = "$newID.$field.$'";
            wbMoveFile( "$destDir/$srcFile", "$destDir/$destFile" );
            $rec{$field} = $destFile;

        }

    }    # each field

}

#############################################
# core database functions
#
#	recRead
#	recWrite
#	recExists
#	recDelete
#	recRandID
#	recAll
#
#	recOpen
#	recClose
#	recCloseAll
#
#############################################
#
#	Files:
#		_data/records.txt contains all records
#		separated with \033\n
#		_data/fat.txt
#
#
# repeated calls to Read/Write/Delete
# keep open at all times?
#
# open/close ... fast FLD ???
#############################################

# return handle names

sub recHandles {

    local ($x) = @_;
    return ( "HD$x", "HF$x" );
}

# touch data

sub recTouch {

    open( h, ">>$DATADIR/_data/fat.txt" );
    close(h);
}

# return handles to data or opens Fat & Records

sub recOpen {

    local ($table) = @_;
    local ($dir);

    local ( $HData, $HFat );

    if ( !$Opened{$table} ) {

        ( $HData, $HFat ) = &recHandles($table);
        $dir = &wbTables( 'DIR', $table );

        open( $HFat, "$dir/_data/fat.txt" ) || return;
        binmode($HFat);
        open( $HData, "$dir/_data/records.txt" ) || return;
        binmode($HData);

        $Opened{$table} = 1;
        $Log .= "Opened $dir\n";

        return ( $HData, $HFat );

    }
    else {

        return &recHandles($table);

    }
}

# close Fat & Records

sub recClose {
    local ($table) = @_;

    if ( $Opened{$table} ) {
        ( $HData, $HFat ) = &recHandles($table);
        close($HData);    # || wbFail("not closed1");
        close($HFat);     # || wbFail("not closed2");
        $Log .= "Closed $table\n";
    }

    $Opened{$table} = 0;
}

# close all tables

sub recCloseAll {

    $Log .= "Close all\n";
    foreach $table ( keys(%Opened) ) {
        recClose($table);
    }
}

# reads a record with given id into %rec
# also sets $rec{'_id'}

sub recRead {
    local ( $id, $table ) = @_;
    local ( $err, $handle, $p, $meta, $l, $t );
    undef %rec;

    ( $HData, $HFat ) = &recOpen($table);
    $p = &recExists( $id, $table );
    if ( $p >= 0 ) {

        # same as in recReadAt
        seek( $HData, $p, 0 ) || wbFail("Cannot seek $p in $HData.");
        $meta = <$HData>;
        chop($meta);
        ( $l, $time ) = split( /\t/, $meta );
        read( $HData, $string, $l )
          || wbFail("Cannot read $l bytes at $p in $Hdata.");
        recFromVector($string);
        if (%rec) {
            $rec{'_id'} = $id;
            return '';
        }
    }
    else {
        return "No such record";
    }
}

# same as above in recRead
# reads a record into %rec in recRead

sub recReadAt {

    local ($p) = @_;

    seek( $HData, $p, 0 ) || wbFail("Cannot seek $p in $HData.");
    $meta = <$HData>;
    chop($meta);
    ( $l, $time ) = split( /\t/, $meta );
    read( $HData, $string, $l )
      || wbFail("Cannot read $l bytes at $p in $Hdata.");
    recFromVector($string);
}

# writes a record

sub recWrite {

    local ($table) = @_;
    local ($dir)   = &wbTables( 'DIR', $table );
    local ($id)    = $rec{'_id'};
    local ( $vector, $start, $len );

    if ( $id eq '' ) {
        wbFail( "Error: Blank record ID", "It should not be blank!!!" );
    }

    open( h, ">>$dir/_data/records.txt" );
    flock( h, 2 );
    binmode(h);
    seek( h, 0, 2 );
    $start  = tell(h);
    $vector = &recToVector();
    $len    = length($vector);
    $t      = time;
    print h "$len\t$t\n";
    print h $vector;
    close(h);

    wbCheckIdPos( $id, $start );

    open( h, ">>$dir/_data/fat.txt" );
    flock( h, 2 );
    binmode(h);
    print h "$id\t$start\n";
    close(h);
}

# does a record exist? where? at what offset ... test must be to empty string!

sub recExists {

    local ( $id, $table ) = @_;
    local ( $idd, @x );

    $idd = -1;

    if ( defined $IDS && $IDS eq $table ) {
        $idd = $IDS{$id};
    }
    else {
        ( $HData, $HFat ) = &recOpen($table);
        seek( $HFat, 0, 0 );
        while (<$HFat>) {
            chop;
            if (m/^$id\t/) {
                $idd = $';
            }
            else {

                # $Log .= "$_";
            }
        }
    }

    return $idd;

}

# deletes a record
# returns error string
# does not delete attachments

sub recDelete {

    local ( $id, $keepAttach ) = @_;

    if ( $id eq '' || $id =~ m/^[\*\?\.]/ || $id =~ m/\t|\n/ ) {
        wbFail( 'Bad id', "Cannot delete record with id='$id'" );
    }

    wbCheckIdPos( $id, '-1' );

    open( h, ">>$DATADIR/_data/fat.txt" );
    flock( h, 2 );
    binmode(h);
    print h "$id\t-1\n";
    close(h);

    if ($keepAttach) {

        # nothing
    }
    else {
        chdir("$HTMLDIR/att") || wbFail("Cannot chdir to $dir");
        local ( @all, $x );
        @all = &glob("$id.*");
        foreach $x (@all) {
            unlink "$x" || wbFail("Could not delete $x");
        }
    }
}

# returns next free ID in this table

sub NEXT {

    local ( $first, $step ) = @_;

    $first = 1 if $first eq '';
    $step  = 1 if $step  eq '';

    local ( $last, $file );

    $file = "$DATADIR/_cache/lastid.txt";
    if ( -e $file ) {

        # nothing
    }
    else {
        open( h, ">$file" );
        print h ( $first - $step );
        close(h);
    }

    open( h, "+<$file" );
    $last = <h>;
    $last = $last + $step;

    seek( h, 0, 0 );
    print h $last . "\n" || wbFail("Cannot get new record ID");
    close(h);

    return $last;
}

# returns a valid ID for a record
# NEXT available number or a random number!

sub recRandID {

    local ( $id, $x );

    $x = $WBB{'key'};
    if ( $x =~ m/\&NEXT/ || $x =~ m/NEXT\(/ ) {
        return &evalSafe($x);
    }
    else {
        while (1) {
            $id = &wbRandKey();
            last unless &recExists($id) >= 0;
        }
        return $id;
    }
}

# prepares data structures for recNext

sub recAll {

    local ($table) = @_;

    undef %IDS;
    ( $HData, $HFat ) = &recOpen($table);
    seek( $HFat, 0, 0 );
    while (<$HFat>) {
        ( $id, $rest ) = split( /\t/, $_, 2 );
        if ( $rest == -1 ) {
            delete $IDS{$id};
        }
        else {
            $IDS{$id} = $rest;
        }
    }

    $IDS = $table;

    return wantarray ? keys(%IDS) : '';
}

# return id of the next record

sub recNext {

    local ( $n, $v ) = each(%IDS);
    return $n;
}

# returns % of database sizes

sub ddQuota {

    local ( $text, $bin ) = ( $WBB{'quotaText'}, $WBB{'quotaBin'} );
    local ( $ptext, $pbin );

    if ( $text == 0 ) {
        $ptext = 0;
    }
    else {
        $ptext = ( -s "$DATADIR/_cache/default.tbl" ) / ( 1024 * 1024 ) / $text;
    }

    if ( $bin == 0 ) {
        $pbin = 0;
    }
    else {
        local (@files);
        local ( $size, $file );
        $size  = 0;
        @files = &glob("$HTMLDIR/att/*");
        foreach $file (@files) {
            $size += -s "$HTMLDIR/att/$file";
        }
        $pbin = $size / ( 1024 * 1024 ) / $bin;
    }

    return ( $ptext, $pbin );
}

# if string is perl it is evaled
# string is assumed perl if
# it starts with a ',
# it starts with a "
# or it starts with a perl comment (hash)
# or includes semicolon-newline combination
# if string is string it is printed

sub evalMaybe {

    local ($string) = @_;

    if ( $string =~ m/^\s*qq\(|^\s*"|^\s*'|^#|;\n/ ) {
        return &evalSafe($string);
    }
    else {
        return $string;
    }
}

# do a safe eval
# allow unsafe eval sometimes!

sub evalSafe {

    local ($ev) = @_;
    local ( $x, $e );

    return '' if $ev eq '';

    if ( $x = &evalTrouble($ev) ) {
        $x  = &htQuote($x);
        $ev = &htQuote($ev);
        $e  = qq(<em class="ERROR"><a href=\"$SCRIPT/AdmHelpSecurity\">Security risk:</a>'$x' in <pre>$ev</pre></em>\n);
        $Log .= "EVAL ERROR: UNSAFE\n$x\n";
        $@ = $x;
    }
    else {
        $e = eval $ev;
        $e = qq(<em class="ERROR">Eval error:$@</em>\n) if $@;
    }

    return $e;
}

# returns '' if ok

sub evalTrouble {

    local ($_) = $_[0];

    return '';
}

# print a warning about perl code

sub cgiAdmHelpSecurity {

    printHead("Secure Perl Code");

    print <<EOM;
Only a subset of Perl is allowed in expressions. The offending string is shown in error report.
Some tips:
<ul>
<li>Comments with # are not allowed at all
</li><li>&lt;&lt;EOM strings are not allowed.
</li><li>matching and substituting only works as m/.../ or s/.../
</li><li>avoid single quotes
</li><li>do not use () in qq( strings
</li></ul>
EOM
}

# can the admin modify this setting

sub ddAdminCanModify {

    local ( $type, $name ) = @_;

    if ( $type eq 'B' ) {
        return 1 if $WBB{ $name, 'config' } eq 'yes';
        return 0 if $WBB{ $name, 'config' } eq 'no';
        return $name !~ m/$HIDDENWBB/i if $HIDDENWBB ne '';
    }
    else {
        return 1 if $WBF{ $name, 'config' } eq 'yes';
        return 0 if $WBF{ $name, 'config' } eq 'no';
        return $name !~ m/$HIDDENWBF/i if $HIDDENWBF ne '';
    }

    return 1;
}

# validates a table setting
# s ... string
# r ... regular expression
# l ... logical

sub admMetaValidate {

    local ( $value, $type ) = @_;

    return ( '', $value ) if $value eq '' || $type eq '' || $type eq ' ';

    if ( $type eq 'r' ) {

        # m/$something/ if $something='/';
    }
    elsif ( $type eq 'l' || $type eq 's' ) {
        $x = &evalTrouble($value);
        if ( $x ne '' ) {
            return ( "Eval contains malign code: $x", $value );
        }
        else {
            eval $value;
            if ($@) {
                return ( "Eval contains an error: $@", $value );
            }
            else {

                # OK
            }
        }
    }
    else {

        # anything goes
    }

    return ( '', $value );
}

################
################
################
################
################
################

# prints new cgi definition

sub cgiAdmCreateCGI {

    if ( $CGI{'filename'} ne '' ) {
        printHead("Printing new definition to file $CGI{'filename'}");
    }
    else {
        print "Content-type: text/plain\n\n";
        $HttpHeadPrinted = 1;
    }

    ddModify();
}

# change fields in a cgi definition

sub cgiAdmModifyFieldsForm {

    printHead("Add, delete, reorder fields of a table");

    wbSortFields();
    $fields = join( "\n", @Fields );
    $rows = $#Fields + 3;

    print <<EOM;
<form class="FORM" action="$SCRIPT/AdmModifyFields">
<table cellspacing="0">
<tr valign="top">
<td align="right" width="40%">
The box on the right lists all the fields defined in this
table, one per row. To add, delete, reorder etc. the fields,
edit the list.
<p>
You cannot rename the fields! 
</td>
<td>
<textarea name="fields" rows=$rows>$fields</textarea>
</td>
</tr>
</table>
EOM

    admChangeDefinitionButton();

    print "</form>\n";

}

# change fields in cgi definition

sub cgiAdmModifyFields {

    wbSortFields();
    $fields = join( "\n", @Fields );
    $newFields = $CGI{'fields'};
    $newFields =~ s/\r//g;
    $newFields =~ s/\n\n/\n/;
    $newFields =~ s/\n$//g;

    if ( $fields eq $newFields ) {
        printHead("The fields are the same");
        return;
    }
    else {
        $Fch = 1;
    }

    if ( defined $CGI{'filename'} ) {
        print "Content-type: text/plain\n\n";
    }
    else {
        printHead("Fields have changed");
    }

    @fields = split( /\n/, $newFields );

    foreach $field (@Fields) {
        $WBF{ $field, 'srt' } = 0;
    }

    $i = 1000;
    foreach $field (@fields) {
        $WBF{ $field, 'srt' } = $i--;
        if ( $WBF{$field} eq '' ) {
            print "<br>New field <b>$field</b>\n"
              unless defined $CGI{'filename'};
            $Fch = 1;
            $WBF{$field} = '1;';
        }
    }

    foreach $field (@Fields) {
        if ( $WBF{ $field, 'srt' } == 0 ) {
            delete $WBF{$field};
            @x = grep( /^$field$;/, keys(%WBF) );
            foreach $x (@x) {
                delete $WBF{$x};
            }
            print "<br>Deleted field <b>$field</b>\n"
              unless defined $CGI{'filename'};
            $Fch = 1;
        }
    }

    wbSortFields();    # the new version
    ddModify();
}

# validates database settings
# in CGI variables
# and modifies them
# returns 1 if they can be written
# returns 0 of not

sub ddModify {

    local ($setting);

    foreach $x ( keys(%CGI) ) {
        $CGI{$x} =~ s/\r//g;
    }

    # save current values

    admMeta();

    $CGI{'language'} =~ tr/[A-Z]/[a-z]/;

    $errors = 0;

    foreach $setting (@MetaB) {
        next if $MetaB{ $setting, 'what' } eq 'H';
        if ( defined $CGI{$setting} ) {

            $new = $CGI{$setting};
            $old = $WBB{$setting};

            if ( $new ne $old ) {

                if ( &ddAdminCanModify( 'B', $setting ) ) {

                    #OK
                }
                else {
                    wbError("You are not allowed to modify setting '$setting'");
                }

                ( $error, $new ) = &admMetaValidate( $new, $MetaB{ $setting, 'type' } );
                if ( $error ne '' ) {
                    $errors++;
                    print q(<p><em class="ERROR">$setting:</em><br>$error</p>\n);
                    next;
                }

                if ( $CGI{$setting} eq '' ) {
                    $changed++;
                    print "<br>Deleted table setting $setting\n"
                      if $WBB{$setting} && $HttpHeadPrinted;
                    undef $WBB{$setting};
                }
                else {
                    $changed++;
                    print "<br>Changed table setting $setting\n"
                      if $HttpHeadPrinted;
                    $WBB{$setting} = $new;
                }
            }
        }
    }

    # fields

    $i = 1;
    while ( $CGI{"_f$i"} ) {

        $fName = $CGI{"_f$i"};

        foreach $setting (@MetaF) {

            if ($setting) {
                $key = "$fName$;$setting";
            }
            else {
                $key = "$fName";
            }

            $new = $CGI{"$setting$i"};
            $old = $WBF{$key};

            if ( defined $CGI{"$setting$i"} && $new ne $old ) {

                if ( &ddAdminCanModify( 'F', $setting ) ) {

                    #OK
                }
                else {
                    wbError("You are not allowed to modify setting '$setting'");
                }

                print "<hr><br>CGI $fName: " . &htEscape($new);
                print "<br>WBF $fName: " . &htEscape($old);

                ( $error, $new ) = &admMetaValidate( $new, $MetaF{ $setting, 'type' } );
                if ( $error ne '' ) {
                    $errors++;
                    print qq(<p class="ERROR">$setting: $error</p>\n);
                    next;
                }

                if ( $new eq '' ) {
                    $changed++;
                    print "<br>Deleted setting $fName,$setting\n";
                    undef $WBF{ $fName, $setting };
                }
                else {
                    $changed++;
                    print "<br>Changed setting {$fName,$setting} to '$new'\n"
                      if $HttpHeadPrinted;
                    $WBF{$key} = $new;
                }
            }
        }
        $i++;
    }

    if ($errors) {
        wbError( "$errors errors", "Please go back and correct them" );
    }
    elsif ( $changed <= 0 && !$Fch ) {
        wbWarn( "WARNING:", "Nothing has changed" );
    }
    else {
        if ( defined $CGI{'filename'} ) {
            admPrintScript();
        }
        else {
        }
    }
}

# --- add some information to the debug file ---

sub logDebug {

    local ($x);

    open( h, ">>$DATADIR/_log/debug.log" );
    binmode(h);

    print h "--- request ------------------------------------------\n";
    foreach $x ( sort( keys(%ENV) ) ) {
        print h "$x=$ENV{$x}\n";
    }
    print h $_[0];

    close(h);

}

# check if the swish search engine is installed
# return true or false if $SWISH

sub canFullTextSearch {

    return ( -x $SWISH && -s "$DATADIR/_swish/index" );

}

# Search in full text (experimental)

sub cgiFullTextSearch {

    printHead("Search in full text (experimental)");    #UK

    wbFail( "Not available", $SWISH ) unless &canFullTextSearch();    #UK

    $search = $CGI{'search'};

    $x = "Search full text for:";                                     #UK
    $y = "Search";                                                    #UK
    $m = "Search in metadata";                                        #UK
    $s = &htQuote($search);
    $h = <<EOM;                                                       #UK
<p><b>Help:</b>
<ul>
<li> foo bar ... finds documents with foo AND bar
</li><li> foo or bar ... finds documents with foo OR bar
</li><li> foo not bar ... finds documents with foo but NOT with bar
</li></ul>
EOM

    print <<EOM;
<form action="$SCRIPT/FullTextSearch" name="searchForm">
<table class="SEARCHBOX" cellspacing="0" width="100%">
<tr><td>
$x
<input name="search" value="$s">
<input class="BUTTON" type="submit"  value="$y">
</td></tr>
</table>
</form>
$h
<a href="$SCRIPT/Search?search=$s">$m</a>
EOM

    return unless $search;

    if ( $search =~ m/[<>{}|;:]/ ) {
        wbError("Bad character in search expression");
        return;
    }

    $search =~ s/\W/\\$&/g;

    $x = qq($SWISH -f $DATADIR/_swish/index -H 0 -x "%r %p\\n" -w $search);
    $Log .= "Exec $x\n";

    @results = `$x`;

    $Log .= join( '', @results );

    foreach $result (@results) {
        chop($result);
        ( $rank, $file ) = split( / /, $result );
        $file =~ m|([^/]*)\.html$|;
        $file = $1;
        $file =~ m/^[^\.]*/;
        push( @ids, $& ) unless $have{$&};
        $have{$&} = 1;
    }

    printRecords( 'DEFAULT', *ids, $#ids );
}

# full text indexing of .pdf files in att

sub cgiAdmIndexPDF {

    printHead("Converting records to indexable html");

    print "<p>This will take some time!\n";

    unless ( -x $PDFTOHTML ) {
        wbFail( "Converter missing", "pdftohtml program not installed at '$PDFTOHTML'" );
        return;
    }

    print "<pre>\n";

    opendir( D, "$DATADIR/att" );
    @atts = readdir(D);
    closedir(D);

    @pdfs = grep( /pdf$/i, @atts );

    mkdir( "$DATADIR/_full",  0777 );
    mkdir( "$DATADIR/_swish", 0777 );

    foreach $pdf (@pdfs) {
        $in  = $pdf;
        $out = $in;
        $out =~ s/\.pdf$//i;

        $in  = "$DATADIR/att/$in";
        $out = "$DATADIR/_full/$out.html";

        print "$pdf ...";
        if ( &wbMtime($in) > &wbMtime($out) ) {
            print "converting\n";
            system "$PDFTOHTML -noframes -i $in $out 2>&1";
        }
        else {
            print "old.";
        }
        print "\n";
    }

    # export records to html

    recAll();
    while ( $id = &recNext() ) {
        $hp = "$DATADIR/_full/$id.html";
        print "metadata for $id.html";
        if (1) {
            recRead($id);
            $then = $rec{_at};
            $now  = ( stat($hp) )[9];
            if ( $then < $now ) {
                print "is old\n";
                next;
            }
            print "IS NEW!!!\n";
            open( cache, ">$hp" ) || wbFail("Cannot open file $hp");
            binmode(cache);
            $selected     = select(cache);
            $HTMLFILE     = 1;
            $CGI{_id}     = $id;
            $group        = $Group;
            $Group        = 'guest';
            $IGNOREERRORS = 1;
            wbDo("cgiShow");
            undef @Toolbar;
            delete $CGI{_id};
            $Group    = $group;
            $HTMLFILE = 0;
            close(cache);
            select($selected);
        }
        else {
            recRead($id);
            open( h, ">$hp" );
            binmode(h);
            $HTMLFILE     = 1;
            $IGNOREERRORS = 1;
            print h &wbRecHtml();
            close(h);
        }
    }

    print "<HR><B>Indexing ...<BR>\n";
    print `$SWISH -i $DATADIR/_full -f $DATADIR/_swish/index`;
    print "</pre>Done!";
}

# support for the Office 2003 research function

sub cgiOffice {

    logDebug( $CGI{'XML'} );

    if ( $ENV{'HTTP_SOAPACTION'} eq "urn:Microsoft.Search/Registration" ) {

        officeRegistration();

    }
    elsif ( $ENV{'HTTP_SOAPACTION'} eq "urn:Microsoft.Search/Query" ) {

        officeQuery();

    }
    else {

        printHead("Office Research Task Pane");    #UK

        print <<EOM;                               #UK
<p>
This application also acts as a Web Service feeding data into the 
<a href="http://office.microsoft.com/en-us/assistance/HP010002811033.aspx">Microsoft Office research task pane</A>.

<p>
To insert it into the Office environment, please follow
<a href="http://office.microsoft.com/en-us/assistance/HP010377881033.aspx">these instructions</A>.

<p>
For more information about Research Services in Microsoft Office browse
<a href="http://office.microsoft.com/en-us/assistance/CH010716751033.aspx">this list of articles</A>.

<p>
Here is the <a href="$SCRIPT/OfficeWSDL">WSDL</A> of this service.
EOM

    }
}

# support for the MS Office 2003 research function
# print the MS Office WSDL

sub cgiOfficeWSDL {

    $string = <<EOM;
<?xml version="1.0" encoding="utf-8"?>
<definitions xmlns:http="http://schemas.xmlsoap.org/wsdl/http/" xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/" xmlns:s="http://www.w3.org/2001/XMLSchema" xmlns:s0="urn:Microsoft.Search" xmlns:soapenc="http://schemas.xmlsoap.org/soap/encoding/" xmlns:tm="http://microsoft.com/wsdl/mime/textMatching/" xmlns:mime="http://schemas.xmlsoap.org/wsdl/mime/" targetNamespace="urn:Microsoft.Search" xmlns="http://schemas.xmlsoap.org/wsdl/">
  <types>
	<s:schema elementFormDefault="qualified" targetNamespace="urn:Microsoft.Search">
	  <s:element name="Registration">
		<s:complexType>
		  <s:sequence>
			<s:element minOccurs="0" maxOccurs="1" name="registrationXml" type="s:string" />
		  </s:sequence>
		</s:complexType>
	  </s:element>
	  <s:element name="RegistrationResponse">
		<s:complexType>
		  <s:sequence>
			<s:element minOccurs="0" maxOccurs="1" name="RegistrationResult" type="s:string" />
		  </s:sequence>
		</s:complexType>
	  </s:element>
	  <s:element name="Query">
		<s:complexType>
		  <s:sequence>
			<s:element minOccurs="0" maxOccurs="1" name="queryXml" type="s:string" />
		  </s:sequence>
		</s:complexType>
	  </s:element>
	  <s:element name="QueryResponse">
		<s:complexType>
		  <s:sequence>
			<s:element minOccurs="0" maxOccurs="1" name="QueryResult" type="s:string" />
		  </s:sequence>
		</s:complexType>
	  </s:element>
	</s:schema>
  </types>
  <message name="RegistrationSoapIn">
	<part name="parameters" element="s0:Registration" />
  </message>
  <message name="RegistrationSoapOut">
	<part name="parameters" element="s0:RegistrationResponse" />
  </message>
  <message name="QuerySoapIn">
	<part name="parameters" element="s0:Query" />
  </message>
  <message name="QuerySoapOut">
	<part name="parameters" element="s0:QueryResponse" />
  </message>
  <portType name="ResearchSoap">
	<operation name="Registration">
	  <documentation>Registration for the Woda Search Service.</documentation>
	  <input message="s0:RegistrationSoapIn" />
	  <output message="s0:RegistrationSoapOut" />
	</operation>
	<operation name="Query">
	  <documentation>Runs a search in Woda.</documentation>
	  <input message="s0:QuerySoapIn" />
	  <output message="s0:QuerySoapOut" />
	</operation>
  </portType>
  <binding name="ResearchSoap" type="s0:ResearchSoap">
	<soap:binding transport="http://schemas.xmlsoap.org/soap/http" style="document" />
	<operation name="Registration">
	  <soap:operation soapAction="urn:Microsoft.Search/Registration" style="document" />
	  <input>
		<soap:body use="literal" />
	  </input>
	  <output>
		<soap:body use="literal" />
	  </output>
	</operation>
	<operation name="Query">
	  <soap:operation soapAction="urn:Microsoft.Search/Query" style="document" />
	  <input>
		<soap:body use="literal" />
	  </input>
	  <output>
		<soap:body use="literal" />
	  </output>
	</operation>
  </binding>
  <service name="Research">
	<documentation>Woda Research Web Service</documentation>
	<port name="ResearchSoap" binding="s0:ResearchSoap">
	  <soap:address location="$SCRIPTBASE$SCRIPT/Office" />
	</port>
  </service>
</definitions>
EOM

    $len = length($string);
    binmode(STDOUT);
    print "Content-type: text/xml; charset=utf-8\nContent-Length: $len\n\n$string";

}

# support for the Office 2003 research function

sub officeRegistration {

    $description = &htUntag( $WBB{'about'} );
    $name        = &htUntag( $WBB{'apTitle'} . $WBB{'dbTitle'} );
    $name =~ s/\s//g;

    $provider = $WBB{'provider'} || $ENV{'SERVER_NAME'};
    $license = $WBB{'license'}
      || 'This is a free service. You do not have to pay anything, anytime.';    #UK
    $category = $WBB{'officeCategory'} || 'RESEARCH_GENERAL';

    $string = <<EOM;
<ProviderUpdate xmlns='urn:Microsoft.Search.Registration.Response'>
<Status>SUCCESS</Status>
<Providers revision='2'>
<Provider action='UPDATE'>
<Message>
This is a WODA based service for Office 2003, provided by $provider. 
</Message>
<License acceptRequired='true'><LicenseText>$license</LicenseText></License>
<Id>{C818DC43-F71C-11d6-9039-00B0D019A5D1}</Id>
<Name>$provider</Name>
<QueryPath>$SCRIPTBASE$SCRIPT/Office</QueryPath>
<RegistrationPath>$SCRIPTBASE$SCRIPT/Office</RegistrationPath>
<Type>SOAP</Type>
<AboutPath>$SCRIPTBASE$SCRIPT/About</aboutPath>
<Revision>2</Revision>
<Services>
<Service>
<Id>{C818DC43-F71C-11d6-9039-00B0D019A5D1}</Id>
<Name>$name</Name>
<Description>$description</Description>
<AboutPath>$SCRIPTBASE$SCRIPT/About</aboutPath>
<Copyright>Copyright (c) $provider.  All Rights Reserved.</Copyright>
<TermsOfUsePath>$SCRIPTBASE$SCRIPT/About</TermsOfUsePath>
<Display>On</Display>
<Category>$category</Category>
<Parental>Unsupported</Parental>
</Service>
</Services>
</Provider>
</Providers>
</ProviderUpdate>
EOM

    $string = &htQuote($string);

    $string = <<EOM;
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
	<RegistrationResponse xmlns="urn:Microsoft.Search">
	  <RegistrationResult>$string</RegistrationResult>
	</RegistrationResponse>
  </soap:Body>
</soap:Envelope>	
EOM

    $len = length($string);
    binmode(STDOUT);
    print "Content-type: text/xml; charset=utf-8\nContent-Length: $len\n\n$string";

}

# support for the Office 2003 research function

sub officeQuery {

    $CGI{'xml'} =~ m|([^;]*)\&lt\;/QueryText|;
    $search = $1;

    local (@list);

    ( $allHits, *list, *fields ) = &recSelect( '', $search, 'DEFAULT', 'DEFAULT' );
    $dataHits = $#list + 1;

    $string = <<EOM;
<ResponsePacket revision="1" xmlns="urn:Microsoft.Search.Response">
<Response domain="{C818DC43-F71C-11D6-9039-00B0D019A5D1}">
	<Range>
		<StartAt>1</StartAt>
		<Count>$dataHits</Count>
		<TotalAvailable>$allHits</TotalAvailable>
		<Results>
		<Content xmlns="urn:Microsoft.Search.Response.Content">
		<Image source="$SCRIPTBASE$ICONURL/poweredby.gif"/>
EOM

    foreach $hit (@list) {

        $hit =~ s/\n/ /g;
        $hit =~ m|_id\=([^\&]+)|i;
        $url = "$SCRIPTBASE$SCRIPT/Show?_id=$1";

        $hit = &htUntag($hit);

        $hit =~ s/^[\s\d\.]+//g;

        $hit =~ s/&nbsp;//g;
        $hit =~ s/[ \t]+/ /g;
        $hit =~ s/\n+/ /g;
        $hit =~ s/^\s*//g;

        $hit =~ m/\(/;
        $title = $`;

        $string .= <<EOM;
		<Heading collapsible="true">
		<Text>$title</Text>
		<p>$hit</p>
		<Hyperlink url="$url">
			<Text> more </Text>
		</Hyperlink>
		</Heading>
EOM

    }

    $string .= <<EOM;
		<HorizontalRule/>
			<Hyperlink url="$SCRIPTBASE$SCRIPT/Search?search=$search">
			<Text>Continue your search in Browser ...</Text>
			</Hyperlink>
		</Content>
		</Results>
	</Range>
	<Status>SUCCESS</Status>
	</Response>
</ResponsePacket>
EOM

    $string = &htQuote($string);
    $string =~ &htUTF8($string);

    $string = <<EOM;
<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
	<QueryResponse xmlns="urn:Microsoft.Search">
	  <QueryResult>$string</QueryResult>
	</QueryResponse>
  </soap:Body>
</soap:Envelope>
EOM

    $len = length($string);
    binmode(STDOUT);
    print "Content-type: text/xml; charset=utf-8\nContent-Length: $len\n\n$string";
}

# print a list of all subroutines

sub cgiAdmPrintSubs {

    my $woda    = "$WodaFileDir/$WodaFile";
    my $type    = lc $CGI{'type'};
    my $inFrame = $CGI{'inframe'};
    my $menu    = $CGI{'menu'} eq 'yes' ? 'yes' : 'no';
    $type = '[A-Z][A-Z]' if $CGI{'type'} =~ m/api/i;
    $type = 'cgiAdm'     if $CGI{'type'} =~ m/cgiadm/i;

    $Log .= qq~\nSearch for:  m/^sub\s+$type/i\n~;
    my @subs;
    openFile( H, $woda );
    if ( $CGI{'type'} eq 'cgi' ) {
        while (<H>) {
            next if m/^sub\s+cgiAdm/;
            push( @subs, $_ ) if m/^sub\s+cgi/;
        }
    }
    else {
        while (<H>) {
            push( @subs, $_ ) if m/^sub\s+$type/;
        }
    }

    close(H);

    @subs = sort { $a cmp $b } @subs;

    if ( !$inFrame ) {
        print <<EOM;
Content-type: text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">
<head>
<title>Modify Setting</title>
</head>
<frameset cols="150,*" border="1" bordercolor="#000000">
	<frame src="$SCRIPT/AdmPrintSubs?inframe=1&amp;menu=yes" name="woda_menu" marginwidth="0" marginheight="0">
	<frame src="$SCRIPT/AdmPrintSubs?inframe=1&amp;type="api"" name="woda_right">
</frameset>
EOM
    }

    if ( $menu eq 'yes' ) {
        printFrameHead('Index');
    }
    else {
        printFrameHead('WODA subroutines');
    }

    my $last;
    my $curr;
    my @prefix;
    foreach $subs (@subs) {
        $subs =~ m/^sub\s+([a-z]+)/;
        $curr = $1;
        $subs =~ s#\{#</dt><dd>#g;
        $subs =~ s/#/<br>/g;
        $subs =~ s/<dd>\s+<br>/<dd>/g;
        $subs =~ s/^sub\s+//g;
        $subs = qq~<dt>$subs</dd>~;
        if ( $last ne $curr ) {
            $last = $curr;
            if ( $curr eq 'cgi' ) {
                $curr = uc $curr;
            }
            else {
                push( @prefix, $curr );
            }
            $subs = qq~<h2><a name="$curr"></a><a href="#top"><img src="$ICONURL/up.gif" $SMALLICONPAR></a> $curr </h2>~ . $subs;
        }
    }
    s/<dt>cgi/<dt>/ for @subs;
    s/<dt>my/<dt>/  for @subs;

    if ( $CGI{'menu'} eq 'yes' ) {
        print qq~
<div class="PADDED-H"><a name="top"></a> <a href="$SCRIPTNAME/$Page?type=api&amp;inframe=1" target="woda_right">Perl API</a> 
| <a name="top"></a> <a href="$SCRIPTNAME/$Page?type=cgi&amp;inframe=1" target="woda_right">User CGI</a> 
| <a name="top"></a> <a href="$SCRIPTNAME/$Page?type=cgiAdm&amp;inframe=1" target="woda_right">Admin&nbsp;CGI</a>
| <a name="top"></a> <a href="$SCRIPTNAME/$Page?type=&amp;inframe=1" target="woda_right">Show All</a><hr>Others by prefix<br>~;
        print qq~ <a href="$SCRIPTNAME/$Page?type=$_&amp;inframe=1" target="woda_right">$_</a><br>~ for @prefix;
        print '</div>';
    }
    else {
        print '<dl>', join( '', @subs ), '</dl>';
    }

    printFootOnly() if $menu eq 'yes';
}

# ---------------------------------------------------------------------

$xtime6 = time;    ##----------------------------------------------------- END OF WODA

1;                 ##----------------------------------------------------- NOTHING AFTER THIS LINE
